/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "pyFAI/ext/src/isnan.h"
        ],
        "extra_compile_args": [
            "-fopenmp"
        ],
        "extra_link_args": [
            "-fopenmp"
        ],
        "include_dirs": [
            "./pyFAI/ext",
            "/opt/python/cp36-cp36m/lib/python3.6/site-packages/numpy/core/include"
        ],
        "language": "c++",
        "name": "pyFAI.ext._distortion",
        "sources": [
            "pyFAI/ext/_distortion.pyx"
        ]
    },
    "module_name": "pyFAI.ext._distortion"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyFAI__ext___distortion
#define __PYX_HAVE_API__pyFAI__ext___distortion
/* Early includes */
#include "src/isnan.h"
#include <math.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <list>
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "pyFAI/ext/_distortion.pyx",
  "stringsource",
  "type.pxd",
  "pyFAI/ext/sparse_builder.pxd",
  "pyFAI/ext/regrid_common.pxi",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* None.proto */
#if defined(__GNUC__)
#define __Pyx_PACKED __attribute__((__packed__))
#else
#define __Pyx_PACKED
#endif

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "shared_types.pxd":6
 *                          int32_t, uint32_t, int64_t, uint64_t
 * 
 * ctypedef double float64_t             # <<<<<<<<<<<<<<
 * ctypedef float float32_t
 */
typedef double __pyx_t_5pyFAI_3ext_12shared_types_float64_t;

/* "shared_types.pxd":7
 * 
 * ctypedef double float64_t
 * ctypedef float float32_t             # <<<<<<<<<<<<<<
 */
typedef float __pyx_t_5pyFAI_3ext_12shared_types_float32_t;

/* "pyFAI/ext/regrid_common.pxi":60
 * 
 * # How position are stored
 * ctypedef float64_t position_t             # <<<<<<<<<<<<<<
 * position_d = numpy.float64
 * 
 */
typedef __pyx_t_5pyFAI_3ext_12shared_types_float64_t __pyx_t_5pyFAI_3ext_11_distortion_position_t;

/* "pyFAI/ext/regrid_common.pxi":64
 * 
 * # How weights or data are stored
 * ctypedef float32_t data_t             # <<<<<<<<<<<<<<
 * data_d = numpy.float32
 * 
 */
typedef __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_t_5pyFAI_3ext_11_distortion_data_t;

/* "pyFAI/ext/regrid_common.pxi":68
 * 
 * # how data are accumulated
 * ctypedef float64_t acc_t             # <<<<<<<<<<<<<<
 * acc_d = numpy.float64
 * 
 */
typedef __pyx_t_5pyFAI_3ext_12shared_types_float64_t __pyx_t_5pyFAI_3ext_11_distortion_acc_t;

/* "pyFAI/ext/regrid_common.pxi":72
 * 
 * # type of the mask:
 * ctypedef int8_t mask_t             # <<<<<<<<<<<<<<
 * mask_d = numpy.int8
 * 
 */
typedef int8_t __pyx_t_5pyFAI_3ext_11_distortion_mask_t;

/* "pyFAI/ext/regrid_common.pxi":76
 * 
 * # type of the indexes:
 * ctypedef int32_t index_t             # <<<<<<<<<<<<<<
 * index_d = numpy.int32
 * 
 */
typedef int32_t __pyx_t_5pyFAI_3ext_11_distortion_index_t;

/*--- Type declarations ---*/
struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;
struct __pyx_t_5pyFAI_3ext_14sparse_builder_pixel_t;
struct __pyx_t_5pyFAI_3ext_14sparse_builder_sparse_builder_private_t;

/* "sparse_builder.pxd":33
 * from .shared_types cimport int32_t, float32_t
 * 
 * cdef packed struct pixel_t:             # <<<<<<<<<<<<<<
 *     int32_t index
 *     float32_t coef
 */
#if defined(__SUNPRO_C)
  #pragma pack(1)
#elif !defined(__GNUC__)
  #pragma pack(push, 1)
#endif
struct __Pyx_PACKED __pyx_t_5pyFAI_3ext_14sparse_builder_pixel_t {
  int32_t index;
  __pyx_t_5pyFAI_3ext_12shared_types_float32_t coef;
};
#if defined(__SUNPRO_C)
  #pragma pack()
#elif !defined(__GNUC__)
  #pragma pack(pop)
#endif

/* "sparse_builder.pxd":38
 * 
 * 
 * cdef struct sparse_builder_private_t:             # <<<<<<<<<<<<<<
 *     void **_bins
 *     void *_compact_bins
 */
struct __pyx_t_5pyFAI_3ext_14sparse_builder_sparse_builder_private_t {
  void **_bins;
  void *_compact_bins;
  void *_heap;
};
struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t;
struct __pyx_t_5pyFAI_3ext_11_distortion_preproc_t;
struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_1d_accumulator;
struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_2d_accumulator;
struct __pyx_fuse_0__pyx_opt_args_5pyFAI_3ext_11_distortion_preproc_value_inplace;
struct __pyx_fuse_1__pyx_opt_args_5pyFAI_3ext_11_distortion_preproc_value_inplace;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;
struct __pyx_defaults2;
typedef struct __pyx_defaults2 __pyx_defaults2;
struct __pyx_defaults3;
typedef struct __pyx_defaults3 __pyx_defaults3;
struct __pyx_defaults4;
typedef struct __pyx_defaults4 __pyx_defaults4;
struct __pyx_defaults5;
typedef struct __pyx_defaults5 __pyx_defaults5;
struct __pyx_defaults6;
typedef struct __pyx_defaults6 __pyx_defaults6;
struct __pyx_defaults7;
typedef struct __pyx_defaults7 __pyx_defaults7;

/* "pyFAI/ext/regrid_common.pxi":79
 * index_d = numpy.int32
 * 
 * cdef struct lut_t:             # <<<<<<<<<<<<<<
 *     index_t idx
 *     data_t coef
 */
struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t {
  __pyx_t_5pyFAI_3ext_11_distortion_index_t idx;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t coef;
};

/* "pyFAI/ext/regrid_common.pxi":115
 * 
 * cdef:
 *     struct preproc_t:             # <<<<<<<<<<<<<<
 *         data_t signal
 *         data_t variance
 */
struct __pyx_t_5pyFAI_3ext_11_distortion_preproc_t {
  __pyx_t_5pyFAI_3ext_11_distortion_data_t signal;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t variance;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t norm;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t count;
};

/* "pyFAI/ext/regrid_common.pxi":236
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_1d_accumulator(acc_t[:, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin,
 *                                        preproc_t value,
 */
struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_1d_accumulator {
  int __pyx_n;
  double weight;
};

/* "pyFAI/ext/regrid_common.pxi":255
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_2d_accumulator(acc_t[:, :, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin0,
 *                                        int bin1,
 */
struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_2d_accumulator {
  int __pyx_n;
  double weight;
};

/* "pyFAI/ext/regrid_common.pxi":156
 * 
 * 
 * cdef inline bint preproc_value_inplace(preproc_t* result,             # <<<<<<<<<<<<<<
 *                                        floating data,
 *                                        floating variance=0.0,
 */
struct __pyx_fuse_0__pyx_opt_args_5pyFAI_3ext_11_distortion_preproc_value_inplace {
  int __pyx_n;
  float variance;
  float dark;
  float flat;
  float solidangle;
  float polarization;
  float absorption;
  __pyx_t_5pyFAI_3ext_11_distortion_mask_t mask;
  float dummy;
  float delta_dummy;
  int check_dummy;
  float normalization_factor;
  float dark_variance;
};
struct __pyx_fuse_1__pyx_opt_args_5pyFAI_3ext_11_distortion_preproc_value_inplace {
  int __pyx_n;
  double variance;
  double dark;
  double flat;
  double solidangle;
  double polarization;
  double absorption;
  __pyx_t_5pyFAI_3ext_11_distortion_mask_t mask;
  double dummy;
  double delta_dummy;
  int check_dummy;
  double normalization_factor;
  double dark_variance;
};
struct __pyx_defaults {
  PyObject *__pyx_arg_shape_out;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_shape_out;
};
struct __pyx_defaults2 {
  PyObject *__pyx_arg_shape_out;
};
struct __pyx_defaults3 {
  PyObject *__pyx_arg_shape_out;
};
struct __pyx_defaults4 {
  PyObject *__pyx_arg_offset;
  __Pyx_memviewslice __pyx_arg_mask;
};
struct __pyx_defaults5 {
  PyObject *__pyx_arg_offset;
  __Pyx_memviewslice __pyx_arg_mask;
};
struct __pyx_defaults6 {
  PyObject *__pyx_arg_offset;
  __Pyx_memviewslice __pyx_arg_mask;
};
struct __pyx_defaults7 {
  PyObject *__pyx_arg_offset;
  __Pyx_memviewslice __pyx_arg_mask;
};

/* "sparse_builder.pxd":44
 * 
 * 
 * cdef class SparseBuilder(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef sparse_builder_private_t _data
 */
struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyFAI_3ext_14sparse_builder_SparseBuilder *__pyx_vtab;
  struct __pyx_t_5pyFAI_3ext_14sparse_builder_sparse_builder_private_t _data;
  int _nbin;
  int _block_size;
  int *_sizes;
  bool _use_linked_list;
  bool _use_blocks;
  bool _use_heap_linked_list;
  bool _use_packed_list;
  PyObject *_mode;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "sparse_builder.pxd":44
 * 
 * 
 * cdef class SparseBuilder(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef sparse_builder_private_t _data
 */

struct __pyx_vtabstruct_5pyFAI_3ext_14sparse_builder_SparseBuilder {
  void *(*_create_bin)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *);
  void (*_copy_bin_indexes_to)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *, int, int32_t *);
  void (*_copy_bin_coefs_to)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *, int, __pyx_t_5pyFAI_3ext_12shared_types_float32_t *);
  void (*_copy_bin_data_to)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *, int, struct __pyx_t_5pyFAI_3ext_14sparse_builder_pixel_t *);
  int (*cget_bin_size)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *, int);
  void (*cinsert)(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *, int, int, __pyx_t_5pyFAI_3ext_12shared_types_float32_t);
};
static struct __pyx_vtabstruct_5pyFAI_3ext_14sparse_builder_SparseBuilder *__pyx_vtabptr_5pyFAI_3ext_14sparse_builder_SparseBuilder;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyIntFromDouble.proto */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value);
#else
#define __Pyx_PyInt_FromDouble(value) PyLong_FromDouble(value)
#endif

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_AddObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_AddCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_int8_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn_int8_t(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int8_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_int32_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn_int32_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(const char *itemp, PyObject *obj);

struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t;
static PyObject* __pyx_convert__to_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t s);
/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_int(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_int(const char *itemp, PyObject *obj);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* BytesContains.proto */
static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_int(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int8_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_int8_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *, int writable_flag);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'pyFAI.ext.isnan' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'pyFAI.ext.shared_types' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libcpp' */

/* Module declarations from 'libcpp.list' */

/* Module declarations from 'pyFAI.ext.sparse_builder' */
static PyTypeObject *__pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder = 0;

/* Module declarations from 'pyFAI.ext._distortion' */
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static float __pyx_v_5pyFAI_3ext_11_distortion_pi;
static float __pyx_v_5pyFAI_3ext_11_distortion_piover2;
static float __pyx_v_5pyFAI_3ext_11_distortion_onef;
static float __pyx_v_5pyFAI_3ext_11_distortion_zerof;
static double __pyx_v_5pyFAI_3ext_11_distortion_EPS32;
static int __pyx_v_5pyFAI_3ext_11_distortion_NEED_DECREF;
static PyObject *__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion_calc_area(float, float, float, float, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE int __pyx_f_5pyFAI_3ext_11_distortion_clip(int, int, int, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(float, float, float, float); /*proto*/
static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(float, float, float, float); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyFAI_3ext_11_distortion_integrate(__Pyx_memviewslice, float, float, float, float); /*proto*/
static struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_float = { "float", NULL, sizeof(float), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_int8_t = { "int8_t", NULL, sizeof(int8_t), { 0 }, 0, IS_UNSIGNED(int8_t) ? 'U' : 'I', IS_UNSIGNED(int8_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t = { "float32_t", NULL, sizeof(__pyx_t_5pyFAI_3ext_12shared_types_float32_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_11_distortion_index_t = { "index_t", NULL, sizeof(__pyx_t_5pyFAI_3ext_11_distortion_index_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5pyFAI_3ext_11_distortion_index_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5pyFAI_3ext_11_distortion_index_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_11_distortion_data_t = { "data_t", NULL, sizeof(__pyx_t_5pyFAI_3ext_11_distortion_data_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_StructField __Pyx_StructFields_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t[] = {
  {&__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_11_distortion_index_t, "idx", offsetof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t, idx)},
  {&__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_11_distortion_data_t, "coef", offsetof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t, coef)},
  {NULL, NULL, 0}
};
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t = { "lut_t", __Pyx_StructFields_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, sizeof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t), { 0 }, 0, 'S', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_int32_t = { "int32_t", NULL, sizeof(int32_t), { 0 }, 0, IS_UNSIGNED(int32_t) ? 'U' : 'I', IS_UNSIGNED(int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5pyFAI_3ext_12shared_types_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_int = { "int", NULL, sizeof(int), { 0 }, 0, IS_UNSIGNED(int) ? 'U' : 'I', IS_UNSIGNED(int), 0 };
#define __Pyx_MODULE_NAME "pyFAI.ext._distortion"
extern int __pyx_module_is_main_pyFAI__ext___distortion;
int __pyx_module_is_main_pyFAI__ext___distortion = 0;

/* Implementation of 'pyFAI.ext._distortion' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_A0[] = " A0=";
static const char __pyx_k_A1[] = "; A1=";
static const char __pyx_k_B0[] = "; B0=";
static const char __pyx_k_B1[] = "; B1=";
static const char __pyx_k_C0[] = "; C0=";
static const char __pyx_k_C1[] = "; C1=";
static const char __pyx_k_D0[] = "; D0=";
static const char __pyx_k_D1[] = "; D1=";
static const char __pyx_k_I1[] = "I1";
static const char __pyx_k_I2[] = "I2";
static const char __pyx_k__2[] = "()";
static const char __pyx_k__3[] = "|";
static const char __pyx_k_d1[] = "d1";
static const char __pyx_k_d2[] = "d2";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_if[] = "if";
static const char __pyx_k_ml[] = "ml";
static const char __pyx_k_ms[] = "ms";
static const char __pyx_k_nl[] = "nl";
static const char __pyx_k_ns[] = "ns";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_p0[] = "p0";
static const char __pyx_k_p1[] = "p1";
static const char __pyx_k_1_5[] = "1.5";
static const char __pyx_k_BIG[] = "BIG";
static const char __pyx_k_LUT[] = "LUT";
static const char __pyx_k_MIT[] = "MIT";
static const char __pyx_k_NaN[] = "NaN";
static const char __pyx_k_Nan[] = "Nan";
static const char __pyx_k__10[] = ", ";
static const char __pyx_k__11[] = ")";
static const char __pyx_k__18[] = "";
static const char __pyx_k__64[] = ".";
static const char __pyx_k_cAB[] = "cAB";
static const char __pyx_k_cBC[] = "cBC";
static const char __pyx_k_cCD[] = "cCD";
static const char __pyx_k_cDA[] = "cDA";
static const char __pyx_k_csr[] = "csr";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_eps[] = "eps";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_lin[] = "lin";
static const char __pyx_k_lut[] = "lut";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_pAB[] = "pAB";
static const char __pyx_k_pBC[] = "pBC";
static const char __pyx_k_pCD[] = "pCD";
static const char __pyx_k_pDA[] = "pDA";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_sem[] = "_sem";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_tot[] = "tot";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_A0_2[] = "A0";
static const char __pyx_k_A1_2[] = "A1";
static const char __pyx_k_B0_2[] = "B0";
static const char __pyx_k_B1_2[] = "B1";
static const char __pyx_k_C0_2[] = "C0";
static const char __pyx_k_C1_2[] = "C1";
static const char __pyx_k_D0_2[] = "D0";
static const char __pyx_k_D1_2[] = "D1";
static const char __pyx_k_area[] = "area";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_bins[] = "bins";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_clip[] = "clip";
static const char __pyx_k_coef[] = "coef";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_core[] = "core";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_date[] = "__date__";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_dim0[] = "dim0";
static const char __pyx_k_dim1[] = "dim1";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_int8[] = "int8";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_lerr[] = "lerr";
static const char __pyx_k_lout[] = "lout";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_max0[] = "max0";
static const char __pyx_k_max1[] = "max1";
static const char __pyx_k_min0[] = "min0";
static const char __pyx_k_min1[] = "min1";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_norm[] = "norm";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_prod[] = "prod";
static const char __pyx_k_prop[] = "prop";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_view[] = "view";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_acc_d[] = "acc_d";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_cmask[] = "cmask";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_dummy[] = "dummy";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_fabio[] = "fabio";
static const char __pyx_k_finfo[] = "finfo";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_image[] = "image";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_kahan[] = "kahan";
static const char __pyx_k_lmask[] = "lmask";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_lprop[] = "lprop";
static const char __pyx_k_lut_d[] = "lut_d";
static const char __pyx_k_major[] = "major";
static const char __pyx_k_nbins[] = "nbins";
static const char __pyx_k_nchan[] = "nchan";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ravel[] = "ravel";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_slope[] = "slope";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_total[] = "total";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_utils[] = "utils";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_author[] = "__author__";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_cdummy[] = "cdummy";
static const char __pyx_k_cumsum[] = "cumsum";
static const char __pyx_k_data_d[] = "data_d";
static const char __pyx_k_delta0[] = "delta0";
static const char __pyx_k_delta1[] = "delta1";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indptr[] = "indptr";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_mask_d[] = "mask_d";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_nbytes[] = "nbytes";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_outMax[] = "outMax";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pixel1[] = "pixel1";
static const char __pyx_k_pixel2[] = "pixel2";
static const char __pyx_k_prop_d[] = "prop_d";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_shape0[] = "shape0";
static const char __pyx_k_shape1[] = "shape1";
static const char __pyx_k_signal[] = "signal";
static const char __pyx_k_stable[] = "stable";
static const char __pyx_k_status[] = "__status__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_to_csr[] = "to_csr";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_builder[] = "builder";
static const char __pyx_k_contact[] = "__contact__";
static const char __pyx_k_correct[] = "correct";
static const char __pyx_k_counter[] = "counter";
static const char __pyx_k_do_mask[] = "do_mask";
static const char __pyx_k_err_cnt[] = "err_cnt";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_idx_bin[] = "idx_bin";
static const char __pyx_k_index_d[] = "index_d";
static const char __pyx_k_indices[] = "indices";
static const char __pyx_k_license[] = "__license__";
static const char __pyx_k_linesep[] = "linesep";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_max_val[] = "max_val";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_min_val[] = "min_val";
static const char __pyx_k_offset0[] = "offset0";
static const char __pyx_k_offset1[] = "offset1";
static const char __pyx_k_pos0max[] = "pos0max";
static const char __pyx_k_pos0min[] = "pos0min";
static const char __pyx_k_pos1max[] = "pos1max";
static const char __pyx_k_pos1min[] = "pos1min";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_records[] = "records";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_size_in[] = "size_in";
static const char __pyx_k_sum_sig[] = "sum_sig";
static const char __pyx_k_sum_var[] = "sum_var";
static const char __pyx_k_tmp_ary[] = "tmp_ary";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_all_max0[] = "all_max0";
static const char __pyx_k_all_max1[] = "all_max1";
static const char __pyx_k_all_min0[] = "all_min0";
static const char __pyx_k_all_min1[] = "all_min1";
static const char __pyx_k_bin_size[] = "bin_size";
static const char __pyx_k_calc_CSR[] = "calc_CSR";
static const char __pyx_k_calc_LUT[] = "calc_LUT";
static const char __pyx_k_calc_pos[] = "calc_pos";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_detector[] = "detector";
static const char __pyx_k_do_dummy[] = "do_dummy";
static const char __pyx_k_do_shape[] = "do_shape";
static const char __pyx_k_expand2d[] = "expand2d";
static const char __pyx_k_foffset0[] = "foffset0";
static const char __pyx_k_foffset1[] = "foffset1";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_lut_size[] = "lut_size";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_recarray[] = "recarray";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_shape_in[] = "shape_in";
static const char __pyx_k_sum_norm[] = "sum_norm";
static const char __pyx_k_variance[] = "variance";
static const char __pyx_k_Semaphore[] = "Semaphore";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_box_size0[] = "box_size0";
static const char __pyx_k_box_size1[] = "box_size1";
static const char __pyx_k_calc_size[] = "calc_size";
static const char __pyx_k_copyright[] = "__copyright__";
static const char __pyx_k_detectors[] = "detectors";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_heap_size[] = "heap_size";
static const char __pyx_k_idx_pixel[] = "idx_pixel";
static const char __pyx_k_img_shape[] = "img_shape";
static const char __pyx_k_intercept[] = "intercept";
static const char __pyx_k_max_shape[] = "max_shape";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_new_image[] = "new_image";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_shape_in0[] = "shape_in0";
static const char __pyx_k_shape_in1[] = "shape_in1";
static const char __pyx_k_shape_out[] = "shape_out";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_tmp_index[] = "tmp_index";
static const char __pyx_k_uncorrect[] = "uncorrect";
static const char __pyx_k_14_01_2021[] = "14/01/2021";
static const char __pyx_k_Distortion[] = "Distortion";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_bin_number[] = "bin_number";
static const char __pyx_k_block_size[] = "block_size";
static const char __pyx_k_containers[] = "containers";
static const char __pyx_k_large_data[] = "large_data";
static const char __pyx_k_large_size[] = "large_size";
static const char __pyx_k_namedtuple[] = "namedtuple";
static const char __pyx_k_nb_corners[] = "nb_corners";
static const char __pyx_k_position_d[] = "position_d";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_shape_img0[] = "shape_img0";
static const char __pyx_k_shape_img1[] = "shape_img1";
static const char __pyx_k_shape_out0[] = "shape_out0";
static const char __pyx_k_shape_out1[] = "shape_out1";
static const char __pyx_k_signatures[] = "signatures";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_calc_sparse[] = "calc_sparse";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_correct_CSR[] = "correct_CSR";
static const char __pyx_k_correct_LUT[] = "correct_LUT";
static const char __pyx_k_delta_dummy[] = "delta_dummy";
static const char __pyx_k_image_is_2D[] = "image is 2D";
static const char __pyx_k_image_is_3D[] = "image is 3D";
static const char __pyx_k_np_lut_size[] = "np_lut_size";
static const char __pyx_k_pixel_count[] = "pixel_count";
static const char __pyx_k_pos_corners[] = "pos_corners";
static const char __pyx_k_LUT_ITEMSIZE[] = "LUT_ITEMSIZE";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_cdelta_dummy[] = "cdelta_dummy";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_Accessing_i_i[] = "Accessing %i >= %i !!!";
static const char __pyx_k_calc_LUT_size[] = "calc_LUT_size";
static const char __pyx_k_pixel_corners[] = "pixel_corners";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_uncorrect_CSR[] = "uncorrect_CSR";
static const char __pyx_k_uncorrect_LUT[] = "uncorrect_LUT";
static const char __pyx_k_2011_2021_ESRF[] = "2011-2021, ESRF";
static const char __pyx_k_Integrate1dtpl[] = "Integrate1dtpl";
static const char __pyx_k_Integrate2dtpl[] = "Integrate2dtpl";
static const char __pyx_k_Jerome_Kieffer[] = "Jerome Kieffer";
static const char __pyx_k_bins_per_pixel[] = "bins_per_pixel";
static const char __pyx_k_builder_config[] = "builder_config";
static const char __pyx_k_calc_sparse_v2[] = "calc_sparse_v2";
static const char __pyx_k_lut_total_size[] = "lut_total_size";
static const char __pyx_k_max_pixel_size[] = "max_pixel_size";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_mask_shape_dim0[] = "mask shape dim0";
static const char __pyx_k_mask_shape_dim1[] = "mask shape dim1";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_resize_image_2D[] = "resize_image_2D";
static const char __pyx_k_resize_image_3D[] = "resize_image_3D";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_detector_factory[] = "detector_factory";
static const char __pyx_k_Distortion___init[] = "Distortion.__init__";
static const char __pyx_k_Distortion___repr[] = "Distortion.__repr__";
static const char __pyx_k_and_shape_of_mask[] = ") and shape of mask (";
static const char __pyx_k_ascontiguousarray[] = "ascontiguousarray";
static const char __pyx_k_correct_CSR_kahan[] = "correct_CSR_kahan";
static const char __pyx_k_correct_LUT_kahan[] = "correct_LUT_kahan";
static const char __pyx_k_preprocessed_data[] = "preprocessed_data";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_Distortion_correct[] = "Distortion.correct";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_correct_CSR_double[] = "correct_CSR_double";
static const char __pyx_k_correct_LUT_double[] = "correct_LUT_double";
static const char __pyx_k_numpy_1_12_py2_bug[] = "_numpy_1_12_py2_bug";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_CSR_matrix_3f_MByte[] = "CSR matrix: %.3f MByte";
static const char __pyx_k_Distortion_calc_LUT[] = "Distortion.calc_LUT";
static const char __pyx_k_Distortion_calc_pos[] = "Distortion.calc_pos";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_Distortion_uncorrect[] = "Distortion.uncorrect";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_pyFAI_ext__distortion[] = "pyFAI.ext._distortion";
static const char __pyx_k_Jerome_kieffer_esrf_fr[] = "Jerome.kieffer@esrf.fr";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_Temporary_storage_3fMB[] = "Temporary storage: %.3fMB";
static const char __pyx_k_jerome_kieffer_esrf_fr[] = "jerome.kieffer@esrf.fr";
static const char __pyx_k_shape_in0_mask_shape_0[] = "shape_in0 == mask.shape[0]";
static const char __pyx_k_shape_in1_mask_shape_1[] = "shape_in1 == mask.shape[1]";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Distortion_calc_LUT_size[] = "Distortion.calc_LUT_size";
static const char __pyx_k_LUT_shape_i_i_i_3f_MByte[] = "LUT shape: (%i,%i,%i) %.3f MByte";
static const char __pyx_k_calc_cartesian_positions[] = "calc_cartesian_positions";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_pyFAI_ext__distortion_pyx[] = "pyFAI/ext/_distortion.pyx";
static const char __pyx_k_correct_CSR_preproc_double[] = "correct_CSR_preproc_double";
static const char __pyx_k_correct_LUT_preproc_double[] = "correct_LUT_preproc_double";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_Distortion_correction_are_corre[] = "\nDistortion correction are correction are applied by look-up table (or CSR)\n";
static const char __pyx_k_Patching_image_as_image_is_ix_i[] = "Patching image as image is %ix%i and spline is %ix%i";
static const char __pyx_k_Patching_image_of_shape_ix_i_on[] = "Patching image of shape %ix%i on expected size of %ix%i";
static const char __pyx_k_The_look_up_table_has_dimension[] = "The look-up table has dimension 0 which is a non-sense. Did you mask out all pixel or is your image out of the geometry range?";
static const char __pyx_k_This_class_applies_a_distortion[] = "\n\n    This class applies a distortion correction on an image.\n\n    It is also able to apply an inversion of the correction.\n\n    ";
static const char __pyx_k_shape_out0_shape_out1_LUT_shape[] = "shape_out0 * shape_out1 == LUT.shape[0]";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_CSR_matrix_3f_MByte_Max_source_p[] = "CSR matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Distortion_correction_for_detect[] = "Distortion correction for detector:";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_LUT_matrix_3f_MByte_Max_source_p[] = "LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f";
static const char __pyx_k_Max_pixel_size_ix_i_Max_source_p[] = "Max pixel size: %ix%i; Max source pixel in target: %i";
static const char __pyx_k_Mismatch_between_shape_of_detect[] = "Mismatch between shape of detector (";
static const char __pyx_k_No_value_specified_for_struct_at[] = "No value specified for struct attribute 'idx'";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Pixel_size_cannot_be_null_Zero_d[] = "Pixel size cannot be null -> Zero division error";
static const char __pyx_k_Please_increase_bins_per_pixel_s[] = "Please increase bins_per_pixel=%s. ";
static const char __pyx_k_Provided_temporary_space_for_sto[] = "Provided temporary space for storage is not enough. ";
static const char __pyx_k_The_suggested_value_is_i_or_grea[] = "The suggested value is %i or greater.";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_Unimplemented_sparse_matrix_form[] = "Unimplemented sparse matrix format: %s";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_makes_little_sense_to_correct_fo[] = "makes little sense to correct for distortion non-flat detectors: %s";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_number_of_elements_s_average_per[] = "number of elements: %s, average per bin %.3f allocated max: %s";
static const char __pyx_k_shape_out0_shape_out1_indptr_siz[] = "shape_out0*shape_out1 == indptr.size-1";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_No_value_specified_for_struct_at_2[] = "No value specified for struct attribute 'coef'";
static PyObject *__pyx_kp_u_14_01_2021;
static PyObject *__pyx_kp_u_1_5;
static PyObject *__pyx_kp_u_2011_2021_ESRF;
static PyObject *__pyx_kp_u_A0;
static PyObject *__pyx_n_s_A0_2;
static PyObject *__pyx_kp_u_A1;
static PyObject *__pyx_n_s_A1_2;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_u_Accessing_i_i;
static PyObject *__pyx_kp_u_B0;
static PyObject *__pyx_n_s_B0_2;
static PyObject *__pyx_kp_u_B1;
static PyObject *__pyx_n_s_B1_2;
static PyObject *__pyx_n_s_BIG;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_u_C0;
static PyObject *__pyx_n_s_C0_2;
static PyObject *__pyx_kp_u_C1;
static PyObject *__pyx_n_s_C1_2;
static PyObject *__pyx_kp_u_CSR_matrix_3f_MByte;
static PyObject *__pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_kp_u_D0;
static PyObject *__pyx_n_s_D0_2;
static PyObject *__pyx_kp_u_D1;
static PyObject *__pyx_n_s_D1_2;
static PyObject *__pyx_n_s_Distortion;
static PyObject *__pyx_n_s_Distortion___init;
static PyObject *__pyx_n_s_Distortion___repr;
static PyObject *__pyx_n_s_Distortion_calc_LUT;
static PyObject *__pyx_n_s_Distortion_calc_LUT_size;
static PyObject *__pyx_n_s_Distortion_calc_pos;
static PyObject *__pyx_n_s_Distortion_correct;
static PyObject *__pyx_kp_u_Distortion_correction_for_detect;
static PyObject *__pyx_n_s_Distortion_uncorrect;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_I1;
static PyObject *__pyx_n_s_I2;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_n_s_Integrate1dtpl;
static PyObject *__pyx_n_s_Integrate2dtpl;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_kp_u_Jerome_Kieffer;
static PyObject *__pyx_kp_u_Jerome_kieffer_esrf_fr;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_LUT;
static PyObject *__pyx_n_s_LUT_ITEMSIZE;
static PyObject *__pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p;
static PyObject *__pyx_kp_u_LUT_shape_i_i_i_3f_MByte;
static PyObject *__pyx_n_u_MIT;
static PyObject *__pyx_kp_u_Max_pixel_size_ix_i_Max_source_p;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_kp_u_Mismatch_between_shape_of_detect;
static PyObject *__pyx_n_s_NaN;
static PyObject *__pyx_n_s_Nan;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_2;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_kp_u_Patching_image_as_image_is_ix_i;
static PyObject *__pyx_kp_u_Patching_image_of_shape_ix_i_on;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_kp_u_Pixel_size_cannot_be_null_Zero_d;
static PyObject *__pyx_kp_u_Please_increase_bins_per_pixel_s;
static PyObject *__pyx_kp_u_Provided_temporary_space_for_sto;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_Semaphore;
static PyObject *__pyx_kp_u_Temporary_storage_3fMB;
static PyObject *__pyx_kp_u_The_look_up_table_has_dimension;
static PyObject *__pyx_kp_u_The_suggested_value_is_i_or_grea;
static PyObject *__pyx_kp_s_This_class_applies_a_distortion;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_kp_u_Unimplemented_sparse_matrix_form;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_kp_u__18;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_u__64;
static PyObject *__pyx_n_s_acc_d;
static PyObject *__pyx_n_s_all_max0;
static PyObject *__pyx_n_s_all_max1;
static PyObject *__pyx_n_s_all_min0;
static PyObject *__pyx_n_s_all_min1;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_kp_u_and_shape_of_mask;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_area;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_ascontiguousarray;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_author;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_bin_number;
static PyObject *__pyx_n_s_bin_size;
static PyObject *__pyx_n_s_bins;
static PyObject *__pyx_n_s_bins_per_pixel;
static PyObject *__pyx_n_s_block_size;
static PyObject *__pyx_n_s_box_size0;
static PyObject *__pyx_n_s_box_size1;
static PyObject *__pyx_n_s_buffer;
static PyObject *__pyx_n_s_builder;
static PyObject *__pyx_n_s_builder_config;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_cAB;
static PyObject *__pyx_n_s_cBC;
static PyObject *__pyx_n_s_cCD;
static PyObject *__pyx_n_s_cDA;
static PyObject *__pyx_n_s_calc_CSR;
static PyObject *__pyx_n_s_calc_LUT;
static PyObject *__pyx_n_s_calc_LUT_size;
static PyObject *__pyx_n_s_calc_cartesian_positions;
static PyObject *__pyx_n_s_calc_pos;
static PyObject *__pyx_n_s_calc_size;
static PyObject *__pyx_n_s_calc_sparse;
static PyObject *__pyx_n_s_calc_sparse_v2;
static PyObject *__pyx_n_s_cdelta_dummy;
static PyObject *__pyx_n_s_cdummy;
static PyObject *__pyx_n_s_ceil;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clip;
static PyObject *__pyx_n_s_cmask;
static PyObject *__pyx_n_b_coef;
static PyObject *__pyx_n_s_coef;
static PyObject *__pyx_n_u_coef;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_concatenate;
static PyObject *__pyx_n_s_contact;
static PyObject *__pyx_n_s_containers;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_copyright;
static PyObject *__pyx_n_s_core;
static PyObject *__pyx_n_s_correct;
static PyObject *__pyx_n_s_correct_CSR;
static PyObject *__pyx_n_s_correct_CSR_double;
static PyObject *__pyx_n_s_correct_CSR_kahan;
static PyObject *__pyx_n_s_correct_CSR_preproc_double;
static PyObject *__pyx_n_s_correct_LUT;
static PyObject *__pyx_n_s_correct_LUT_double;
static PyObject *__pyx_n_s_correct_LUT_kahan;
static PyObject *__pyx_n_s_correct_LUT_preproc_double;
static PyObject *__pyx_n_b_count;
static PyObject *__pyx_n_u_count;
static PyObject *__pyx_n_s_counter;
static PyObject *__pyx_n_u_csr;
static PyObject *__pyx_n_s_cumsum;
static PyObject *__pyx_n_s_d1;
static PyObject *__pyx_n_s_d2;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_d;
static PyObject *__pyx_n_s_date;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_delta0;
static PyObject *__pyx_n_s_delta1;
static PyObject *__pyx_n_s_delta_dummy;
static PyObject *__pyx_n_s_detector;
static PyObject *__pyx_n_u_detector;
static PyObject *__pyx_n_s_detector_factory;
static PyObject *__pyx_n_s_detectors;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dim0;
static PyObject *__pyx_n_s_dim1;
static PyObject *__pyx_n_s_do_dummy;
static PyObject *__pyx_n_s_do_mask;
static PyObject *__pyx_n_s_do_shape;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_double;
static PyObject *__pyx_n_u_double;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_dummy;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eps;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_err_cnt;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_expand2d;
static PyObject *__pyx_n_s_fabio;
static PyObject *__pyx_n_s_finfo;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_floor;
static PyObject *__pyx_n_s_foffset0;
static PyObject *__pyx_n_s_foffset1;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getLogger;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_heap_size;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_b_idx;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_u_idx;
static PyObject *__pyx_n_s_idx_bin;
static PyObject *__pyx_n_s_idx_pixel;
static PyObject *__pyx_n_u_if;
static PyObject *__pyx_n_s_image;
static PyObject *__pyx_kp_u_image_is_2D;
static PyObject *__pyx_kp_u_image_is_3D;
static PyObject *__pyx_n_s_img_shape;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index_d;
static PyObject *__pyx_n_s_indices;
static PyObject *__pyx_n_s_indptr;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_int8;
static PyObject *__pyx_n_s_intercept;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_kp_u_jerome_kieffer_esrf_fr;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_u_kahan;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_large_data;
static PyObject *__pyx_n_s_large_size;
static PyObject *__pyx_n_s_lerr;
static PyObject *__pyx_n_s_license;
static PyObject *__pyx_n_s_lin;
static PyObject *__pyx_n_s_linesep;
static PyObject *__pyx_n_s_lmask;
static PyObject *__pyx_n_s_logger;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_n_s_lout;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_lprop;
static PyObject *__pyx_n_s_lut;
static PyObject *__pyx_n_u_lut;
static PyObject *__pyx_n_s_lut_d;
static PyObject *__pyx_n_s_lut_size;
static PyObject *__pyx_n_s_lut_total_size;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_major;
static PyObject *__pyx_kp_u_makes_little_sense_to_correct_fo;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask_d;
static PyObject *__pyx_kp_u_mask_shape_dim0;
static PyObject *__pyx_kp_u_mask_shape_dim1;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max0;
static PyObject *__pyx_n_s_max1;
static PyObject *__pyx_n_s_max_pixel_size;
static PyObject *__pyx_n_s_max_shape;
static PyObject *__pyx_n_u_max_shape;
static PyObject *__pyx_n_s_max_val;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_method;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_min0;
static PyObject *__pyx_n_s_min1;
static PyObject *__pyx_n_s_min_val;
static PyObject *__pyx_n_s_ml;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_ms;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_namedtuple;
static PyObject *__pyx_n_s_nb_corners;
static PyObject *__pyx_n_s_nbins;
static PyObject *__pyx_n_s_nbytes;
static PyObject *__pyx_n_s_nchan;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_new_image;
static PyObject *__pyx_n_s_nl;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_b_norm;
static PyObject *__pyx_n_u_norm;
static PyObject *__pyx_n_s_np_lut_size;
static PyObject *__pyx_n_s_ns;
static PyObject *__pyx_kp_u_number_of_elements_s_average_per;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_numpy_1_12_py2_bug;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_offset0;
static PyObject *__pyx_n_s_offset1;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_outMax;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_p0;
static PyObject *__pyx_n_s_p1;
static PyObject *__pyx_n_s_pAB;
static PyObject *__pyx_n_s_pBC;
static PyObject *__pyx_n_s_pCD;
static PyObject *__pyx_n_s_pDA;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pixel1;
static PyObject *__pyx_n_s_pixel2;
static PyObject *__pyx_n_s_pixel_corners;
static PyObject *__pyx_n_s_pixel_count;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_pos0max;
static PyObject *__pyx_n_s_pos0min;
static PyObject *__pyx_n_s_pos1max;
static PyObject *__pyx_n_s_pos1min;
static PyObject *__pyx_n_s_pos_corners;
static PyObject *__pyx_n_s_position_d;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_preprocessed_data;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_prod;
static PyObject *__pyx_n_s_prop;
static PyObject *__pyx_n_s_prop_d;
static PyObject *__pyx_n_s_pyFAI_ext__distortion;
static PyObject *__pyx_kp_s_pyFAI_ext__distortion_pyx;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ravel;
static PyObject *__pyx_n_s_recarray;
static PyObject *__pyx_n_s_records;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_resize_image_2D;
static PyObject *__pyx_n_s_resize_image_3D;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_sem;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_shape0;
static PyObject *__pyx_n_s_shape1;
static PyObject *__pyx_n_s_shape_img0;
static PyObject *__pyx_n_s_shape_img1;
static PyObject *__pyx_n_s_shape_in;
static PyObject *__pyx_n_s_shape_in0;
static PyObject *__pyx_kp_u_shape_in0_mask_shape_0;
static PyObject *__pyx_n_s_shape_in1;
static PyObject *__pyx_kp_u_shape_in1_mask_shape_1;
static PyObject *__pyx_n_s_shape_out;
static PyObject *__pyx_n_s_shape_out0;
static PyObject *__pyx_kp_u_shape_out0_shape_out1_LUT_shape;
static PyObject *__pyx_kp_u_shape_out0_shape_out1_indptr_siz;
static PyObject *__pyx_n_s_shape_out1;
static PyObject *__pyx_n_b_signal;
static PyObject *__pyx_n_u_signal;
static PyObject *__pyx_n_s_signatures;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_size_in;
static PyObject *__pyx_n_s_slope;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_u_stable;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_sum_norm;
static PyObject *__pyx_n_s_sum_sig;
static PyObject *__pyx_n_s_sum_var;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_threading;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_tmp_ary;
static PyObject *__pyx_n_s_tmp_index;
static PyObject *__pyx_n_s_to_csr;
static PyObject *__pyx_n_s_tot;
static PyObject *__pyx_n_s_total;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_uncorrect;
static PyObject *__pyx_n_s_uncorrect_CSR;
static PyObject *__pyx_n_s_uncorrect_LUT;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_utils;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_b_variance;
static PyObject *__pyx_n_s_variance;
static PyObject *__pyx_n_u_variance;
static PyObject *__pyx_n_s_version;
static PyObject *__pyx_n_s_version_info;
static PyObject *__pyx_n_s_view;
static PyObject *__pyx_n_s_warning;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_calc_area(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_I1, float __pyx_v_I2, float __pyx_v_slope, float __pyx_v_intercept); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_2clip(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value, int __pyx_v_min_val, int __pyx_v_max_val); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_4calc_pos(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_58__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_42calc_pos(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pixel_corners, float __pyx_v_pixel1, float __pyx_v_pixel2, PyObject *__pyx_v_shape_out); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_60__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_44calc_pos(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pixel_corners, float __pyx_v_pixel1, float __pyx_v_pixel2, PyObject *__pyx_v_shape_out); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_6calc_size(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_66__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_48calc_size(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_offset); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_68__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_50calc_size(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_offset); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_8calc_LUT(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_bin_size, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10calc_CSR(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_bin_size, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_12calc_sparse(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_format, int __pyx_v_bins_per_pixel); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_14calc_sparse_v2(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_format, int __pyx_v_bins_per_pixel, PyObject *__pyx_v_builder_config); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_16resize_image_2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_18resize_image_3D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_20correct(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_method); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_22correct_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_method); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_24correct_LUT_kahan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_26correct_LUT_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_28correct_CSR(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, CYTHON_UNUSED PyObject *__pyx_v_variance, PyObject *__pyx_v_method); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_30correct_CSR_kahan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_32correct_CSR_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_34correct_LUT_preproc_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_empty); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_36correct_CSR_preproc_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_empty); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_38uncorrect_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_LUT); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_40uncorrect_CSR(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape, PyObject *__pyx_v_LUT); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_detector, PyObject *__pyx_v_shape); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_4calc_pos(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_6calc_LUT_size(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_8calc_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_10correct(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_image); /* proto */
static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_12uncorrect(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_image); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_1e6;
static PyObject *__pyx_float_1_0e6;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k_;
static PyObject *__pyx_k__8;
static PyObject *__pyx_k__9;
static __Pyx_memviewslice __pyx_k__12;
static __Pyx_memviewslice __pyx_k__14;
static __Pyx_memviewslice __pyx_k__16;
static __Pyx_memviewslice __pyx_k__17;
static PyObject *__pyx_k__23;
static PyObject *__pyx_k__24;
static float __pyx_k__66;
static float __pyx_k__67;
static float __pyx_k__68;
static float __pyx_k__69;
static float __pyx_k__70;
static float __pyx_k__71;
static __pyx_t_5pyFAI_3ext_11_distortion_mask_t __pyx_k__72;
static float __pyx_k__73;
static float __pyx_k__74;
static int __pyx_k__75;
static float __pyx_k__76;
static float __pyx_k__77;
static double __pyx_k__78;
static double __pyx_k__79;
static double __pyx_k__80;
static double __pyx_k__81;
static double __pyx_k__82;
static double __pyx_k__83;
static __pyx_t_5pyFAI_3ext_11_distortion_mask_t __pyx_k__84;
static double __pyx_k__85;
static double __pyx_k__86;
static int __pyx_k__87;
static double __pyx_k__88;
static double __pyx_k__89;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_slice__20;
static PyObject *__pyx_slice__25;
static PyObject *__pyx_slice__28;
static PyObject *__pyx_slice__31;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__121;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__98;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__108;
static PyObject *__pyx_codeobj__110;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__116;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__120;
static PyObject *__pyx_codeobj__122;
static PyObject *__pyx_codeobj__124;
static PyObject *__pyx_codeobj__126;
static PyObject *__pyx_codeobj__128;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__134;
static PyObject *__pyx_codeobj__136;
static PyObject *__pyx_codeobj__138;
static PyObject *__pyx_codeobj__140;
static PyObject *__pyx_codeobj__142;
static PyObject *__pyx_codeobj__144;
static PyObject *__pyx_codeobj__151;
/* Late includes */

/* "pyFAI/ext/regrid_common.pxi":236
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_1d_accumulator(acc_t[:, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin,
 *                                        preproc_t value,
 */

static CYTHON_INLINE void __pyx_f_5pyFAI_3ext_11_distortion_update_1d_accumulator(__Pyx_memviewslice __pyx_v_out_data, int __pyx_v_bin, struct __pyx_t_5pyFAI_3ext_11_distortion_preproc_t __pyx_v_value, struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_1d_accumulator *__pyx_optional_args) {
  double __pyx_v_weight = ((double)1.0);
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_weight = __pyx_optional_args->weight;
    }
  }

  /* "pyFAI/ext/regrid_common.pxi":248
 *     :return: Nothing
 *     """
 *     out_data[bin, 0] += value.signal * weight             # <<<<<<<<<<<<<<
 *     out_data[bin, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin, 2] += value.norm * weight
 */
  __pyx_t_1 = __pyx_v_bin;
  __pyx_t_2 = 0;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_1 * __pyx_v_out_data.strides[0]) )) + __pyx_t_2)) )) += (__pyx_v_value.signal * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":249
 *     """
 *     out_data[bin, 0] += value.signal * weight
 *     out_data[bin, 1] += value.variance * weight * weight  # Important for variance propagation             # <<<<<<<<<<<<<<
 *     out_data[bin, 2] += value.norm * weight
 *     out_data[bin, 3] += value.count * weight
 */
  __pyx_t_2 = __pyx_v_bin;
  __pyx_t_1 = 1;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_2 * __pyx_v_out_data.strides[0]) )) + __pyx_t_1)) )) += ((__pyx_v_value.variance * __pyx_v_weight) * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":250
 *     out_data[bin, 0] += value.signal * weight
 *     out_data[bin, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin, 2] += value.norm * weight             # <<<<<<<<<<<<<<
 *     out_data[bin, 3] += value.count * weight
 * 
 */
  __pyx_t_1 = __pyx_v_bin;
  __pyx_t_2 = 2;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_1 * __pyx_v_out_data.strides[0]) )) + __pyx_t_2)) )) += (__pyx_v_value.norm * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":251
 *     out_data[bin, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin, 2] += value.norm * weight
 *     out_data[bin, 3] += value.count * weight             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_bin;
  __pyx_t_1 = 3;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_2 * __pyx_v_out_data.strides[0]) )) + __pyx_t_1)) )) += (__pyx_v_value.count * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":236
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_1d_accumulator(acc_t[:, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin,
 *                                        preproc_t value,
 */

  /* function exit code */
}

/* "pyFAI/ext/regrid_common.pxi":255
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_2d_accumulator(acc_t[:, :, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin0,
 *                                        int bin1,
 */

static CYTHON_INLINE void __pyx_f_5pyFAI_3ext_11_distortion_update_2d_accumulator(__Pyx_memviewslice __pyx_v_out_data, int __pyx_v_bin0, int __pyx_v_bin1, struct __pyx_t_5pyFAI_3ext_11_distortion_preproc_t __pyx_v_value, struct __pyx_opt_args_5pyFAI_3ext_11_distortion_update_2d_accumulator *__pyx_optional_args) {
  double __pyx_v_weight = ((double)1.0);
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_weight = __pyx_optional_args->weight;
    }
  }

  /* "pyFAI/ext/regrid_common.pxi":266
 *     :return: Nothing
 *     """
 *     out_data[bin0, bin1, 0] += value.signal * weight             # <<<<<<<<<<<<<<
 *     out_data[bin0, bin1, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin0, bin1, 2] += value.norm * weight
 */
  __pyx_t_1 = __pyx_v_bin0;
  __pyx_t_2 = __pyx_v_bin1;
  __pyx_t_3 = 0;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=2 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_1 * __pyx_v_out_data.strides[0]) ) + __pyx_t_2 * __pyx_v_out_data.strides[1]) )) + __pyx_t_3)) )) += (__pyx_v_value.signal * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":267
 *     """
 *     out_data[bin0, bin1, 0] += value.signal * weight
 *     out_data[bin0, bin1, 1] += value.variance * weight * weight  # Important for variance propagation             # <<<<<<<<<<<<<<
 *     out_data[bin0, bin1, 2] += value.norm * weight
 *     out_data[bin0, bin1, 3] += value.count * weight
 */
  __pyx_t_3 = __pyx_v_bin0;
  __pyx_t_2 = __pyx_v_bin1;
  __pyx_t_1 = 1;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=2 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_3 * __pyx_v_out_data.strides[0]) ) + __pyx_t_2 * __pyx_v_out_data.strides[1]) )) + __pyx_t_1)) )) += ((__pyx_v_value.variance * __pyx_v_weight) * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":268
 *     out_data[bin0, bin1, 0] += value.signal * weight
 *     out_data[bin0, bin1, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin0, bin1, 2] += value.norm * weight             # <<<<<<<<<<<<<<
 *     out_data[bin0, bin1, 3] += value.count * weight
 */
  __pyx_t_1 = __pyx_v_bin0;
  __pyx_t_2 = __pyx_v_bin1;
  __pyx_t_3 = 2;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=2 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_1 * __pyx_v_out_data.strides[0]) ) + __pyx_t_2 * __pyx_v_out_data.strides[1]) )) + __pyx_t_3)) )) += (__pyx_v_value.norm * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":269
 *     out_data[bin0, bin1, 1] += value.variance * weight * weight  # Important for variance propagation
 *     out_data[bin0, bin1, 2] += value.norm * weight
 *     out_data[bin0, bin1, 3] += value.count * weight             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __pyx_v_bin0;
  __pyx_t_2 = __pyx_v_bin1;
  __pyx_t_1 = 3;
  *((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=2 */ ((char *) (((__pyx_t_5pyFAI_3ext_11_distortion_acc_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_out_data.data + __pyx_t_3 * __pyx_v_out_data.strides[0]) ) + __pyx_t_2 * __pyx_v_out_data.strides[1]) )) + __pyx_t_1)) )) += (__pyx_v_value.count * __pyx_v_weight);

  /* "pyFAI/ext/regrid_common.pxi":255
 * 
 * @cython.boundscheck(False)
 * cdef inline void update_2d_accumulator(acc_t[:, :, ::1] out_data,             # <<<<<<<<<<<<<<
 *                                        int bin0,
 *                                        int bin1,
 */

  /* function exit code */
}

/* "pyFAI/ext/_distortion.pyx":66
 * 
 * 
 * cpdef inline float calc_area(float I1, float I2, float slope, float intercept) nogil:             # <<<<<<<<<<<<<<
 *     "Calculate the area between I1 and I2 of a line with a given slope & intercept"
 *     return 0.5 * (I2 - I1) * (slope * (I2 + I1) + 2 * intercept)
 */

static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_1calc_area(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion_calc_area(float __pyx_v_I1, float __pyx_v_I2, float __pyx_v_slope, float __pyx_v_intercept, CYTHON_UNUSED int __pyx_skip_dispatch) {
  float __pyx_r;

  /* "pyFAI/ext/_distortion.pyx":68
 * cpdef inline float calc_area(float I1, float I2, float slope, float intercept) nogil:
 *     "Calculate the area between I1 and I2 of a line with a given slope & intercept"
 *     return 0.5 * (I2 - I1) * (slope * (I2 + I1) + 2 * intercept)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((0.5 * (__pyx_v_I2 - __pyx_v_I1)) * ((__pyx_v_slope * (__pyx_v_I2 + __pyx_v_I1)) + (2.0 * __pyx_v_intercept)));
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":66
 * 
 * 
 * cpdef inline float calc_area(float I1, float I2, float slope, float intercept) nogil:             # <<<<<<<<<<<<<<
 *     "Calculate the area between I1 and I2 of a line with a given slope & intercept"
 *     return 0.5 * (I2 - I1) * (slope * (I2 + I1) + 2 * intercept)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_1calc_area(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_calc_area[] = "calc_area(float I1, float I2, float slope, float intercept) -> float\nCalculate the area between I1 and I2 of a line with a given slope & intercept";
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_1calc_area(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  float __pyx_v_I1;
  float __pyx_v_I2;
  float __pyx_v_slope;
  float __pyx_v_intercept;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_area (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_I1,&__pyx_n_s_I2,&__pyx_n_s_slope,&__pyx_n_s_intercept,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_I1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_I2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_area", 1, 4, 4, 1); __PYX_ERR(0, 66, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_slope)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_area", 1, 4, 4, 2); __PYX_ERR(0, 66, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intercept)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_area", 1, 4, 4, 3); __PYX_ERR(0, 66, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_area") < 0)) __PYX_ERR(0, 66, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_I1 = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_I1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
    __pyx_v_I2 = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_I2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
    __pyx_v_slope = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_slope == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
    __pyx_v_intercept = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_intercept == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_area", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 66, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_area", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_calc_area(__pyx_self, __pyx_v_I1, __pyx_v_I2, __pyx_v_slope, __pyx_v_intercept);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_calc_area(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_I1, float __pyx_v_I2, float __pyx_v_slope, float __pyx_v_intercept) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_area", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_I1, __pyx_v_I2, __pyx_v_slope, __pyx_v_intercept, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_area", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":71
 * 
 * 
 * cpdef inline int clip(int value, int min_val, int max_val) nogil:             # <<<<<<<<<<<<<<
 *     "Limits the value to bounds"
 *     if value < min_val:
 */

static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_3clip(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE int __pyx_f_5pyFAI_3ext_11_distortion_clip(int __pyx_v_value, int __pyx_v_min_val, int __pyx_v_max_val, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":73
 * cpdef inline int clip(int value, int min_val, int max_val) nogil:
 *     "Limits the value to bounds"
 *     if value < min_val:             # <<<<<<<<<<<<<<
 *         return min_val
 *     elif value > max_val:
 */
  __pyx_t_1 = ((__pyx_v_value < __pyx_v_min_val) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":74
 *     "Limits the value to bounds"
 *     if value < min_val:
 *         return min_val             # <<<<<<<<<<<<<<
 *     elif value > max_val:
 *         return max_val
 */
    __pyx_r = __pyx_v_min_val;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":73
 * cpdef inline int clip(int value, int min_val, int max_val) nogil:
 *     "Limits the value to bounds"
 *     if value < min_val:             # <<<<<<<<<<<<<<
 *         return min_val
 *     elif value > max_val:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":75
 *     if value < min_val:
 *         return min_val
 *     elif value > max_val:             # <<<<<<<<<<<<<<
 *         return max_val
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_value > __pyx_v_max_val) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":76
 *         return min_val
 *     elif value > max_val:
 *         return max_val             # <<<<<<<<<<<<<<
 *     else:
 *         return value
 */
    __pyx_r = __pyx_v_max_val;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":75
 *     if value < min_val:
 *         return min_val
 *     elif value > max_val:             # <<<<<<<<<<<<<<
 *         return max_val
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":78
 *         return max_val
 *     else:
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }

  /* "pyFAI/ext/_distortion.pyx":71
 * 
 * 
 * cpdef inline int clip(int value, int min_val, int max_val) nogil:             # <<<<<<<<<<<<<<
 *     "Limits the value to bounds"
 *     if value < min_val:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_3clip(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_2clip[] = "clip(int value, int min_val, int max_val) -> int\nLimits the value to bounds";
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_3clip(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_value;
  int __pyx_v_min_val;
  int __pyx_v_max_val;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clip (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_min_val,&__pyx_n_s_max_val,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_val)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("clip", 1, 3, 3, 1); __PYX_ERR(0, 71, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_val)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("clip", 1, 3, 3, 2); __PYX_ERR(0, 71, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clip") < 0)) __PYX_ERR(0, 71, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_value = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L3_error)
    __pyx_v_min_val = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_min_val == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L3_error)
    __pyx_v_max_val = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_max_val == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clip", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 71, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.clip", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_2clip(__pyx_self, __pyx_v_value, __pyx_v_min_val, __pyx_v_max_val);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_2clip(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value, int __pyx_v_min_val, int __pyx_v_max_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clip", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_5pyFAI_3ext_11_distortion_clip(__pyx_v_value, __pyx_v_min_val, __pyx_v_max_val, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.clip", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":81
 * 
 * 
 * cdef inline float _floor_min4(float a, float b, float c, float d) nogil:             # <<<<<<<<<<<<<<
 *     "return floor(min(a,b,c,d))"
 *     cdef float res
 */

static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(float __pyx_v_a, float __pyx_v_b, float __pyx_v_c, float __pyx_v_d) {
  float __pyx_v_res;
  float __pyx_r;
  int __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":84
 *     "return floor(min(a,b,c,d))"
 *     cdef float res
 *     if (b < a):             # <<<<<<<<<<<<<<
 *         res = b
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_b < __pyx_v_a) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":85
 *     cdef float res
 *     if (b < a):
 *         res = b             # <<<<<<<<<<<<<<
 *     else:
 *         res = a
 */
    __pyx_v_res = __pyx_v_b;

    /* "pyFAI/ext/_distortion.pyx":84
 *     "return floor(min(a,b,c,d))"
 *     cdef float res
 *     if (b < a):             # <<<<<<<<<<<<<<
 *         res = b
 *     else:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":87
 *         res = b
 *     else:
 *         res = a             # <<<<<<<<<<<<<<
 *     if (c < res):
 *         res = c
 */
  /*else*/ {
    __pyx_v_res = __pyx_v_a;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":88
 *     else:
 *         res = a
 *     if (c < res):             # <<<<<<<<<<<<<<
 *         res = c
 *     if (d < res):
 */
  __pyx_t_1 = ((__pyx_v_c < __pyx_v_res) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":89
 *         res = a
 *     if (c < res):
 *         res = c             # <<<<<<<<<<<<<<
 *     if (d < res):
 *         res = d
 */
    __pyx_v_res = __pyx_v_c;

    /* "pyFAI/ext/_distortion.pyx":88
 *     else:
 *         res = a
 *     if (c < res):             # <<<<<<<<<<<<<<
 *         res = c
 *     if (d < res):
 */
  }

  /* "pyFAI/ext/_distortion.pyx":90
 *     if (c < res):
 *         res = c
 *     if (d < res):             # <<<<<<<<<<<<<<
 *         res = d
 *     return floor(res)
 */
  __pyx_t_1 = ((__pyx_v_d < __pyx_v_res) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":91
 *         res = c
 *     if (d < res):
 *         res = d             # <<<<<<<<<<<<<<
 *     return floor(res)
 * 
 */
    __pyx_v_res = __pyx_v_d;

    /* "pyFAI/ext/_distortion.pyx":90
 *     if (c < res):
 *         res = c
 *     if (d < res):             # <<<<<<<<<<<<<<
 *         res = d
 *     return floor(res)
 */
  }

  /* "pyFAI/ext/_distortion.pyx":92
 *     if (d < res):
 *         res = d
 *     return floor(res)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = floor(__pyx_v_res);
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":81
 * 
 * 
 * cdef inline float _floor_min4(float a, float b, float c, float d) nogil:             # <<<<<<<<<<<<<<
 *     "return floor(min(a,b,c,d))"
 *     cdef float res
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":95
 * 
 * 
 * cdef inline float _ceil_max4(float a, float b, float c, float d) nogil:             # <<<<<<<<<<<<<<
 *     "return  ceil(max(a,b,c,d))"
 *     cdef float res
 */

static CYTHON_INLINE float __pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(float __pyx_v_a, float __pyx_v_b, float __pyx_v_c, float __pyx_v_d) {
  float __pyx_v_res;
  float __pyx_r;
  int __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":98
 *     "return  ceil(max(a,b,c,d))"
 *     cdef float res
 *     if (b > a):             # <<<<<<<<<<<<<<
 *         res = b
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_b > __pyx_v_a) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":99
 *     cdef float res
 *     if (b > a):
 *         res = b             # <<<<<<<<<<<<<<
 *     else:
 *         res = a
 */
    __pyx_v_res = __pyx_v_b;

    /* "pyFAI/ext/_distortion.pyx":98
 *     "return  ceil(max(a,b,c,d))"
 *     cdef float res
 *     if (b > a):             # <<<<<<<<<<<<<<
 *         res = b
 *     else:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":101
 *         res = b
 *     else:
 *         res = a             # <<<<<<<<<<<<<<
 *     if (c > res):
 *         res = c
 */
  /*else*/ {
    __pyx_v_res = __pyx_v_a;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":102
 *     else:
 *         res = a
 *     if (c > res):             # <<<<<<<<<<<<<<
 *         res = c
 *     if (d > res):
 */
  __pyx_t_1 = ((__pyx_v_c > __pyx_v_res) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":103
 *         res = a
 *     if (c > res):
 *         res = c             # <<<<<<<<<<<<<<
 *     if (d > res):
 *         res = d
 */
    __pyx_v_res = __pyx_v_c;

    /* "pyFAI/ext/_distortion.pyx":102
 *     else:
 *         res = a
 *     if (c > res):             # <<<<<<<<<<<<<<
 *         res = c
 *     if (d > res):
 */
  }

  /* "pyFAI/ext/_distortion.pyx":104
 *     if (c > res):
 *         res = c
 *     if (d > res):             # <<<<<<<<<<<<<<
 *         res = d
 *     return ceil(res)
 */
  __pyx_t_1 = ((__pyx_v_d > __pyx_v_res) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":105
 *         res = c
 *     if (d > res):
 *         res = d             # <<<<<<<<<<<<<<
 *     return ceil(res)
 * 
 */
    __pyx_v_res = __pyx_v_d;

    /* "pyFAI/ext/_distortion.pyx":104
 *     if (c > res):
 *         res = c
 *     if (d > res):             # <<<<<<<<<<<<<<
 *         res = d
 *     return ceil(res)
 */
  }

  /* "pyFAI/ext/_distortion.pyx":106
 *     if (d > res):
 *         res = d
 *     return ceil(res)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ceil(__pyx_v_res);
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":95
 * 
 * 
 * cdef inline float _ceil_max4(float a, float b, float c, float d) nogil:             # <<<<<<<<<<<<<<
 *     "return  ceil(max(a,b,c,d))"
 *     cdef float res
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":109
 * 
 * 
 * cdef inline void integrate(float32_t[:, ::1] box, float start, float stop, float slope, float intercept) nogil:             # <<<<<<<<<<<<<<
 *     """Integrate in a box a line between start and stop, line defined by its slope & intercept
 * 
 */

static CYTHON_INLINE void __pyx_f_5pyFAI_3ext_11_distortion_integrate(__Pyx_memviewslice __pyx_v_box, float __pyx_v_start, float __pyx_v_stop, float __pyx_v_slope, float __pyx_v_intercept) {
  int __pyx_v_i;
  int __pyx_v_h;
  float __pyx_v_P;
  float __pyx_v_dP;
  float __pyx_v_segment_area;
  float __pyx_v_abs_area;
  float __pyx_v_dA;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":115
 *     """
 *     cdef:
 *         int i, h = 0             # <<<<<<<<<<<<<<
 *         float P, dP, segment_area, abs_area, dA
 *         # , sign
 */
  __pyx_v_h = 0;

  /* "pyFAI/ext/_distortion.pyx":118
 *         float P, dP, segment_area, abs_area, dA
 *         # , sign
 *     if start < stop:  # positive contribution             # <<<<<<<<<<<<<<
 *         P = ceil(start)
 *         dP = P - start
 */
  __pyx_t_1 = ((__pyx_v_start < __pyx_v_stop) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":119
 *         # , sign
 *     if start < stop:  # positive contribution
 *         P = ceil(start)             # <<<<<<<<<<<<<<
 *         dP = P - start
 *         if P > stop:  # start and stop are in the same unit
 */
    __pyx_v_P = ceil(__pyx_v_start);

    /* "pyFAI/ext/_distortion.pyx":120
 *     if start < stop:  # positive contribution
 *         P = ceil(start)
 *         dP = P - start             # <<<<<<<<<<<<<<
 *         if P > stop:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 */
    __pyx_v_dP = (__pyx_v_P - __pyx_v_start);

    /* "pyFAI/ext/_distortion.pyx":121
 *         P = ceil(start)
 *         dP = P - start
 *         if P > stop:  # start and stop are in the same unit             # <<<<<<<<<<<<<<
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0.0:
 */
    __pyx_t_1 = ((__pyx_v_P > __pyx_v_stop) != 0);
    if (__pyx_t_1) {

      /* "pyFAI/ext/_distortion.pyx":122
 *         dP = P - start
 *         if P > stop:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)             # <<<<<<<<<<<<<<
 *             if segment_area != 0.0:
 *                 abs_area = fabs(segment_area)
 */
      __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_start, __pyx_v_stop, __pyx_v_slope, __pyx_v_intercept, 0);

      /* "pyFAI/ext/_distortion.pyx":123
 *         if P > stop:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0.0:             # <<<<<<<<<<<<<<
 *                 abs_area = fabs(segment_area)
 *                 dA = (stop - start)  # always positive
 */
      __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":124
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0.0:
 *                 abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                 dA = (stop - start)  # always positive
 *                 h = 0
 */
        __pyx_v_abs_area = fabs(__pyx_v_segment_area);

        /* "pyFAI/ext/_distortion.pyx":125
 *             if segment_area != 0.0:
 *                 abs_area = fabs(segment_area)
 *                 dA = (stop - start)  # always positive             # <<<<<<<<<<<<<<
 *                 h = 0
 *                 while abs_area > 0:
 */
        __pyx_v_dA = (__pyx_v_stop - __pyx_v_start);

        /* "pyFAI/ext/_distortion.pyx":126
 *                 abs_area = fabs(segment_area)
 *                 dA = (stop - start)  # always positive
 *                 h = 0             # <<<<<<<<<<<<<<
 *                 while abs_area > 0:
 *                     if dA > abs_area:
 */
        __pyx_v_h = 0;

        /* "pyFAI/ext/_distortion.pyx":127
 *                 dA = (stop - start)  # always positive
 *                 h = 0
 *                 while abs_area > 0:             # <<<<<<<<<<<<<<
 *                     if dA > abs_area:
 *                         dA = abs_area
 */
        while (1) {
          __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
          if (!__pyx_t_1) break;

          /* "pyFAI/ext/_distortion.pyx":128
 *                 h = 0
 *                 while abs_area > 0:
 *                     if dA > abs_area:             # <<<<<<<<<<<<<<
 *                         dA = abs_area
 *                         abs_area = -1
 */
          __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
          if (__pyx_t_1) {

            /* "pyFAI/ext/_distortion.pyx":129
 *                 while abs_area > 0:
 *                     if dA > abs_area:
 *                         dA = abs_area             # <<<<<<<<<<<<<<
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 */
            __pyx_v_dA = __pyx_v_abs_area;

            /* "pyFAI/ext/_distortion.pyx":130
 *                     if dA > abs_area:
 *                         dA = abs_area
 *                         abs_area = -1             # <<<<<<<<<<<<<<
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA
 */
            __pyx_v_abs_area = -1.0;

            /* "pyFAI/ext/_distortion.pyx":128
 *                 h = 0
 *                 while abs_area > 0:
 *                     if dA > abs_area:             # <<<<<<<<<<<<<<
 *                         dA = abs_area
 *                         abs_area = -1
 */
          }

          /* "pyFAI/ext/_distortion.pyx":131
 *                         dA = abs_area
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                     abs_area -= dA
 *                     h += 1
 */
          __pyx_t_2 = ((int)__pyx_v_start);
          __pyx_t_3 = __pyx_v_h;
          *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_2 * __pyx_v_box.strides[0]) )) + __pyx_t_3)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":132
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA             # <<<<<<<<<<<<<<
 *                     h += 1
 *         else:
 */
          __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

          /* "pyFAI/ext/_distortion.pyx":133
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA
 *                     h += 1             # <<<<<<<<<<<<<<
 *         else:
 *             if dP > 0:
 */
          __pyx_v_h = (__pyx_v_h + 1);
        }

        /* "pyFAI/ext/_distortion.pyx":123
 *         if P > stop:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0.0:             # <<<<<<<<<<<<<<
 *                 abs_area = fabs(segment_area)
 *                 dA = (stop - start)  # always positive
 */
      }

      /* "pyFAI/ext/_distortion.pyx":121
 *         P = ceil(start)
 *         dP = P - start
 *         if P > stop:  # start and stop are in the same unit             # <<<<<<<<<<<<<<
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0.0:
 */
      goto __pyx_L4;
    }

    /* "pyFAI/ext/_distortion.pyx":135
 *                     h += 1
 *         else:
 *             if dP > 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0.0:
 */
    /*else*/ {
      __pyx_t_1 = ((__pyx_v_dP > 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":136
 *         else:
 *             if dP > 0:
 *                 segment_area = calc_area(start, P, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0.0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_start, __pyx_v_P, __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":137
 *             if dP > 0:
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0.0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":138
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0.0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = dP
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":139
 *                 if segment_area != 0.0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = dP
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":140
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = dP             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = __pyx_v_dP;

          /* "pyFAI/ext/_distortion.pyx":141
 *                     h = 0
 *                     dA = dP
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":142
 *                     dA = dP
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":143
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[(<int> P) - 1, h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":144
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[(<int> P) - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":142
 *                     dA = dP
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":145
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[(<int> P) - 1, h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_3 = (((int)__pyx_v_P) - 1);
            __pyx_t_2 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_3 * __pyx_v_box.strides[0]) )) + __pyx_t_2)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":146
 *                             abs_area = -1
 *                         box[(<int> P) - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 *             # subsection P1->Pn
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":147
 *                         box[(<int> P) - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 *             # subsection P1->Pn
 *             for i in range((<int> floor(P)), (<int> floor(stop))):
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":137
 *             if dP > 0:
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0.0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }

        /* "pyFAI/ext/_distortion.pyx":135
 *                     h += 1
 *         else:
 *             if dP > 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0.0:
 */
      }

      /* "pyFAI/ext/_distortion.pyx":149
 *                         h += 1
 *             # subsection P1->Pn
 *             for i in range((<int> floor(P)), (<int> floor(stop))):             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(i, i + 1, slope, intercept)
 *                 if segment_area != 0:
 */
      __pyx_t_4 = ((int)floor(__pyx_v_stop));
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = ((int)floor(__pyx_v_P)); __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "pyFAI/ext/_distortion.pyx":150
 *             # subsection P1->Pn
 *             for i in range((<int> floor(P)), (<int> floor(stop))):
 *                 segment_area = calc_area(i, i + 1, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_i, (__pyx_v_i + 1), __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":151
 *             for i in range((<int> floor(P)), (<int> floor(stop))):
 *                 segment_area = calc_area(i, i + 1, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":152
 *                 segment_area = calc_area(i, i + 1, slope, intercept)
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = 1.0
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":153
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = 1.0
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":154
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = 1.0             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = 1.0;

          /* "pyFAI/ext/_distortion.pyx":155
 *                     h = 0
 *                     dA = 1.0
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":156
 *                     dA = 1.0
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":157
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[i, h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":158
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[i, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":156
 *                     dA = 1.0
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":159
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[i, h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_2 = __pyx_v_i;
            __pyx_t_3 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_2 * __pyx_v_box.strides[0]) )) + __pyx_t_3)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":160
 *                             abs_area = -1
 *                         box[i, h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 *             # Section Pn->B
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":161
 *                         box[i, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 *             # Section Pn->B
 *             P = floor(stop)
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":151
 *             for i in range((<int> floor(P)), (<int> floor(stop))):
 *                 segment_area = calc_area(i, i + 1, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }
      }

      /* "pyFAI/ext/_distortion.pyx":163
 *                         h += 1
 *             # Section Pn->B
 *             P = floor(stop)             # <<<<<<<<<<<<<<
 *             dP = stop - P
 *             if dP > 0:
 */
      __pyx_v_P = floor(__pyx_v_stop);

      /* "pyFAI/ext/_distortion.pyx":164
 *             # Section Pn->B
 *             P = floor(stop)
 *             dP = stop - P             # <<<<<<<<<<<<<<
 *             if dP > 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 */
      __pyx_v_dP = (__pyx_v_stop - __pyx_v_P);

      /* "pyFAI/ext/_distortion.pyx":165
 *             P = floor(stop)
 *             dP = stop - P
 *             if dP > 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 */
      __pyx_t_1 = ((__pyx_v_dP > 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":166
 *             dP = stop - P
 *             if dP > 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_P, __pyx_v_stop, __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":167
 *             if dP > 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":168
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = fabs(dP)
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":169
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":170
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = fabs(dP)             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = fabs(__pyx_v_dP);

          /* "pyFAI/ext/_distortion.pyx":171
 *                     h = 0
 *                     dA = fabs(dP)
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":172
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":173
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":174
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":172
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":175
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_3 = ((int)__pyx_v_P);
            __pyx_t_2 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_3 * __pyx_v_box.strides[0]) )) + __pyx_t_2)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":176
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":177
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution
 *         P = floor(start)
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":167
 *             if dP > 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }

        /* "pyFAI/ext/_distortion.pyx":165
 *             P = floor(stop)
 *             dP = stop - P
 *             if dP > 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 */
      }
    }
    __pyx_L4:;

    /* "pyFAI/ext/_distortion.pyx":118
 *         float P, dP, segment_area, abs_area, dA
 *         # , sign
 *     if start < stop:  # positive contribution             # <<<<<<<<<<<<<<
 *         P = ceil(start)
 *         dP = P - start
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":178
 *                         abs_area -= dA
 *                         h += 1
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution             # <<<<<<<<<<<<<<
 *         P = floor(start)
 *         if stop > P:  # start and stop are in the same unit
 */
  __pyx_t_1 = ((__pyx_v_start > __pyx_v_stop) != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":179
 *                         h += 1
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution
 *         P = floor(start)             # <<<<<<<<<<<<<<
 *         if stop > P:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 */
    __pyx_v_P = floor(__pyx_v_start);

    /* "pyFAI/ext/_distortion.pyx":180
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution
 *         P = floor(start)
 *         if stop > P:  # start and stop are in the same unit             # <<<<<<<<<<<<<<
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0:
 */
    __pyx_t_1 = ((__pyx_v_stop > __pyx_v_P) != 0);
    if (__pyx_t_1) {

      /* "pyFAI/ext/_distortion.pyx":181
 *         P = floor(start)
 *         if stop > P:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)             # <<<<<<<<<<<<<<
 *             if segment_area != 0:
 *                 abs_area = fabs(segment_area)
 */
      __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_start, __pyx_v_stop, __pyx_v_slope, __pyx_v_intercept, 0);

      /* "pyFAI/ext/_distortion.pyx":182
 *         if stop > P:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0:             # <<<<<<<<<<<<<<
 *                 abs_area = fabs(segment_area)
 *                 # sign = segment_area / abs_area
 */
      __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":183
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0:
 *                 abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                 # sign = segment_area / abs_area
 *                 dA = (start - stop)  # always positive
 */
        __pyx_v_abs_area = fabs(__pyx_v_segment_area);

        /* "pyFAI/ext/_distortion.pyx":185
 *                 abs_area = fabs(segment_area)
 *                 # sign = segment_area / abs_area
 *                 dA = (start - stop)  # always positive             # <<<<<<<<<<<<<<
 *                 h = 0
 *                 while abs_area > 0:
 */
        __pyx_v_dA = (__pyx_v_start - __pyx_v_stop);

        /* "pyFAI/ext/_distortion.pyx":186
 *                 # sign = segment_area / abs_area
 *                 dA = (start - stop)  # always positive
 *                 h = 0             # <<<<<<<<<<<<<<
 *                 while abs_area > 0:
 *                     if dA > abs_area:
 */
        __pyx_v_h = 0;

        /* "pyFAI/ext/_distortion.pyx":187
 *                 dA = (start - stop)  # always positive
 *                 h = 0
 *                 while abs_area > 0:             # <<<<<<<<<<<<<<
 *                     if dA > abs_area:
 *                         dA = abs_area
 */
        while (1) {
          __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
          if (!__pyx_t_1) break;

          /* "pyFAI/ext/_distortion.pyx":188
 *                 h = 0
 *                 while abs_area > 0:
 *                     if dA > abs_area:             # <<<<<<<<<<<<<<
 *                         dA = abs_area
 *                         abs_area = -1
 */
          __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
          if (__pyx_t_1) {

            /* "pyFAI/ext/_distortion.pyx":189
 *                 while abs_area > 0:
 *                     if dA > abs_area:
 *                         dA = abs_area             # <<<<<<<<<<<<<<
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 */
            __pyx_v_dA = __pyx_v_abs_area;

            /* "pyFAI/ext/_distortion.pyx":190
 *                     if dA > abs_area:
 *                         dA = abs_area
 *                         abs_area = -1             # <<<<<<<<<<<<<<
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA
 */
            __pyx_v_abs_area = -1.0;

            /* "pyFAI/ext/_distortion.pyx":188
 *                 h = 0
 *                 while abs_area > 0:
 *                     if dA > abs_area:             # <<<<<<<<<<<<<<
 *                         dA = abs_area
 *                         abs_area = -1
 */
          }

          /* "pyFAI/ext/_distortion.pyx":191
 *                         dA = abs_area
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                     abs_area -= dA
 *                     h += 1
 */
          __pyx_t_2 = ((int)__pyx_v_start);
          __pyx_t_3 = __pyx_v_h;
          *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_2 * __pyx_v_box.strides[0]) )) + __pyx_t_3)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":192
 *                         abs_area = -1
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA             # <<<<<<<<<<<<<<
 *                     h += 1
 *         else:
 */
          __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

          /* "pyFAI/ext/_distortion.pyx":193
 *                     box[(<int> start), h] += copysign(dA, segment_area)
 *                     abs_area -= dA
 *                     h += 1             # <<<<<<<<<<<<<<
 *         else:
 *             dP = P - start
 */
          __pyx_v_h = (__pyx_v_h + 1);
        }

        /* "pyFAI/ext/_distortion.pyx":182
 *         if stop > P:  # start and stop are in the same unit
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0:             # <<<<<<<<<<<<<<
 *                 abs_area = fabs(segment_area)
 *                 # sign = segment_area / abs_area
 */
      }

      /* "pyFAI/ext/_distortion.pyx":180
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution
 *         P = floor(start)
 *         if stop > P:  # start and stop are in the same unit             # <<<<<<<<<<<<<<
 *             segment_area = calc_area(start, stop, slope, intercept)
 *             if segment_area != 0:
 */
      goto __pyx_L25;
    }

    /* "pyFAI/ext/_distortion.pyx":195
 *                     h += 1
 *         else:
 *             dP = P - start             # <<<<<<<<<<<<<<
 *             if dP < 0:
 *                 segment_area = calc_area(start, P, slope, intercept)
 */
    /*else*/ {
      __pyx_v_dP = (__pyx_v_P - __pyx_v_start);

      /* "pyFAI/ext/_distortion.pyx":196
 *         else:
 *             dP = P - start
 *             if dP < 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0:
 */
      __pyx_t_1 = ((__pyx_v_dP < 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":197
 *             dP = P - start
 *             if dP < 0:
 *                 segment_area = calc_area(start, P, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_start, __pyx_v_P, __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":198
 *             if dP < 0:
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":199
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = fabs(dP)
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":200
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":201
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = fabs(dP)             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = fabs(__pyx_v_dP);

          /* "pyFAI/ext/_distortion.pyx":202
 *                     h = 0
 *                     dA = fabs(dP)
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":203
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":204
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":205
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":203
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":206
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_3 = ((int)__pyx_v_P);
            __pyx_t_2 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_3 * __pyx_v_box.strides[0]) )) + __pyx_t_2)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":207
 *                             abs_area = -1
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 *             # subsection P1->Pn
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":208
 *                         box[(<int> P), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 *             # subsection P1->Pn
 *             for i in range((<int> start), (<int> ceil(stop)), -1):
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":198
 *             if dP < 0:
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }

        /* "pyFAI/ext/_distortion.pyx":196
 *         else:
 *             dP = P - start
 *             if dP < 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(start, P, slope, intercept)
 *                 if segment_area != 0:
 */
      }

      /* "pyFAI/ext/_distortion.pyx":210
 *                         h += 1
 *             # subsection P1->Pn
 *             for i in range((<int> start), (<int> ceil(stop)), -1):             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(i, i - 1, slope, intercept)
 *                 if segment_area != 0:
 */
      __pyx_t_4 = ((int)ceil(__pyx_v_stop));
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = ((int)__pyx_v_start); __pyx_t_6 > __pyx_t_5; __pyx_t_6-=1) {
        __pyx_v_i = __pyx_t_6;

        /* "pyFAI/ext/_distortion.pyx":211
 *             # subsection P1->Pn
 *             for i in range((<int> start), (<int> ceil(stop)), -1):
 *                 segment_area = calc_area(i, i - 1, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_i, (__pyx_v_i - 1), __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":212
 *             for i in range((<int> start), (<int> ceil(stop)), -1):
 *                 segment_area = calc_area(i, i - 1, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":213
 *                 segment_area = calc_area(i, i - 1, slope, intercept)
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = 1
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":214
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = 1
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":215
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = 1             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = 1.0;

          /* "pyFAI/ext/_distortion.pyx":216
 *                     h = 0
 *                     dA = 1
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":217
 *                     dA = 1
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":218
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[i - 1, h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":219
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[i - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":217
 *                     dA = 1
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":220
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[i - 1, h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_2 = (__pyx_v_i - 1);
            __pyx_t_3 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_2 * __pyx_v_box.strides[0]) )) + __pyx_t_3)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":221
 *                             abs_area = -1
 *                         box[i - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 *             # Section Pn->B
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":222
 *                         box[i - 1, h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 *             # Section Pn->B
 *             P = ceil(stop)
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":212
 *             for i in range((<int> start), (<int> ceil(stop)), -1):
 *                 segment_area = calc_area(i, i - 1, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }
      }

      /* "pyFAI/ext/_distortion.pyx":224
 *                         h += 1
 *             # Section Pn->B
 *             P = ceil(stop)             # <<<<<<<<<<<<<<
 *             dP = stop - P
 *             if dP < 0:
 */
      __pyx_v_P = ceil(__pyx_v_stop);

      /* "pyFAI/ext/_distortion.pyx":225
 *             # Section Pn->B
 *             P = ceil(stop)
 *             dP = stop - P             # <<<<<<<<<<<<<<
 *             if dP < 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 */
      __pyx_v_dP = (__pyx_v_stop - __pyx_v_P);

      /* "pyFAI/ext/_distortion.pyx":226
 *             P = ceil(stop)
 *             dP = stop - P
 *             if dP < 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 */
      __pyx_t_1 = ((__pyx_v_dP < 0.0) != 0);
      if (__pyx_t_1) {

        /* "pyFAI/ext/_distortion.pyx":227
 *             dP = stop - P
 *             if dP < 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)             # <<<<<<<<<<<<<<
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 */
        __pyx_v_segment_area = __pyx_f_5pyFAI_3ext_11_distortion_calc_area(__pyx_v_P, __pyx_v_stop, __pyx_v_slope, __pyx_v_intercept, 0);

        /* "pyFAI/ext/_distortion.pyx":228
 *             if dP < 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        __pyx_t_1 = ((__pyx_v_segment_area != 0.0) != 0);
        if (__pyx_t_1) {

          /* "pyFAI/ext/_distortion.pyx":229
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)             # <<<<<<<<<<<<<<
 *                     h = 0
 *                     dA = fabs(dP)
 */
          __pyx_v_abs_area = fabs(__pyx_v_segment_area);

          /* "pyFAI/ext/_distortion.pyx":230
 *                 if segment_area != 0:
 *                     abs_area = fabs(segment_area)
 *                     h = 0             # <<<<<<<<<<<<<<
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 */
          __pyx_v_h = 0;

          /* "pyFAI/ext/_distortion.pyx":231
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 *                     dA = fabs(dP)             # <<<<<<<<<<<<<<
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 */
          __pyx_v_dA = fabs(__pyx_v_dP);

          /* "pyFAI/ext/_distortion.pyx":232
 *                     h = 0
 *                     dA = fabs(dP)
 *                     while abs_area > 0:             # <<<<<<<<<<<<<<
 *                         if dA > abs_area:
 *                             dA = abs_area
 */
          while (1) {
            __pyx_t_1 = ((__pyx_v_abs_area > 0.0) != 0);
            if (!__pyx_t_1) break;

            /* "pyFAI/ext/_distortion.pyx":233
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            __pyx_t_1 = ((__pyx_v_dA > __pyx_v_abs_area) != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":234
 *                     while abs_area > 0:
 *                         if dA > abs_area:
 *                             dA = abs_area             # <<<<<<<<<<<<<<
 *                             abs_area = -1
 *                         box[(<int> stop), h] += copysign(dA, segment_area)
 */
              __pyx_v_dA = __pyx_v_abs_area;

              /* "pyFAI/ext/_distortion.pyx":235
 *                         if dA > abs_area:
 *                             dA = abs_area
 *                             abs_area = -1             # <<<<<<<<<<<<<<
 *                         box[(<int> stop), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 */
              __pyx_v_abs_area = -1.0;

              /* "pyFAI/ext/_distortion.pyx":233
 *                     dA = fabs(dP)
 *                     while abs_area > 0:
 *                         if dA > abs_area:             # <<<<<<<<<<<<<<
 *                             dA = abs_area
 *                             abs_area = -1
 */
            }

            /* "pyFAI/ext/_distortion.pyx":236
 *                             dA = abs_area
 *                             abs_area = -1
 *                         box[(<int> stop), h] += copysign(dA, segment_area)             # <<<<<<<<<<<<<<
 *                         abs_area -= dA
 *                         h += 1
 */
            __pyx_t_3 = ((int)__pyx_v_stop);
            __pyx_t_2 = __pyx_v_h;
            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_3 * __pyx_v_box.strides[0]) )) + __pyx_t_2)) )) += copysign(__pyx_v_dA, __pyx_v_segment_area);

            /* "pyFAI/ext/_distortion.pyx":237
 *                             abs_area = -1
 *                         box[(<int> stop), h] += copysign(dA, segment_area)
 *                         abs_area -= dA             # <<<<<<<<<<<<<<
 *                         h += 1
 * 
 */
            __pyx_v_abs_area = (__pyx_v_abs_area - __pyx_v_dA);

            /* "pyFAI/ext/_distortion.pyx":238
 *                         box[(<int> stop), h] += copysign(dA, segment_area)
 *                         abs_area -= dA
 *                         h += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_v_h = (__pyx_v_h + 1);
          }

          /* "pyFAI/ext/_distortion.pyx":228
 *             if dP < 0:
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:             # <<<<<<<<<<<<<<
 *                     abs_area = fabs(segment_area)
 *                     h = 0
 */
        }

        /* "pyFAI/ext/_distortion.pyx":226
 *             P = ceil(stop)
 *             dP = stop - P
 *             if dP < 0:             # <<<<<<<<<<<<<<
 *                 segment_area = calc_area(P, stop, slope, intercept)
 *                 if segment_area != 0:
 */
      }
    }
    __pyx_L25:;

    /* "pyFAI/ext/_distortion.pyx":178
 *                         abs_area -= dA
 *                         h += 1
 *     elif start > stop:  # negative contribution. Nota if start==stop: no contribution             # <<<<<<<<<<<<<<
 *         P = floor(start)
 *         if stop > P:  # start and stop are in the same unit
 */
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":109
 * 
 * 
 * cdef inline void integrate(float32_t[:, ::1] box, float start, float stop, float slope, float intercept) nogil:             # <<<<<<<<<<<<<<
 *     """Integrate in a box a line between start and stop, line defined by its slope & intercept
 * 
 */

  /* function exit code */
}

/* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_5calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_4calc_pos[] = "calc_pos(signatures, args, kwargs, defaults)\nCalculate the pixel boundary position on the regular grid\n\n    :param pixel_corners: pixel corner coordinate as detector.get_pixel_corner()\n    :param shape: requested output shape. If None, it is calculated\n    :param pixel1, pixel2: pixel size along row and column coordinates\n    :return: pos, delta1, delta2, shape_out, offset\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_5calc_pos = {"calc_pos", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_5calc_pos, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_4calc_pos};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_5calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = __pyx_k_;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 2); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 245, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_signatures = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_defaults = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_4calc_pos(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_4calc_pos(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_pos", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 0);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_pixel_corners, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_pixel_corners); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 245, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 4) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 245, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 4) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(__pyx_v_signatures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__2);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__3);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 245, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_signatures == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_58__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_shape_out);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_shape_out);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_shape_out);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_43calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_5pyFAI_3ext_11_distortion_43calc_pos = {"__pyx_fuse_0calc_pos", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_43calc_pos, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_4calc_pos};
static PyObject *__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_43calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pixel_corners = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_v_pixel1;
  float __pyx_v_pixel2;
  PyObject *__pyx_v_shape_out = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_pos (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pixel_corners,&__pyx_n_s_pixel1,&__pyx_n_s_pixel2,&__pyx_n_s_shape_out,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults2 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_shape_out;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel_corners)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, 1); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, 2); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_pos") < 0)) __PYX_ERR(0, 245, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pixel_corners = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pixel_corners.memview)) __PYX_ERR(0, 245, __pyx_L3_error)
    __pyx_v_pixel1 = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_pixel1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    __pyx_v_pixel2 = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_pixel2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    __pyx_v_shape_out = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pixel_corners.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pixel_corners"); __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_42calc_pos(__pyx_self, __pyx_v_pixel_corners, __pyx_v_pixel1, __pyx_v_pixel2, __pyx_v_shape_out);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_42calc_pos(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pixel_corners, float __pyx_v_pixel1, float __pyx_v_pixel2, PyObject *__pyx_v_shape_out) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_dim0;
  int __pyx_v_dim1;
  int __pyx_v_nb_corners;
  int __pyx_v_do_shape;
  float __pyx_v_BIG;
  float __pyx_v_min0;
  float __pyx_v_min1;
  float __pyx_v_max0;
  float __pyx_v_max1;
  float __pyx_v_delta0;
  float __pyx_v_delta1;
  float __pyx_v_all_min0;
  float __pyx_v_all_max0;
  float __pyx_v_all_max1;
  float __pyx_v_all_min1;
  float __pyx_v_p0;
  float __pyx_v_p1;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  double __pyx_t_23;
  double __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0calc_pos", 0);

  /* "pyFAI/ext/_distortion.pyx":257
 *         float32_t[:, :, :, ::1] pos
 *         int i, j, k, dim0, dim1, nb_corners
 *         bint do_shape = (shape_out is None)             # <<<<<<<<<<<<<<
 *         float BIG = numpy.finfo(numpy.float32).max
 *         float min0, min1, max0, max1, delta0, delta1
 */
  __pyx_t_1 = (__pyx_v_shape_out == Py_None);
  __pyx_v_do_shape = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":258
 *         int i, j, k, dim0, dim1, nb_corners
 *         bint do_shape = (shape_out is None)
 *         float BIG = numpy.finfo(numpy.float32).max             # <<<<<<<<<<<<<<
 *         float min0, min1, max0, max1, delta0, delta1
 *         float all_min0, all_max0, all_max1, all_min1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_finfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_BIG = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":263
 *         float p0, p1
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 */
  __pyx_t_7 = ((__pyx_v_pixel1 == 0.0) != 0);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_pixel2 == 0.0) != 0);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyFAI/ext/_distortion.pyx":264
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")             # <<<<<<<<<<<<<<
 * 
 *     dim0 = pixel_corners.shape[0]
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 264, __pyx_L1_error)

    /* "pyFAI/ext/_distortion.pyx":263
 *         float p0, p1
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":266
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 *     dim0 = pixel_corners.shape[0]             # <<<<<<<<<<<<<<
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]
 */
  __pyx_v_dim0 = (__pyx_v_pixel_corners.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":267
 * 
 *     dim0 = pixel_corners.shape[0]
 *     dim1 = pixel_corners.shape[1]             # <<<<<<<<<<<<<<
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 */
  __pyx_v_dim1 = (__pyx_v_pixel_corners.shape[1]);

  /* "pyFAI/ext/_distortion.pyx":268
 *     dim0 = pixel_corners.shape[0]
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]             # <<<<<<<<<<<<<<
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:
 */
  __pyx_v_nb_corners = (__pyx_v_pixel_corners.shape[2]);

  /* "pyFAI/ext/_distortion.pyx":269
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     with nogil:
 *         delta0 = -BIG
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_dim0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_dim1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_int_2);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_pos = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":270
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         delta0 = -BIG
 *         delta1 = -BIG
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":271
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:
 *         delta0 = -BIG             # <<<<<<<<<<<<<<
 *         delta1 = -BIG
 *         all_min0 = BIG
 */
        __pyx_v_delta0 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":272
 *     with nogil:
 *         delta0 = -BIG
 *         delta1 = -BIG             # <<<<<<<<<<<<<<
 *         all_min0 = BIG
 *         all_min1 = BIG
 */
        __pyx_v_delta1 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":273
 *         delta0 = -BIG
 *         delta1 = -BIG
 *         all_min0 = BIG             # <<<<<<<<<<<<<<
 *         all_min1 = BIG
 *         all_max0 = -BIG
 */
        __pyx_v_all_min0 = __pyx_v_BIG;

        /* "pyFAI/ext/_distortion.pyx":274
 *         delta1 = -BIG
 *         all_min0 = BIG
 *         all_min1 = BIG             # <<<<<<<<<<<<<<
 *         all_max0 = -BIG
 *         all_max1 = -BIG
 */
        __pyx_v_all_min1 = __pyx_v_BIG;

        /* "pyFAI/ext/_distortion.pyx":275
 *         all_min0 = BIG
 *         all_min1 = BIG
 *         all_max0 = -BIG             # <<<<<<<<<<<<<<
 *         all_max1 = -BIG
 *         for i in range(dim0):
 */
        __pyx_v_all_max0 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":276
 *         all_min1 = BIG
 *         all_max0 = -BIG
 *         all_max1 = -BIG             # <<<<<<<<<<<<<<
 *         for i in range(dim0):
 *             for j in range(dim1):
 */
        __pyx_v_all_max1 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":277
 *         all_max0 = -BIG
 *         all_max1 = -BIG
 *         for i in range(dim0):             # <<<<<<<<<<<<<<
 *             for j in range(dim1):
 *                 min0 = BIG
 */
        __pyx_t_10 = __pyx_v_dim0;
        __pyx_t_11 = __pyx_t_10;
        for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
          __pyx_v_i = __pyx_t_12;

          /* "pyFAI/ext/_distortion.pyx":278
 *         all_max1 = -BIG
 *         for i in range(dim0):
 *             for j in range(dim1):             # <<<<<<<<<<<<<<
 *                 min0 = BIG
 *                 min1 = BIG
 */
          __pyx_t_13 = __pyx_v_dim1;
          __pyx_t_14 = __pyx_t_13;
          for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
            __pyx_v_j = __pyx_t_15;

            /* "pyFAI/ext/_distortion.pyx":279
 *         for i in range(dim0):
 *             for j in range(dim1):
 *                 min0 = BIG             # <<<<<<<<<<<<<<
 *                 min1 = BIG
 *                 max0 = -BIG
 */
            __pyx_v_min0 = __pyx_v_BIG;

            /* "pyFAI/ext/_distortion.pyx":280
 *             for j in range(dim1):
 *                 min0 = BIG
 *                 min1 = BIG             # <<<<<<<<<<<<<<
 *                 max0 = -BIG
 *                 max1 = -BIG
 */
            __pyx_v_min1 = __pyx_v_BIG;

            /* "pyFAI/ext/_distortion.pyx":281
 *                 min0 = BIG
 *                 min1 = BIG
 *                 max0 = -BIG             # <<<<<<<<<<<<<<
 *                 max1 = -BIG
 *                 for k in range(nb_corners):
 */
            __pyx_v_max0 = (-__pyx_v_BIG);

            /* "pyFAI/ext/_distortion.pyx":282
 *                 min1 = BIG
 *                 max0 = -BIG
 *                 max1 = -BIG             # <<<<<<<<<<<<<<
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 */
            __pyx_v_max1 = (-__pyx_v_BIG);

            /* "pyFAI/ext/_distortion.pyx":283
 *                 max0 = -BIG
 *                 max1 = -BIG
 *                 for k in range(nb_corners):             # <<<<<<<<<<<<<<
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 */
            __pyx_t_16 = __pyx_v_nb_corners;
            __pyx_t_17 = __pyx_t_16;
            for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
              __pyx_v_k = __pyx_t_18;

              /* "pyFAI/ext/_distortion.pyx":284
 *                 max1 = -BIG
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1             # <<<<<<<<<<<<<<
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0
 */
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_20 = __pyx_v_j;
              __pyx_t_21 = __pyx_v_k;
              __pyx_t_22 = 1;
              __pyx_v_p0 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pixel_corners.data + __pyx_t_19 * __pyx_v_pixel_corners.strides[0]) ) + __pyx_t_20 * __pyx_v_pixel_corners.strides[1]) ) + __pyx_t_21 * __pyx_v_pixel_corners.strides[2]) )) + __pyx_t_22)) ))) / __pyx_v_pixel1);

              /* "pyFAI/ext/_distortion.pyx":285
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2             # <<<<<<<<<<<<<<
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1
 */
              __pyx_t_22 = __pyx_v_i;
              __pyx_t_21 = __pyx_v_j;
              __pyx_t_20 = __pyx_v_k;
              __pyx_t_19 = 2;
              __pyx_v_p1 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pixel_corners.data + __pyx_t_22 * __pyx_v_pixel_corners.strides[0]) ) + __pyx_t_21 * __pyx_v_pixel_corners.strides[1]) ) + __pyx_t_20 * __pyx_v_pixel_corners.strides[2]) )) + __pyx_t_19)) ))) / __pyx_v_pixel2);

              /* "pyFAI/ext/_distortion.pyx":286
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0             # <<<<<<<<<<<<<<
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0
 */
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_20 = __pyx_v_j;
              __pyx_t_21 = __pyx_v_k;
              __pyx_t_22 = 0;
              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_19 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_21 * __pyx_v_pos.strides[2]) )) + __pyx_t_22)) )) = __pyx_v_p0;

              /* "pyFAI/ext/_distortion.pyx":287
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1             # <<<<<<<<<<<<<<
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1
 */
              __pyx_t_22 = __pyx_v_i;
              __pyx_t_21 = __pyx_v_j;
              __pyx_t_20 = __pyx_v_k;
              __pyx_t_19 = 1;
              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_21 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_19)) )) = __pyx_v_p1;

              /* "pyFAI/ext/_distortion.pyx":288
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0             # <<<<<<<<<<<<<<
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0
 */
              if (((__pyx_v_p0 < __pyx_v_min0) != 0)) {
                __pyx_t_6 = __pyx_v_p0;
              } else {
                __pyx_t_6 = __pyx_v_min0;
              }
              __pyx_v_min0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":289
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1             # <<<<<<<<<<<<<<
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1
 */
              if (((__pyx_v_p1 < __pyx_v_min1) != 0)) {
                __pyx_t_6 = __pyx_v_p1;
              } else {
                __pyx_t_6 = __pyx_v_min1;
              }
              __pyx_v_min1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":290
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0             # <<<<<<<<<<<<<<
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 */
              if (((__pyx_v_p0 > __pyx_v_max0) != 0)) {
                __pyx_t_6 = __pyx_v_p0;
              } else {
                __pyx_t_6 = __pyx_v_max0;
              }
              __pyx_v_max0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":291
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1             # <<<<<<<<<<<<<<
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 */
              if (((__pyx_v_p1 > __pyx_v_max1) != 0)) {
                __pyx_t_6 = __pyx_v_p1;
              } else {
                __pyx_t_6 = __pyx_v_max1;
              }
              __pyx_v_max1 = __pyx_t_6;
            }

            /* "pyFAI/ext/_distortion.pyx":292
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))             # <<<<<<<<<<<<<<
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:
 */
            __pyx_t_23 = (ceil(__pyx_v_max0) - floor(__pyx_v_min0));
            __pyx_t_6 = __pyx_v_delta0;
            if (((__pyx_t_23 > __pyx_t_6) != 0)) {
              __pyx_t_24 = __pyx_t_23;
            } else {
              __pyx_t_24 = __pyx_t_6;
            }
            __pyx_v_delta0 = __pyx_t_24;

            /* "pyFAI/ext/_distortion.pyx":293
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))             # <<<<<<<<<<<<<<
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 */
            __pyx_t_24 = (ceil(__pyx_v_max1) - floor(__pyx_v_min1));
            __pyx_t_6 = __pyx_v_delta1;
            if (((__pyx_t_24 > __pyx_t_6) != 0)) {
              __pyx_t_23 = __pyx_t_24;
            } else {
              __pyx_t_23 = __pyx_t_6;
            }
            __pyx_v_delta1 = __pyx_t_23;

            /* "pyFAI/ext/_distortion.pyx":294
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:             # <<<<<<<<<<<<<<
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 */
            __pyx_t_1 = (__pyx_v_do_shape != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":295
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0             # <<<<<<<<<<<<<<
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 */
              if (((__pyx_v_min0 < __pyx_v_all_min0) != 0)) {
                __pyx_t_6 = __pyx_v_min0;
              } else {
                __pyx_t_6 = __pyx_v_all_min0;
              }
              __pyx_v_all_min0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":296
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1             # <<<<<<<<<<<<<<
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 */
              if (((__pyx_v_min1 < __pyx_v_all_min1) != 0)) {
                __pyx_t_6 = __pyx_v_min1;
              } else {
                __pyx_t_6 = __pyx_v_all_min1;
              }
              __pyx_v_all_min1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":297
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0             # <<<<<<<<<<<<<<
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 */
              if (((__pyx_v_max0 > __pyx_v_all_max0) != 0)) {
                __pyx_t_6 = __pyx_v_max0;
              } else {
                __pyx_t_6 = __pyx_v_all_max0;
              }
              __pyx_v_all_max0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":298
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 *                     all_max1 = max1 if max1 > all_max1 else all_max1             # <<<<<<<<<<<<<<
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 */
              if (((__pyx_v_max1 > __pyx_v_all_max1) != 0)) {
                __pyx_t_6 = __pyx_v_max1;
              } else {
                __pyx_t_6 = __pyx_v_all_max1;
              }
              __pyx_v_all_max1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":294
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:             # <<<<<<<<<<<<<<
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 */
            }
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":270
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         delta0 = -BIG
 *         delta1 = -BIG
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":300
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \             # <<<<<<<<<<<<<<
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_pos, 4, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromDouble(__pyx_v_delta0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyInt_FromDouble(__pyx_v_delta1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyFAI/ext/_distortion.pyx":301
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \             # <<<<<<<<<<<<<<
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 *     return res
 */
  if ((__pyx_v_do_shape != 0)) {
    __pyx_t_4 = __Pyx_PyInt_FromDouble(ceil((__pyx_v_all_max0 - __pyx_v_all_min0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_25 = __Pyx_PyInt_FromDouble(ceil((__pyx_v_all_max1 - __pyx_v_all_min1))); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_25);
    __pyx_t_4 = 0;
    __pyx_t_25 = 0;
    __pyx_t_2 = __pyx_t_26;
    __pyx_t_26 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_shape_out);
    __pyx_t_2 = __pyx_v_shape_out;
  }

  /* "pyFAI/ext/_distortion.pyx":302
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  if ((__pyx_v_do_shape != 0)) {
    __pyx_t_25 = PyFloat_FromDouble(((double)__pyx_v_all_min0)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_4 = PyFloat_FromDouble(((double)__pyx_v_all_min1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_4);
    __pyx_t_25 = 0;
    __pyx_t_4 = 0;
    __pyx_t_26 = __pyx_t_27;
    __pyx_t_27 = 0;
  } else {
    __Pyx_INCREF(__pyx_tuple__7);
    __pyx_t_26 = __pyx_tuple__7;
  }

  /* "pyFAI/ext/_distortion.pyx":300
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \             # <<<<<<<<<<<<<<
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 */
  __pyx_t_27 = PyTuple_New(5); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_27, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_27, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_27, 4, __pyx_t_26);
  __pyx_t_8 = 0;
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_26 = 0;
  __pyx_v_res = ((PyObject*)__pyx_t_27);
  __pyx_t_27 = 0;

  /* "pyFAI/ext/_distortion.pyx":303
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __Pyx_XDECREF(__pyx_v_res);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pixel_corners, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_60__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_shape_out);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_shape_out);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_shape_out);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_45calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_5pyFAI_3ext_11_distortion_45calc_pos = {"__pyx_fuse_1calc_pos", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_45calc_pos, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_4calc_pos};
static PyObject *__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_45calc_pos(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pixel_corners = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_v_pixel1;
  float __pyx_v_pixel2;
  PyObject *__pyx_v_shape_out = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_pos (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pixel_corners,&__pyx_n_s_pixel1,&__pyx_n_s_pixel2,&__pyx_n_s_shape_out,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults3 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_shape_out;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel_corners)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, 1); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, 2); __PYX_ERR(0, 245, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_pos") < 0)) __PYX_ERR(0, 245, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pixel_corners = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pixel_corners.memview)) __PYX_ERR(0, 245, __pyx_L3_error)
    __pyx_v_pixel1 = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_pixel1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    __pyx_v_pixel2 = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_pixel2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L3_error)
    __pyx_v_shape_out = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_pos", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pixel_corners.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pixel_corners"); __PYX_ERR(0, 245, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_44calc_pos(__pyx_self, __pyx_v_pixel_corners, __pyx_v_pixel1, __pyx_v_pixel2, __pyx_v_shape_out);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_44calc_pos(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pixel_corners, float __pyx_v_pixel1, float __pyx_v_pixel2, PyObject *__pyx_v_shape_out) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_dim0;
  int __pyx_v_dim1;
  int __pyx_v_nb_corners;
  int __pyx_v_do_shape;
  float __pyx_v_BIG;
  float __pyx_v_min0;
  float __pyx_v_min1;
  float __pyx_v_max0;
  float __pyx_v_max1;
  float __pyx_v_delta0;
  float __pyx_v_delta1;
  float __pyx_v_all_min0;
  float __pyx_v_all_max0;
  float __pyx_v_all_max1;
  float __pyx_v_all_min1;
  float __pyx_v_p0;
  float __pyx_v_p1;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  double __pyx_t_23;
  double __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1calc_pos", 0);

  /* "pyFAI/ext/_distortion.pyx":257
 *         float32_t[:, :, :, ::1] pos
 *         int i, j, k, dim0, dim1, nb_corners
 *         bint do_shape = (shape_out is None)             # <<<<<<<<<<<<<<
 *         float BIG = numpy.finfo(numpy.float32).max
 *         float min0, min1, max0, max1, delta0, delta1
 */
  __pyx_t_1 = (__pyx_v_shape_out == Py_None);
  __pyx_v_do_shape = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":258
 *         int i, j, k, dim0, dim1, nb_corners
 *         bint do_shape = (shape_out is None)
 *         float BIG = numpy.finfo(numpy.float32).max             # <<<<<<<<<<<<<<
 *         float min0, min1, max0, max1, delta0, delta1
 *         float all_min0, all_max0, all_max1, all_min1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_finfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_BIG = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":263
 *         float p0, p1
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 */
  __pyx_t_7 = ((__pyx_v_pixel1 == 0.0) != 0);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_pixel2 == 0.0) != 0);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyFAI/ext/_distortion.pyx":264
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")             # <<<<<<<<<<<<<<
 * 
 *     dim0 = pixel_corners.shape[0]
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 264, __pyx_L1_error)

    /* "pyFAI/ext/_distortion.pyx":263
 *         float p0, p1
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":266
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")
 * 
 *     dim0 = pixel_corners.shape[0]             # <<<<<<<<<<<<<<
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]
 */
  __pyx_v_dim0 = (__pyx_v_pixel_corners.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":267
 * 
 *     dim0 = pixel_corners.shape[0]
 *     dim1 = pixel_corners.shape[1]             # <<<<<<<<<<<<<<
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 */
  __pyx_v_dim1 = (__pyx_v_pixel_corners.shape[1]);

  /* "pyFAI/ext/_distortion.pyx":268
 *     dim0 = pixel_corners.shape[0]
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]             # <<<<<<<<<<<<<<
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:
 */
  __pyx_v_nb_corners = (__pyx_v_pixel_corners.shape[2]);

  /* "pyFAI/ext/_distortion.pyx":269
 *     dim1 = pixel_corners.shape[1]
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     with nogil:
 *         delta0 = -BIG
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_dim0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_dim1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_int_2);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_pos = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":270
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         delta0 = -BIG
 *         delta1 = -BIG
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":271
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:
 *         delta0 = -BIG             # <<<<<<<<<<<<<<
 *         delta1 = -BIG
 *         all_min0 = BIG
 */
        __pyx_v_delta0 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":272
 *     with nogil:
 *         delta0 = -BIG
 *         delta1 = -BIG             # <<<<<<<<<<<<<<
 *         all_min0 = BIG
 *         all_min1 = BIG
 */
        __pyx_v_delta1 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":273
 *         delta0 = -BIG
 *         delta1 = -BIG
 *         all_min0 = BIG             # <<<<<<<<<<<<<<
 *         all_min1 = BIG
 *         all_max0 = -BIG
 */
        __pyx_v_all_min0 = __pyx_v_BIG;

        /* "pyFAI/ext/_distortion.pyx":274
 *         delta1 = -BIG
 *         all_min0 = BIG
 *         all_min1 = BIG             # <<<<<<<<<<<<<<
 *         all_max0 = -BIG
 *         all_max1 = -BIG
 */
        __pyx_v_all_min1 = __pyx_v_BIG;

        /* "pyFAI/ext/_distortion.pyx":275
 *         all_min0 = BIG
 *         all_min1 = BIG
 *         all_max0 = -BIG             # <<<<<<<<<<<<<<
 *         all_max1 = -BIG
 *         for i in range(dim0):
 */
        __pyx_v_all_max0 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":276
 *         all_min1 = BIG
 *         all_max0 = -BIG
 *         all_max1 = -BIG             # <<<<<<<<<<<<<<
 *         for i in range(dim0):
 *             for j in range(dim1):
 */
        __pyx_v_all_max1 = (-__pyx_v_BIG);

        /* "pyFAI/ext/_distortion.pyx":277
 *         all_max0 = -BIG
 *         all_max1 = -BIG
 *         for i in range(dim0):             # <<<<<<<<<<<<<<
 *             for j in range(dim1):
 *                 min0 = BIG
 */
        __pyx_t_10 = __pyx_v_dim0;
        __pyx_t_11 = __pyx_t_10;
        for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
          __pyx_v_i = __pyx_t_12;

          /* "pyFAI/ext/_distortion.pyx":278
 *         all_max1 = -BIG
 *         for i in range(dim0):
 *             for j in range(dim1):             # <<<<<<<<<<<<<<
 *                 min0 = BIG
 *                 min1 = BIG
 */
          __pyx_t_13 = __pyx_v_dim1;
          __pyx_t_14 = __pyx_t_13;
          for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
            __pyx_v_j = __pyx_t_15;

            /* "pyFAI/ext/_distortion.pyx":279
 *         for i in range(dim0):
 *             for j in range(dim1):
 *                 min0 = BIG             # <<<<<<<<<<<<<<
 *                 min1 = BIG
 *                 max0 = -BIG
 */
            __pyx_v_min0 = __pyx_v_BIG;

            /* "pyFAI/ext/_distortion.pyx":280
 *             for j in range(dim1):
 *                 min0 = BIG
 *                 min1 = BIG             # <<<<<<<<<<<<<<
 *                 max0 = -BIG
 *                 max1 = -BIG
 */
            __pyx_v_min1 = __pyx_v_BIG;

            /* "pyFAI/ext/_distortion.pyx":281
 *                 min0 = BIG
 *                 min1 = BIG
 *                 max0 = -BIG             # <<<<<<<<<<<<<<
 *                 max1 = -BIG
 *                 for k in range(nb_corners):
 */
            __pyx_v_max0 = (-__pyx_v_BIG);

            /* "pyFAI/ext/_distortion.pyx":282
 *                 min1 = BIG
 *                 max0 = -BIG
 *                 max1 = -BIG             # <<<<<<<<<<<<<<
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 */
            __pyx_v_max1 = (-__pyx_v_BIG);

            /* "pyFAI/ext/_distortion.pyx":283
 *                 max0 = -BIG
 *                 max1 = -BIG
 *                 for k in range(nb_corners):             # <<<<<<<<<<<<<<
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 */
            __pyx_t_16 = __pyx_v_nb_corners;
            __pyx_t_17 = __pyx_t_16;
            for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
              __pyx_v_k = __pyx_t_18;

              /* "pyFAI/ext/_distortion.pyx":284
 *                 max1 = -BIG
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1             # <<<<<<<<<<<<<<
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0
 */
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_20 = __pyx_v_j;
              __pyx_t_21 = __pyx_v_k;
              __pyx_t_22 = 1;
              __pyx_v_p0 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pixel_corners.data + __pyx_t_19 * __pyx_v_pixel_corners.strides[0]) ) + __pyx_t_20 * __pyx_v_pixel_corners.strides[1]) ) + __pyx_t_21 * __pyx_v_pixel_corners.strides[2]) )) + __pyx_t_22)) ))) / ((double)__pyx_v_pixel1));

              /* "pyFAI/ext/_distortion.pyx":285
 *                 for k in range(nb_corners):
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2             # <<<<<<<<<<<<<<
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1
 */
              __pyx_t_22 = __pyx_v_i;
              __pyx_t_21 = __pyx_v_j;
              __pyx_t_20 = __pyx_v_k;
              __pyx_t_19 = 2;
              __pyx_v_p1 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pixel_corners.data + __pyx_t_22 * __pyx_v_pixel_corners.strides[0]) ) + __pyx_t_21 * __pyx_v_pixel_corners.strides[1]) ) + __pyx_t_20 * __pyx_v_pixel_corners.strides[2]) )) + __pyx_t_19)) ))) / ((double)__pyx_v_pixel2));

              /* "pyFAI/ext/_distortion.pyx":286
 *                     p0 = pixel_corners[i, j, k, 1] / pixel1
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0             # <<<<<<<<<<<<<<
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0
 */
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_20 = __pyx_v_j;
              __pyx_t_21 = __pyx_v_k;
              __pyx_t_22 = 0;
              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_19 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_21 * __pyx_v_pos.strides[2]) )) + __pyx_t_22)) )) = __pyx_v_p0;

              /* "pyFAI/ext/_distortion.pyx":287
 *                     p1 = pixel_corners[i, j, k, 2] / pixel2
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1             # <<<<<<<<<<<<<<
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1
 */
              __pyx_t_22 = __pyx_v_i;
              __pyx_t_21 = __pyx_v_j;
              __pyx_t_20 = __pyx_v_k;
              __pyx_t_19 = 1;
              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_21 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_19)) )) = __pyx_v_p1;

              /* "pyFAI/ext/_distortion.pyx":288
 *                     pos[i, j, k, 0] = p0
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0             # <<<<<<<<<<<<<<
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0
 */
              if (((__pyx_v_p0 < __pyx_v_min0) != 0)) {
                __pyx_t_6 = __pyx_v_p0;
              } else {
                __pyx_t_6 = __pyx_v_min0;
              }
              __pyx_v_min0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":289
 *                     pos[i, j, k, 1] = p1
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1             # <<<<<<<<<<<<<<
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1
 */
              if (((__pyx_v_p1 < __pyx_v_min1) != 0)) {
                __pyx_t_6 = __pyx_v_p1;
              } else {
                __pyx_t_6 = __pyx_v_min1;
              }
              __pyx_v_min1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":290
 *                     min0 = p0 if p0 < min0 else min0
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0             # <<<<<<<<<<<<<<
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 */
              if (((__pyx_v_p0 > __pyx_v_max0) != 0)) {
                __pyx_t_6 = __pyx_v_p0;
              } else {
                __pyx_t_6 = __pyx_v_max0;
              }
              __pyx_v_max0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":291
 *                     min1 = p1 if p1 < min1 else min1
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1             # <<<<<<<<<<<<<<
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 */
              if (((__pyx_v_p1 > __pyx_v_max1) != 0)) {
                __pyx_t_6 = __pyx_v_p1;
              } else {
                __pyx_t_6 = __pyx_v_max1;
              }
              __pyx_v_max1 = __pyx_t_6;
            }

            /* "pyFAI/ext/_distortion.pyx":292
 *                     max0 = p0 if p0 > max0 else max0
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))             # <<<<<<<<<<<<<<
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:
 */
            __pyx_t_23 = (ceil(__pyx_v_max0) - floor(__pyx_v_min0));
            __pyx_t_6 = __pyx_v_delta0;
            if (((__pyx_t_23 > __pyx_t_6) != 0)) {
              __pyx_t_24 = __pyx_t_23;
            } else {
              __pyx_t_24 = __pyx_t_6;
            }
            __pyx_v_delta0 = __pyx_t_24;

            /* "pyFAI/ext/_distortion.pyx":293
 *                     max1 = p1 if p1 > max1 else max1
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))             # <<<<<<<<<<<<<<
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 */
            __pyx_t_24 = (ceil(__pyx_v_max1) - floor(__pyx_v_min1));
            __pyx_t_6 = __pyx_v_delta1;
            if (((__pyx_t_24 > __pyx_t_6) != 0)) {
              __pyx_t_23 = __pyx_t_24;
            } else {
              __pyx_t_23 = __pyx_t_6;
            }
            __pyx_v_delta1 = __pyx_t_23;

            /* "pyFAI/ext/_distortion.pyx":294
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:             # <<<<<<<<<<<<<<
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 */
            __pyx_t_1 = (__pyx_v_do_shape != 0);
            if (__pyx_t_1) {

              /* "pyFAI/ext/_distortion.pyx":295
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0             # <<<<<<<<<<<<<<
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 */
              if (((__pyx_v_min0 < __pyx_v_all_min0) != 0)) {
                __pyx_t_6 = __pyx_v_min0;
              } else {
                __pyx_t_6 = __pyx_v_all_min0;
              }
              __pyx_v_all_min0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":296
 *                 if do_shape:
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1             # <<<<<<<<<<<<<<
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 */
              if (((__pyx_v_min1 < __pyx_v_all_min1) != 0)) {
                __pyx_t_6 = __pyx_v_min1;
              } else {
                __pyx_t_6 = __pyx_v_all_min1;
              }
              __pyx_v_all_min1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":297
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0             # <<<<<<<<<<<<<<
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 */
              if (((__pyx_v_max0 > __pyx_v_all_max0) != 0)) {
                __pyx_t_6 = __pyx_v_max0;
              } else {
                __pyx_t_6 = __pyx_v_all_max0;
              }
              __pyx_v_all_max0 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":298
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 *                     all_max0 = max0 if max0 > all_max0 else all_max0
 *                     all_max1 = max1 if max1 > all_max1 else all_max1             # <<<<<<<<<<<<<<
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 */
              if (((__pyx_v_max1 > __pyx_v_all_max1) != 0)) {
                __pyx_t_6 = __pyx_v_max1;
              } else {
                __pyx_t_6 = __pyx_v_all_max1;
              }
              __pyx_v_all_max1 = __pyx_t_6;

              /* "pyFAI/ext/_distortion.pyx":294
 *                 delta0 = max(delta0, ceil(max0) - floor(min0))
 *                 delta1 = max(delta1, ceil(max1) - floor(min1))
 *                 if do_shape:             # <<<<<<<<<<<<<<
 *                     all_min0 = min0 if min0 < all_min0 else all_min0
 *                     all_min1 = min1 if min1 < all_min1 else all_min1
 */
            }
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":270
 *     nb_corners = pixel_corners.shape[2]
 *     pos = numpy.zeros((dim0, dim1, 4, 2), dtype=numpy.float32)
 *     with nogil:             # <<<<<<<<<<<<<<
 *         delta0 = -BIG
 *         delta1 = -BIG
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":300
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \             # <<<<<<<<<<<<<<
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_pos, 4, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromDouble(__pyx_v_delta0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyInt_FromDouble(__pyx_v_delta1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyFAI/ext/_distortion.pyx":301
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \             # <<<<<<<<<<<<<<
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 *     return res
 */
  if ((__pyx_v_do_shape != 0)) {
    __pyx_t_4 = __Pyx_PyInt_FromDouble(ceil((__pyx_v_all_max0 - __pyx_v_all_min0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_25 = __Pyx_PyInt_FromDouble(ceil((__pyx_v_all_max1 - __pyx_v_all_min1))); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_25);
    __pyx_t_4 = 0;
    __pyx_t_25 = 0;
    __pyx_t_2 = __pyx_t_26;
    __pyx_t_26 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_shape_out);
    __pyx_t_2 = __pyx_v_shape_out;
  }

  /* "pyFAI/ext/_distortion.pyx":302
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  if ((__pyx_v_do_shape != 0)) {
    __pyx_t_25 = PyFloat_FromDouble(((double)__pyx_v_all_min0)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_4 = PyFloat_FromDouble(((double)__pyx_v_all_min1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_4);
    __pyx_t_25 = 0;
    __pyx_t_4 = 0;
    __pyx_t_26 = __pyx_t_27;
    __pyx_t_27 = 0;
  } else {
    __Pyx_INCREF(__pyx_tuple__7);
    __pyx_t_26 = __pyx_tuple__7;
  }

  /* "pyFAI/ext/_distortion.pyx":300
 *                     all_max1 = max1 if max1 > all_max1 else all_max1
 * 
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \             # <<<<<<<<<<<<<<
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 */
  __pyx_t_27 = PyTuple_New(5); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_27, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_27, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_27, 4, __pyx_t_26);
  __pyx_t_8 = 0;
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_26 = 0;
  __pyx_v_res = ((PyObject*)__pyx_t_27);
  __pyx_t_27 = 0;

  /* "pyFAI/ext/_distortion.pyx":303
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __Pyx_XDECREF(__pyx_v_res);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pixel_corners, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_7calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_6calc_size[] = "calc_size(signatures, args, kwargs, defaults)\nCalculate the number of items per output pixel\n\n    :param pos: 4D array with position in space\n    :param shape: shape of the output array\n    :param mask: input data mask\n    :param offset: 2-tuple of float with the minimal index of\n    :return: number of input element per output elements\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_7calc_size = {"calc_size", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_7calc_size, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_6calc_size};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_7calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_signatures = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signatures,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = __pyx_k__8;
    values[3] = __pyx_k__9;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signatures)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, 1); __PYX_ERR(0, 306, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 306, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_signatures = values[0];
    __pyx_v_args = values[1];
    __pyx_v_kwargs = values[2];
    __pyx_v_defaults = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 306, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_6calc_size(__pyx_self, __pyx_v_signatures, __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_6calc_size(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_size", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_2 = ((0 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 306, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 0);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_pos, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 306, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 306, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 4) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 4) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(__pyx_v_signatures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__2);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__3);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 306, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_signatures == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 306, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_66__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_fromslice(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_mask, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int8_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_offset);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_offset);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_offset);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_49calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_5pyFAI_3ext_11_distortion_49calc_size = {"__pyx_fuse_0calc_size", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_49calc_size, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_6calc_size};
static PyObject *__pyx_fuse_0__pyx_pw_5pyFAI_3ext_11_distortion_49calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_offset = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_size (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_mask,&__pyx_n_s_offset,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults6 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_offset;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_size", 0, 2, 4, 1); __PYX_ERR(0, 306, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_size") < 0)) __PYX_ERR(0, 306, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 306, __pyx_L3_error)
    __pyx_v_shape = values[1];
    if (values[2]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 308, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_dynamic_args->__pyx_arg_mask;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
    __pyx_v_offset = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_size", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 306, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_48calc_size(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_mask, __pyx_v_offset);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_48calc_size(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_offset) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_l;
  int __pyx_v_shape_out0;
  int __pyx_v_shape_out1;
  int __pyx_v_shape_in0;
  int __pyx_v_shape_in1;
  int __pyx_v_min0;
  int __pyx_v_min1;
  int __pyx_v_max0;
  int __pyx_v_max1;
  __Pyx_memviewslice __pyx_v_lut_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_offset0;
  float __pyx_v_offset1;
  int __pyx_v_do_mask;
  __Pyx_memviewslice __pyx_v_cmask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_UCS4 __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_t_16;
  float __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0calc_size", 0);

  /* "pyFAI/ext/_distortion.pyx":320
 *     cdef:
 *         int i, j, k, l, shape_out0, shape_out1, shape_in0, shape_in1, min0, min1, max0, max1
 *         int32_t[:, ::1] lut_size = numpy.zeros(shape, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *         float A0, A1, B0, B1, C0, C1, D0, D1, offset0, offset1
 *         bint do_mask = mask is not None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_lut_size = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":322
 *         int32_t[:, ::1] lut_size = numpy.zeros(shape, dtype=numpy.int32)
 *         float A0, A1, B0, B1, C0, C1, D0, D1, offset0, offset1
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":324
 *         bint do_mask = mask is not None
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape
 * 
 */
  __pyx_t_7 = (__pyx_v_pos.shape[0]);
  __pyx_t_8 = (__pyx_v_pos.shape[1]);
  __pyx_v_shape_in0 = __pyx_t_7;
  __pyx_v_shape_in1 = __pyx_t_8;

  /* "pyFAI/ext/_distortion.pyx":325
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]
 *     shape_out0, shape_out1 = shape             # <<<<<<<<<<<<<<
 * 
 *     if do_mask:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 325, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_1); if (unlikely(!__pyx_t_5)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_3 = __pyx_t_9(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_1), 2) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_shape_out0 = __pyx_t_10;
  __pyx_v_shape_out1 = __pyx_t_11;

  /* "pyFAI/ext/_distortion.pyx":327
 *     shape_out0, shape_out1 = shape
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 */
  __pyx_t_12 = (__pyx_v_do_mask != 0);
  if (__pyx_t_12) {

    /* "pyFAI/ext/_distortion.pyx":329
 *     if do_mask:
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):             # <<<<<<<<<<<<<<
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 */
    __pyx_t_13 = (((__pyx_v_mask.shape[0]) != __pyx_v_shape_in0) != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_12 = __pyx_t_13;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_13 = (((__pyx_v_mask.shape[1]) != __pyx_v_shape_in1) != 0);
    __pyx_t_12 = __pyx_t_13;
    __pyx_L7_bool_binop_done:;
    if (unlikely(__pyx_t_12)) {

      /* "pyFAI/ext/_distortion.pyx":330
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])             # <<<<<<<<<<<<<<
 *             logger.error(err)
 *             raise RuntimeError(err)
 */
      __pyx_t_3 = PyTuple_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = 0;
      __pyx_t_14 = 127;
      __Pyx_INCREF(__pyx_kp_u_Mismatch_between_shape_of_detect);
      __pyx_t_8 += 36;
      __Pyx_GIVEREF(__pyx_kp_u_Mismatch_between_shape_of_detect);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Mismatch_between_shape_of_detect);
      __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_shape_in0, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_t_8 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__10);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__10);
      __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_shape_in1, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_and_shape_of_mask);
      __pyx_t_8 += 21;
      __Pyx_GIVEREF(__pyx_kp_u_and_shape_of_mask);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_and_shape_of_mask);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t((__pyx_v_mask.shape[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_t_8 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__10);
      PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u__10);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t((__pyx_v_mask.shape[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__11);
      __pyx_t_8 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__11);
      PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u__11);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 9, __pyx_t_8, __pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_err = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":331
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)             # <<<<<<<<<<<<<<
 *             raise RuntimeError(err)
 *         else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_err) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_err);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":332
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 *             raise RuntimeError(err)             # <<<<<<<<<<<<<<
 *         else:
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 332, __pyx_L1_error)

      /* "pyFAI/ext/_distortion.pyx":329
 *     if do_mask:
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):             # <<<<<<<<<<<<<<
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 */
    }

    /* "pyFAI/ext/_distortion.pyx":334
 *             raise RuntimeError(err)
 *         else:
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)             # <<<<<<<<<<<<<<
 * 
 *     if offset is not None:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_mask, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int8_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_15 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_15.memview)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_cmask = __pyx_t_15;
      __pyx_t_15.memview = NULL;
      __pyx_t_15.data = NULL;
    }

    /* "pyFAI/ext/_distortion.pyx":327
 *     shape_out0, shape_out1 = shape
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 */
  }

  /* "pyFAI/ext/_distortion.pyx":336
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 * 
 *     if offset is not None:             # <<<<<<<<<<<<<<
 *         offset0, offset1 = offset
 * 
 */
  __pyx_t_12 = (__pyx_v_offset != Py_None);
  __pyx_t_13 = (__pyx_t_12 != 0);
  if (__pyx_t_13) {

    /* "pyFAI/ext/_distortion.pyx":337
 * 
 *     if offset is not None:
 *         offset0, offset1 = offset             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_offset))) || (PyList_CheckExact(__pyx_v_offset))) {
      PyObject* sequence = __pyx_v_offset;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 337, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_9(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_9(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_3), 2) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 337, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __pyx_t_16 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_16 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_17 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_17 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_offset0 = __pyx_t_16;
    __pyx_v_offset1 = __pyx_t_17;

    /* "pyFAI/ext/_distortion.pyx":336
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 * 
 *     if offset is not None:             # <<<<<<<<<<<<<<
 *         offset0, offset1 = offset
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":339
 *         offset0, offset1 = offset
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":340
 * 
 *     with nogil:
 *         for i in range(shape_in0):             # <<<<<<<<<<<<<<
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:
 */
        __pyx_t_11 = __pyx_v_shape_in0;
        __pyx_t_10 = __pyx_t_11;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_10; __pyx_t_18+=1) {
          __pyx_v_i = __pyx_t_18;

          /* "pyFAI/ext/_distortion.pyx":341
 *     with nogil:
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):             # <<<<<<<<<<<<<<
 *                 if do_mask and cmask[i, j]:
 *                     continue
 */
          __pyx_t_19 = __pyx_v_shape_in1;
          __pyx_t_20 = __pyx_t_19;
          for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
            __pyx_v_j = __pyx_t_21;

            /* "pyFAI/ext/_distortion.pyx":342
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 */
            __pyx_t_12 = (__pyx_v_do_mask != 0);
            if (__pyx_t_12) {
            } else {
              __pyx_t_13 = __pyx_t_12;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_12 = ((*((int8_t *) ( /* dim=1 */ ((char *) (((int8_t *) ( /* dim=0 */ (__pyx_v_cmask.data + __pyx_t_22 * __pyx_v_cmask.strides[0]) )) + __pyx_t_23)) ))) != 0);
            __pyx_t_13 = __pyx_t_12;
            __pyx_L20_bool_binop_done:;
            if (__pyx_t_13) {

              /* "pyFAI/ext/_distortion.pyx":343
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:
 *                     continue             # <<<<<<<<<<<<<<
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1
 */
              goto __pyx_L17_continue;

              /* "pyFAI/ext/_distortion.pyx":342
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 */
            }

            /* "pyFAI/ext/_distortion.pyx":344
 *                 if do_mask and cmask[i, j]:
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0             # <<<<<<<<<<<<<<
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 0;
            __pyx_t_25 = 0;
            __pyx_v_A0 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":345
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1             # <<<<<<<<<<<<<<
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 0;
            __pyx_t_23 = 1;
            __pyx_v_A1 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":346
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0             # <<<<<<<<<<<<<<
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 1;
            __pyx_t_25 = 0;
            __pyx_v_B0 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":347
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1             # <<<<<<<<<<<<<<
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 1;
            __pyx_t_23 = 1;
            __pyx_v_B1 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":348
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0             # <<<<<<<<<<<<<<
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 2;
            __pyx_t_25 = 0;
            __pyx_v_C0 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":349
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1             # <<<<<<<<<<<<<<
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 2;
            __pyx_t_23 = 1;
            __pyx_v_C1 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":350
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0             # <<<<<<<<<<<<<<
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 3;
            __pyx_t_25 = 0;
            __pyx_v_D0 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":351
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1             # <<<<<<<<<<<<<<
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 3;
            __pyx_t_23 = 1;
            __pyx_v_D1 = ((*((float *) ( /* dim=3 */ ((char *) (((float *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":352
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)             # <<<<<<<<<<<<<<
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 */
            __pyx_v_min0 = __pyx_f_5pyFAI_3ext_11_distortion_clip(((int)__pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)), 0, __pyx_v_shape_out0, 0);

            /* "pyFAI/ext/_distortion.pyx":353
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)             # <<<<<<<<<<<<<<
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 */
            __pyx_v_min1 = __pyx_f_5pyFAI_3ext_11_distortion_clip(((int)__pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)), 0, __pyx_v_shape_out1, 0);

            /* "pyFAI/ext/_distortion.pyx":354
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)             # <<<<<<<<<<<<<<
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):
 */
            __pyx_v_max0 = __pyx_f_5pyFAI_3ext_11_distortion_clip((((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) + 1), 0, __pyx_v_shape_out0, 0);

            /* "pyFAI/ext/_distortion.pyx":355
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)             # <<<<<<<<<<<<<<
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):
 */
            __pyx_v_max1 = __pyx_f_5pyFAI_3ext_11_distortion_clip((((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) + 1), 0, __pyx_v_shape_out1, 0);

            /* "pyFAI/ext/_distortion.pyx":356
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):             # <<<<<<<<<<<<<<
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1
 */
            __pyx_t_26 = __pyx_v_max0;
            __pyx_t_27 = __pyx_t_26;
            for (__pyx_t_28 = __pyx_v_min0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
              __pyx_v_k = __pyx_t_28;

              /* "pyFAI/ext/_distortion.pyx":357
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):             # <<<<<<<<<<<<<<
 *                         lut_size[k, l] += 1
 *     return numpy.asarray(lut_size)
 */
              __pyx_t_29 = __pyx_v_max1;
              __pyx_t_30 = __pyx_t_29;
              for (__pyx_t_31 = __pyx_v_min1; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
                __pyx_v_l = __pyx_t_31;

                /* "pyFAI/ext/_distortion.pyx":358
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1             # <<<<<<<<<<<<<<
 *     return numpy.asarray(lut_size)
 * 
 */
                __pyx_t_23 = __pyx_v_k;
                __pyx_t_22 = __pyx_v_l;
                *((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_lut_size.data + __pyx_t_23 * __pyx_v_lut_size.strides[0]) )) + __pyx_t_22)) )) += 1;
              }
            }
            __pyx_L17_continue:;
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":339
 *         offset0, offset1 = offset
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L14;
        }
        __pyx_L14:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":359
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1
 *     return numpy.asarray(lut_size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_lut_size, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut_size, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cmask, 1);
  __Pyx_XDECREF(__pyx_v_err);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_68__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_fromslice(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_mask, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int8_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_offset);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_offset);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_offset);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyFAI.ext._distortion.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_51calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_5pyFAI_3ext_11_distortion_51calc_size = {"__pyx_fuse_1calc_size", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_51calc_size, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_6calc_size};
static PyObject *__pyx_fuse_1__pyx_pw_5pyFAI_3ext_11_distortion_51calc_size(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_offset = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_size (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_mask,&__pyx_n_s_offset,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults7 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_offset;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_size", 0, 2, 4, 1); __PYX_ERR(0, 306, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_size") < 0)) __PYX_ERR(0, 306, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 306, __pyx_L3_error)
    __pyx_v_shape = values[1];
    if (values[2]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 308, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_dynamic_args->__pyx_arg_mask;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
    __pyx_v_offset = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_size", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 306, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 306, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_50calc_size(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_mask, __pyx_v_offset);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_50calc_size(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_offset) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_l;
  int __pyx_v_shape_out0;
  int __pyx_v_shape_out1;
  int __pyx_v_shape_in0;
  int __pyx_v_shape_in1;
  int __pyx_v_min0;
  int __pyx_v_min1;
  int __pyx_v_max0;
  int __pyx_v_max1;
  __Pyx_memviewslice __pyx_v_lut_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_offset0;
  float __pyx_v_offset1;
  int __pyx_v_do_mask;
  __Pyx_memviewslice __pyx_v_cmask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_UCS4 __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  float __pyx_t_16;
  float __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1calc_size", 0);

  /* "pyFAI/ext/_distortion.pyx":320
 *     cdef:
 *         int i, j, k, l, shape_out0, shape_out1, shape_in0, shape_in1, min0, min1, max0, max1
 *         int32_t[:, ::1] lut_size = numpy.zeros(shape, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *         float A0, A1, B0, B1, C0, C1, D0, D1, offset0, offset1
 *         bint do_mask = mask is not None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_lut_size = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":322
 *         int32_t[:, ::1] lut_size = numpy.zeros(shape, dtype=numpy.int32)
 *         float A0, A1, B0, B1, C0, C1, D0, D1, offset0, offset1
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":324
 *         bint do_mask = mask is not None
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape
 * 
 */
  __pyx_t_7 = (__pyx_v_pos.shape[0]);
  __pyx_t_8 = (__pyx_v_pos.shape[1]);
  __pyx_v_shape_in0 = __pyx_t_7;
  __pyx_v_shape_in1 = __pyx_t_8;

  /* "pyFAI/ext/_distortion.pyx":325
 *         int8_t[:, ::1] cmask
 *     shape_in0, shape_in1 = pos.shape[0], pos.shape[1]
 *     shape_out0, shape_out1 = shape             # <<<<<<<<<<<<<<
 * 
 *     if do_mask:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 325, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_1); if (unlikely(!__pyx_t_5)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_3 = __pyx_t_9(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_1), 2) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_shape_out0 = __pyx_t_10;
  __pyx_v_shape_out1 = __pyx_t_11;

  /* "pyFAI/ext/_distortion.pyx":327
 *     shape_out0, shape_out1 = shape
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 */
  __pyx_t_12 = (__pyx_v_do_mask != 0);
  if (__pyx_t_12) {

    /* "pyFAI/ext/_distortion.pyx":329
 *     if do_mask:
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):             # <<<<<<<<<<<<<<
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 */
    __pyx_t_13 = (((__pyx_v_mask.shape[0]) != __pyx_v_shape_in0) != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_12 = __pyx_t_13;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_13 = (((__pyx_v_mask.shape[1]) != __pyx_v_shape_in1) != 0);
    __pyx_t_12 = __pyx_t_13;
    __pyx_L7_bool_binop_done:;
    if (unlikely(__pyx_t_12)) {

      /* "pyFAI/ext/_distortion.pyx":330
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])             # <<<<<<<<<<<<<<
 *             logger.error(err)
 *             raise RuntimeError(err)
 */
      __pyx_t_3 = PyTuple_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = 0;
      __pyx_t_14 = 127;
      __Pyx_INCREF(__pyx_kp_u_Mismatch_between_shape_of_detect);
      __pyx_t_8 += 36;
      __Pyx_GIVEREF(__pyx_kp_u_Mismatch_between_shape_of_detect);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Mismatch_between_shape_of_detect);
      __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_shape_in0, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_t_8 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__10);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__10);
      __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_shape_in1, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_and_shape_of_mask);
      __pyx_t_8 += 21;
      __Pyx_GIVEREF(__pyx_kp_u_and_shape_of_mask);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_and_shape_of_mask);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t((__pyx_v_mask.shape[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_t_8 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__10);
      PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u__10);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t((__pyx_v_mask.shape[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__11);
      __pyx_t_8 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__11);
      PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u__11);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 9, __pyx_t_8, __pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_err = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":331
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)             # <<<<<<<<<<<<<<
 *             raise RuntimeError(err)
 *         else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_err) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_err);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 331, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":332
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 *             raise RuntimeError(err)             # <<<<<<<<<<<<<<
 *         else:
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 332, __pyx_L1_error)

      /* "pyFAI/ext/_distortion.pyx":329
 *     if do_mask:
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):             # <<<<<<<<<<<<<<
 *             err = 'Mismatch between shape of detector (%s, %s) and shape of mask (%s, %s)' % (shape_in0, shape_in1, mask.shape[0], mask.shape[1])
 *             logger.error(err)
 */
    }

    /* "pyFAI/ext/_distortion.pyx":334
 *             raise RuntimeError(err)
 *         else:
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)             # <<<<<<<<<<<<<<
 * 
 *     if offset is not None:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_mask, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int8_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_15 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_15.memview)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_cmask = __pyx_t_15;
      __pyx_t_15.memview = NULL;
      __pyx_t_15.data = NULL;
    }

    /* "pyFAI/ext/_distortion.pyx":327
 *     shape_out0, shape_out1 = shape
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 * 
 *         if ((mask.shape[0] != shape_in0) or (mask.shape[1] != shape_in1)):
 */
  }

  /* "pyFAI/ext/_distortion.pyx":336
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 * 
 *     if offset is not None:             # <<<<<<<<<<<<<<
 *         offset0, offset1 = offset
 * 
 */
  __pyx_t_12 = (__pyx_v_offset != Py_None);
  __pyx_t_13 = (__pyx_t_12 != 0);
  if (__pyx_t_13) {

    /* "pyFAI/ext/_distortion.pyx":337
 * 
 *     if offset is not None:
 *         offset0, offset1 = offset             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_offset))) || (PyList_CheckExact(__pyx_v_offset))) {
      PyObject* sequence = __pyx_v_offset;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 337, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_9(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_9(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_3), 2) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 337, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __pyx_t_16 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_16 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_17 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_17 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_offset0 = __pyx_t_16;
    __pyx_v_offset1 = __pyx_t_17;

    /* "pyFAI/ext/_distortion.pyx":336
 *             cmask = numpy.ascontiguousarray(mask, dtype=numpy.int8)
 * 
 *     if offset is not None:             # <<<<<<<<<<<<<<
 *         offset0, offset1 = offset
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":339
 *         offset0, offset1 = offset
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":340
 * 
 *     with nogil:
 *         for i in range(shape_in0):             # <<<<<<<<<<<<<<
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:
 */
        __pyx_t_11 = __pyx_v_shape_in0;
        __pyx_t_10 = __pyx_t_11;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_10; __pyx_t_18+=1) {
          __pyx_v_i = __pyx_t_18;

          /* "pyFAI/ext/_distortion.pyx":341
 *     with nogil:
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):             # <<<<<<<<<<<<<<
 *                 if do_mask and cmask[i, j]:
 *                     continue
 */
          __pyx_t_19 = __pyx_v_shape_in1;
          __pyx_t_20 = __pyx_t_19;
          for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
            __pyx_v_j = __pyx_t_21;

            /* "pyFAI/ext/_distortion.pyx":342
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 */
            __pyx_t_12 = (__pyx_v_do_mask != 0);
            if (__pyx_t_12) {
            } else {
              __pyx_t_13 = __pyx_t_12;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_12 = ((*((int8_t *) ( /* dim=1 */ ((char *) (((int8_t *) ( /* dim=0 */ (__pyx_v_cmask.data + __pyx_t_22 * __pyx_v_cmask.strides[0]) )) + __pyx_t_23)) ))) != 0);
            __pyx_t_13 = __pyx_t_12;
            __pyx_L20_bool_binop_done:;
            if (__pyx_t_13) {

              /* "pyFAI/ext/_distortion.pyx":343
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:
 *                     continue             # <<<<<<<<<<<<<<
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1
 */
              goto __pyx_L17_continue;

              /* "pyFAI/ext/_distortion.pyx":342
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 *                 if do_mask and cmask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 */
            }

            /* "pyFAI/ext/_distortion.pyx":344
 *                 if do_mask and cmask[i, j]:
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0             # <<<<<<<<<<<<<<
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 0;
            __pyx_t_25 = 0;
            __pyx_v_A0 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":345
 *                     continue
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1             # <<<<<<<<<<<<<<
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 0;
            __pyx_t_23 = 1;
            __pyx_v_A1 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":346
 *                 A0 = pos[i, j, 0, 0] - offset0
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0             # <<<<<<<<<<<<<<
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 1;
            __pyx_t_25 = 0;
            __pyx_v_B0 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":347
 *                 A1 = pos[i, j, 0, 1] - offset1
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1             # <<<<<<<<<<<<<<
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 1;
            __pyx_t_23 = 1;
            __pyx_v_B1 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":348
 *                 B0 = pos[i, j, 1, 0] - offset0
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0             # <<<<<<<<<<<<<<
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 2;
            __pyx_t_25 = 0;
            __pyx_v_C0 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":349
 *                 B1 = pos[i, j, 1, 1] - offset1
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1             # <<<<<<<<<<<<<<
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 2;
            __pyx_t_23 = 1;
            __pyx_v_C1 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":350
 *                 C0 = pos[i, j, 2, 0] - offset0
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0             # <<<<<<<<<<<<<<
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 */
            __pyx_t_23 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_24 = 3;
            __pyx_t_25 = 0;
            __pyx_v_D0 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_23 * __pyx_v_pos.strides[0]) ) + __pyx_t_22 * __pyx_v_pos.strides[1]) ) + __pyx_t_24 * __pyx_v_pos.strides[2]) )) + __pyx_t_25)) ))) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":351
 *                 C1 = pos[i, j, 2, 1] - offset1
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1             # <<<<<<<<<<<<<<
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 */
            __pyx_t_25 = __pyx_v_i;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_22 = 3;
            __pyx_t_23 = 1;
            __pyx_v_D1 = ((*((double *) ( /* dim=3 */ ((char *) (((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_25 * __pyx_v_pos.strides[0]) ) + __pyx_t_24 * __pyx_v_pos.strides[1]) ) + __pyx_t_22 * __pyx_v_pos.strides[2]) )) + __pyx_t_23)) ))) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":352
 *                 D0 = pos[i, j, 3, 0] - offset0
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)             # <<<<<<<<<<<<<<
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 */
            __pyx_v_min0 = __pyx_f_5pyFAI_3ext_11_distortion_clip(((int)__pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)), 0, __pyx_v_shape_out0, 0);

            /* "pyFAI/ext/_distortion.pyx":353
 *                 D1 = pos[i, j, 3, 1] - offset1
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)             # <<<<<<<<<<<<<<
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 */
            __pyx_v_min1 = __pyx_f_5pyFAI_3ext_11_distortion_clip(((int)__pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)), 0, __pyx_v_shape_out1, 0);

            /* "pyFAI/ext/_distortion.pyx":354
 *                 min0 = clip(<int> _floor_min4(A0, B0, C0, D0), 0, shape_out0)
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)             # <<<<<<<<<<<<<<
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):
 */
            __pyx_v_max0 = __pyx_f_5pyFAI_3ext_11_distortion_clip((((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) + 1), 0, __pyx_v_shape_out0, 0);

            /* "pyFAI/ext/_distortion.pyx":355
 *                 min1 = clip(<int> _floor_min4(A1, B1, C1, D1), 0, shape_out1)
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)             # <<<<<<<<<<<<<<
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):
 */
            __pyx_v_max1 = __pyx_f_5pyFAI_3ext_11_distortion_clip((((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) + 1), 0, __pyx_v_shape_out1, 0);

            /* "pyFAI/ext/_distortion.pyx":356
 *                 max0 = clip(<int> _ceil_max4(A0, B0, C0, D0) + 1, 0, shape_out0)
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):             # <<<<<<<<<<<<<<
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1
 */
            __pyx_t_26 = __pyx_v_max0;
            __pyx_t_27 = __pyx_t_26;
            for (__pyx_t_28 = __pyx_v_min0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
              __pyx_v_k = __pyx_t_28;

              /* "pyFAI/ext/_distortion.pyx":357
 *                 max1 = clip(<int> _ceil_max4(A1, B1, C1, D1) + 1, 0, shape_out1)
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):             # <<<<<<<<<<<<<<
 *                         lut_size[k, l] += 1
 *     return numpy.asarray(lut_size)
 */
              __pyx_t_29 = __pyx_v_max1;
              __pyx_t_30 = __pyx_t_29;
              for (__pyx_t_31 = __pyx_v_min1; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
                __pyx_v_l = __pyx_t_31;

                /* "pyFAI/ext/_distortion.pyx":358
 *                 for k in range(min0, max0):
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1             # <<<<<<<<<<<<<<
 *     return numpy.asarray(lut_size)
 * 
 */
                __pyx_t_23 = __pyx_v_k;
                __pyx_t_22 = __pyx_v_l;
                *((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_lut_size.data + __pyx_t_23 * __pyx_v_lut_size.strides[0]) )) + __pyx_t_22)) )) += 1;
              }
            }
            __pyx_L17_continue:;
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":339
 *         offset0, offset1 = offset
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         for i in range(shape_in0):
 *             for j in range(shape_in1):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L14;
        }
        __pyx_L14:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":359
 *                     for l in range(min1, max1):
 *                         lut_size[k, l] += 1
 *     return numpy.asarray(lut_size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_lut_size, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut_size, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cmask, 1);
  __Pyx_XDECREF(__pyx_v_err);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":362
 * 
 * 
 * def calc_LUT(float32_t[:, :, :, ::1] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, :] mask=None):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_9calc_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_8calc_LUT[] = "calc_LUT(float32_t[:, :, :, ::1] pos, shape, bin_size, max_pixel_size, int8_t[:, :] mask=None)\n\n    :param pos: 4D position array\n    :param shape: output shape\n    :param bin_size: number of input element per output element (numpy array)\n    :param max_pixel_size: (2-tuple of int) size of a buffer covering the largest pixel\n    :param mask: arry with bad pixels marked as True\n    :return: look-up table\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_9calc_LUT = {"calc_LUT", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_9calc_LUT, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_8calc_LUT};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_9calc_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_bin_size = 0;
  PyObject *__pyx_v_max_pixel_size = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_LUT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_bin_size,&__pyx_n_s_max_pixel_size,&__pyx_n_s_mask,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_LUT", 0, 4, 5, 1); __PYX_ERR(0, 362, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bin_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_LUT", 0, 4, 5, 2); __PYX_ERR(0, 362, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_pixel_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_LUT", 0, 4, 5, 3); __PYX_ERR(0, 362, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_LUT") < 0)) __PYX_ERR(0, 362, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 362, __pyx_L3_error)
    __pyx_v_shape = values[1];
    __pyx_v_bin_size = values[2];
    __pyx_v_max_pixel_size = values[3];
    if (values[4]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int8_t(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 363, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_k__12;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_LUT", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 362, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 362, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_8calc_LUT(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_bin_size, __pyx_v_max_pixel_size, __pyx_v_mask);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_8calc_LUT(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_bin_size, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_ms;
  int __pyx_v_ml;
  int __pyx_v_ns;
  int __pyx_v_nl;
  int __pyx_v_shape0;
  int __pyx_v_shape1;
  int __pyx_v_delta0;
  int __pyx_v_delta1;
  int __pyx_v_offset0;
  int __pyx_v_offset1;
  int __pyx_v_box_size0;
  int __pyx_v_box_size1;
  int __pyx_v_size;
  int __pyx_v_k;
  int32_t __pyx_v_idx;
  int __pyx_v_err_cnt;
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_pAB;
  float __pyx_v_pBC;
  float __pyx_v_pCD;
  float __pyx_v_pDA;
  float __pyx_v_cAB;
  float __pyx_v_cBC;
  float __pyx_v_cCD;
  float __pyx_v_cDA;
  float __pyx_v_area;
  float __pyx_v_value;
  float __pyx_v_foffset0;
  float __pyx_v_foffset1;
  __Pyx_memviewslice __pyx_v_lut = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_mask;
  __Pyx_memviewslice __pyx_v_buffer = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_outMax = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_lut_total_size = NULL;
  __Pyx_memviewslice __pyx_v_tmp_ary = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  size_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  Py_ssize_t __pyx_t_38;
  Py_UCS4 __pyx_t_39;
  __Pyx_memviewslice __pyx_t_40 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_41;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_LUT", 0);

  /* "pyFAI/ext/_distortion.pyx":375
 *         int i, j, ms, ml, ns, nl, shape0, shape1, delta0, delta1
 *         int offset0, offset1, box_size0, box_size1, size, k
 *         int32_t idx = 0             # <<<<<<<<<<<<<<
 *         int err_cnt = 0
 *         float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA,
 */
  __pyx_v_idx = 0;

  /* "pyFAI/ext/_distortion.pyx":376
 *         int offset0, offset1, box_size0, box_size1, size, k
 *         int32_t idx = 0
 *         int err_cnt = 0             # <<<<<<<<<<<<<<
 *         float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA,
 *         float area, value, foffset0, foffset1
 */
  __pyx_v_err_cnt = 0;

  /* "pyFAI/ext/_distortion.pyx":380
 *         float area, value, foffset0, foffset1
 *         lut_t[:, :, :] lut
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 *         float32_t[:, ::1] buffer
 *     size = bin_size.max()
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":382
 *         bint do_mask = mask is not None
 *         float32_t[:, ::1] buffer
 *     size = bin_size.max()             # <<<<<<<<<<<<<<
 *     shape0, shape1 = shape
 *     if do_mask:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_bin_size, __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_size = __pyx_t_4;

  /* "pyFAI/ext/_distortion.pyx":383
 *         float32_t[:, ::1] buffer
 *     size = bin_size.max()
 *     shape0, shape1 = shape             # <<<<<<<<<<<<<<
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 383, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 2) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 383, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_shape0 = __pyx_t_4;
  __pyx_v_shape1 = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":384
 *     size = bin_size.max()
 *     shape0, shape1 = shape
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 */
  __pyx_t_7 = (__pyx_v_do_mask != 0);
  if (__pyx_t_7) {

    /* "pyFAI/ext/_distortion.pyx":385
 *     shape0, shape1 = shape
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"             # <<<<<<<<<<<<<<
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 *     delta0, delta1 = max_pixel_size
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape0 == (__pyx_v_mask.shape[0])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_mask_shape_dim0);
        __PYX_ERR(0, 385, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":386
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"             # <<<<<<<<<<<<<<
 *     delta0, delta1 = max_pixel_size
 *     cdef int32_t[:, :] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape1 == (__pyx_v_mask.shape[1])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_mask_shape_dim1);
        __PYX_ERR(0, 386, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":384
 *     size = bin_size.max()
 *     shape0, shape1 = shape
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":387
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 *     delta0, delta1 = max_pixel_size             # <<<<<<<<<<<<<<
 *     cdef int32_t[:, :] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_max_pixel_size))) || (PyList_CheckExact(__pyx_v_max_pixel_size))) {
    PyObject* sequence = __pyx_v_max_pixel_size;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 387, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_max_pixel_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 2) < 0) __PYX_ERR(0, 387, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 387, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_delta0 = __pyx_t_6;
  __pyx_v_delta1 = __pyx_t_4;

  /* "pyFAI/ext/_distortion.pyx":388
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 *     delta0, delta1 = max_pixel_size
 *     cdef int32_t[:, :] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     #buffer_nbytes = buffer.nbytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_shape0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_shape1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_outMax = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":389
 *     delta0, delta1 = max_pixel_size
 *     cdef int32_t[:, :] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     #buffer_nbytes = buffer.nbytes
 *     if (size == 0):  # fix 271
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_9 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_float32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_buffer = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":391
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     #buffer_nbytes = buffer.nbytes
 *     if (size == 0):  # fix 271             # <<<<<<<<<<<<<<
 *         raise RuntimeError("The look-up table has dimension 0 which is a non-sense." +
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 */
  __pyx_t_7 = ((__pyx_v_size == 0) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pyFAI/ext/_distortion.pyx":392
 *     #buffer_nbytes = buffer.nbytes
 *     if (size == 0):  # fix 271
 *         raise RuntimeError("The look-up table has dimension 0 which is a non-sense." +             # <<<<<<<<<<<<<<
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 *     lut = view.array(shape=(shape0, shape1, size), itemsize=sizeof(lut_t), format="if")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 392, __pyx_L1_error)

    /* "pyFAI/ext/_distortion.pyx":391
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     #buffer_nbytes = buffer.nbytes
 *     if (size == 0):  # fix 271             # <<<<<<<<<<<<<<
 *         raise RuntimeError("The look-up table has dimension 0 which is a non-sense." +
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 */
  }

  /* "pyFAI/ext/_distortion.pyx":394
 *         raise RuntimeError("The look-up table has dimension 0 which is a non-sense." +
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 *     lut = view.array(shape=(shape0, shape1, size), itemsize=sizeof(lut_t), format="if")             # <<<<<<<<<<<<<<
 *     lut_total_size = shape0 * shape1 * size * sizeof(lut_t)
 *     memset(&lut[0, 0, 0], 0, lut_total_size)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_shape0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_shape1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_8);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_9) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_FromSize_t((sizeof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_itemsize, __pyx_t_9) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_format, __pyx_n_u_if) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lut = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":395
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 *     lut = view.array(shape=(shape0, shape1, size), itemsize=sizeof(lut_t), format="if")
 *     lut_total_size = shape0 * shape1 * size * sizeof(lut_t)             # <<<<<<<<<<<<<<
 *     memset(&lut[0, 0, 0], 0, lut_total_size)
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))
 */
  __pyx_t_9 = __Pyx_PyInt_FromSize_t((((__pyx_v_shape0 * __pyx_v_shape1) * __pyx_v_size) * (sizeof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t)))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_lut_total_size = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":396
 *     lut = view.array(shape=(shape0, shape1, size), itemsize=sizeof(lut_t), format="if")
 *     lut_total_size = shape0 * shape1 * size * sizeof(lut_t)
 *     memset(&lut[0, 0, 0], 0, lut_total_size)             # <<<<<<<<<<<<<<
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (delta1, delta0, size))
 */
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_size_t(__pyx_v_lut_total_size); if (unlikely((__pyx_t_16 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L1_error)
  (void)(memset((&(*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_13 * __pyx_v_lut.strides[0]) ) + __pyx_t_14 * __pyx_v_lut.strides[1]) ) + __pyx_t_15 * __pyx_v_lut.strides[2]) )))), 0, __pyx_t_16));

  /* "pyFAI/ext/_distortion.pyx":397
 *     lut_total_size = shape0 * shape1 * size * sizeof(lut_t)
 *     memset(&lut[0, 0, 0], 0, lut_total_size)
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))             # <<<<<<<<<<<<<<
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (delta1, delta0, size))
 *     with nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_info); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_lut.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_lut.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_lut.shape[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = __Pyx_PyFloat_TrueDivideObjC(__pyx_v_lut_total_size, __pyx_float_1_0e6, 1.0e6, 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyTuple_New(4); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_18, 3, __pyx_t_17);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_17 = 0;
  __pyx_t_17 = PyUnicode_Format(__pyx_kp_u_LUT_shape_i_i_i_3f_MByte, __pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_18)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_18);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_9 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":398
 *     memset(&lut[0, 0, 0], 0, lut_total_size)
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (delta1, delta0, size))             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # i,j, idx are indexes of the raw image uncorrected
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_logger); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_info); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_8 = 0;
  __pyx_t_18 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_Max_pixel_size_ix_i_Max_source_p, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_17);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_17, function);
    }
  }
  __pyx_t_9 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":399
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (delta1, delta0, size))
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i,j, idx are indexes of the raw image uncorrected
 *         for i in range(shape0):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":401
 *     with nogil:
 *         # i,j, idx are indexes of the raw image uncorrected
 *         for i in range(shape0):             # <<<<<<<<<<<<<<
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:
 */
        __pyx_t_4 = __pyx_v_shape0;
        __pyx_t_6 = __pyx_t_4;
        for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_6; __pyx_t_19+=1) {
          __pyx_v_i = __pyx_t_19;

          /* "pyFAI/ext/_distortion.pyx":402
 *         # i,j, idx are indexes of the raw image uncorrected
 *         for i in range(shape0):
 *             for j in range(shape1):             # <<<<<<<<<<<<<<
 *                 if do_mask and mask[i, j]:
 *                     continue
 */
          __pyx_t_20 = __pyx_v_shape1;
          __pyx_t_21 = __pyx_t_20;
          for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
            __pyx_v_j = __pyx_t_22;

            /* "pyFAI/ext/_distortion.pyx":403
 *         for i in range(shape0):
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # reset buffer
 */
            __pyx_t_23 = (__pyx_v_do_mask != 0);
            if (__pyx_t_23) {
            } else {
              __pyx_t_7 = __pyx_t_23;
              goto __pyx_L17_bool_binop_done;
            }
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_j;
            __pyx_t_23 = ((*((int8_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_15 * __pyx_v_mask.strides[0]) ) + __pyx_t_14 * __pyx_v_mask.strides[1]) ))) != 0);
            __pyx_t_7 = __pyx_t_23;
            __pyx_L17_bool_binop_done:;
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":404
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:
 *                     continue             # <<<<<<<<<<<<<<
 *                 # reset buffer
 *                 buffer[:, :] = 0.0
 */
              goto __pyx_L14_continue;

              /* "pyFAI/ext/_distortion.pyx":403
 *         for i in range(shape0):
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # reset buffer
 */
            }

            /* "pyFAI/ext/_distortion.pyx":406
 *                     continue
 *                 # reset buffer
 *                 buffer[:, :] = 0.0             # <<<<<<<<<<<<<<
 * 
 *                 A0 = pos[i, j, 0, 0]
 */
            {
                __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_temp_scalar = 0.0;
                {
                    Py_ssize_t __pyx_temp_extent = __pyx_v_buffer.shape[0] * __pyx_v_buffer.shape[1];
                    Py_ssize_t __pyx_temp_idx;
                    __pyx_t_5pyFAI_3ext_12shared_types_float32_t *__pyx_temp_pointer = (__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_buffer.data;
                    for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
                      *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
                      __pyx_temp_pointer += 1;
                    }
                }
            }

            /* "pyFAI/ext/_distortion.pyx":408
 *                 buffer[:, :] = 0.0
 * 
 *                 A0 = pos[i, j, 0, 0]             # <<<<<<<<<<<<<<
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_15 = __pyx_v_j;
            __pyx_t_13 = 0;
            __pyx_t_24 = 0;
            __pyx_v_A0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_24)) )));

            /* "pyFAI/ext/_distortion.pyx":409
 * 
 *                 A0 = pos[i, j, 0, 0]
 *                 A1 = pos[i, j, 0, 1]             # <<<<<<<<<<<<<<
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_13 = __pyx_v_j;
            __pyx_t_15 = 0;
            __pyx_t_14 = 1;
            __pyx_v_A1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

            /* "pyFAI/ext/_distortion.pyx":410
 *                 A0 = pos[i, j, 0, 0]
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]             # <<<<<<<<<<<<<<
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_15 = __pyx_v_j;
            __pyx_t_13 = 1;
            __pyx_t_24 = 0;
            __pyx_v_B0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_24)) )));

            /* "pyFAI/ext/_distortion.pyx":411
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]             # <<<<<<<<<<<<<<
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_13 = __pyx_v_j;
            __pyx_t_15 = 1;
            __pyx_t_14 = 1;
            __pyx_v_B1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

            /* "pyFAI/ext/_distortion.pyx":412
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]             # <<<<<<<<<<<<<<
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_15 = __pyx_v_j;
            __pyx_t_13 = 2;
            __pyx_t_24 = 0;
            __pyx_v_C0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_24)) )));

            /* "pyFAI/ext/_distortion.pyx":413
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]             # <<<<<<<<<<<<<<
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_13 = __pyx_v_j;
            __pyx_t_15 = 2;
            __pyx_t_14 = 1;
            __pyx_v_C1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

            /* "pyFAI/ext/_distortion.pyx":414
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]             # <<<<<<<<<<<<<<
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_15 = __pyx_v_j;
            __pyx_t_13 = 3;
            __pyx_t_24 = 0;
            __pyx_v_D0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_24)) )));

            /* "pyFAI/ext/_distortion.pyx":415
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]             # <<<<<<<<<<<<<<
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_13 = __pyx_v_j;
            __pyx_t_15 = 3;
            __pyx_t_14 = 1;
            __pyx_v_D1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

            /* "pyFAI/ext/_distortion.pyx":416
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)             # <<<<<<<<<<<<<<
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)
 */
            __pyx_v_foffset0 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0);

            /* "pyFAI/ext/_distortion.pyx":417
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)             # <<<<<<<<<<<<<<
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)
 */
            __pyx_v_foffset1 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1);

            /* "pyFAI/ext/_distortion.pyx":418
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)             # <<<<<<<<<<<<<<
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 */
            __pyx_v_offset0 = ((int)__pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":419
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)             # <<<<<<<<<<<<<<
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 */
            __pyx_v_offset1 = ((int)__pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":420
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0             # <<<<<<<<<<<<<<
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *                 if (box_size0 > delta0) or (box_size1 > delta1):
 */
            __pyx_v_box_size0 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":421
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1             # <<<<<<<<<<<<<<
 *                 if (box_size0 > delta0) or (box_size1 > delta1):
 *                     # Increase size of the buffer
 */
            __pyx_v_box_size1 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":422
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *                 if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 */
            __pyx_t_23 = ((__pyx_v_box_size0 > __pyx_v_delta0) != 0);
            if (!__pyx_t_23) {
            } else {
              __pyx_t_7 = __pyx_t_23;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_23 = ((__pyx_v_box_size1 > __pyx_v_delta1) != 0);
            __pyx_t_7 = __pyx_t_23;
            __pyx_L20_bool_binop_done:;
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":424
 *                 if (box_size0 > delta0) or (box_size1 > delta1):
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0             # <<<<<<<<<<<<<<
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:
 */
              if (((__pyx_v_offset0 > __pyx_v_delta0) != 0)) {
                __pyx_t_25 = __pyx_v_offset0;
              } else {
                __pyx_t_25 = __pyx_v_delta0;
              }
              __pyx_v_delta0 = __pyx_t_25;

              /* "pyFAI/ext/_distortion.pyx":425
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1             # <<<<<<<<<<<<<<
 *                     with gil:
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 */
              if (((__pyx_v_offset1 > __pyx_v_delta1) != 0)) {
                __pyx_t_25 = __pyx_v_offset1;
              } else {
                __pyx_t_25 = __pyx_v_delta1;
              }
              __pyx_v_delta1 = __pyx_t_25;

              /* "pyFAI/ext/_distortion.pyx":426
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
              {
                  #ifdef WITH_THREAD
                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                  #endif
                  /*try:*/ {

                    /* "pyFAI/ext/_distortion.pyx":427
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *                 A0 -= foffset0
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_9);
                    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_9);
                    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_GIVEREF(__pyx_t_9);
                    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_9);
                    __Pyx_GIVEREF(__pyx_t_2);
                    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
                    __pyx_t_9 = 0;
                    __pyx_t_2 = 0;
                    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __Pyx_GIVEREF(__pyx_t_3);
                    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
                    __pyx_t_3 = 0;
                    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_9);
                    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_float32); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_18) < 0) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_18, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 427, __pyx_L25_error)
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
                    __pyx_v_buffer = __pyx_t_11;
                    __pyx_t_11.memview = NULL;
                    __pyx_t_11.data = NULL;
                  }

                  /* "pyFAI/ext/_distortion.pyx":426
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      goto __pyx_L26;
                    }
                    __pyx_L25_error: {
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      goto __pyx_L10_error;
                    }
                    __pyx_L26:;
                  }
              }

              /* "pyFAI/ext/_distortion.pyx":422
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *                 if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 */
            }

            /* "pyFAI/ext/_distortion.pyx":429
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 *                 A0 -= foffset0             # <<<<<<<<<<<<<<
 *                 A1 -= foffset1
 *                 B0 -= foffset0
 */
            __pyx_v_A0 = (__pyx_v_A0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":430
 * 
 *                 A0 -= foffset0
 *                 A1 -= foffset1             # <<<<<<<<<<<<<<
 *                 B0 -= foffset0
 *                 B1 -= foffset1
 */
            __pyx_v_A1 = (__pyx_v_A1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":431
 *                 A0 -= foffset0
 *                 A1 -= foffset1
 *                 B0 -= foffset0             # <<<<<<<<<<<<<<
 *                 B1 -= foffset1
 *                 C0 -= foffset0
 */
            __pyx_v_B0 = (__pyx_v_B0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":432
 *                 A1 -= foffset1
 *                 B0 -= foffset0
 *                 B1 -= foffset1             # <<<<<<<<<<<<<<
 *                 C0 -= foffset0
 *                 C1 -= foffset1
 */
            __pyx_v_B1 = (__pyx_v_B1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":433
 *                 B0 -= foffset0
 *                 B1 -= foffset1
 *                 C0 -= foffset0             # <<<<<<<<<<<<<<
 *                 C1 -= foffset1
 *                 D0 -= foffset0
 */
            __pyx_v_C0 = (__pyx_v_C0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":434
 *                 B1 -= foffset1
 *                 C0 -= foffset0
 *                 C1 -= foffset1             # <<<<<<<<<<<<<<
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 */
            __pyx_v_C1 = (__pyx_v_C1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":435
 *                 C0 -= foffset0
 *                 C1 -= foffset1
 *                 D0 -= foffset0             # <<<<<<<<<<<<<<
 *                 D1 -= foffset1
 *                 if B0 != A0:
 */
            __pyx_v_D0 = (__pyx_v_D0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":436
 *                 C1 -= foffset1
 *                 D0 -= foffset0
 *                 D1 -= foffset1             # <<<<<<<<<<<<<<
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)
 */
            __pyx_v_D1 = (__pyx_v_D1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":437
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 *                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0
 */
            __pyx_t_7 = ((__pyx_v_B0 != __pyx_v_A0) != 0);
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":438
 *                 D1 -= foffset1
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)             # <<<<<<<<<<<<<<
 *                     cAB = A1 - pAB * A0
 *                 else:
 */
              __pyx_v_pAB = ((__pyx_v_B1 - __pyx_v_A1) / (__pyx_v_B0 - __pyx_v_A0));

              /* "pyFAI/ext/_distortion.pyx":439
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pAB = cAB = 0.0
 */
              __pyx_v_cAB = (__pyx_v_A1 - (__pyx_v_pAB * __pyx_v_A0));

              /* "pyFAI/ext/_distortion.pyx":437
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 *                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0
 */
              goto __pyx_L27;
            }

            /* "pyFAI/ext/_distortion.pyx":441
 *                     cAB = A1 - pAB * A0
 *                 else:
 *                     pAB = cAB = 0.0             # <<<<<<<<<<<<<<
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)
 */
            /*else*/ {
              __pyx_v_pAB = 0.0;
              __pyx_v_cAB = 0.0;
            }
            __pyx_L27:;

            /* "pyFAI/ext/_distortion.pyx":442
 *                 else:
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0
 */
            __pyx_t_7 = ((__pyx_v_C0 != __pyx_v_B0) != 0);
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":443
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)             # <<<<<<<<<<<<<<
 *                     cBC = B1 - pBC * B0
 *                 else:
 */
              __pyx_v_pBC = ((__pyx_v_C1 - __pyx_v_B1) / (__pyx_v_C0 - __pyx_v_B0));

              /* "pyFAI/ext/_distortion.pyx":444
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pBC = cBC = 0.0
 */
              __pyx_v_cBC = (__pyx_v_B1 - (__pyx_v_pBC * __pyx_v_B0));

              /* "pyFAI/ext/_distortion.pyx":442
 *                 else:
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0
 */
              goto __pyx_L28;
            }

            /* "pyFAI/ext/_distortion.pyx":446
 *                     cBC = B1 - pBC * B0
 *                 else:
 *                     pBC = cBC = 0.0             # <<<<<<<<<<<<<<
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)
 */
            /*else*/ {
              __pyx_v_pBC = 0.0;
              __pyx_v_cBC = 0.0;
            }
            __pyx_L28:;

            /* "pyFAI/ext/_distortion.pyx":447
 *                 else:
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0
 */
            __pyx_t_7 = ((__pyx_v_D0 != __pyx_v_C0) != 0);
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":448
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)             # <<<<<<<<<<<<<<
 *                     cCD = C1 - pCD * C0
 *                 else:
 */
              __pyx_v_pCD = ((__pyx_v_D1 - __pyx_v_C1) / (__pyx_v_D0 - __pyx_v_C0));

              /* "pyFAI/ext/_distortion.pyx":449
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pCD = cCD = 0.0
 */
              __pyx_v_cCD = (__pyx_v_C1 - (__pyx_v_pCD * __pyx_v_C0));

              /* "pyFAI/ext/_distortion.pyx":447
 *                 else:
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0
 */
              goto __pyx_L29;
            }

            /* "pyFAI/ext/_distortion.pyx":451
 *                     cCD = C1 - pCD * C0
 *                 else:
 *                     pCD = cCD = 0.0             # <<<<<<<<<<<<<<
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)
 */
            /*else*/ {
              __pyx_v_pCD = 0.0;
              __pyx_v_cCD = 0.0;
            }
            __pyx_L29:;

            /* "pyFAI/ext/_distortion.pyx":452
 *                 else:
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0
 */
            __pyx_t_7 = ((__pyx_v_A0 != __pyx_v_D0) != 0);
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":453
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)             # <<<<<<<<<<<<<<
 *                     cDA = D1 - pDA * D0
 *                 else:
 */
              __pyx_v_pDA = ((__pyx_v_A1 - __pyx_v_D1) / (__pyx_v_A0 - __pyx_v_D0));

              /* "pyFAI/ext/_distortion.pyx":454
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pDA = cDA = 0.0
 */
              __pyx_v_cDA = (__pyx_v_D1 - (__pyx_v_pDA * __pyx_v_D0));

              /* "pyFAI/ext/_distortion.pyx":452
 *                 else:
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0
 */
              goto __pyx_L30;
            }

            /* "pyFAI/ext/_distortion.pyx":456
 *                     cDA = D1 - pDA * D0
 *                 else:
 *                     pDA = cDA = 0.0             # <<<<<<<<<<<<<<
 * 
 *                 # ABCD is trigonometric order: order input position accordingly
 */
            /*else*/ {
              __pyx_v_pDA = 0.0;
              __pyx_v_cDA = 0.0;
            }
            __pyx_L30:;

            /* "pyFAI/ext/_distortion.pyx":459
 * 
 *                 # ABCD is trigonometric order: order input position accordingly
 *                 integrate(buffer, B0, A0, pAB, cAB)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 integrate(buffer, D0, C0, pCD, cCD)
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_B0, __pyx_v_A0, __pyx_v_pAB, __pyx_v_cAB);

            /* "pyFAI/ext/_distortion.pyx":460
 *                 # ABCD is trigonometric order: order input position accordingly
 *                 integrate(buffer, B0, A0, pAB, cAB)
 *                 integrate(buffer, C0, B0, pBC, cBC)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, D0, C0, pCD, cCD)
 *                 integrate(buffer, A0, D0, pDA, cDA)
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_C0, __pyx_v_B0, __pyx_v_pBC, __pyx_v_cBC);

            /* "pyFAI/ext/_distortion.pyx":461
 *                 integrate(buffer, B0, A0, pAB, cAB)
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 integrate(buffer, D0, C0, pCD, cCD)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, A0, D0, pDA, cDA)
 * 
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_D0, __pyx_v_C0, __pyx_v_pCD, __pyx_v_cCD);

            /* "pyFAI/ext/_distortion.pyx":462
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 integrate(buffer, D0, C0, pCD, cCD)
 *                 integrate(buffer, A0, D0, pDA, cDA)             # <<<<<<<<<<<<<<
 * 
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_A0, __pyx_v_D0, __pyx_v_pDA, __pyx_v_cDA);

            /* "pyFAI/ext/_distortion.pyx":464
 *                 integrate(buffer, A0, D0, pDA, cDA)
 * 
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))             # <<<<<<<<<<<<<<
 * 
 *                 for ms in range(box_size0):
 */
            __pyx_v_area = (0.5 * (((__pyx_v_C0 - __pyx_v_A0) * (__pyx_v_D1 - __pyx_v_B1)) - ((__pyx_v_C1 - __pyx_v_A1) * (__pyx_v_D0 - __pyx_v_B0))));

            /* "pyFAI/ext/_distortion.pyx":466
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 * 
 *                 for ms in range(box_size0):             # <<<<<<<<<<<<<<
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:
 */
            __pyx_t_25 = __pyx_v_box_size0;
            __pyx_t_26 = __pyx_t_25;
            for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
              __pyx_v_ms = __pyx_t_27;

              /* "pyFAI/ext/_distortion.pyx":467
 * 
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0             # <<<<<<<<<<<<<<
 *                     if ml < 0 or ml >= shape0:
 *                         continue
 */
              __pyx_v_ml = (__pyx_v_ms + __pyx_v_offset0);

              /* "pyFAI/ext/_distortion.pyx":468
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     for ns in range(box_size1):
 */
              __pyx_t_23 = ((__pyx_v_ml < 0) != 0);
              if (!__pyx_t_23) {
              } else {
                __pyx_t_7 = __pyx_t_23;
                goto __pyx_L34_bool_binop_done;
              }
              __pyx_t_23 = ((__pyx_v_ml >= __pyx_v_shape0) != 0);
              __pyx_t_7 = __pyx_t_23;
              __pyx_L34_bool_binop_done:;
              if (__pyx_t_7) {

                /* "pyFAI/ext/_distortion.pyx":469
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     for ns in range(box_size1):
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 */
                goto __pyx_L31_continue;

                /* "pyFAI/ext/_distortion.pyx":468
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     for ns in range(box_size1):
 */
              }

              /* "pyFAI/ext/_distortion.pyx":470
 *                     if ml < 0 or ml >= shape0:
 *                         continue
 *                     for ns in range(box_size1):             # <<<<<<<<<<<<<<
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 */
              __pyx_t_28 = __pyx_v_box_size1;
              __pyx_t_29 = __pyx_t_28;
              for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                __pyx_v_ns = __pyx_t_30;

                /* "pyFAI/ext/_distortion.pyx":472
 *                     for ns in range(box_size1):
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1             # <<<<<<<<<<<<<<
 *                         if nl < 0 or nl >= shape1:
 *                             continue
 */
                __pyx_v_nl = (__pyx_v_ns + __pyx_v_offset1);

                /* "pyFAI/ext/_distortion.pyx":473
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                             continue
 *                         value = buffer[ms, ns] / area
 */
                __pyx_t_23 = ((__pyx_v_nl < 0) != 0);
                if (!__pyx_t_23) {
                } else {
                  __pyx_t_7 = __pyx_t_23;
                  goto __pyx_L39_bool_binop_done;
                }
                __pyx_t_23 = ((__pyx_v_nl >= __pyx_v_shape1) != 0);
                __pyx_t_7 = __pyx_t_23;
                __pyx_L39_bool_binop_done:;
                if (__pyx_t_7) {

                  /* "pyFAI/ext/_distortion.pyx":474
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:
 *                             continue             # <<<<<<<<<<<<<<
 *                         value = buffer[ms, ns] / area
 *                         if value == 0:
 */
                  goto __pyx_L36_continue;

                  /* "pyFAI/ext/_distortion.pyx":473
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                             continue
 *                         value = buffer[ms, ns] / area
 */
                }

                /* "pyFAI/ext/_distortion.pyx":475
 *                         if nl < 0 or nl >= shape1:
 *                             continue
 *                         value = buffer[ms, ns] / area             # <<<<<<<<<<<<<<
 *                         if value == 0:
 *                             continue
 */
                __pyx_t_14 = __pyx_v_ms;
                __pyx_t_15 = __pyx_v_ns;
                __pyx_v_value = ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_area));

                /* "pyFAI/ext/_distortion.pyx":476
 *                             continue
 *                         value = buffer[ms, ns] / area
 *                         if value == 0:             # <<<<<<<<<<<<<<
 *                             continue
 *                         if value < 0 or value > 1.0001:
 */
                __pyx_t_7 = ((__pyx_v_value == 0.0) != 0);
                if (__pyx_t_7) {

                  /* "pyFAI/ext/_distortion.pyx":477
 *                         value = buffer[ms, ns] / area
 *                         if value == 0:
 *                             continue             # <<<<<<<<<<<<<<
 *                         if value < 0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 */
                  goto __pyx_L36_continue;

                  /* "pyFAI/ext/_distortion.pyx":476
 *                             continue
 *                         value = buffer[ms, ns] / area
 *                         if value == 0:             # <<<<<<<<<<<<<<
 *                             continue
 *                         if value < 0 or value > 1.0001:
 */
                }

                /* "pyFAI/ext/_distortion.pyx":478
 *                         if value == 0:
 *                             continue
 *                         if value < 0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 */
                __pyx_t_23 = ((__pyx_v_value < 0.0) != 0);
                if (!__pyx_t_23) {
                } else {
                  __pyx_t_7 = __pyx_t_23;
                  goto __pyx_L43_bool_binop_done;
                }
                __pyx_t_23 = ((__pyx_v_value > 1.0001) != 0);
                __pyx_t_7 = __pyx_t_23;
                __pyx_L43_bool_binop_done:;
                if (__pyx_t_7) {

                  /* "pyFAI/ext/_distortion.pyx":480
 *                         if value < 0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                  __pyx_t_7 = ((__pyx_v_err_cnt < 0x3E8) != 0);
                  if (__pyx_t_7) {

                    /* "pyFAI/ext/_distortion.pyx":481
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 *                                 with gil:             # <<<<<<<<<<<<<<
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                    {
                        #ifdef WITH_THREAD
                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                        #endif
                        /*try:*/ {

                          /* "pyFAI/ext/_distortion.pyx":482
 *                             if err_cnt < 1000:
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])             # <<<<<<<<<<<<<<
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1
 */
                          __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_18);
                          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_3);
                          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_ms); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_2);
                          __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_box_size0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_17);
                          __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_ns); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_9);
                          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_box_size1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_8);
                          __pyx_t_15 = __pyx_v_ms;
                          __pyx_t_14 = __pyx_v_ns;
                          __pyx_t_1 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_1);
                          __pyx_t_31 = PyFloat_FromDouble(__pyx_v_area); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_31);
                          __pyx_t_32 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_32);
                          __pyx_t_14 = 0;
                          __pyx_t_15 = 0;
                          __pyx_t_33 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) )))); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_33);
                          __pyx_t_15 = 0;
                          __pyx_t_14 = 1;
                          __pyx_t_34 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_34);
                          __pyx_t_14 = 1;
                          __pyx_t_15 = 0;
                          __pyx_t_35 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) )))); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_15 = 1;
                          __pyx_t_14 = 1;
                          __pyx_t_36 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __pyx_t_37 = PyTuple_New(13); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_18);
                          PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_18);
                          __Pyx_GIVEREF(__pyx_t_3);
                          PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_3);
                          __Pyx_GIVEREF(__pyx_t_2);
                          PyTuple_SET_ITEM(__pyx_t_37, 2, __pyx_t_2);
                          __Pyx_GIVEREF(__pyx_t_17);
                          PyTuple_SET_ITEM(__pyx_t_37, 3, __pyx_t_17);
                          __Pyx_GIVEREF(__pyx_t_9);
                          PyTuple_SET_ITEM(__pyx_t_37, 4, __pyx_t_9);
                          __Pyx_GIVEREF(__pyx_t_8);
                          PyTuple_SET_ITEM(__pyx_t_37, 5, __pyx_t_8);
                          __Pyx_GIVEREF(__pyx_t_1);
                          PyTuple_SET_ITEM(__pyx_t_37, 6, __pyx_t_1);
                          __Pyx_GIVEREF(__pyx_t_31);
                          PyTuple_SET_ITEM(__pyx_t_37, 7, __pyx_t_31);
                          __Pyx_GIVEREF(__pyx_t_32);
                          PyTuple_SET_ITEM(__pyx_t_37, 8, __pyx_t_32);
                          __Pyx_GIVEREF(__pyx_t_33);
                          PyTuple_SET_ITEM(__pyx_t_37, 9, __pyx_t_33);
                          __Pyx_GIVEREF(__pyx_t_34);
                          PyTuple_SET_ITEM(__pyx_t_37, 10, __pyx_t_34);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_37, 11, __pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_36);
                          PyTuple_SET_ITEM(__pyx_t_37, 12, __pyx_t_36);
                          __pyx_t_18 = 0;
                          __pyx_t_3 = 0;
                          __pyx_t_2 = 0;
                          __pyx_t_17 = 0;
                          __pyx_t_9 = 0;
                          __pyx_t_8 = 0;
                          __pyx_t_1 = 0;
                          __pyx_t_31 = 0;
                          __pyx_t_32 = 0;
                          __pyx_t_33 = 0;
                          __pyx_t_34 = 0;
                          __pyx_t_35 = 0;
                          __pyx_t_36 = 0;
                          __pyx_t_36 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_37, NULL); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 482, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

                          /* "pyFAI/ext/_distortion.pyx":483
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))             # <<<<<<<<<<<<<<
 *                                 err_cnt += 1
 *                             continue
 */
                          __pyx_t_36 = PyTuple_New(16); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __pyx_t_38 = 0;
                          __pyx_t_39 = 127;
                          __Pyx_INCREF(__pyx_kp_u_A0);
                          __pyx_t_38 += 4;
                          __Pyx_GIVEREF(__pyx_kp_u_A0);
                          PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_kp_u_A0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_A0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_A1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_A1);
                          PyTuple_SET_ITEM(__pyx_t_36, 2, __pyx_kp_u_A1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_A1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 3, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_B0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_B0);
                          PyTuple_SET_ITEM(__pyx_t_36, 4, __pyx_kp_u_B0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_B0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 5, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_B1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_B1);
                          PyTuple_SET_ITEM(__pyx_t_36, 6, __pyx_kp_u_B1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_B1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 7, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_C0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_C0);
                          PyTuple_SET_ITEM(__pyx_t_36, 8, __pyx_kp_u_C0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_C0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 9, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_C1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_C1);
                          PyTuple_SET_ITEM(__pyx_t_36, 10, __pyx_kp_u_C1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_C1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 11, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_D0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_D0);
                          PyTuple_SET_ITEM(__pyx_t_36, 12, __pyx_kp_u_D0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_D0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 13, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_D1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_D1);
                          PyTuple_SET_ITEM(__pyx_t_36, 14, __pyx_kp_u_D1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_D1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 15, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __pyx_t_37 = __Pyx_PyUnicode_Join(__pyx_t_36, 16, __pyx_t_38, __pyx_t_39); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
                          __pyx_t_36 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_37); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 483, __pyx_L49_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
                        }

                        /* "pyFAI/ext/_distortion.pyx":481
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 *                                 with gil:             # <<<<<<<<<<<<<<
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                        /*finally:*/ {
                          /*normal exit:*/{
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L50;
                          }
                          __pyx_L49_error: {
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L10_error;
                          }
                          __pyx_L50:;
                        }
                    }

                    /* "pyFAI/ext/_distortion.pyx":484
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1             # <<<<<<<<<<<<<<
 *                             continue
 *                         k = outMax[ml, nl]
 */
                    __pyx_v_err_cnt = (__pyx_v_err_cnt + 1);

                    /* "pyFAI/ext/_distortion.pyx":480
 *                         if value < 0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                  }

                  /* "pyFAI/ext/_distortion.pyx":485
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1
 *                             continue             # <<<<<<<<<<<<<<
 *                         k = outMax[ml, nl]
 *                         lut[ml, nl, k].idx = idx
 */
                  goto __pyx_L36_continue;

                  /* "pyFAI/ext/_distortion.pyx":478
 *                         if value == 0:
 *                             continue
 *                         if value < 0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 */
                }

                /* "pyFAI/ext/_distortion.pyx":486
 *                                 err_cnt += 1
 *                             continue
 *                         k = outMax[ml, nl]             # <<<<<<<<<<<<<<
 *                         lut[ml, nl, k].idx = idx
 *                         lut[ml, nl, k].coef = value
 */
                __pyx_t_14 = __pyx_v_ml;
                __pyx_t_15 = __pyx_v_nl;
                __pyx_v_k = (*((int32_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_14 * __pyx_v_outMax.strides[0]) ) + __pyx_t_15 * __pyx_v_outMax.strides[1]) )));

                /* "pyFAI/ext/_distortion.pyx":487
 *                             continue
 *                         k = outMax[ml, nl]
 *                         lut[ml, nl, k].idx = idx             # <<<<<<<<<<<<<<
 *                         lut[ml, nl, k].coef = value
 *                         outMax[ml, nl] = k + 1
 */
                __pyx_t_15 = __pyx_v_ml;
                __pyx_t_14 = __pyx_v_nl;
                __pyx_t_13 = __pyx_v_k;
                (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_15 * __pyx_v_lut.strides[0]) ) + __pyx_t_14 * __pyx_v_lut.strides[1]) ) + __pyx_t_13 * __pyx_v_lut.strides[2]) ))).idx = __pyx_v_idx;

                /* "pyFAI/ext/_distortion.pyx":488
 *                         k = outMax[ml, nl]
 *                         lut[ml, nl, k].idx = idx
 *                         lut[ml, nl, k].coef = value             # <<<<<<<<<<<<<<
 *                         outMax[ml, nl] = k + 1
 *                 idx += 1
 */
                __pyx_t_13 = __pyx_v_ml;
                __pyx_t_14 = __pyx_v_nl;
                __pyx_t_15 = __pyx_v_k;
                (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_13 * __pyx_v_lut.strides[0]) ) + __pyx_t_14 * __pyx_v_lut.strides[1]) ) + __pyx_t_15 * __pyx_v_lut.strides[2]) ))).coef = __pyx_v_value;

                /* "pyFAI/ext/_distortion.pyx":489
 *                         lut[ml, nl, k].idx = idx
 *                         lut[ml, nl, k].coef = value
 *                         outMax[ml, nl] = k + 1             # <<<<<<<<<<<<<<
 *                 idx += 1
 * 
 */
                __pyx_t_15 = __pyx_v_ml;
                __pyx_t_14 = __pyx_v_nl;
                *((int32_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_15 * __pyx_v_outMax.strides[0]) ) + __pyx_t_14 * __pyx_v_outMax.strides[1]) )) = (__pyx_v_k + 1);
                __pyx_L36_continue:;
              }
              __pyx_L31_continue:;
            }

            /* "pyFAI/ext/_distortion.pyx":490
 *                         lut[ml, nl, k].coef = value
 *                         outMax[ml, nl] = k + 1
 *                 idx += 1             # <<<<<<<<<<<<<<
 * 
 *     # Hack to prevent memory leak !!!
 */
            __pyx_v_idx = (__pyx_v_idx + 1);
            __pyx_L14_continue:;
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":399
 *     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], lut_total_size / 1.0e6))
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (delta1, delta0, size))
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i,j, idx are indexes of the raw image uncorrected
 *         for i in range(shape0):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L11;
        }
        __pyx_L10_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L11:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":493
 * 
 *     # Hack to prevent memory leak !!!
 *     cdef float64_t[:, ::1] tmp_ary = numpy.empty(shape=(shape0 * shape1, size), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 *     memcpy(&tmp_ary[0, 0], &lut[0, 0, 0], tmp_ary.nbytes)
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_36, __pyx_n_s_numpy); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_empty); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_35 = __Pyx_PyInt_From_int((__pyx_v_shape0 * __pyx_v_shape1)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_34 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_34);
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_shape, __pyx_t_33) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_33, __pyx_n_s_numpy); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = __Pyx_PyObject_GetAttrStr(__pyx_t_33, __pyx_n_s_float64); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_dtype, __pyx_t_34) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_PyObject_Call(__pyx_t_37, __pyx_empty_tuple, __pyx_t_36); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_40 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(__pyx_t_34, PyBUF_WRITABLE); if (unlikely(!__pyx_t_40.memview)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_v_tmp_ary = __pyx_t_40;
  __pyx_t_40.memview = NULL;
  __pyx_t_40.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":494
 *     # Hack to prevent memory leak !!!
 *     cdef float64_t[:, ::1] tmp_ary = numpy.empty(shape=(shape0 * shape1, size), dtype=numpy.float64)
 *     memcpy(&tmp_ary[0, 0], &lut[0, 0, 0], tmp_ary.nbytes)             # <<<<<<<<<<<<<<
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),
 *                                     shape=(shape0 * shape1, size), dtype=lut_d,
 */
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_13 = 0;
  __pyx_t_24 = 0;
  __pyx_t_41 = 0;
  __pyx_t_34 = __pyx_memoryview_fromslice(__pyx_v_tmp_ary, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t, 0);; if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_36 = __Pyx_PyObject_GetAttrStr(__pyx_t_34, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_size_t(__pyx_t_36); if (unlikely((__pyx_t_16 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  (void)(memcpy((&(*((__pyx_t_5pyFAI_3ext_12shared_types_float64_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float64_t *) ( /* dim=0 */ (__pyx_v_tmp_ary.data + __pyx_t_14 * __pyx_v_tmp_ary.strides[0]) )) + __pyx_t_15)) )))), (&(*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_13 * __pyx_v_lut.strides[0]) ) + __pyx_t_24 * __pyx_v_lut.strides[1]) ) + __pyx_t_41 * __pyx_v_lut.strides[2]) )))), __pyx_t_16));

  /* "pyFAI/ext/_distortion.pyx":495
 *     cdef float64_t[:, ::1] tmp_ary = numpy.empty(shape=(shape0 * shape1, size), dtype=numpy.float64)
 *     memcpy(&tmp_ary[0, 0], &lut[0, 0, 0], tmp_ary.nbytes)
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),             # <<<<<<<<<<<<<<
 *                                     shape=(shape0 * shape1, size), dtype=lut_d,
 *                                     copy=True)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_36, __pyx_n_s_numpy); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_34 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_core); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_PyObject_GetAttrStr(__pyx_t_34, __pyx_n_s_records); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_array); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_37, __pyx_n_s_numpy); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_33 = __Pyx_PyObject_GetAttrStr(__pyx_t_37, __pyx_n_s_asarray); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_memoryview_fromslice(__pyx_v_tmp_ary, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t, 0);; if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_35 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_33))) {
    __pyx_t_35 = PyMethod_GET_SELF(__pyx_t_33);
    if (likely(__pyx_t_35)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_33);
      __Pyx_INCREF(__pyx_t_35);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_33, function);
    }
  }
  __pyx_t_36 = (__pyx_t_35) ? __Pyx_PyObject_Call2Args(__pyx_t_33, __pyx_t_35, __pyx_t_37) : __Pyx_PyObject_CallOneArg(__pyx_t_33, __pyx_t_37);
  __Pyx_XDECREF(__pyx_t_35); __pyx_t_35 = 0;
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_PyObject_GetAttrStr(__pyx_t_36, __pyx_n_s_view); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_GetModuleGlobalName(__pyx_t_37, __pyx_n_s_lut_d); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_dtype, __pyx_t_37) < 0) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __Pyx_PyObject_Call(__pyx_t_33, __pyx_empty_tuple, __pyx_t_36); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_t_37);
  __pyx_t_37 = 0;

  /* "pyFAI/ext/_distortion.pyx":496
 *     memcpy(&tmp_ary[0, 0], &lut[0, 0, 0], tmp_ary.nbytes)
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),
 *                                     shape=(shape0 * shape1, size), dtype=lut_d,             # <<<<<<<<<<<<<<
 *                                     copy=True)
 * 
 */
  __pyx_t_37 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_33 = __Pyx_PyInt_From_int((__pyx_v_shape0 * __pyx_v_shape1)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_35 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_35);
  __pyx_t_33 = 0;
  __pyx_t_35 = 0;
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_shape, __pyx_t_32) < 0) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_lut_d); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_dtype, __pyx_t_32) < 0) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pyFAI/ext/_distortion.pyx":497
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),
 *                                     shape=(shape0 * shape1, size), dtype=lut_d,
 *                                     copy=True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_copy, Py_True) < 0) __PYX_ERR(0, 496, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":495
 *     cdef float64_t[:, ::1] tmp_ary = numpy.empty(shape=(shape0 * shape1, size), dtype=numpy.float64)
 *     memcpy(&tmp_ary[0, 0], &lut[0, 0, 0], tmp_ary.nbytes)
 *     return numpy.core.records.array(numpy.asarray(tmp_ary).view(dtype=lut_d),             # <<<<<<<<<<<<<<
 *                                     shape=(shape0 * shape1, size), dtype=lut_d,
 *                                     copy=True)
 */
  __pyx_t_32 = __Pyx_PyObject_Call(__pyx_t_34, __pyx_t_36, __pyx_t_37); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_r = __pyx_t_32;
  __pyx_t_32 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":362
 * 
 * 
 * def calc_LUT(float32_t[:, :, :, ::1] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, :] mask=None):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __PYX_XDEC_MEMVIEW(&__pyx_t_40, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_outMax, 1);
  __Pyx_XDECREF(__pyx_v_lut_total_size);
  __PYX_XDEC_MEMVIEW(&__pyx_v_tmp_ary, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":500
 * 
 * 
 * def calc_CSR(float32_t[:, :, :, :] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, ::1] mask=None):
 *     """Calculate the Look-up table as CSR format
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_11calc_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10calc_CSR[] = "calc_CSR(float32_t[:, :, :, :] pos, shape, bin_size, max_pixel_size, int8_t[:, ::1] mask=None)\nCalculate the Look-up table as CSR format\n\n    :param pos: 4D position array\n    :param shape: output shape\n    :param bin_size: number of input element per output element (as numpy array)\n    :param max_pixel_size: (2-tuple of int) size of a buffer covering the largest pixel\n    :return: look-up table in CSR format: 3-tuple of array";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_11calc_CSR = {"calc_CSR", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_11calc_CSR, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_10calc_CSR};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_11calc_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_bin_size = 0;
  PyObject *__pyx_v_max_pixel_size = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_CSR (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_bin_size,&__pyx_n_s_max_pixel_size,&__pyx_n_s_mask,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_CSR", 0, 4, 5, 1); __PYX_ERR(0, 500, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bin_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_CSR", 0, 4, 5, 2); __PYX_ERR(0, 500, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_pixel_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_CSR", 0, 4, 5, 3); __PYX_ERR(0, 500, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_CSR") < 0)) __PYX_ERR(0, 500, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_shape = values[1];
    __pyx_v_bin_size = values[2];
    __pyx_v_max_pixel_size = values[3];
    if (values[4]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 501, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_k__14;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_CSR", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 500, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 500, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10calc_CSR(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_bin_size, __pyx_v_max_pixel_size, __pyx_v_mask);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10calc_CSR(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_bin_size, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask) {
  int __pyx_v_shape0;
  int __pyx_v_shape1;
  int __pyx_v_delta0;
  int __pyx_v_delta1;
  int __pyx_v_bins;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ms;
  int __pyx_v_ml;
  int __pyx_v_ns;
  int __pyx_v_nl;
  int __pyx_v_idx;
  int __pyx_v_tmp_index;
  int __pyx_v_err_cnt;
  int __pyx_v_lut_size;
  int __pyx_v_offset0;
  int __pyx_v_offset1;
  int __pyx_v_box_size0;
  int __pyx_v_box_size1;
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_pAB;
  float __pyx_v_pBC;
  float __pyx_v_pCD;
  float __pyx_v_pDA;
  float __pyx_v_cAB;
  float __pyx_v_cBC;
  float __pyx_v_cCD;
  float __pyx_v_cDA;
  float __pyx_v_area;
  float __pyx_v_value;
  float __pyx_v_foffset0;
  float __pyx_v_foffset1;
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_outMax = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_buffer = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  Py_ssize_t __pyx_t_38;
  Py_UCS4 __pyx_t_39;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_CSR", 0);

  /* "pyFAI/ext/_distortion.pyx":511
 *     cdef:
 *         int shape0, shape1, delta0, delta1, bins
 *     shape0, shape1 = shape             # <<<<<<<<<<<<<<
 *     delta0, delta1 = max_pixel_size
 *     bins = shape0 * shape1
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 511, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 511, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_shape0 = __pyx_t_5;
  __pyx_v_shape1 = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":512
 *         int shape0, shape1, delta0, delta1, bins
 *     shape0, shape1 = shape
 *     delta0, delta1 = max_pixel_size             # <<<<<<<<<<<<<<
 *     bins = shape0 * shape1
 *     cdef:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_max_pixel_size))) || (PyList_CheckExact(__pyx_v_max_pixel_size))) {
    PyObject* sequence = __pyx_v_max_pixel_size;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 512, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_max_pixel_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 512, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_delta0 = __pyx_t_6;
  __pyx_v_delta1 = __pyx_t_5;

  /* "pyFAI/ext/_distortion.pyx":513
 *     shape0, shape1 = shape
 *     delta0, delta1 = max_pixel_size
 *     bins = shape0 * shape1             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, j, k, ms, ml, ns, nl, idx = 0, tmp_index, err_cnt = 0
 */
  __pyx_v_bins = (__pyx_v_shape0 * __pyx_v_shape1);

  /* "pyFAI/ext/_distortion.pyx":515
 *     bins = shape0 * shape1
 *     cdef:
 *         int i, j, k, ms, ml, ns, nl, idx = 0, tmp_index, err_cnt = 0             # <<<<<<<<<<<<<<
 *         int lut_size, offset0, offset1, box_size0, box_size1
 *         float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA,
 */
  __pyx_v_idx = 0;
  __pyx_v_err_cnt = 0;

  /* "pyFAI/ext/_distortion.pyx":521
 *         int32_t[::1] indptr, indices
 *         float32_t[::1] data
 *         int32_t[:, ::1] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_shape0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_shape1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_outMax = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":523
 *         int32_t[:, ::1] outMax = numpy.zeros((shape0, shape1), dtype=numpy.int32)
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":524
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 */
  __pyx_t_10 = (__pyx_v_do_mask != 0);
  if (__pyx_t_10) {

    /* "pyFAI/ext/_distortion.pyx":525
 *         bint do_mask = mask is not None
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"             # <<<<<<<<<<<<<<
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape0 == (__pyx_v_mask.shape[0])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_mask_shape_dim0);
        __PYX_ERR(0, 525, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":526
 *     if do_mask:
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"             # <<<<<<<<<<<<<<
 * 
 *     indptr = numpy.concatenate(([numpy.int32(0)], bin_size.cumsum(dtype=numpy.int32)))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape1 == (__pyx_v_mask.shape[1])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_mask_shape_dim1);
        __PYX_ERR(0, 526, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":524
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape0 == mask.shape[0], "mask shape dim0"
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":528
 *         assert shape1 == mask.shape[1], "mask shape dim1"
 * 
 *     indptr = numpy.concatenate(([numpy.int32(0)], bin_size.cumsum(dtype=numpy.int32)))             # <<<<<<<<<<<<<<
 *     lut_size = indptr[bins]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_int_0) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_0);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_bin_size, __pyx_n_s_cumsum); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_numpy); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_int32); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_12);
  __pyx_t_1 = 0;
  __pyx_t_12 = 0;
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_12, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_13.memview)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_indptr = __pyx_t_13;
  __pyx_t_13.memview = NULL;
  __pyx_t_13.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":529
 * 
 *     indptr = numpy.concatenate(([numpy.int32(0)], bin_size.cumsum(dtype=numpy.int32)))
 *     lut_size = indptr[bins]             # <<<<<<<<<<<<<<
 * 
 *     indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)
 */
  __pyx_t_14 = __pyx_v_bins;
  __pyx_v_lut_size = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_14)) )));

  /* "pyFAI/ext/_distortion.pyx":531
 *     lut_size = indptr[bins]
 * 
 *     indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *     data = numpy.zeros(shape=lut_size, dtype=numpy.float32)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_13 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_12, PyBUF_WRITABLE); if (unlikely(!__pyx_t_13.memview)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_v_indices = __pyx_t_13;
  __pyx_t_13.memview = NULL;
  __pyx_t_13.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":532
 * 
 *     indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)
 *     data = numpy.zeros(shape=lut_size, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *     logger.info("CSR matrix: %.3f MByte" % ((indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_15 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_15.memview)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_data = __pyx_t_15;
  __pyx_t_15.memview = NULL;
  __pyx_t_15.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":534
 *     data = numpy.zeros(shape=lut_size, dtype=numpy.float32)
 * 
 *     logger.info("CSR matrix: %.3f MByte" % ((indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6))             # <<<<<<<<<<<<<<
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], lut_size))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_logger); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_info); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __pyx_memoryview_fromslice(__pyx_v_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_12, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_1_0e6, 1.0e6, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_CSR_matrix_3f_MByte, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyFAI/ext/_distortion.pyx":535
 * 
 *     logger.info("CSR matrix: %.3f MByte" % ((indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6))
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], lut_size))
 *     with nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_12, PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_v_buffer = __pyx_t_16;
  __pyx_t_16.memview = NULL;
  __pyx_t_16.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":536
 *     logger.info("CSR matrix: %.3f MByte" % ((indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6))
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], lut_size))             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # i,j, idx are indices of the raw image uncorrected
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_buffer.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = PyInt_FromSsize_t((__pyx_v_buffer.shape[0])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_Max_pixel_size_ix_i_Max_source_p, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_12 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pyFAI/ext/_distortion.pyx":537
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], lut_size))
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i,j, idx are indices of the raw image uncorrected
 *         for i in range(shape0):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":539
 *     with nogil:
 *         # i,j, idx are indices of the raw image uncorrected
 *         for i in range(shape0):             # <<<<<<<<<<<<<<
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:
 */
        __pyx_t_5 = __pyx_v_shape0;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_6; __pyx_t_17+=1) {
          __pyx_v_i = __pyx_t_17;

          /* "pyFAI/ext/_distortion.pyx":540
 *         # i,j, idx are indices of the raw image uncorrected
 *         for i in range(shape0):
 *             for j in range(shape1):             # <<<<<<<<<<<<<<
 *                 if do_mask and mask[i, j]:
 *                     continue
 */
          __pyx_t_18 = __pyx_v_shape1;
          __pyx_t_19 = __pyx_t_18;
          for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
            __pyx_v_j = __pyx_t_20;

            /* "pyFAI/ext/_distortion.pyx":541
 *         for i in range(shape0):
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # reinit of buffer
 */
            __pyx_t_21 = (__pyx_v_do_mask != 0);
            if (__pyx_t_21) {
            } else {
              __pyx_t_10 = __pyx_t_21;
              goto __pyx_L16_bool_binop_done;
            }
            __pyx_t_14 = __pyx_v_i;
            __pyx_t_22 = __pyx_v_j;
            __pyx_t_21 = ((*((int8_t *) ( /* dim=1 */ ((char *) (((int8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_14 * __pyx_v_mask.strides[0]) )) + __pyx_t_22)) ))) != 0);
            __pyx_t_10 = __pyx_t_21;
            __pyx_L16_bool_binop_done:;
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":542
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:
 *                     continue             # <<<<<<<<<<<<<<
 *                 # reinit of buffer
 *                 buffer[:, :] = 0
 */
              goto __pyx_L13_continue;

              /* "pyFAI/ext/_distortion.pyx":541
 *         for i in range(shape0):
 *             for j in range(shape1):
 *                 if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # reinit of buffer
 */
            }

            /* "pyFAI/ext/_distortion.pyx":544
 *                     continue
 *                 # reinit of buffer
 *                 buffer[:, :] = 0             # <<<<<<<<<<<<<<
 *                 A0 = pos[i, j, 0, 0]
 *                 A1 = pos[i, j, 0, 1]
 */
            {
                __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_temp_scalar = 0.0;
                {
                    Py_ssize_t __pyx_temp_extent = __pyx_v_buffer.shape[0] * __pyx_v_buffer.shape[1];
                    Py_ssize_t __pyx_temp_idx;
                    __pyx_t_5pyFAI_3ext_12shared_types_float32_t *__pyx_temp_pointer = (__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_buffer.data;
                    for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
                      *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
                      __pyx_temp_pointer += 1;
                    }
                }
            }

            /* "pyFAI/ext/_distortion.pyx":545
 *                 # reinit of buffer
 *                 buffer[:, :] = 0
 *                 A0 = pos[i, j, 0, 0]             # <<<<<<<<<<<<<<
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]
 */
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_j;
            __pyx_t_23 = 0;
            __pyx_t_24 = 0;
            __pyx_v_A0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_14 * __pyx_v_pos.strides[1]) ) + __pyx_t_23 * __pyx_v_pos.strides[2]) ) + __pyx_t_24 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":546
 *                 buffer[:, :] = 0
 *                 A0 = pos[i, j, 0, 0]
 *                 A1 = pos[i, j, 0, 1]             # <<<<<<<<<<<<<<
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_14 = 0;
            __pyx_t_22 = 1;
            __pyx_v_A1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_23 * __pyx_v_pos.strides[1]) ) + __pyx_t_14 * __pyx_v_pos.strides[2]) ) + __pyx_t_22 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":547
 *                 A0 = pos[i, j, 0, 0]
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]             # <<<<<<<<<<<<<<
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]
 */
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_j;
            __pyx_t_23 = 1;
            __pyx_t_24 = 0;
            __pyx_v_B0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_14 * __pyx_v_pos.strides[1]) ) + __pyx_t_23 * __pyx_v_pos.strides[2]) ) + __pyx_t_24 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":548
 *                 A1 = pos[i, j, 0, 1]
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]             # <<<<<<<<<<<<<<
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_14 = 1;
            __pyx_t_22 = 1;
            __pyx_v_B1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_23 * __pyx_v_pos.strides[1]) ) + __pyx_t_14 * __pyx_v_pos.strides[2]) ) + __pyx_t_22 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":549
 *                 B0 = pos[i, j, 1, 0]
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]             # <<<<<<<<<<<<<<
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]
 */
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_j;
            __pyx_t_23 = 2;
            __pyx_t_24 = 0;
            __pyx_v_C0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_14 * __pyx_v_pos.strides[1]) ) + __pyx_t_23 * __pyx_v_pos.strides[2]) ) + __pyx_t_24 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":550
 *                 B1 = pos[i, j, 1, 1]
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]             # <<<<<<<<<<<<<<
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_14 = 2;
            __pyx_t_22 = 1;
            __pyx_v_C1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_23 * __pyx_v_pos.strides[1]) ) + __pyx_t_14 * __pyx_v_pos.strides[2]) ) + __pyx_t_22 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":551
 *                 C0 = pos[i, j, 2, 0]
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]             # <<<<<<<<<<<<<<
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 */
            __pyx_t_22 = __pyx_v_i;
            __pyx_t_14 = __pyx_v_j;
            __pyx_t_23 = 3;
            __pyx_t_24 = 0;
            __pyx_v_D0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_22 * __pyx_v_pos.strides[0]) ) + __pyx_t_14 * __pyx_v_pos.strides[1]) ) + __pyx_t_23 * __pyx_v_pos.strides[2]) ) + __pyx_t_24 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":552
 *                 C1 = pos[i, j, 2, 1]
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]             # <<<<<<<<<<<<<<
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 */
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_23 = __pyx_v_j;
            __pyx_t_14 = 3;
            __pyx_t_22 = 1;
            __pyx_v_D1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_24 * __pyx_v_pos.strides[0]) ) + __pyx_t_23 * __pyx_v_pos.strides[1]) ) + __pyx_t_14 * __pyx_v_pos.strides[2]) ) + __pyx_t_22 * __pyx_v_pos.strides[3]) )));

            /* "pyFAI/ext/_distortion.pyx":553
 *                 D0 = pos[i, j, 3, 0]
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)             # <<<<<<<<<<<<<<
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)
 */
            __pyx_v_foffset0 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0);

            /* "pyFAI/ext/_distortion.pyx":554
 *                 D1 = pos[i, j, 3, 1]
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)             # <<<<<<<<<<<<<<
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)
 */
            __pyx_v_foffset1 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1);

            /* "pyFAI/ext/_distortion.pyx":555
 *                 foffset0 = _floor_min4(A0, B0, C0, D0)
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)             # <<<<<<<<<<<<<<
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 */
            __pyx_v_offset0 = ((int)__pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":556
 *                 foffset1 = _floor_min4(A1, B1, C1, D1)
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)             # <<<<<<<<<<<<<<
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 */
            __pyx_v_offset1 = ((int)__pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":557
 *                 offset0 = (<int> foffset0)
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0             # <<<<<<<<<<<<<<
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 * 
 */
            __pyx_v_box_size0 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) - __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":558
 *                 offset1 = (<int> foffset1)
 *                 box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1             # <<<<<<<<<<<<<<
 * 
 *                 if (box_size0 > delta0) or (box_size1 > delta1):
 */
            __pyx_v_box_size1 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) - __pyx_v_offset1);

            /* "pyFAI/ext/_distortion.pyx":560
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 * 
 *                 if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 */
            __pyx_t_21 = ((__pyx_v_box_size0 > __pyx_v_delta0) != 0);
            if (!__pyx_t_21) {
            } else {
              __pyx_t_10 = __pyx_t_21;
              goto __pyx_L19_bool_binop_done;
            }
            __pyx_t_21 = ((__pyx_v_box_size1 > __pyx_v_delta1) != 0);
            __pyx_t_10 = __pyx_t_21;
            __pyx_L19_bool_binop_done:;
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":562
 *                 if (box_size0 > delta0) or (box_size1 > delta1):
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0             # <<<<<<<<<<<<<<
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:
 */
              if (((__pyx_v_offset0 > __pyx_v_delta0) != 0)) {
                __pyx_t_25 = __pyx_v_offset0;
              } else {
                __pyx_t_25 = __pyx_v_delta0;
              }
              __pyx_v_delta0 = __pyx_t_25;

              /* "pyFAI/ext/_distortion.pyx":563
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1             # <<<<<<<<<<<<<<
 *                     with gil:
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 */
              if (((__pyx_v_offset1 > __pyx_v_delta1) != 0)) {
                __pyx_t_25 = __pyx_v_offset1;
              } else {
                __pyx_t_25 = __pyx_v_delta1;
              }
              __pyx_v_delta1 = __pyx_t_25;

              /* "pyFAI/ext/_distortion.pyx":564
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
              {
                  #ifdef WITH_THREAD
                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                  #endif
                  /*try:*/ {

                    /* "pyFAI/ext/_distortion.pyx":565
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *                 A0 -= foffset0
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_7);
                    __Pyx_GIVEREF(__pyx_t_12);
                    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_12);
                    __Pyx_GIVEREF(__pyx_t_2);
                    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
                    __pyx_t_12 = 0;
                    __pyx_t_2 = 0;
                    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __Pyx_GIVEREF(__pyx_t_7);
                    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
                    __pyx_t_7 = 0;
                    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_7);
                    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                    __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 565, __pyx_L24_error)
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
                    __pyx_v_buffer = __pyx_t_16;
                    __pyx_t_16.memview = NULL;
                    __pyx_t_16.data = NULL;
                  }

                  /* "pyFAI/ext/_distortion.pyx":564
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 *                     delta1 = offset1 if offset1 > delta1 else delta1
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      goto __pyx_L25;
                    }
                    __pyx_L24_error: {
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      goto __pyx_L9_error;
                    }
                    __pyx_L25:;
                  }
              }

              /* "pyFAI/ext/_distortion.pyx":560
 *                 box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 * 
 *                 if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                     # Increase size of the buffer
 *                     delta0 = offset0 if offset0 > delta0 else delta0
 */
            }

            /* "pyFAI/ext/_distortion.pyx":567
 *                         buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 *                 A0 -= foffset0             # <<<<<<<<<<<<<<
 *                 A1 -= foffset1
 *                 B0 -= foffset0
 */
            __pyx_v_A0 = (__pyx_v_A0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":568
 * 
 *                 A0 -= foffset0
 *                 A1 -= foffset1             # <<<<<<<<<<<<<<
 *                 B0 -= foffset0
 *                 B1 -= foffset1
 */
            __pyx_v_A1 = (__pyx_v_A1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":569
 *                 A0 -= foffset0
 *                 A1 -= foffset1
 *                 B0 -= foffset0             # <<<<<<<<<<<<<<
 *                 B1 -= foffset1
 *                 C0 -= foffset0
 */
            __pyx_v_B0 = (__pyx_v_B0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":570
 *                 A1 -= foffset1
 *                 B0 -= foffset0
 *                 B1 -= foffset1             # <<<<<<<<<<<<<<
 *                 C0 -= foffset0
 *                 C1 -= foffset1
 */
            __pyx_v_B1 = (__pyx_v_B1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":571
 *                 B0 -= foffset0
 *                 B1 -= foffset1
 *                 C0 -= foffset0             # <<<<<<<<<<<<<<
 *                 C1 -= foffset1
 *                 D0 -= foffset0
 */
            __pyx_v_C0 = (__pyx_v_C0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":572
 *                 B1 -= foffset1
 *                 C0 -= foffset0
 *                 C1 -= foffset1             # <<<<<<<<<<<<<<
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 */
            __pyx_v_C1 = (__pyx_v_C1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":573
 *                 C0 -= foffset0
 *                 C1 -= foffset1
 *                 D0 -= foffset0             # <<<<<<<<<<<<<<
 *                 D1 -= foffset1
 *                 if B0 != A0:
 */
            __pyx_v_D0 = (__pyx_v_D0 - __pyx_v_foffset0);

            /* "pyFAI/ext/_distortion.pyx":574
 *                 C1 -= foffset1
 *                 D0 -= foffset0
 *                 D1 -= foffset1             # <<<<<<<<<<<<<<
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)
 */
            __pyx_v_D1 = (__pyx_v_D1 - __pyx_v_foffset1);

            /* "pyFAI/ext/_distortion.pyx":575
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 *                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0
 */
            __pyx_t_10 = ((__pyx_v_B0 != __pyx_v_A0) != 0);
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":576
 *                 D1 -= foffset1
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)             # <<<<<<<<<<<<<<
 *                     cAB = A1 - pAB * A0
 *                 else:
 */
              __pyx_v_pAB = ((__pyx_v_B1 - __pyx_v_A1) / (__pyx_v_B0 - __pyx_v_A0));

              /* "pyFAI/ext/_distortion.pyx":577
 *                 if B0 != A0:
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pAB = cAB = 0.0
 */
              __pyx_v_cAB = (__pyx_v_A1 - (__pyx_v_pAB * __pyx_v_A0));

              /* "pyFAI/ext/_distortion.pyx":575
 *                 D0 -= foffset0
 *                 D1 -= foffset1
 *                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                     pAB = (B1 - A1) / (B0 - A0)
 *                     cAB = A1 - pAB * A0
 */
              goto __pyx_L26;
            }

            /* "pyFAI/ext/_distortion.pyx":579
 *                     cAB = A1 - pAB * A0
 *                 else:
 *                     pAB = cAB = 0.0             # <<<<<<<<<<<<<<
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)
 */
            /*else*/ {
              __pyx_v_pAB = 0.0;
              __pyx_v_cAB = 0.0;
            }
            __pyx_L26:;

            /* "pyFAI/ext/_distortion.pyx":580
 *                 else:
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0
 */
            __pyx_t_10 = ((__pyx_v_C0 != __pyx_v_B0) != 0);
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":581
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)             # <<<<<<<<<<<<<<
 *                     cBC = B1 - pBC * B0
 *                 else:
 */
              __pyx_v_pBC = ((__pyx_v_C1 - __pyx_v_B1) / (__pyx_v_C0 - __pyx_v_B0));

              /* "pyFAI/ext/_distortion.pyx":582
 *                 if C0 != B0:
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pBC = cBC = 0.0
 */
              __pyx_v_cBC = (__pyx_v_B1 - (__pyx_v_pBC * __pyx_v_B0));

              /* "pyFAI/ext/_distortion.pyx":580
 *                 else:
 *                     pAB = cAB = 0.0
 *                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                     pBC = (C1 - B1) / (C0 - B0)
 *                     cBC = B1 - pBC * B0
 */
              goto __pyx_L27;
            }

            /* "pyFAI/ext/_distortion.pyx":584
 *                     cBC = B1 - pBC * B0
 *                 else:
 *                     pBC = cBC = 0.0             # <<<<<<<<<<<<<<
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)
 */
            /*else*/ {
              __pyx_v_pBC = 0.0;
              __pyx_v_cBC = 0.0;
            }
            __pyx_L27:;

            /* "pyFAI/ext/_distortion.pyx":585
 *                 else:
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0
 */
            __pyx_t_10 = ((__pyx_v_D0 != __pyx_v_C0) != 0);
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":586
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)             # <<<<<<<<<<<<<<
 *                     cCD = C1 - pCD * C0
 *                 else:
 */
              __pyx_v_pCD = ((__pyx_v_D1 - __pyx_v_C1) / (__pyx_v_D0 - __pyx_v_C0));

              /* "pyFAI/ext/_distortion.pyx":587
 *                 if D0 != C0:
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pCD = cCD = 0.0
 */
              __pyx_v_cCD = (__pyx_v_C1 - (__pyx_v_pCD * __pyx_v_C0));

              /* "pyFAI/ext/_distortion.pyx":585
 *                 else:
 *                     pBC = cBC = 0.0
 *                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                     pCD = (D1 - C1) / (D0 - C0)
 *                     cCD = C1 - pCD * C0
 */
              goto __pyx_L28;
            }

            /* "pyFAI/ext/_distortion.pyx":589
 *                     cCD = C1 - pCD * C0
 *                 else:
 *                     pCD = cCD = 0.0             # <<<<<<<<<<<<<<
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)
 */
            /*else*/ {
              __pyx_v_pCD = 0.0;
              __pyx_v_cCD = 0.0;
            }
            __pyx_L28:;

            /* "pyFAI/ext/_distortion.pyx":590
 *                 else:
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0
 */
            __pyx_t_10 = ((__pyx_v_A0 != __pyx_v_D0) != 0);
            if (__pyx_t_10) {

              /* "pyFAI/ext/_distortion.pyx":591
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)             # <<<<<<<<<<<<<<
 *                     cDA = D1 - pDA * D0
 *                 else:
 */
              __pyx_v_pDA = ((__pyx_v_A1 - __pyx_v_D1) / (__pyx_v_A0 - __pyx_v_D0));

              /* "pyFAI/ext/_distortion.pyx":592
 *                 if A0 != D0:
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0             # <<<<<<<<<<<<<<
 *                 else:
 *                     pDA = cDA = 0.0
 */
              __pyx_v_cDA = (__pyx_v_D1 - (__pyx_v_pDA * __pyx_v_D0));

              /* "pyFAI/ext/_distortion.pyx":590
 *                 else:
 *                     pCD = cCD = 0.0
 *                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                     pDA = (A1 - D1) / (A0 - D0)
 *                     cDA = D1 - pDA * D0
 */
              goto __pyx_L29;
            }

            /* "pyFAI/ext/_distortion.pyx":594
 *                     cDA = D1 - pDA * D0
 *                 else:
 *                     pDA = cDA = 0.0             # <<<<<<<<<<<<<<
 *                 integrate(buffer, B0, A0, pAB, cAB)
 *                 integrate(buffer, A0, D0, pDA, cDA)
 */
            /*else*/ {
              __pyx_v_pDA = 0.0;
              __pyx_v_cDA = 0.0;
            }
            __pyx_L29:;

            /* "pyFAI/ext/_distortion.pyx":595
 *                 else:
 *                     pDA = cDA = 0.0
 *                 integrate(buffer, B0, A0, pAB, cAB)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, A0, D0, pDA, cDA)
 *                 integrate(buffer, D0, C0, pCD, cCD)
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_B0, __pyx_v_A0, __pyx_v_pAB, __pyx_v_cAB);

            /* "pyFAI/ext/_distortion.pyx":596
 *                     pDA = cDA = 0.0
 *                 integrate(buffer, B0, A0, pAB, cAB)
 *                 integrate(buffer, A0, D0, pDA, cDA)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, D0, C0, pCD, cCD)
 *                 integrate(buffer, C0, B0, pBC, cBC)
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_A0, __pyx_v_D0, __pyx_v_pDA, __pyx_v_cDA);

            /* "pyFAI/ext/_distortion.pyx":597
 *                 integrate(buffer, B0, A0, pAB, cAB)
 *                 integrate(buffer, A0, D0, pDA, cDA)
 *                 integrate(buffer, D0, C0, pCD, cCD)             # <<<<<<<<<<<<<<
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_D0, __pyx_v_C0, __pyx_v_pCD, __pyx_v_cCD);

            /* "pyFAI/ext/_distortion.pyx":598
 *                 integrate(buffer, A0, D0, pDA, cDA)
 *                 integrate(buffer, D0, C0, pCD, cCD)
 *                 integrate(buffer, C0, B0, pBC, cBC)             # <<<<<<<<<<<<<<
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                 for ms in range(box_size0):
 */
            __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_C0, __pyx_v_B0, __pyx_v_pBC, __pyx_v_cBC);

            /* "pyFAI/ext/_distortion.pyx":599
 *                 integrate(buffer, D0, C0, pCD, cCD)
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))             # <<<<<<<<<<<<<<
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0
 */
            __pyx_v_area = (0.5 * (((__pyx_v_C0 - __pyx_v_A0) * (__pyx_v_D1 - __pyx_v_B1)) - ((__pyx_v_C1 - __pyx_v_A1) * (__pyx_v_D0 - __pyx_v_B0))));

            /* "pyFAI/ext/_distortion.pyx":600
 *                 integrate(buffer, C0, B0, pBC, cBC)
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                 for ms in range(box_size0):             # <<<<<<<<<<<<<<
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:
 */
            __pyx_t_25 = __pyx_v_box_size0;
            __pyx_t_26 = __pyx_t_25;
            for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
              __pyx_v_ms = __pyx_t_27;

              /* "pyFAI/ext/_distortion.pyx":601
 *                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0             # <<<<<<<<<<<<<<
 *                     if ml < 0 or ml >= shape0:
 *                         continue
 */
              __pyx_v_ml = (__pyx_v_ms + __pyx_v_offset0);

              /* "pyFAI/ext/_distortion.pyx":602
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     for ns in range(box_size1):
 */
              __pyx_t_21 = ((__pyx_v_ml < 0) != 0);
              if (!__pyx_t_21) {
              } else {
                __pyx_t_10 = __pyx_t_21;
                goto __pyx_L33_bool_binop_done;
              }
              __pyx_t_21 = ((__pyx_v_ml >= __pyx_v_shape0) != 0);
              __pyx_t_10 = __pyx_t_21;
              __pyx_L33_bool_binop_done:;
              if (__pyx_t_10) {

                /* "pyFAI/ext/_distortion.pyx":603
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     for ns in range(box_size1):
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 */
                goto __pyx_L30_continue;

                /* "pyFAI/ext/_distortion.pyx":602
 *                 for ms in range(box_size0):
 *                     ml = ms + offset0
 *                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     for ns in range(box_size1):
 */
              }

              /* "pyFAI/ext/_distortion.pyx":604
 *                     if ml < 0 or ml >= shape0:
 *                         continue
 *                     for ns in range(box_size1):             # <<<<<<<<<<<<<<
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 */
              __pyx_t_28 = __pyx_v_box_size1;
              __pyx_t_29 = __pyx_t_28;
              for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                __pyx_v_ns = __pyx_t_30;

                /* "pyFAI/ext/_distortion.pyx":606
 *                     for ns in range(box_size1):
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1             # <<<<<<<<<<<<<<
 *                         if nl < 0 or nl >= shape1:
 *                             continue
 */
                __pyx_v_nl = (__pyx_v_ns + __pyx_v_offset1);

                /* "pyFAI/ext/_distortion.pyx":607
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                             continue
 *                         value = buffer[ms, ns] / area
 */
                __pyx_t_21 = ((__pyx_v_nl < 0) != 0);
                if (!__pyx_t_21) {
                } else {
                  __pyx_t_10 = __pyx_t_21;
                  goto __pyx_L38_bool_binop_done;
                }
                __pyx_t_21 = ((__pyx_v_nl >= __pyx_v_shape1) != 0);
                __pyx_t_10 = __pyx_t_21;
                __pyx_L38_bool_binop_done:;
                if (__pyx_t_10) {

                  /* "pyFAI/ext/_distortion.pyx":608
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:
 *                             continue             # <<<<<<<<<<<<<<
 *                         value = buffer[ms, ns] / area
 *                         if value == 0.0:
 */
                  goto __pyx_L35_continue;

                  /* "pyFAI/ext/_distortion.pyx":607
 *                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                         nl = ns + offset1
 *                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                             continue
 *                         value = buffer[ms, ns] / area
 */
                }

                /* "pyFAI/ext/_distortion.pyx":609
 *                         if nl < 0 or nl >= shape1:
 *                             continue
 *                         value = buffer[ms, ns] / area             # <<<<<<<<<<<<<<
 *                         if value == 0.0:
 *                             continue
 */
                __pyx_t_22 = __pyx_v_ms;
                __pyx_t_14 = __pyx_v_ns;
                __pyx_v_value = ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_22 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_area));

                /* "pyFAI/ext/_distortion.pyx":610
 *                             continue
 *                         value = buffer[ms, ns] / area
 *                         if value == 0.0:             # <<<<<<<<<<<<<<
 *                             continue
 *                         if value < 0.0 or value > 1.0001:
 */
                __pyx_t_10 = ((__pyx_v_value == 0.0) != 0);
                if (__pyx_t_10) {

                  /* "pyFAI/ext/_distortion.pyx":611
 *                         value = buffer[ms, ns] / area
 *                         if value == 0.0:
 *                             continue             # <<<<<<<<<<<<<<
 *                         if value < 0.0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 */
                  goto __pyx_L35_continue;

                  /* "pyFAI/ext/_distortion.pyx":610
 *                             continue
 *                         value = buffer[ms, ns] / area
 *                         if value == 0.0:             # <<<<<<<<<<<<<<
 *                             continue
 *                         if value < 0.0 or value > 1.0001:
 */
                }

                /* "pyFAI/ext/_distortion.pyx":612
 *                         if value == 0.0:
 *                             continue
 *                         if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 */
                __pyx_t_21 = ((__pyx_v_value < 0.0) != 0);
                if (!__pyx_t_21) {
                } else {
                  __pyx_t_10 = __pyx_t_21;
                  goto __pyx_L42_bool_binop_done;
                }
                __pyx_t_21 = ((__pyx_v_value > 1.0001) != 0);
                __pyx_t_10 = __pyx_t_21;
                __pyx_L42_bool_binop_done:;
                if (__pyx_t_10) {

                  /* "pyFAI/ext/_distortion.pyx":614
 *                         if value < 0.0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                  __pyx_t_10 = ((__pyx_v_err_cnt < 0x3E8) != 0);
                  if (__pyx_t_10) {

                    /* "pyFAI/ext/_distortion.pyx":615
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 *                                 with gil:             # <<<<<<<<<<<<<<
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                    {
                        #ifdef WITH_THREAD
                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                        #endif
                        /*try:*/ {

                          /* "pyFAI/ext/_distortion.pyx":616
 *                             if err_cnt < 1000:
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])             # <<<<<<<<<<<<<<
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1
 */
                          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_8);
                          __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_7);
                          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_ms); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_2);
                          __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_box_size0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_1);
                          __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_ns); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_12);
                          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_box_size1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_3);
                          __pyx_t_14 = __pyx_v_ms;
                          __pyx_t_22 = __pyx_v_ns;
                          __pyx_t_11 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_22)) )))); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_11);
                          __pyx_t_31 = PyFloat_FromDouble(__pyx_v_area); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_31);
                          __pyx_t_32 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_32);
                          __pyx_t_22 = 0;
                          __pyx_t_14 = 0;
                          __pyx_t_33 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_22 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_33);
                          __pyx_t_14 = 0;
                          __pyx_t_22 = 1;
                          __pyx_t_34 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_22)) )))); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_34);
                          __pyx_t_22 = 1;
                          __pyx_t_14 = 0;
                          __pyx_t_35 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_22 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_14 = 1;
                          __pyx_t_22 = 1;
                          __pyx_t_36 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_22)) )))); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __pyx_t_37 = PyTuple_New(13); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_8);
                          PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_8);
                          __Pyx_GIVEREF(__pyx_t_7);
                          PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_7);
                          __Pyx_GIVEREF(__pyx_t_2);
                          PyTuple_SET_ITEM(__pyx_t_37, 2, __pyx_t_2);
                          __Pyx_GIVEREF(__pyx_t_1);
                          PyTuple_SET_ITEM(__pyx_t_37, 3, __pyx_t_1);
                          __Pyx_GIVEREF(__pyx_t_12);
                          PyTuple_SET_ITEM(__pyx_t_37, 4, __pyx_t_12);
                          __Pyx_GIVEREF(__pyx_t_3);
                          PyTuple_SET_ITEM(__pyx_t_37, 5, __pyx_t_3);
                          __Pyx_GIVEREF(__pyx_t_11);
                          PyTuple_SET_ITEM(__pyx_t_37, 6, __pyx_t_11);
                          __Pyx_GIVEREF(__pyx_t_31);
                          PyTuple_SET_ITEM(__pyx_t_37, 7, __pyx_t_31);
                          __Pyx_GIVEREF(__pyx_t_32);
                          PyTuple_SET_ITEM(__pyx_t_37, 8, __pyx_t_32);
                          __Pyx_GIVEREF(__pyx_t_33);
                          PyTuple_SET_ITEM(__pyx_t_37, 9, __pyx_t_33);
                          __Pyx_GIVEREF(__pyx_t_34);
                          PyTuple_SET_ITEM(__pyx_t_37, 10, __pyx_t_34);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_37, 11, __pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_36);
                          PyTuple_SET_ITEM(__pyx_t_37, 12, __pyx_t_36);
                          __pyx_t_8 = 0;
                          __pyx_t_7 = 0;
                          __pyx_t_2 = 0;
                          __pyx_t_1 = 0;
                          __pyx_t_12 = 0;
                          __pyx_t_3 = 0;
                          __pyx_t_11 = 0;
                          __pyx_t_31 = 0;
                          __pyx_t_32 = 0;
                          __pyx_t_33 = 0;
                          __pyx_t_34 = 0;
                          __pyx_t_35 = 0;
                          __pyx_t_36 = 0;
                          __pyx_t_36 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_37, NULL); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 616, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

                          /* "pyFAI/ext/_distortion.pyx":617
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))             # <<<<<<<<<<<<<<
 *                                 err_cnt += 1
 *                             continue
 */
                          __pyx_t_36 = PyTuple_New(16); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __pyx_t_38 = 0;
                          __pyx_t_39 = 127;
                          __Pyx_INCREF(__pyx_kp_u_A0);
                          __pyx_t_38 += 4;
                          __Pyx_GIVEREF(__pyx_kp_u_A0);
                          PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_kp_u_A0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_A0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_A1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_A1);
                          PyTuple_SET_ITEM(__pyx_t_36, 2, __pyx_kp_u_A1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_A1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 3, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_B0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_B0);
                          PyTuple_SET_ITEM(__pyx_t_36, 4, __pyx_kp_u_B0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_B0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 5, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_B1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_B1);
                          PyTuple_SET_ITEM(__pyx_t_36, 6, __pyx_kp_u_B1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_B1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 7, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_C0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_C0);
                          PyTuple_SET_ITEM(__pyx_t_36, 8, __pyx_kp_u_C0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_C0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 9, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_C1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_C1);
                          PyTuple_SET_ITEM(__pyx_t_36, 10, __pyx_kp_u_C1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_C1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 11, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __Pyx_INCREF(__pyx_kp_u_D0);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_D0);
                          PyTuple_SET_ITEM(__pyx_t_36, 12, __pyx_kp_u_D0);
                          __pyx_t_37 = PyFloat_FromDouble(__pyx_v_D0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __pyx_t_35 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_37), __pyx_empty_unicode); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_35) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_35);
                          __Pyx_GIVEREF(__pyx_t_35);
                          PyTuple_SET_ITEM(__pyx_t_36, 13, __pyx_t_35);
                          __pyx_t_35 = 0;
                          __Pyx_INCREF(__pyx_kp_u_D1);
                          __pyx_t_38 += 5;
                          __Pyx_GIVEREF(__pyx_kp_u_D1);
                          PyTuple_SET_ITEM(__pyx_t_36, 14, __pyx_kp_u_D1);
                          __pyx_t_35 = PyFloat_FromDouble(__pyx_v_D1); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_35);
                          __pyx_t_37 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_35), __pyx_empty_unicode); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
                          __pyx_t_39 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) > __pyx_t_39) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_37) : __pyx_t_39;
                          __pyx_t_38 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_37);
                          __Pyx_GIVEREF(__pyx_t_37);
                          PyTuple_SET_ITEM(__pyx_t_36, 15, __pyx_t_37);
                          __pyx_t_37 = 0;
                          __pyx_t_37 = __Pyx_PyUnicode_Join(__pyx_t_36, 16, __pyx_t_38, __pyx_t_39); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_37);
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
                          __pyx_t_36 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_37); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 617, __pyx_L48_error)
                          __Pyx_GOTREF(__pyx_t_36);
                          __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
                          __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
                        }

                        /* "pyFAI/ext/_distortion.pyx":615
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 *                                 with gil:             # <<<<<<<<<<<<<<
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                        /*finally:*/ {
                          /*normal exit:*/{
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L49;
                          }
                          __pyx_L48_error: {
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L9_error;
                          }
                          __pyx_L49:;
                        }
                    }

                    /* "pyFAI/ext/_distortion.pyx":618
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                    __pyx_v_err_cnt = (__pyx_v_err_cnt + 1);

                    /* "pyFAI/ext/_distortion.pyx":614
 *                         if value < 0.0 or value > 1.0001:
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                                 with gil:
 *                                     print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                  }

                  /* "pyFAI/ext/_distortion.pyx":619
 *                                     print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                                 err_cnt += 1
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         k = outMax[ml, nl]
 */
                  goto __pyx_L35_continue;

                  /* "pyFAI/ext/_distortion.pyx":612
 *                         if value == 0.0:
 *                             continue
 *                         if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                             # here we print pathological cases for debugging
 *                             if err_cnt < 1000:
 */
                }

                /* "pyFAI/ext/_distortion.pyx":621
 *                             continue
 * 
 *                         k = outMax[ml, nl]             # <<<<<<<<<<<<<<
 *                         tmp_index = indptr[ml * shape1 + nl]
 *                         indices[tmp_index + k] = idx
 */
                __pyx_t_22 = __pyx_v_ml;
                __pyx_t_14 = __pyx_v_nl;
                __pyx_v_k = (*((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_22 * __pyx_v_outMax.strides[0]) )) + __pyx_t_14)) )));

                /* "pyFAI/ext/_distortion.pyx":622
 * 
 *                         k = outMax[ml, nl]
 *                         tmp_index = indptr[ml * shape1 + nl]             # <<<<<<<<<<<<<<
 *                         indices[tmp_index + k] = idx
 *                         data[tmp_index + k] = value
 */
                __pyx_t_14 = ((__pyx_v_ml * __pyx_v_shape1) + __pyx_v_nl);
                __pyx_v_tmp_index = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_14)) )));

                /* "pyFAI/ext/_distortion.pyx":623
 *                         k = outMax[ml, nl]
 *                         tmp_index = indptr[ml * shape1 + nl]
 *                         indices[tmp_index + k] = idx             # <<<<<<<<<<<<<<
 *                         data[tmp_index + k] = value
 *                         outMax[ml, nl] = k + 1
 */
                __pyx_t_14 = (__pyx_v_tmp_index + __pyx_v_k);
                *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indices.data) + __pyx_t_14)) )) = __pyx_v_idx;

                /* "pyFAI/ext/_distortion.pyx":624
 *                         tmp_index = indptr[ml * shape1 + nl]
 *                         indices[tmp_index + k] = idx
 *                         data[tmp_index + k] = value             # <<<<<<<<<<<<<<
 *                         outMax[ml, nl] = k + 1
 *                 idx += 1
 */
                __pyx_t_14 = (__pyx_v_tmp_index + __pyx_v_k);
                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_14)) )) = __pyx_v_value;

                /* "pyFAI/ext/_distortion.pyx":625
 *                         indices[tmp_index + k] = idx
 *                         data[tmp_index + k] = value
 *                         outMax[ml, nl] = k + 1             # <<<<<<<<<<<<<<
 *                 idx += 1
 *     return (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 */
                __pyx_t_14 = __pyx_v_ml;
                __pyx_t_22 = __pyx_v_nl;
                *((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_14 * __pyx_v_outMax.strides[0]) )) + __pyx_t_22)) )) = (__pyx_v_k + 1);
                __pyx_L35_continue:;
              }
              __pyx_L30_continue:;
            }

            /* "pyFAI/ext/_distortion.pyx":626
 *                         data[tmp_index + k] = value
 *                         outMax[ml, nl] = k + 1
 *                 idx += 1             # <<<<<<<<<<<<<<
 *     return (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 * 
 */
            __pyx_v_idx = (__pyx_v_idx + 1);
            __pyx_L13_continue:;
          }
        }
      }

      /* "pyFAI/ext/_distortion.pyx":537
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], lut_size))
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i,j, idx are indices of the raw image uncorrected
 *         for i in range(shape0):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L10;
        }
        __pyx_L9_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L10:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":627
 *                         outMax[ml, nl] = k + 1
 *                 idx += 1
 *     return (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_37, __pyx_n_s_numpy); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_35 = __Pyx_PyObject_GetAttrStr(__pyx_t_37, __pyx_n_s_asarray); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_34 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_35))) {
    __pyx_t_34 = PyMethod_GET_SELF(__pyx_t_35);
    if (likely(__pyx_t_34)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_35);
      __Pyx_INCREF(__pyx_t_34);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_35, function);
    }
  }
  __pyx_t_36 = (__pyx_t_34) ? __Pyx_PyObject_Call2Args(__pyx_t_35, __pyx_t_34, __pyx_t_37) : __Pyx_PyObject_CallOneArg(__pyx_t_35, __pyx_t_37);
  __Pyx_XDECREF(__pyx_t_34); __pyx_t_34 = 0;
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_37, __pyx_n_s_numpy); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_34 = __Pyx_PyObject_GetAttrStr(__pyx_t_37, __pyx_n_s_asarray); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_memoryview_fromslice(__pyx_v_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_33 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_34))) {
    __pyx_t_33 = PyMethod_GET_SELF(__pyx_t_34);
    if (likely(__pyx_t_33)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_34);
      __Pyx_INCREF(__pyx_t_33);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_34, function);
    }
  }
  __pyx_t_35 = (__pyx_t_33) ? __Pyx_PyObject_Call2Args(__pyx_t_34, __pyx_t_33, __pyx_t_37) : __Pyx_PyObject_CallOneArg(__pyx_t_34, __pyx_t_37);
  __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_37, __pyx_n_s_numpy); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_33 = __Pyx_PyObject_GetAttrStr(__pyx_t_37, __pyx_n_s_asarray); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_32 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_33))) {
    __pyx_t_32 = PyMethod_GET_SELF(__pyx_t_33);
    if (likely(__pyx_t_32)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_33);
      __Pyx_INCREF(__pyx_t_32);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_33, function);
    }
  }
  __pyx_t_34 = (__pyx_t_32) ? __Pyx_PyObject_Call2Args(__pyx_t_33, __pyx_t_32, __pyx_t_37) : __Pyx_PyObject_CallOneArg(__pyx_t_33, __pyx_t_37);
  __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = PyTuple_New(3); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_33, 2, __pyx_t_34);
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_r = __pyx_t_33;
  __pyx_t_33 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":500
 * 
 * 
 * def calc_CSR(float32_t[:, :, :, :] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, ::1] mask=None):
 *     """Calculate the Look-up table as CSR format
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_outMax, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":630
 * 
 * 
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                 shape,
 *                 max_pixel_size=(8, 8),
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_13calc_sparse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_12calc_sparse[] = "calc_sparse(float32_t[:, :, :, ::1] pos, shape, max_pixel_size=(8, 8), int8_t[:, ::1] mask=None, format=u'csr', int bins_per_pixel=8)\nCalculate the look-up table (or CSR) using OpenMP\n\n    :param pos: 4D position array\n    :param shape: output shape\n    :param max_pixel_size: (2-tuple of int) size of a buffer covering the largest pixel\n    :param format: can be \"CSR\" or \"LUT\"\n    :param bins_per_pixel: average splitting factor (number of pixels per bin)\n    :return: look-up table in CSR/LUT format\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_13calc_sparse = {"calc_sparse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_13calc_sparse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_12calc_sparse};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_13calc_sparse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_max_pixel_size = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_format = 0;
  int __pyx_v_bins_per_pixel;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_sparse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_max_pixel_size,&__pyx_n_s_mask,&__pyx_n_s_format,&__pyx_n_s_bins_per_pixel,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":632
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,
 *                 shape,
 *                 max_pixel_size=(8, 8),             # <<<<<<<<<<<<<<
 *                 int8_t[:, ::1] mask=None,
 *                 format="csr",
 */
    values[2] = ((PyObject *)__pyx_tuple__15);
    values[4] = ((PyObject *)__pyx_n_u_csr);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_sparse", 0, 2, 6, 1); __PYX_ERR(0, 630, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_pixel_size);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bins_per_pixel);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_sparse") < 0)) __PYX_ERR(0, 630, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_shape = values[1];
    __pyx_v_max_pixel_size = values[2];
    if (values[3]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 633, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_k__16;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
    __pyx_v_format = values[4];
    if (values[5]) {
      __pyx_v_bins_per_pixel = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_bins_per_pixel == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L3_error)
    } else {
      __pyx_v_bins_per_pixel = ((int)8);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_sparse", 0, 2, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 630, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 630, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_12calc_sparse(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_max_pixel_size, __pyx_v_mask, __pyx_v_format, __pyx_v_bins_per_pixel);

  /* "pyFAI/ext/_distortion.pyx":630
 * 
 * 
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                 shape,
 *                 max_pixel_size=(8, 8),
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_12calc_sparse(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_format, int __pyx_v_bins_per_pixel) {
  int __pyx_v_shape_in0;
  int __pyx_v_shape_in1;
  int __pyx_v_shape_out0;
  int __pyx_v_shape_out1;
  int __pyx_v_size_in;
  int __pyx_v_delta0;
  int __pyx_v_delta1;
  int __pyx_v_bins;
  int __pyx_v_large_size;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_ms;
  int __pyx_v_ml;
  int __pyx_v_ns;
  int __pyx_v_nl;
  int __pyx_v_lut_size;
  int __pyx_v_offset0;
  int __pyx_v_offset1;
  int __pyx_v_box_size0;
  int __pyx_v_box_size1;
  int __pyx_v_counter;
  int __pyx_v_bin_number;
  int __pyx_v_idx;
  int __pyx_v_err_cnt;
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_pAB;
  float __pyx_v_pBC;
  float __pyx_v_pCD;
  float __pyx_v_pDA;
  float __pyx_v_cAB;
  float __pyx_v_cBC;
  float __pyx_v_cCD;
  float __pyx_v_cDA;
  float __pyx_v_area;
  float __pyx_v_value;
  float __pyx_v_foffset0;
  float __pyx_v_foffset1;
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_idx_bin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_idx_pixel = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_pixel_count = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_large_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_buffer = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_mask;
  __Pyx_memviewslice __pyx_v_lut = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  Py_ssize_t __pyx_t_34;
  Py_UCS4 __pyx_t_35;
  long __pyx_t_36;
  long __pyx_t_37;
  __Pyx_memviewslice __pyx_t_38 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_sparse", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "pyFAI/ext/_distortion.pyx":647
 *     cdef:
 *         int shape_in0, shape_in1, shape_out0, shape_out1, size_in, delta0, delta1, bins, large_size
 *     format = format.lower()             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_lower); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":648
 *         int shape_in0, shape_in1, shape_out0, shape_out1, size_in, delta0, delta1, bins, large_size
 *     format = format.lower()
 *     shape_out0, shape_out1 = shape             # <<<<<<<<<<<<<<
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 648, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 648, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_shape_out0 = __pyx_t_5;
  __pyx_v_shape_out1 = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":649
 *     format = format.lower()
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size             # <<<<<<<<<<<<<<
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_max_pixel_size))) || (PyList_CheckExact(__pyx_v_max_pixel_size))) {
    PyObject* sequence = __pyx_v_max_pixel_size;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 649, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_max_pixel_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 649, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 649, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_delta0 = __pyx_t_6;
  __pyx_v_delta1 = __pyx_t_5;

  /* "pyFAI/ext/_distortion.pyx":650
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1             # <<<<<<<<<<<<<<
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]
 */
  __pyx_v_bins = (__pyx_v_shape_out0 * __pyx_v_shape_out1);

  /* "pyFAI/ext/_distortion.pyx":651
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel             # <<<<<<<<<<<<<<
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]
 */
  __pyx_v_large_size = (__pyx_v_bins * __pyx_v_bins_per_pixel);

  /* "pyFAI/ext/_distortion.pyx":652
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]             # <<<<<<<<<<<<<<
 *     shape_in1 = pos.shape[1]
 *     size_in = shape_in0 * shape_in1
 */
  __pyx_v_shape_in0 = (__pyx_v_pos.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":653
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]             # <<<<<<<<<<<<<<
 *     size_in = shape_in0 * shape_in1
 *     cdef:
 */
  __pyx_v_shape_in1 = (__pyx_v_pos.shape[1]);

  /* "pyFAI/ext/_distortion.pyx":654
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]
 *     size_in = shape_in0 * shape_in1             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, j, ms, ml, ns, nl
 */
  __pyx_v_size_in = (__pyx_v_shape_in0 * __pyx_v_shape_in1);

  /* "pyFAI/ext/_distortion.pyx":659
 *         int lut_size, offset0, offset1, box_size0, box_size1
 *         int counter, bin_number
 *         int idx, err_cnt = 0             # <<<<<<<<<<<<<<
 *         float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA,
 *         float area, value, foffset0, foffset1
 */
  __pyx_v_err_cnt = 0;

  /* "pyFAI/ext/_distortion.pyx":665
 *         float32_t[::1] data, large_data
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 *         lut_t[:, :] lut
 *     if do_mask:
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":667
 *         bint do_mask = mask is not None
 *         lut_t[:, :] lut
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 */
  __pyx_t_7 = (__pyx_v_do_mask != 0);
  if (__pyx_t_7) {

    /* "pyFAI/ext/_distortion.pyx":668
 *         lut_t[:, :] lut
 *     if do_mask:
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"             # <<<<<<<<<<<<<<
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape_in0 == (__pyx_v_mask.shape[0])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_in0_mask_shape_0);
        __PYX_ERR(0, 668, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":669
 *     if do_mask:
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"             # <<<<<<<<<<<<<<
 * 
 *     #  count the number of pixel falling into every single bin
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape_in1 == (__pyx_v_mask.shape[1])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_in1_mask_shape_1);
        __PYX_ERR(0, 669, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":667
 *         bint do_mask = mask is not None
 *         lut_t[:, :] lut
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":672
 * 
 *     #  count the number of pixel falling into every single bin
 *     pixel_count = numpy.zeros(bins, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *     idx_pixel = numpy.zeros(large_size, dtype=numpy.int32)
 *     idx_bin = numpy.zeros(large_size, dtype=numpy.int32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_pixel_count = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":673
 *     #  count the number of pixel falling into every single bin
 *     pixel_count = numpy.zeros(bins, dtype=numpy.int32)
 *     idx_pixel = numpy.zeros(large_size, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *     idx_bin = numpy.zeros(large_size, dtype=numpy.int32)
 *     large_data = numpy.zeros(large_size, dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_large_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_idx_pixel = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":674
 *     pixel_count = numpy.zeros(bins, dtype=numpy.int32)
 *     idx_pixel = numpy.zeros(large_size, dtype=numpy.int32)
 *     idx_bin = numpy.zeros(large_size, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *     large_data = numpy.zeros(large_size, dtype=numpy.float32)
 *     logger.info("Temporary storage: %.3fMB",
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_large_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_idx_bin = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":675
 *     idx_pixel = numpy.zeros(large_size, dtype=numpy.int32)
 *     idx_bin = numpy.zeros(large_size, dtype=numpy.int32)
 *     large_data = numpy.zeros(large_size, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     logger.info("Temporary storage: %.3fMB",
 *                 (large_data.nbytes + pixel_count.nbytes + idx_pixel.nbytes + idx_bin.nbytes) / 1e6)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_large_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_float32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_large_data = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":676
 *     idx_bin = numpy.zeros(large_size, dtype=numpy.int32)
 *     large_data = numpy.zeros(large_size, dtype=numpy.float32)
 *     logger.info("Temporary storage: %.3fMB",             # <<<<<<<<<<<<<<
 *                 (large_data.nbytes + pixel_count.nbytes + idx_pixel.nbytes + idx_bin.nbytes) / 1e6)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyFAI/ext/_distortion.pyx":677
 *     large_data = numpy.zeros(large_size, dtype=numpy.float32)
 *     logger.info("Temporary storage: %.3fMB",
 *                 (large_data.nbytes + pixel_count.nbytes + idx_pixel.nbytes + idx_bin.nbytes) / 1e6)             # <<<<<<<<<<<<<<
 * 
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 */
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_large_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_pixel_count, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_memoryview_fromslice(__pyx_v_idx_pixel, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_idx_bin, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyNumber_Add(__pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_8, __pyx_float_1e6, 1e6, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_kp_u_Temporary_storage_3fMB, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_kp_u_Temporary_storage_3fMB, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_Temporary_storage_3fMB);
    __Pyx_GIVEREF(__pyx_kp_u_Temporary_storage_3fMB);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_5, __pyx_kp_u_Temporary_storage_3fMB);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_5, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":679
 *                 (large_data.nbytes + pixel_count.nbytes + idx_pixel.nbytes + idx_bin.nbytes) / 1e6)
 * 
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     counter = -1  # bin index
 *     with nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_buffer = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":680
 * 
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # i, j, idx are indices of the raw image uncorrected
 */
  __pyx_v_counter = -1;

  /* "pyFAI/ext/_distortion.pyx":681
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":683
 *     with nogil:
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):             # <<<<<<<<<<<<<<
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 */
        __pyx_t_5 = __pyx_v_size_in;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_6; __pyx_t_13+=1) {
          __pyx_v_idx = __pyx_t_13;

          /* "pyFAI/ext/_distortion.pyx":684
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 *             i = idx // shape_in1             # <<<<<<<<<<<<<<
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:
 */
          __pyx_v_i = (__pyx_v_idx / __pyx_v_shape_in1);

          /* "pyFAI/ext/_distortion.pyx":685
 *         for idx in range(size_in):
 *             i = idx // shape_in1
 *             j = idx % shape_in1             # <<<<<<<<<<<<<<
 *             if do_mask and mask[i, j]:
 *                 continue
 */
          __pyx_v_j = (__pyx_v_idx % __pyx_v_shape_in1);

          /* "pyFAI/ext/_distortion.pyx":686
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 */
          __pyx_t_14 = (__pyx_v_do_mask != 0);
          if (__pyx_t_14) {
          } else {
            __pyx_t_7 = __pyx_t_14;
            goto __pyx_L14_bool_binop_done;
          }
          __pyx_t_15 = __pyx_v_i;
          __pyx_t_16 = __pyx_v_j;
          __pyx_t_14 = ((*((int8_t *) ( /* dim=1 */ ((char *) (((int8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_15 * __pyx_v_mask.strides[0]) )) + __pyx_t_16)) ))) != 0);
          __pyx_t_7 = __pyx_t_14;
          __pyx_L14_bool_binop_done:;
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":687
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:
 *                 continue             # <<<<<<<<<<<<<<
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0
 */
            goto __pyx_L11_continue;

            /* "pyFAI/ext/_distortion.pyx":686
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 */
          }

          /* "pyFAI/ext/_distortion.pyx":688
 *             if do_mask and mask[i, j]:
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index             # <<<<<<<<<<<<<<
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]
 */
          __pyx_v_idx = ((__pyx_v_i * __pyx_v_shape_in1) + __pyx_v_j);

          /* "pyFAI/ext/_distortion.pyx":689
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0             # <<<<<<<<<<<<<<
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]
 */
          {
              __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_temp_scalar = 0.0;
              {
                  Py_ssize_t __pyx_temp_extent = __pyx_v_buffer.shape[0] * __pyx_v_buffer.shape[1];
                  Py_ssize_t __pyx_temp_idx;
                  __pyx_t_5pyFAI_3ext_12shared_types_float32_t *__pyx_temp_pointer = (__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_buffer.data;
                  for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
                    *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
                    __pyx_temp_pointer += 1;
                  }
              }
          }

          /* "pyFAI/ext/_distortion.pyx":690
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]             # <<<<<<<<<<<<<<
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_17 = 0;
          __pyx_t_18 = 0;
          __pyx_v_A0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_17 * __pyx_v_pos.strides[2]) )) + __pyx_t_18)) )));

          /* "pyFAI/ext/_distortion.pyx":691
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]             # <<<<<<<<<<<<<<
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]
 */
          __pyx_t_18 = __pyx_v_i;
          __pyx_t_17 = __pyx_v_j;
          __pyx_t_15 = 0;
          __pyx_t_16 = 1;
          __pyx_v_A1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_18 * __pyx_v_pos.strides[0]) ) + __pyx_t_17 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":692
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]             # <<<<<<<<<<<<<<
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_17 = 1;
          __pyx_t_18 = 0;
          __pyx_v_B0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_17 * __pyx_v_pos.strides[2]) )) + __pyx_t_18)) )));

          /* "pyFAI/ext/_distortion.pyx":693
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]             # <<<<<<<<<<<<<<
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]
 */
          __pyx_t_18 = __pyx_v_i;
          __pyx_t_17 = __pyx_v_j;
          __pyx_t_15 = 1;
          __pyx_t_16 = 1;
          __pyx_v_B1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_18 * __pyx_v_pos.strides[0]) ) + __pyx_t_17 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":694
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]             # <<<<<<<<<<<<<<
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_17 = 2;
          __pyx_t_18 = 0;
          __pyx_v_C0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_17 * __pyx_v_pos.strides[2]) )) + __pyx_t_18)) )));

          /* "pyFAI/ext/_distortion.pyx":695
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]             # <<<<<<<<<<<<<<
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]
 */
          __pyx_t_18 = __pyx_v_i;
          __pyx_t_17 = __pyx_v_j;
          __pyx_t_15 = 2;
          __pyx_t_16 = 1;
          __pyx_v_C1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_18 * __pyx_v_pos.strides[0]) ) + __pyx_t_17 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":696
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]             # <<<<<<<<<<<<<<
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_17 = 3;
          __pyx_t_18 = 0;
          __pyx_v_D0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_17 * __pyx_v_pos.strides[2]) )) + __pyx_t_18)) )));

          /* "pyFAI/ext/_distortion.pyx":697
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]             # <<<<<<<<<<<<<<
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 */
          __pyx_t_18 = __pyx_v_i;
          __pyx_t_17 = __pyx_v_j;
          __pyx_t_15 = 3;
          __pyx_t_16 = 1;
          __pyx_v_D1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_18 * __pyx_v_pos.strides[0]) ) + __pyx_t_17 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":698
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)             # <<<<<<<<<<<<<<
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0
 */
          __pyx_v_foffset0 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0);

          /* "pyFAI/ext/_distortion.pyx":699
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)             # <<<<<<<<<<<<<<
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1
 */
          __pyx_v_foffset1 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1);

          /* "pyFAI/ext/_distortion.pyx":700
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0             # <<<<<<<<<<<<<<
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 */
          __pyx_v_offset0 = ((int)__pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":701
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1             # <<<<<<<<<<<<<<
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 */
          __pyx_v_offset1 = ((int)__pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":702
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0             # <<<<<<<<<<<<<<
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 */
          __pyx_v_box_size0 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) - __pyx_v_offset0);

          /* "pyFAI/ext/_distortion.pyx":703
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1             # <<<<<<<<<<<<<<
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 *                 # Increase size of the buffer
 */
          __pyx_v_box_size1 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) - __pyx_v_offset1);

          /* "pyFAI/ext/_distortion.pyx":704
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 */
          __pyx_t_14 = ((__pyx_v_box_size0 > __pyx_v_delta0) != 0);
          if (!__pyx_t_14) {
          } else {
            __pyx_t_7 = __pyx_t_14;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_t_14 = ((__pyx_v_box_size1 > __pyx_v_delta1) != 0);
          __pyx_t_7 = __pyx_t_14;
          __pyx_L17_bool_binop_done:;
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":706
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0             # <<<<<<<<<<<<<<
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:
 */
            if (((__pyx_v_offset0 > __pyx_v_delta0) != 0)) {
              __pyx_t_19 = __pyx_v_offset0;
            } else {
              __pyx_t_19 = __pyx_v_delta0;
            }
            __pyx_v_delta0 = __pyx_t_19;

            /* "pyFAI/ext/_distortion.pyx":707
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 */
            if (((__pyx_v_offset1 > __pyx_v_delta1) != 0)) {
              __pyx_t_19 = __pyx_v_offset1;
            } else {
              __pyx_t_19 = __pyx_v_delta1;
            }
            __pyx_v_delta1 = __pyx_t_19;

            /* "pyFAI/ext/_distortion.pyx":708
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
            {
                #ifdef WITH_THREAD
                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                #endif
                /*try:*/ {

                  /* "pyFAI/ext/_distortion.pyx":709
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *             A0 = A0 - foffset0
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_8);
                  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8);
                  __Pyx_GIVEREF(__pyx_t_9);
                  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_9);
                  __pyx_t_8 = 0;
                  __pyx_t_9 = 0;
                  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_float32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 709, __pyx_L22_error)
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
                  __pyx_v_buffer = __pyx_t_12;
                  __pyx_t_12.memview = NULL;
                  __pyx_t_12.data = NULL;
                }

                /* "pyFAI/ext/_distortion.pyx":708
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L23;
                  }
                  __pyx_L22_error: {
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L9_error;
                  }
                  __pyx_L23:;
                }
            }

            /* "pyFAI/ext/_distortion.pyx":704
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 */
          }

          /* "pyFAI/ext/_distortion.pyx":711
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 *             A0 = A0 - foffset0             # <<<<<<<<<<<<<<
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0
 */
          __pyx_v_A0 = (__pyx_v_A0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":712
 * 
 *             A0 = A0 - foffset0
 *             A1 = A1 - foffset1             # <<<<<<<<<<<<<<
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1
 */
          __pyx_v_A1 = (__pyx_v_A1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":713
 *             A0 = A0 - foffset0
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0             # <<<<<<<<<<<<<<
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0
 */
          __pyx_v_B0 = (__pyx_v_B0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":714
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1             # <<<<<<<<<<<<<<
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1
 */
          __pyx_v_B1 = (__pyx_v_B1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":715
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0             # <<<<<<<<<<<<<<
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0
 */
          __pyx_v_C0 = (__pyx_v_C0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":716
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1             # <<<<<<<<<<<<<<
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 */
          __pyx_v_C1 = (__pyx_v_C1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":717
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0             # <<<<<<<<<<<<<<
 *             D1 = D1 - foffset1
 *             if B0 != A0:
 */
          __pyx_v_D0 = (__pyx_v_D0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":718
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1             # <<<<<<<<<<<<<<
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)
 */
          __pyx_v_D1 = (__pyx_v_D1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":719
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 *             if B0 != A0:             # <<<<<<<<<<<<<<
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0
 */
          __pyx_t_7 = ((__pyx_v_B0 != __pyx_v_A0) != 0);
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":720
 *             D1 = D1 - foffset1
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)             # <<<<<<<<<<<<<<
 *                 cAB = A1 - pAB * A0
 *             else:
 */
            __pyx_v_pAB = ((__pyx_v_B1 - __pyx_v_A1) / (__pyx_v_B0 - __pyx_v_A0));

            /* "pyFAI/ext/_distortion.pyx":721
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0             # <<<<<<<<<<<<<<
 *             else:
 *                 pAB = cAB = 0.0
 */
            __pyx_v_cAB = (__pyx_v_A1 - (__pyx_v_pAB * __pyx_v_A0));

            /* "pyFAI/ext/_distortion.pyx":719
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 *             if B0 != A0:             # <<<<<<<<<<<<<<
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0
 */
            goto __pyx_L24;
          }

          /* "pyFAI/ext/_distortion.pyx":723
 *                 cAB = A1 - pAB * A0
 *             else:
 *                 pAB = cAB = 0.0             # <<<<<<<<<<<<<<
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)
 */
          /*else*/ {
            __pyx_v_pAB = 0.0;
            __pyx_v_cAB = 0.0;
          }
          __pyx_L24:;

          /* "pyFAI/ext/_distortion.pyx":724
 *             else:
 *                 pAB = cAB = 0.0
 *             if C0 != B0:             # <<<<<<<<<<<<<<
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0
 */
          __pyx_t_7 = ((__pyx_v_C0 != __pyx_v_B0) != 0);
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":725
 *                 pAB = cAB = 0.0
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)             # <<<<<<<<<<<<<<
 *                 cBC = B1 - pBC * B0
 *             else:
 */
            __pyx_v_pBC = ((__pyx_v_C1 - __pyx_v_B1) / (__pyx_v_C0 - __pyx_v_B0));

            /* "pyFAI/ext/_distortion.pyx":726
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0             # <<<<<<<<<<<<<<
 *             else:
 *                 pBC = cBC = 0.0
 */
            __pyx_v_cBC = (__pyx_v_B1 - (__pyx_v_pBC * __pyx_v_B0));

            /* "pyFAI/ext/_distortion.pyx":724
 *             else:
 *                 pAB = cAB = 0.0
 *             if C0 != B0:             # <<<<<<<<<<<<<<
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0
 */
            goto __pyx_L25;
          }

          /* "pyFAI/ext/_distortion.pyx":728
 *                 cBC = B1 - pBC * B0
 *             else:
 *                 pBC = cBC = 0.0             # <<<<<<<<<<<<<<
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)
 */
          /*else*/ {
            __pyx_v_pBC = 0.0;
            __pyx_v_cBC = 0.0;
          }
          __pyx_L25:;

          /* "pyFAI/ext/_distortion.pyx":729
 *             else:
 *                 pBC = cBC = 0.0
 *             if D0 != C0:             # <<<<<<<<<<<<<<
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0
 */
          __pyx_t_7 = ((__pyx_v_D0 != __pyx_v_C0) != 0);
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":730
 *                 pBC = cBC = 0.0
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)             # <<<<<<<<<<<<<<
 *                 cCD = C1 - pCD * C0
 *             else:
 */
            __pyx_v_pCD = ((__pyx_v_D1 - __pyx_v_C1) / (__pyx_v_D0 - __pyx_v_C0));

            /* "pyFAI/ext/_distortion.pyx":731
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0             # <<<<<<<<<<<<<<
 *             else:
 *                 pCD = cCD = 0.0
 */
            __pyx_v_cCD = (__pyx_v_C1 - (__pyx_v_pCD * __pyx_v_C0));

            /* "pyFAI/ext/_distortion.pyx":729
 *             else:
 *                 pBC = cBC = 0.0
 *             if D0 != C0:             # <<<<<<<<<<<<<<
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0
 */
            goto __pyx_L26;
          }

          /* "pyFAI/ext/_distortion.pyx":733
 *                 cCD = C1 - pCD * C0
 *             else:
 *                 pCD = cCD = 0.0             # <<<<<<<<<<<<<<
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)
 */
          /*else*/ {
            __pyx_v_pCD = 0.0;
            __pyx_v_cCD = 0.0;
          }
          __pyx_L26:;

          /* "pyFAI/ext/_distortion.pyx":734
 *             else:
 *                 pCD = cCD = 0.0
 *             if A0 != D0:             # <<<<<<<<<<<<<<
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0
 */
          __pyx_t_7 = ((__pyx_v_A0 != __pyx_v_D0) != 0);
          if (__pyx_t_7) {

            /* "pyFAI/ext/_distortion.pyx":735
 *                 pCD = cCD = 0.0
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)             # <<<<<<<<<<<<<<
 *                 cDA = D1 - pDA * D0
 *             else:
 */
            __pyx_v_pDA = ((__pyx_v_A1 - __pyx_v_D1) / (__pyx_v_A0 - __pyx_v_D0));

            /* "pyFAI/ext/_distortion.pyx":736
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0             # <<<<<<<<<<<<<<
 *             else:
 *                 pDA = cDA = 0.0
 */
            __pyx_v_cDA = (__pyx_v_D1 - (__pyx_v_pDA * __pyx_v_D0));

            /* "pyFAI/ext/_distortion.pyx":734
 *             else:
 *                 pCD = cCD = 0.0
 *             if A0 != D0:             # <<<<<<<<<<<<<<
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0
 */
            goto __pyx_L27;
          }

          /* "pyFAI/ext/_distortion.pyx":738
 *                 cDA = D1 - pDA * D0
 *             else:
 *                 pDA = cDA = 0.0             # <<<<<<<<<<<<<<
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)
 */
          /*else*/ {
            __pyx_v_pDA = 0.0;
            __pyx_v_cDA = 0.0;
          }
          __pyx_L27:;

          /* "pyFAI/ext/_distortion.pyx":740
 *                 pDA = cDA = 0.0
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)             # <<<<<<<<<<<<<<
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_B0, __pyx_v_A0, __pyx_v_pAB, __pyx_v_cAB);

          /* "pyFAI/ext/_distortion.pyx":741
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)
 *             integrate(buffer, A0, D0, pDA, cDA)             # <<<<<<<<<<<<<<
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_A0, __pyx_v_D0, __pyx_v_pDA, __pyx_v_cDA);

          /* "pyFAI/ext/_distortion.pyx":742
 *             integrate(buffer, B0, A0, pAB, cAB)
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)             # <<<<<<<<<<<<<<
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_D0, __pyx_v_C0, __pyx_v_pCD, __pyx_v_cCD);

          /* "pyFAI/ext/_distortion.pyx":743
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)             # <<<<<<<<<<<<<<
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_C0, __pyx_v_B0, __pyx_v_pBC, __pyx_v_cBC);

          /* "pyFAI/ext/_distortion.pyx":744
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))             # <<<<<<<<<<<<<<
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 */
          __pyx_v_area = (0.5 * (((__pyx_v_C0 - __pyx_v_A0) * (__pyx_v_D1 - __pyx_v_B1)) - ((__pyx_v_C1 - __pyx_v_A1) * (__pyx_v_D0 - __pyx_v_B0))));

          /* "pyFAI/ext/_distortion.pyx":745
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):             # <<<<<<<<<<<<<<
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:
 */
          __pyx_t_19 = __pyx_v_box_size0;
          __pyx_t_20 = __pyx_t_19;
          for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
            __pyx_v_ms = __pyx_t_21;

            /* "pyFAI/ext/_distortion.pyx":746
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):
 *                 ml = ms + offset0             # <<<<<<<<<<<<<<
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue
 */
            __pyx_v_ml = (__pyx_v_ms + __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":747
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for ns in range(box_size1):
 */
            __pyx_t_14 = ((__pyx_v_ml < 0) != 0);
            if (!__pyx_t_14) {
            } else {
              __pyx_t_7 = __pyx_t_14;
              goto __pyx_L31_bool_binop_done;
            }
            __pyx_t_14 = ((__pyx_v_ml >= __pyx_v_shape_out0) != 0);
            __pyx_t_7 = __pyx_t_14;
            __pyx_L31_bool_binop_done:;
            if (__pyx_t_7) {

              /* "pyFAI/ext/_distortion.pyx":748
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 for ns in range(box_size1):
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 */
              goto __pyx_L28_continue;

              /* "pyFAI/ext/_distortion.pyx":747
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for ns in range(box_size1):
 */
            }

            /* "pyFAI/ext/_distortion.pyx":749
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue
 *                 for ns in range(box_size1):             # <<<<<<<<<<<<<<
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 */
            __pyx_t_22 = __pyx_v_box_size1;
            __pyx_t_23 = __pyx_t_22;
            for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
              __pyx_v_ns = __pyx_t_24;

              /* "pyFAI/ext/_distortion.pyx":751
 *                 for ns in range(box_size1):
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1             # <<<<<<<<<<<<<<
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue
 */
              __pyx_v_nl = (__pyx_v_ns + __pyx_v_offset1);

              /* "pyFAI/ext/_distortion.pyx":752
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:             # <<<<<<<<<<<<<<
 *                         continue
 *                     value = buffer[ms, ns] / area
 */
              __pyx_t_14 = ((__pyx_v_nl < 0) != 0);
              if (!__pyx_t_14) {
              } else {
                __pyx_t_7 = __pyx_t_14;
                goto __pyx_L36_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_nl >= __pyx_v_shape_out1) != 0);
              __pyx_t_7 = __pyx_t_14;
              __pyx_L36_bool_binop_done:;
              if (__pyx_t_7) {

                /* "pyFAI/ext/_distortion.pyx":753
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue             # <<<<<<<<<<<<<<
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:
 */
                goto __pyx_L33_continue;

                /* "pyFAI/ext/_distortion.pyx":752
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:             # <<<<<<<<<<<<<<
 *                         continue
 *                     value = buffer[ms, ns] / area
 */
              }

              /* "pyFAI/ext/_distortion.pyx":754
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue
 *                     value = buffer[ms, ns] / area             # <<<<<<<<<<<<<<
 *                     if value == 0.0:
 *                         continue
 */
              __pyx_t_16 = __pyx_v_ms;
              __pyx_t_15 = __pyx_v_ns;
              __pyx_v_value = ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_16 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_area));

              /* "pyFAI/ext/_distortion.pyx":755
 *                         continue
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if value < 0.0 or value > 1.0001:
 */
              __pyx_t_7 = ((__pyx_v_value == 0.0) != 0);
              if (__pyx_t_7) {

                /* "pyFAI/ext/_distortion.pyx":756
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 */
                goto __pyx_L33_continue;

                /* "pyFAI/ext/_distortion.pyx":755
 *                         continue
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if value < 0.0 or value > 1.0001:
 */
              }

              /* "pyFAI/ext/_distortion.pyx":757
 *                     if value == 0.0:
 *                         continue
 *                     if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 */
              __pyx_t_14 = ((__pyx_v_value < 0.0) != 0);
              if (!__pyx_t_14) {
              } else {
                __pyx_t_7 = __pyx_t_14;
                goto __pyx_L40_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_value > 1.0001) != 0);
              __pyx_t_7 = __pyx_t_14;
              __pyx_L40_bool_binop_done:;
              if (__pyx_t_7) {

                /* "pyFAI/ext/_distortion.pyx":759
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                __pyx_t_7 = ((__pyx_v_err_cnt < 0x3E8) != 0);
                if (__pyx_t_7) {

                  /* "pyFAI/ext/_distortion.pyx":760
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 *                             with gil:             # <<<<<<<<<<<<<<
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      /*try:*/ {

                        /* "pyFAI/ext/_distortion.pyx":761
 *                         if err_cnt < 1000:
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])             # <<<<<<<<<<<<<<
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1
 */
                        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_1);
                        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_ms); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_9);
                        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_box_size0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_2);
                        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_ns); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_8);
                        __pyx_t_25 = __Pyx_PyInt_From_int(__pyx_v_box_size1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        __pyx_t_15 = __pyx_v_ms;
                        __pyx_t_16 = __pyx_v_ns;
                        __pyx_t_26 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_16)) )))); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_26);
                        __pyx_t_27 = PyFloat_FromDouble(__pyx_v_area); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_27);
                        __pyx_t_28 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_28);
                        __pyx_t_16 = 0;
                        __pyx_t_15 = 0;
                        __pyx_t_29 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_16 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) )))); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_15 = 0;
                        __pyx_t_16 = 1;
                        __pyx_t_30 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_16)) )))); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_30);
                        __pyx_t_16 = 1;
                        __pyx_t_15 = 0;
                        __pyx_t_31 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_16 * __pyx_v_buffer.strides[0]) )) + __pyx_t_15)) )))); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_15 = 1;
                        __pyx_t_16 = 1;
                        __pyx_t_32 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_15 * __pyx_v_buffer.strides[0]) )) + __pyx_t_16)) )))); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_32);
                        __pyx_t_33 = PyTuple_New(13); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_GIVEREF(__pyx_t_1);
                        PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_1);
                        __Pyx_GIVEREF(__pyx_t_3);
                        PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_3);
                        __Pyx_GIVEREF(__pyx_t_9);
                        PyTuple_SET_ITEM(__pyx_t_33, 2, __pyx_t_9);
                        __Pyx_GIVEREF(__pyx_t_2);
                        PyTuple_SET_ITEM(__pyx_t_33, 3, __pyx_t_2);
                        __Pyx_GIVEREF(__pyx_t_8);
                        PyTuple_SET_ITEM(__pyx_t_33, 4, __pyx_t_8);
                        __Pyx_GIVEREF(__pyx_t_25);
                        PyTuple_SET_ITEM(__pyx_t_33, 5, __pyx_t_25);
                        __Pyx_GIVEREF(__pyx_t_26);
                        PyTuple_SET_ITEM(__pyx_t_33, 6, __pyx_t_26);
                        __Pyx_GIVEREF(__pyx_t_27);
                        PyTuple_SET_ITEM(__pyx_t_33, 7, __pyx_t_27);
                        __Pyx_GIVEREF(__pyx_t_28);
                        PyTuple_SET_ITEM(__pyx_t_33, 8, __pyx_t_28);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_33, 9, __pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_30);
                        PyTuple_SET_ITEM(__pyx_t_33, 10, __pyx_t_30);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_33, 11, __pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_32);
                        PyTuple_SET_ITEM(__pyx_t_33, 12, __pyx_t_32);
                        __pyx_t_1 = 0;
                        __pyx_t_3 = 0;
                        __pyx_t_9 = 0;
                        __pyx_t_2 = 0;
                        __pyx_t_8 = 0;
                        __pyx_t_25 = 0;
                        __pyx_t_26 = 0;
                        __pyx_t_27 = 0;
                        __pyx_t_28 = 0;
                        __pyx_t_29 = 0;
                        __pyx_t_30 = 0;
                        __pyx_t_31 = 0;
                        __pyx_t_32 = 0;
                        __pyx_t_32 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_33, NULL); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 761, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_32);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

                        /* "pyFAI/ext/_distortion.pyx":762
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))             # <<<<<<<<<<<<<<
 *                             err_cnt += 1
 *                         continue
 */
                        __pyx_t_32 = PyTuple_New(16); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_32);
                        __pyx_t_34 = 0;
                        __pyx_t_35 = 127;
                        __Pyx_INCREF(__pyx_kp_u_A0);
                        __pyx_t_34 += 4;
                        __Pyx_GIVEREF(__pyx_kp_u_A0);
                        PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_kp_u_A0);
                        __pyx_t_33 = PyFloat_FromDouble(__pyx_v_A0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_33), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_A1);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_A1);
                        PyTuple_SET_ITEM(__pyx_t_32, 2, __pyx_kp_u_A1);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_A1); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_33 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_33);
                        __Pyx_GIVEREF(__pyx_t_33);
                        PyTuple_SET_ITEM(__pyx_t_32, 3, __pyx_t_33);
                        __pyx_t_33 = 0;
                        __Pyx_INCREF(__pyx_kp_u_B0);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_B0);
                        PyTuple_SET_ITEM(__pyx_t_32, 4, __pyx_kp_u_B0);
                        __pyx_t_33 = PyFloat_FromDouble(__pyx_v_B0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_33), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_32, 5, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_B1);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_B1);
                        PyTuple_SET_ITEM(__pyx_t_32, 6, __pyx_kp_u_B1);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_B1); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_33 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_33);
                        __Pyx_GIVEREF(__pyx_t_33);
                        PyTuple_SET_ITEM(__pyx_t_32, 7, __pyx_t_33);
                        __pyx_t_33 = 0;
                        __Pyx_INCREF(__pyx_kp_u_C0);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_C0);
                        PyTuple_SET_ITEM(__pyx_t_32, 8, __pyx_kp_u_C0);
                        __pyx_t_33 = PyFloat_FromDouble(__pyx_v_C0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_33), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_32, 9, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_C1);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_C1);
                        PyTuple_SET_ITEM(__pyx_t_32, 10, __pyx_kp_u_C1);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_C1); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_33 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_33);
                        __Pyx_GIVEREF(__pyx_t_33);
                        PyTuple_SET_ITEM(__pyx_t_32, 11, __pyx_t_33);
                        __pyx_t_33 = 0;
                        __Pyx_INCREF(__pyx_kp_u_D0);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_D0);
                        PyTuple_SET_ITEM(__pyx_t_32, 12, __pyx_kp_u_D0);
                        __pyx_t_33 = PyFloat_FromDouble(__pyx_v_D0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_33), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_32, 13, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_D1);
                        __pyx_t_34 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_D1);
                        PyTuple_SET_ITEM(__pyx_t_32, 14, __pyx_kp_u_D1);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_D1); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_33 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_35 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) > __pyx_t_35) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_33) : __pyx_t_35;
                        __pyx_t_34 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_33);
                        __Pyx_GIVEREF(__pyx_t_33);
                        PyTuple_SET_ITEM(__pyx_t_32, 15, __pyx_t_33);
                        __pyx_t_33 = 0;
                        __pyx_t_33 = __Pyx_PyUnicode_Join(__pyx_t_32, 16, __pyx_t_34, __pyx_t_35); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_33);
                        __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
                        __pyx_t_32 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_33); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 762, __pyx_L46_error)
                        __Pyx_GOTREF(__pyx_t_32);
                        __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                        __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
                      }

                      /* "pyFAI/ext/_distortion.pyx":760
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 *                             with gil:             # <<<<<<<<<<<<<<
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                      /*finally:*/ {
                        /*normal exit:*/{
                          #ifdef WITH_THREAD
                          __Pyx_PyGILState_Release(__pyx_gilstate_save);
                          #endif
                          goto __pyx_L47;
                        }
                        __pyx_L46_error: {
                          #ifdef WITH_THREAD
                          __Pyx_PyGILState_Release(__pyx_gilstate_save);
                          #endif
                          goto __pyx_L9_error;
                        }
                        __pyx_L47:;
                      }
                  }

                  /* "pyFAI/ext/_distortion.pyx":763
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
                  __pyx_v_err_cnt = (__pyx_v_err_cnt + 1);

                  /* "pyFAI/ext/_distortion.pyx":759
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                }

                /* "pyFAI/ext/_distortion.pyx":764
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *                     bin_number = ml * shape_out1 + nl
 */
                goto __pyx_L33_continue;

                /* "pyFAI/ext/_distortion.pyx":757
 *                     if value == 0.0:
 *                         continue
 *                     if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 */
              }

              /* "pyFAI/ext/_distortion.pyx":766
 *                         continue
 * 
 *                     bin_number = ml * shape_out1 + nl             # <<<<<<<<<<<<<<
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1
 */
              __pyx_v_bin_number = ((__pyx_v_ml * __pyx_v_shape_out1) + __pyx_v_nl);

              /* "pyFAI/ext/_distortion.pyx":768
 *                     bin_number = ml * shape_out1 + nl
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1             # <<<<<<<<<<<<<<
 *                     pixel_count[bin_number] += 1
 *                     if counter >= large_size:
 */
              __pyx_v_counter = (__pyx_v_counter + 1);

              /* "pyFAI/ext/_distortion.pyx":769
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1
 *                     pixel_count[bin_number] += 1             # <<<<<<<<<<<<<<
 *                     if counter >= large_size:
 *                         with gil:
 */
              __pyx_t_16 = __pyx_v_bin_number;
              *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_16)) )) += 1;

              /* "pyFAI/ext/_distortion.pyx":770
 *                     counter += 1
 *                     pixel_count[bin_number] += 1
 *                     if counter >= large_size:             # <<<<<<<<<<<<<<
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 */
              __pyx_t_7 = ((__pyx_v_counter >= __pyx_v_large_size) != 0);
              if (__pyx_t_7) {

                /* "pyFAI/ext/_distortion.pyx":771
 *                     pixel_count[bin_number] += 1
 *                     if counter >= large_size:
 *                         with gil:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 */
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    /*try:*/ {

                      /* "pyFAI/ext/_distortion.pyx":773
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +             # <<<<<<<<<<<<<<
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     idx_pixel[counter] += idx
 */
                      __pyx_t_32 = __Pyx_PyInt_From_int(__pyx_v_bins_per_pixel); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 773, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_32);
                      __pyx_t_33 = PyUnicode_Format(__pyx_kp_u_Please_increase_bins_per_pixel_s, __pyx_t_32); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 773, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_33);
                      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

                      /* "pyFAI/ext/_distortion.pyx":772
 *                     if counter >= large_size:
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +             # <<<<<<<<<<<<<<
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 */
                      __pyx_t_32 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Provided_temporary_space_for_sto, __pyx_t_33); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 772, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_32);
                      __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

                      /* "pyFAI/ext/_distortion.pyx":774
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))             # <<<<<<<<<<<<<<
 *                     idx_pixel[counter] += idx
 *                     idx_bin[counter] += bin_number
 */
                      __pyx_t_33 = PyFloat_FromDouble(ceil((((1.1 * __pyx_v_bins_per_pixel) * __pyx_v_size_in) / ((double)__pyx_v_idx)))); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 774, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_33);
                      __pyx_t_31 = PyUnicode_Format(__pyx_kp_u_The_suggested_value_is_i_or_grea, __pyx_t_33); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 774, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_31);
                      __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

                      /* "pyFAI/ext/_distortion.pyx":773
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +             # <<<<<<<<<<<<<<
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     idx_pixel[counter] += idx
 */
                      __pyx_t_33 = __Pyx_PyUnicode_Concat(__pyx_t_32, __pyx_t_31); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 773, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_33);
                      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
                      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

                      /* "pyFAI/ext/_distortion.pyx":772
 *                     if counter >= large_size:
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +             # <<<<<<<<<<<<<<
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 */
                      __pyx_t_31 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_33); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 772, __pyx_L52_error)
                      __Pyx_GOTREF(__pyx_t_31);
                      __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
                      __Pyx_Raise(__pyx_t_31, 0, 0, 0);
                      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                      __PYX_ERR(0, 772, __pyx_L52_error)
                    }

                    /* "pyFAI/ext/_distortion.pyx":771
 *                     pixel_count[bin_number] += 1
 *                     if counter >= large_size:
 *                         with gil:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 */
                    /*finally:*/ {
                      __pyx_L52_error: {
                        #ifdef WITH_THREAD
                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                        #endif
                        goto __pyx_L9_error;
                      }
                    }
                }

                /* "pyFAI/ext/_distortion.pyx":770
 *                     counter += 1
 *                     pixel_count[bin_number] += 1
 *                     if counter >= large_size:             # <<<<<<<<<<<<<<
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 */
              }

              /* "pyFAI/ext/_distortion.pyx":775
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     idx_pixel[counter] += idx             # <<<<<<<<<<<<<<
 *                     idx_bin[counter] += bin_number
 *                     large_data[counter] += value
 */
              __pyx_t_16 = __pyx_v_counter;
              *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_pixel.data) + __pyx_t_16)) )) += __pyx_v_idx;

              /* "pyFAI/ext/_distortion.pyx":776
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     idx_pixel[counter] += idx
 *                     idx_bin[counter] += bin_number             # <<<<<<<<<<<<<<
 *                     large_data[counter] += value
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 */
              __pyx_t_16 = __pyx_v_counter;
              *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_bin.data) + __pyx_t_16)) )) += __pyx_v_bin_number;

              /* "pyFAI/ext/_distortion.pyx":777
 *                     idx_pixel[counter] += idx
 *                     idx_bin[counter] += bin_number
 *                     large_data[counter] += value             # <<<<<<<<<<<<<<
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 *                 counter, counter / size_in, bins_per_pixel)
 */
              __pyx_t_16 = __pyx_v_counter;
              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_large_data.data) + __pyx_t_16)) )) += __pyx_v_value;
              __pyx_L33_continue:;
            }
            __pyx_L28_continue:;
          }
          __pyx_L11_continue:;
        }
      }

      /* "pyFAI/ext/_distortion.pyx":681
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L10;
        }
        __pyx_L9_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L10:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":778
 *                     idx_bin[counter] += bin_number
 *                     large_data[counter] += value
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",             # <<<<<<<<<<<<<<
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_33, __pyx_n_s_logger); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_33, __pyx_n_s_info); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pyFAI/ext/_distortion.pyx":779
 *                     large_data[counter] += value
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 *                 counter, counter / size_in, bins_per_pixel)             # <<<<<<<<<<<<<<
 * 
 *     if format == "csr":
 */
  __pyx_t_33 = __Pyx_PyInt_From_int(__pyx_v_counter); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_30 = __Pyx_PyInt_From_int((__pyx_v_counter / __pyx_v_size_in)); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_29 = __Pyx_PyInt_From_int(__pyx_v_bins_per_pixel); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_28 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_32))) {
    __pyx_t_28 = PyMethod_GET_SELF(__pyx_t_32);
    if (likely(__pyx_t_28)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
      __Pyx_INCREF(__pyx_t_28);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_32, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_32)) {
    PyObject *__pyx_temp[5] = {__pyx_t_28, __pyx_kp_u_number_of_elements_s_average_per, __pyx_t_33, __pyx_t_30, __pyx_t_29};
    __pyx_t_31 = __Pyx_PyFunction_FastCall(__pyx_t_32, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_32)) {
    PyObject *__pyx_temp[5] = {__pyx_t_28, __pyx_kp_u_number_of_elements_s_average_per, __pyx_t_33, __pyx_t_30, __pyx_t_29};
    __pyx_t_31 = __Pyx_PyCFunction_FastCall(__pyx_t_32, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  } else
  #endif
  {
    __pyx_t_27 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    if (__pyx_t_28) {
      __Pyx_GIVEREF(__pyx_t_28); PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28); __pyx_t_28 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_number_of_elements_s_average_per);
    __Pyx_GIVEREF(__pyx_kp_u_number_of_elements_s_average_per);
    PyTuple_SET_ITEM(__pyx_t_27, 0+__pyx_t_5, __pyx_kp_u_number_of_elements_s_average_per);
    __Pyx_GIVEREF(__pyx_t_33);
    PyTuple_SET_ITEM(__pyx_t_27, 1+__pyx_t_5, __pyx_t_33);
    __Pyx_GIVEREF(__pyx_t_30);
    PyTuple_SET_ITEM(__pyx_t_27, 2+__pyx_t_5, __pyx_t_30);
    __Pyx_GIVEREF(__pyx_t_29);
    PyTuple_SET_ITEM(__pyx_t_27, 3+__pyx_t_5, __pyx_t_29);
    __pyx_t_33 = 0;
    __pyx_t_30 = 0;
    __pyx_t_29 = 0;
    __pyx_t_31 = __Pyx_PyObject_Call(__pyx_t_32, __pyx_t_27, NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  }
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pyFAI/ext/_distortion.pyx":781
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 *     if format == "csr":             # <<<<<<<<<<<<<<
 *         indptr = numpy.zeros(bins + 1, dtype=numpy.int32)
 *         # cumsum
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_format, __pyx_n_u_csr, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 781, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "pyFAI/ext/_distortion.pyx":782
 * 
 *     if format == "csr":
 *         indptr = numpy.zeros(bins + 1, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *         # cumsum
 *         j = 0
 */
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_n_s_numpy); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_n_s_zeros); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __pyx_t_31 = __Pyx_PyInt_From_long((__pyx_v_bins + 1)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __pyx_t_27 = PyTuple_New(1); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_31);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_31);
    __pyx_t_31 = 0;
    __pyx_t_31 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_GetModuleGlobalName(__pyx_t_29, __pyx_n_s_numpy); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_int32); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_dtype, __pyx_t_30) < 0) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_32, __pyx_t_27, __pyx_t_31); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_30, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_v_indptr = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":784
 *         indptr = numpy.zeros(bins + 1, dtype=numpy.int32)
 *         # cumsum
 *         j = 0             # <<<<<<<<<<<<<<
 *         for i in range(bins):
 *             indptr[i] = j
 */
    __pyx_v_j = 0;

    /* "pyFAI/ext/_distortion.pyx":785
 *         # cumsum
 *         j = 0
 *         for i in range(bins):             # <<<<<<<<<<<<<<
 *             indptr[i] = j
 *             j += pixel_count[i]
 */
    __pyx_t_5 = __pyx_v_bins;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_6; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "pyFAI/ext/_distortion.pyx":786
 *         j = 0
 *         for i in range(bins):
 *             indptr[i] = j             # <<<<<<<<<<<<<<
 *             j += pixel_count[i]
 *         indptr[bins] = j
 */
      __pyx_t_16 = __pyx_v_i;
      *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_16)) )) = __pyx_v_j;

      /* "pyFAI/ext/_distortion.pyx":787
 *         for i in range(bins):
 *             indptr[i] = j
 *             j += pixel_count[i]             # <<<<<<<<<<<<<<
 *         indptr[bins] = j
 *         # indptr[1:] = numpy.asarray(pixel_count).cumsum(dtype=numpy.int32)
 */
      __pyx_t_16 = __pyx_v_i;
      __pyx_v_j = (__pyx_v_j + (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_16)) ))));
    }

    /* "pyFAI/ext/_distortion.pyx":788
 *             indptr[i] = j
 *             j += pixel_count[i]
 *         indptr[bins] = j             # <<<<<<<<<<<<<<
 *         # indptr[1:] = numpy.asarray(pixel_count).cumsum(dtype=numpy.int32)
 *         pixel_count[:] = 0
 */
    __pyx_t_16 = __pyx_v_bins;
    *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_16)) )) = __pyx_v_j;

    /* "pyFAI/ext/_distortion.pyx":790
 *         indptr[bins] = j
 *         # indptr[1:] = numpy.asarray(pixel_count).cumsum(dtype=numpy.int32)
 *         pixel_count[:] = 0             # <<<<<<<<<<<<<<
 *         lut_size = indptr[bins]
 *         indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)
 */
    {
        int32_t __pyx_temp_scalar = 0;
        {
            Py_ssize_t __pyx_temp_extent = __pyx_v_pixel_count.shape[0];
            Py_ssize_t __pyx_temp_idx;
            int32_t *__pyx_temp_pointer = (int32_t *) __pyx_v_pixel_count.data;
            for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
              *((int32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
              __pyx_temp_pointer += 1;
            }
        }
    }

    /* "pyFAI/ext/_distortion.pyx":791
 *         # indptr[1:] = numpy.asarray(pixel_count).cumsum(dtype=numpy.int32)
 *         pixel_count[:] = 0
 *         lut_size = indptr[bins]             # <<<<<<<<<<<<<<
 *         indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)
 *         data = numpy.zeros(shape=lut_size, dtype=numpy.float32)
 */
    __pyx_t_16 = __pyx_v_bins;
    __pyx_v_lut_size = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_16)) )));

    /* "pyFAI/ext/_distortion.pyx":792
 *         pixel_count[:] = 0
 *         lut_size = indptr[bins]
 *         indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *         data = numpy.zeros(shape=lut_size, dtype=numpy.float32)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_numpy); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_n_s_zeros); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_27 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_shape, __pyx_t_27) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_numpy); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_27, __pyx_n_s_int32); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_dtype, __pyx_t_32) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = __Pyx_PyObject_Call(__pyx_t_31, __pyx_empty_tuple, __pyx_t_30); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_32, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_v_indices = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":793
 *         lut_size = indptr[bins]
 *         indices = numpy.zeros(shape=lut_size, dtype=numpy.int32)
 *         data = numpy.zeros(shape=lut_size, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *         logger.info("CSR matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 */
    __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_numpy); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_zeros); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_31 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_shape, __pyx_t_31) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_n_s_numpy); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_n_s_float32); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_dtype, __pyx_t_27) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_30, __pyx_empty_tuple, __pyx_t_32); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_27, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __pyx_v_data = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":795
 *         data = numpy.zeros(shape=lut_size, dtype=numpy.float32)
 * 
 *         logger.info("CSR matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",             # <<<<<<<<<<<<<<
 *                     (indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_logger); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_info); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

    /* "pyFAI/ext/_distortion.pyx":796
 * 
 *         logger.info("CSR matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 *                     (indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))             # <<<<<<<<<<<<<<
 * 
 *         for idx in range(counter + 1):
 */
    __pyx_t_32 = __pyx_memoryview_fromslice(__pyx_v_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_29 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = PyNumber_Add(__pyx_t_31, __pyx_t_29); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = PyNumber_Add(__pyx_t_32, __pyx_t_31); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __pyx_t_31 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_29, __pyx_float_1_0e6, 1.0e6, 0, 0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_32 = PyFloat_FromDouble(((1.0 * __pyx_v_counter) / ((double)__pyx_v_bins))); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_33 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_30))) {
      __pyx_t_33 = PyMethod_GET_SELF(__pyx_t_30);
      if (likely(__pyx_t_33)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_30);
        __Pyx_INCREF(__pyx_t_33);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_30, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_30)) {
      PyObject *__pyx_temp[5] = {__pyx_t_33, __pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p, __pyx_t_31, __pyx_t_29, __pyx_t_32};
      __pyx_t_27 = __Pyx_PyFunction_FastCall(__pyx_t_30, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 795, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
      __Pyx_GOTREF(__pyx_t_27);
      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_30)) {
      PyObject *__pyx_temp[5] = {__pyx_t_33, __pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p, __pyx_t_31, __pyx_t_29, __pyx_t_32};
      __pyx_t_27 = __Pyx_PyCFunction_FastCall(__pyx_t_30, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 795, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
      __Pyx_GOTREF(__pyx_t_27);
      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    } else
    #endif
    {
      __pyx_t_28 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 795, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      if (__pyx_t_33) {
        __Pyx_GIVEREF(__pyx_t_33); PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_33); __pyx_t_33 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p);
      __Pyx_GIVEREF(__pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p);
      PyTuple_SET_ITEM(__pyx_t_28, 0+__pyx_t_5, __pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p);
      __Pyx_GIVEREF(__pyx_t_31);
      PyTuple_SET_ITEM(__pyx_t_28, 1+__pyx_t_5, __pyx_t_31);
      __Pyx_GIVEREF(__pyx_t_29);
      PyTuple_SET_ITEM(__pyx_t_28, 2+__pyx_t_5, __pyx_t_29);
      __Pyx_GIVEREF(__pyx_t_32);
      PyTuple_SET_ITEM(__pyx_t_28, 3+__pyx_t_5, __pyx_t_32);
      __pyx_t_31 = 0;
      __pyx_t_29 = 0;
      __pyx_t_32 = 0;
      __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_30, __pyx_t_28, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 795, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_27);
      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    }
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "pyFAI/ext/_distortion.pyx":798
 *                     (indices.nbytes + data.nbytes + indptr.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 * 
 *         for idx in range(counter + 1):             # <<<<<<<<<<<<<<
 *             bin_number = idx_bin[idx]
 *             i = indptr[bin_number] + pixel_count[bin_number]
 */
    __pyx_t_36 = (__pyx_v_counter + 1);
    __pyx_t_37 = __pyx_t_36;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_37; __pyx_t_5+=1) {
      __pyx_v_idx = __pyx_t_5;

      /* "pyFAI/ext/_distortion.pyx":799
 * 
 *         for idx in range(counter + 1):
 *             bin_number = idx_bin[idx]             # <<<<<<<<<<<<<<
 *             i = indptr[bin_number] + pixel_count[bin_number]
 *             pixel_count[bin_number] += 1
 */
      __pyx_t_16 = __pyx_v_idx;
      __pyx_v_bin_number = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_bin.data) + __pyx_t_16)) )));

      /* "pyFAI/ext/_distortion.pyx":800
 *         for idx in range(counter + 1):
 *             bin_number = idx_bin[idx]
 *             i = indptr[bin_number] + pixel_count[bin_number]             # <<<<<<<<<<<<<<
 *             pixel_count[bin_number] += 1
 *             indices[i] = idx_pixel[idx]
 */
      __pyx_t_16 = __pyx_v_bin_number;
      __pyx_t_15 = __pyx_v_bin_number;
      __pyx_v_i = ((*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_16)) ))) + (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_15)) ))));

      /* "pyFAI/ext/_distortion.pyx":801
 *             bin_number = idx_bin[idx]
 *             i = indptr[bin_number] + pixel_count[bin_number]
 *             pixel_count[bin_number] += 1             # <<<<<<<<<<<<<<
 *             indices[i] = idx_pixel[idx]
 *             data[i] = large_data[idx]
 */
      __pyx_t_15 = __pyx_v_bin_number;
      *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_15)) )) += 1;

      /* "pyFAI/ext/_distortion.pyx":802
 *             i = indptr[bin_number] + pixel_count[bin_number]
 *             pixel_count[bin_number] += 1
 *             indices[i] = idx_pixel[idx]             # <<<<<<<<<<<<<<
 *             data[i] = large_data[idx]
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 */
      __pyx_t_15 = __pyx_v_idx;
      __pyx_t_16 = __pyx_v_i;
      *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indices.data) + __pyx_t_16)) )) = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_pixel.data) + __pyx_t_15)) )));

      /* "pyFAI/ext/_distortion.pyx":803
 *             pixel_count[bin_number] += 1
 *             indices[i] = idx_pixel[idx]
 *             data[i] = large_data[idx]             # <<<<<<<<<<<<<<
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 *     elif format == "lut":
 */
      __pyx_t_15 = __pyx_v_idx;
      __pyx_t_16 = __pyx_v_i;
      *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_16)) )) = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_large_data.data) + __pyx_t_15)) )));
    }

    /* "pyFAI/ext/_distortion.pyx":804
 *             indices[i] = idx_pixel[idx]
 *             data[i] = large_data[idx]
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))             # <<<<<<<<<<<<<<
 *     elif format == "lut":
 *         lut_size = numpy.asarray(pixel_count).max()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_numpy); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_n_s_asarray); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_32 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_28))) {
      __pyx_t_32 = PyMethod_GET_SELF(__pyx_t_28);
      if (likely(__pyx_t_32)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_28);
        __Pyx_INCREF(__pyx_t_32);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_28, function);
      }
    }
    __pyx_t_27 = (__pyx_t_32) ? __Pyx_PyObject_Call2Args(__pyx_t_28, __pyx_t_32, __pyx_t_30) : __Pyx_PyObject_CallOneArg(__pyx_t_28, __pyx_t_30);
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_numpy); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_n_s_asarray); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __pyx_memoryview_fromslice(__pyx_v_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_29 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_32))) {
      __pyx_t_29 = PyMethod_GET_SELF(__pyx_t_32);
      if (likely(__pyx_t_29)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
        __Pyx_INCREF(__pyx_t_29);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_32, function);
      }
    }
    __pyx_t_28 = (__pyx_t_29) ? __Pyx_PyObject_Call2Args(__pyx_t_32, __pyx_t_29, __pyx_t_30) : __Pyx_PyObject_CallOneArg(__pyx_t_32, __pyx_t_30);
    __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_numpy); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_29 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_n_s_asarray); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_31 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_29))) {
      __pyx_t_31 = PyMethod_GET_SELF(__pyx_t_29);
      if (likely(__pyx_t_31)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_29);
        __Pyx_INCREF(__pyx_t_31);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_29, function);
      }
    }
    __pyx_t_32 = (__pyx_t_31) ? __Pyx_PyObject_Call2Args(__pyx_t_29, __pyx_t_31, __pyx_t_30) : __Pyx_PyObject_CallOneArg(__pyx_t_29, __pyx_t_30);
    __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = PyTuple_New(3); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_32);
    PyTuple_SET_ITEM(__pyx_t_29, 2, __pyx_t_32);
    __pyx_t_27 = 0;
    __pyx_t_28 = 0;
    __pyx_t_32 = 0;
    __pyx_v_res = __pyx_t_29;
    __pyx_t_29 = 0;

    /* "pyFAI/ext/_distortion.pyx":781
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 *     if format == "csr":             # <<<<<<<<<<<<<<
 *         indptr = numpy.zeros(bins + 1, dtype=numpy.int32)
 *         # cumsum
 */
    goto __pyx_L54;
  }

  /* "pyFAI/ext/_distortion.pyx":805
 *             data[i] = large_data[idx]
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 *     elif format == "lut":             # <<<<<<<<<<<<<<
 *         lut_size = numpy.asarray(pixel_count).max()
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_format, __pyx_n_u_lut, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 805, __pyx_L1_error)
  if (likely(__pyx_t_7)) {

    /* "pyFAI/ext/_distortion.pyx":806
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 *     elif format == "lut":
 *         lut_size = numpy.asarray(pixel_count).max()             # <<<<<<<<<<<<<<
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)
 *         pixel_count[:] = 0
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_numpy); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_asarray); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_28 = __pyx_memoryview_fromslice(__pyx_v_pixel_count, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_30 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_27))) {
      __pyx_t_30 = PyMethod_GET_SELF(__pyx_t_27);
      if (likely(__pyx_t_30)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_27);
        __Pyx_INCREF(__pyx_t_30);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_27, function);
      }
    }
    __pyx_t_32 = (__pyx_t_30) ? __Pyx_PyObject_Call2Args(__pyx_t_27, __pyx_t_30, __pyx_t_28) : __Pyx_PyObject_CallOneArg(__pyx_t_27, __pyx_t_28);
    __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_max); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_27))) {
      __pyx_t_32 = PyMethod_GET_SELF(__pyx_t_27);
      if (likely(__pyx_t_32)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_27);
        __Pyx_INCREF(__pyx_t_32);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_27, function);
      }
    }
    __pyx_t_29 = (__pyx_t_32) ? __Pyx_PyObject_CallOneArg(__pyx_t_27, __pyx_t_32) : __Pyx_PyObject_CallNoArg(__pyx_t_27);
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_29); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_v_lut_size = __pyx_t_5;

    /* "pyFAI/ext/_distortion.pyx":807
 *     elif format == "lut":
 *         lut_size = numpy.asarray(pixel_count).max()
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)             # <<<<<<<<<<<<<<
 *         pixel_count[:] = 0
 *         logger.info("LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 */
    __Pyx_GetModuleGlobalName(__pyx_t_29, __pyx_n_s_numpy); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_zeros); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_32 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_28 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_GIVEREF(__pyx_t_32);
    PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_t_32);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_28);
    __pyx_t_32 = 0;
    __pyx_t_28 = 0;
    if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_shape, __pyx_t_30) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_lut_d); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_dtype, __pyx_t_30) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_empty_tuple, __pyx_t_29); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_38 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(__pyx_t_30, PyBUF_WRITABLE); if (unlikely(!__pyx_t_38.memview)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_v_lut = __pyx_t_38;
    __pyx_t_38.memview = NULL;
    __pyx_t_38.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":808
 *         lut_size = numpy.asarray(pixel_count).max()
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)
 *         pixel_count[:] = 0             # <<<<<<<<<<<<<<
 *         logger.info("LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 *                     (lut.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 */
    {
        int32_t __pyx_temp_scalar = 0;
        {
            Py_ssize_t __pyx_temp_extent = __pyx_v_pixel_count.shape[0];
            Py_ssize_t __pyx_temp_idx;
            int32_t *__pyx_temp_pointer = (int32_t *) __pyx_v_pixel_count.data;
            for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
              *((int32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
              __pyx_temp_pointer += 1;
            }
        }
    }

    /* "pyFAI/ext/_distortion.pyx":809
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)
 *         pixel_count[:] = 0
 *         logger.info("LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",             # <<<<<<<<<<<<<<
 *                     (lut.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 *         for idx in range(counter + 1):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_29, __pyx_n_s_logger); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_info); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

    /* "pyFAI/ext/_distortion.pyx":810
 *         pixel_count[:] = 0
 *         logger.info("LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 *                     (lut.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))             # <<<<<<<<<<<<<<
 *         for idx in range(counter + 1):
 *             bin_number = idx_bin[idx]
 */
    __pyx_t_29 = __pyx_memoryview_fromslice(__pyx_v_lut, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, 0);; if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_nbytes); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_28, __pyx_float_1_0e6, 1.0e6, 0, 0); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyInt_From_int(__pyx_v_lut_size); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_32 = PyFloat_FromDouble(((1.0 * __pyx_v_counter) / ((double)__pyx_v_bins))); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_31 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_27))) {
      __pyx_t_31 = PyMethod_GET_SELF(__pyx_t_27);
      if (likely(__pyx_t_31)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_27);
        __Pyx_INCREF(__pyx_t_31);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_27, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_27)) {
      PyObject *__pyx_temp[5] = {__pyx_t_31, __pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p, __pyx_t_29, __pyx_t_28, __pyx_t_32};
      __pyx_t_30 = __Pyx_PyFunction_FastCall(__pyx_t_27, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
      __Pyx_GOTREF(__pyx_t_30);
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_27)) {
      PyObject *__pyx_temp[5] = {__pyx_t_31, __pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p, __pyx_t_29, __pyx_t_28, __pyx_t_32};
      __pyx_t_30 = __Pyx_PyCFunction_FastCall(__pyx_t_27, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
      __Pyx_GOTREF(__pyx_t_30);
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    } else
    #endif
    {
      __pyx_t_33 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_33);
      if (__pyx_t_31) {
        __Pyx_GIVEREF(__pyx_t_31); PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_31); __pyx_t_31 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p);
      __Pyx_GIVEREF(__pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p);
      PyTuple_SET_ITEM(__pyx_t_33, 0+__pyx_t_5, __pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p);
      __Pyx_GIVEREF(__pyx_t_29);
      PyTuple_SET_ITEM(__pyx_t_33, 1+__pyx_t_5, __pyx_t_29);
      __Pyx_GIVEREF(__pyx_t_28);
      PyTuple_SET_ITEM(__pyx_t_33, 2+__pyx_t_5, __pyx_t_28);
      __Pyx_GIVEREF(__pyx_t_32);
      PyTuple_SET_ITEM(__pyx_t_33, 3+__pyx_t_5, __pyx_t_32);
      __pyx_t_29 = 0;
      __pyx_t_28 = 0;
      __pyx_t_32 = 0;
      __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_33, NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_30);
      __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    }
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

    /* "pyFAI/ext/_distortion.pyx":811
 *         logger.info("LUT matrix: %.3f MByte; Max source pixel in target: %i, average splitting: %.2f",
 *                     (lut.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 *         for idx in range(counter + 1):             # <<<<<<<<<<<<<<
 *             bin_number = idx_bin[idx]
 *             i = pixel_count[bin_number]
 */
    __pyx_t_36 = (__pyx_v_counter + 1);
    __pyx_t_37 = __pyx_t_36;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_37; __pyx_t_5+=1) {
      __pyx_v_idx = __pyx_t_5;

      /* "pyFAI/ext/_distortion.pyx":812
 *                     (lut.nbytes) / 1.0e6, lut_size, (1.0 * counter / bins))
 *         for idx in range(counter + 1):
 *             bin_number = idx_bin[idx]             # <<<<<<<<<<<<<<
 *             i = pixel_count[bin_number]
 *             lut[bin_number, i].idx = idx_pixel[idx]
 */
      __pyx_t_15 = __pyx_v_idx;
      __pyx_v_bin_number = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_bin.data) + __pyx_t_15)) )));

      /* "pyFAI/ext/_distortion.pyx":813
 *         for idx in range(counter + 1):
 *             bin_number = idx_bin[idx]
 *             i = pixel_count[bin_number]             # <<<<<<<<<<<<<<
 *             lut[bin_number, i].idx = idx_pixel[idx]
 *             lut[bin_number, i].coef = large_data[idx]
 */
      __pyx_t_15 = __pyx_v_bin_number;
      __pyx_v_i = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_15)) )));

      /* "pyFAI/ext/_distortion.pyx":814
 *             bin_number = idx_bin[idx]
 *             i = pixel_count[bin_number]
 *             lut[bin_number, i].idx = idx_pixel[idx]             # <<<<<<<<<<<<<<
 *             lut[bin_number, i].coef = large_data[idx]
 *             pixel_count[bin_number] += 1
 */
      __pyx_t_15 = __pyx_v_idx;
      __pyx_t_16 = __pyx_v_bin_number;
      __pyx_t_17 = __pyx_v_i;
      (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_16 * __pyx_v_lut.strides[0]) ) + __pyx_t_17 * __pyx_v_lut.strides[1]) ))).idx = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_idx_pixel.data) + __pyx_t_15)) )));

      /* "pyFAI/ext/_distortion.pyx":815
 *             i = pixel_count[bin_number]
 *             lut[bin_number, i].idx = idx_pixel[idx]
 *             lut[bin_number, i].coef = large_data[idx]             # <<<<<<<<<<<<<<
 *             pixel_count[bin_number] += 1
 *         res = numpy.asarray(lut)
 */
      __pyx_t_17 = __pyx_v_idx;
      __pyx_t_16 = __pyx_v_bin_number;
      __pyx_t_15 = __pyx_v_i;
      (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_16 * __pyx_v_lut.strides[0]) ) + __pyx_t_15 * __pyx_v_lut.strides[1]) ))).coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_large_data.data) + __pyx_t_17)) )));

      /* "pyFAI/ext/_distortion.pyx":816
 *             lut[bin_number, i].idx = idx_pixel[idx]
 *             lut[bin_number, i].coef = large_data[idx]
 *             pixel_count[bin_number] += 1             # <<<<<<<<<<<<<<
 *         res = numpy.asarray(lut)
 *     else:
 */
      __pyx_t_15 = __pyx_v_bin_number;
      *((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_pixel_count.data) + __pyx_t_15)) )) += 1;
    }

    /* "pyFAI/ext/_distortion.pyx":817
 *             lut[bin_number, i].coef = large_data[idx]
 *             pixel_count[bin_number] += 1
 *         res = numpy.asarray(lut)             # <<<<<<<<<<<<<<
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_numpy); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_33 = __Pyx_PyObject_GetAttrStr(__pyx_t_27, __pyx_n_s_asarray); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_33);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __pyx_t_27 = __pyx_memoryview_fromslice(__pyx_v_lut, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, 0);; if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_32 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_33))) {
      __pyx_t_32 = PyMethod_GET_SELF(__pyx_t_33);
      if (likely(__pyx_t_32)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_33);
        __Pyx_INCREF(__pyx_t_32);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_33, function);
      }
    }
    __pyx_t_30 = (__pyx_t_32) ? __Pyx_PyObject_Call2Args(__pyx_t_33, __pyx_t_32, __pyx_t_27) : __Pyx_PyObject_CallOneArg(__pyx_t_33, __pyx_t_27);
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __pyx_v_res = __pyx_t_30;
    __pyx_t_30 = 0;

    /* "pyFAI/ext/_distortion.pyx":805
 *             data[i] = large_data[idx]
 *         res = (numpy.asarray(data), numpy.asarray(indices), numpy.asarray(indptr))
 *     elif format == "lut":             # <<<<<<<<<<<<<<
 *         lut_size = numpy.asarray(pixel_count).max()
 *         lut = numpy.zeros(shape=(bins, lut_size), dtype=lut_d)
 */
    goto __pyx_L54;
  }

  /* "pyFAI/ext/_distortion.pyx":819
 *         res = numpy.asarray(lut)
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  /*else*/ {
    __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_INCREF(__pyx_kp_u_Unimplemented_sparse_matrix_form);
    __Pyx_GIVEREF(__pyx_kp_u_Unimplemented_sparse_matrix_form);
    PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_kp_u_Unimplemented_sparse_matrix_form);
    __Pyx_INCREF(__pyx_v_format);
    __Pyx_GIVEREF(__pyx_v_format);
    PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_v_format);
    __pyx_t_33 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_t_30, NULL); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_33);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_Raise(__pyx_t_33, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __PYX_ERR(0, 819, __pyx_L1_error)
  }
  __pyx_L54:;

  /* "pyFAI/ext/_distortion.pyx":820
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":630
 * 
 * 
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                 shape,
 *                 max_pixel_size=(8, 8),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __PYX_XDEC_MEMVIEW(&__pyx_t_38, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_idx_bin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_idx_pixel, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pixel_count, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_large_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut, 1);
  __Pyx_XDECREF(__pyx_v_res);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":823
 * 
 * 
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                    shape,
 *                    max_pixel_size=(8, 8),
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_15calc_sparse_v2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_14calc_sparse_v2[] = "calc_sparse_v2(float32_t[:, :, :, ::1] pos, shape, max_pixel_size=(8, 8), int8_t[:, ::1] mask=None, format=u'csr', int bins_per_pixel=8, builder_config=None)\nCalculate the look-up table (or CSR) using OpenMP\n    :param pos: 4D position array\n    :param shape: output shape\n    :param max_pixel_size: (2-tuple of int) size of a buffer covering the largest pixel\n    :param format: can be \"CSR\" or \"LUT\"\n    :param bins_per_pixel: average splitting factor (number of pixels per bin) #deprecated\n    :return: look-up table in CSR/LUT format\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_15calc_sparse_v2 = {"calc_sparse_v2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_15calc_sparse_v2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_14calc_sparse_v2};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_15calc_sparse_v2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_max_pixel_size = 0;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_format = 0;
  int __pyx_v_bins_per_pixel;
  PyObject *__pyx_v_builder_config = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_sparse_v2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_shape,&__pyx_n_s_max_pixel_size,&__pyx_n_s_mask,&__pyx_n_s_format,&__pyx_n_s_bins_per_pixel,&__pyx_n_s_builder_config,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":825
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,
 *                    shape,
 *                    max_pixel_size=(8, 8),             # <<<<<<<<<<<<<<
 *                    int8_t[:, ::1] mask=None,
 *                    format="csr",
 */
    values[2] = ((PyObject *)__pyx_tuple__15);
    values[4] = ((PyObject *)__pyx_n_u_csr);

    /* "pyFAI/ext/_distortion.pyx":829
 *                    format="csr",
 *                    int bins_per_pixel=8,
 *                    builder_config=None):             # <<<<<<<<<<<<<<
 *     """Calculate the look-up table (or CSR) using OpenMP
 *     :param pos: 4D position array
 */
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_sparse_v2", 0, 2, 7, 1); __PYX_ERR(0, 823, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_pixel_size);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bins_per_pixel);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_builder_config);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_sparse_v2") < 0)) __PYX_ERR(0, 823, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pos = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pos.memview)) __PYX_ERR(0, 823, __pyx_L3_error)
    __pyx_v_shape = values[1];
    __pyx_v_max_pixel_size = values[2];
    if (values[3]) {
      __pyx_v_mask = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_mask.memview)) __PYX_ERR(0, 826, __pyx_L3_error)
    } else {
      __pyx_v_mask = __pyx_k__17;
      __PYX_INC_MEMVIEW(&__pyx_v_mask, 1);
    }
    __pyx_v_format = values[4];
    if (values[5]) {
      __pyx_v_bins_per_pixel = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_bins_per_pixel == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    } else {
      __pyx_v_bins_per_pixel = ((int)8);
    }
    __pyx_v_builder_config = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_sparse_v2", 0, 2, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 823, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_sparse_v2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_pos.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "pos"); __PYX_ERR(0, 823, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_14calc_sparse_v2(__pyx_self, __pyx_v_pos, __pyx_v_shape, __pyx_v_max_pixel_size, __pyx_v_mask, __pyx_v_format, __pyx_v_bins_per_pixel, __pyx_v_builder_config);

  /* "pyFAI/ext/_distortion.pyx":823
 * 
 * 
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                    shape,
 *                    max_pixel_size=(8, 8),
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_14calc_sparse_v2(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_pos, PyObject *__pyx_v_shape, PyObject *__pyx_v_max_pixel_size, __Pyx_memviewslice __pyx_v_mask, PyObject *__pyx_v_format, int __pyx_v_bins_per_pixel, PyObject *__pyx_v_builder_config) {
  int __pyx_v_shape_in0;
  int __pyx_v_shape_in1;
  int __pyx_v_shape_out0;
  int __pyx_v_shape_out1;
  int __pyx_v_size_in;
  int __pyx_v_delta0;
  int __pyx_v_delta1;
  int __pyx_v_bins;
  int __pyx_v_large_size;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_ms;
  int __pyx_v_ml;
  int __pyx_v_ns;
  int __pyx_v_nl;
  int __pyx_v_offset0;
  int __pyx_v_offset1;
  int __pyx_v_box_size0;
  int __pyx_v_box_size1;
  int __pyx_v_counter;
  int __pyx_v_bin_number;
  int __pyx_v_idx;
  int __pyx_v_err_cnt;
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_pAB;
  float __pyx_v_pBC;
  float __pyx_v_pCD;
  float __pyx_v_pDA;
  float __pyx_v_cAB;
  float __pyx_v_cBC;
  float __pyx_v_cCD;
  float __pyx_v_cDA;
  float __pyx_v_area;
  float __pyx_v_value;
  float __pyx_v_foffset0;
  float __pyx_v_foffset1;
  __Pyx_memviewslice __pyx_v_buffer = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_mask;
  struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *__pyx_v_builder = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  Py_ssize_t __pyx_t_32;
  Py_UCS4 __pyx_t_33;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_sparse_v2", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "pyFAI/ext/_distortion.pyx":840
 *     cdef:
 *         int shape_in0, shape_in1, shape_out0, shape_out1, size_in, delta0, delta1, bins, large_size
 *     format = format.lower()             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_lower); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":841
 *         int shape_in0, shape_in1, shape_out0, shape_out1, size_in, delta0, delta1, bins, large_size
 *     format = format.lower()
 *     shape_out0, shape_out1 = shape             # <<<<<<<<<<<<<<
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 841, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 841, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_shape_out0 = __pyx_t_5;
  __pyx_v_shape_out1 = __pyx_t_6;

  /* "pyFAI/ext/_distortion.pyx":842
 *     format = format.lower()
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size             # <<<<<<<<<<<<<<
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_max_pixel_size))) || (PyList_CheckExact(__pyx_v_max_pixel_size))) {
    PyObject* sequence = __pyx_v_max_pixel_size;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 842, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_max_pixel_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 842, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 842, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_delta0 = __pyx_t_6;
  __pyx_v_delta1 = __pyx_t_5;

  /* "pyFAI/ext/_distortion.pyx":843
 *     shape_out0, shape_out1 = shape
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1             # <<<<<<<<<<<<<<
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]
 */
  __pyx_v_bins = (__pyx_v_shape_out0 * __pyx_v_shape_out1);

  /* "pyFAI/ext/_distortion.pyx":844
 *     delta0, delta1 = max_pixel_size
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel             # <<<<<<<<<<<<<<
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]
 */
  __pyx_v_large_size = (__pyx_v_bins * __pyx_v_bins_per_pixel);

  /* "pyFAI/ext/_distortion.pyx":845
 *     bins = shape_out0 * shape_out1
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]             # <<<<<<<<<<<<<<
 *     shape_in1 = pos.shape[1]
 *     size_in = shape_in0 * shape_in1
 */
  __pyx_v_shape_in0 = (__pyx_v_pos.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":846
 *     large_size = bins * bins_per_pixel
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]             # <<<<<<<<<<<<<<
 *     size_in = shape_in0 * shape_in1
 *     cdef:
 */
  __pyx_v_shape_in1 = (__pyx_v_pos.shape[1]);

  /* "pyFAI/ext/_distortion.pyx":847
 *     shape_in0 = pos.shape[0]
 *     shape_in1 = pos.shape[1]
 *     size_in = shape_in0 * shape_in1             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, j, ms, ml, ns, nl
 */
  __pyx_v_size_in = (__pyx_v_shape_in0 * __pyx_v_shape_in1);

  /* "pyFAI/ext/_distortion.pyx":852
 *         int offset0, offset1, box_size0, box_size1
 *         int counter, bin_number
 *         int idx, err_cnt = 0             # <<<<<<<<<<<<<<
 *         float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA,
 *         float area, value, foffset0, foffset1
 */
  __pyx_v_err_cnt = 0;

  /* "pyFAI/ext/_distortion.pyx":856
 *         float area, value, foffset0, foffset1
 *         float32_t[:, ::1] buffer
 *         bint do_mask = mask is not None             # <<<<<<<<<<<<<<
 * 
 *     if do_mask:
 */
  __pyx_v_do_mask = (((PyObject *) __pyx_v_mask.memview) != Py_None);

  /* "pyFAI/ext/_distortion.pyx":858
 *         bint do_mask = mask is not None
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 */
  __pyx_t_7 = (__pyx_v_do_mask != 0);
  if (__pyx_t_7) {

    /* "pyFAI/ext/_distortion.pyx":859
 * 
 *     if do_mask:
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"             # <<<<<<<<<<<<<<
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape_in0 == (__pyx_v_mask.shape[0])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_in0_mask_shape_0);
        __PYX_ERR(0, 859, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":860
 *     if do_mask:
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"             # <<<<<<<<<<<<<<
 * 
 *     # Here we create a builder:
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_shape_in1 == (__pyx_v_mask.shape[1])) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_in1_mask_shape_1);
        __PYX_ERR(0, 860, __pyx_L1_error)
      }
    }
    #endif

    /* "pyFAI/ext/_distortion.pyx":858
 *         bint do_mask = mask is not None
 * 
 *     if do_mask:             # <<<<<<<<<<<<<<
 *         assert shape_in0 == mask.shape[0], "shape_in0 == mask.shape[0]"
 *         assert shape_in1 == mask.shape[1], "shape_in1 == mask.shape[1]"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":863
 * 
 *     # Here we create a builder:
 *     if builder_config is None:             # <<<<<<<<<<<<<<
 *         builder = SparseBuilder(bins, block_size=6, heap_size=bins)
 *     else:
 */
  __pyx_t_7 = (__pyx_v_builder_config == Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "pyFAI/ext/_distortion.pyx":864
 *     # Here we create a builder:
 *     if builder_config is None:
 *         builder = SparseBuilder(bins, block_size=6, heap_size=bins)             # <<<<<<<<<<<<<<
 *     else:
 *         builder = SparseBuilder(bins, **builder_config)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_block_size, __pyx_int_6) < 0) __PYX_ERR(0, 864, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_heap_size, __pyx_t_3) < 0) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder), __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_builder = ((struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyFAI/ext/_distortion.pyx":863
 * 
 *     # Here we create a builder:
 *     if builder_config is None:             # <<<<<<<<<<<<<<
 *         builder = SparseBuilder(bins, block_size=6, heap_size=bins)
 *     else:
 */
    goto __pyx_L8;
  }

  /* "pyFAI/ext/_distortion.pyx":866
 *         builder = SparseBuilder(bins, block_size=6, heap_size=bins)
 *     else:
 *         builder = SparseBuilder(bins, **builder_config)             # <<<<<<<<<<<<<<
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_builder_config == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
      __PYX_ERR(0, 866, __pyx_L1_error)
    }
    if (likely(PyDict_CheckExact(__pyx_v_builder_config))) {
      __pyx_t_3 = PyDict_Copy(__pyx_v_builder_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_3 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_builder_config, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder), __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_builder = ((struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder *)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L8:;

  /* "pyFAI/ext/_distortion.pyx":867
 *     else:
 *         builder = SparseBuilder(bins, **builder_config)
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     counter = -1  # bin index
 *     with nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_buffer = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":868
 *         builder = SparseBuilder(bins, **builder_config)
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index             # <<<<<<<<<<<<<<
 *     with nogil:
 *         # i, j, idx are indices of the raw image uncorrected
 */
  __pyx_v_counter = -1;

  /* "pyFAI/ext/_distortion.pyx":869
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "pyFAI/ext/_distortion.pyx":871
 *     with nogil:
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):             # <<<<<<<<<<<<<<
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 */
        __pyx_t_5 = __pyx_v_size_in;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_6; __pyx_t_12+=1) {
          __pyx_v_idx = __pyx_t_12;

          /* "pyFAI/ext/_distortion.pyx":872
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 *             i = idx // shape_in1             # <<<<<<<<<<<<<<
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:
 */
          __pyx_v_i = (__pyx_v_idx / __pyx_v_shape_in1);

          /* "pyFAI/ext/_distortion.pyx":873
 *         for idx in range(size_in):
 *             i = idx // shape_in1
 *             j = idx % shape_in1             # <<<<<<<<<<<<<<
 *             if do_mask and mask[i, j]:
 *                 continue
 */
          __pyx_v_j = (__pyx_v_idx % __pyx_v_shape_in1);

          /* "pyFAI/ext/_distortion.pyx":874
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 */
          __pyx_t_7 = (__pyx_v_do_mask != 0);
          if (__pyx_t_7) {
          } else {
            __pyx_t_8 = __pyx_t_7;
            goto __pyx_L15_bool_binop_done;
          }
          __pyx_t_13 = __pyx_v_i;
          __pyx_t_14 = __pyx_v_j;
          __pyx_t_7 = ((*((int8_t *) ( /* dim=1 */ ((char *) (((int8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_13 * __pyx_v_mask.strides[0]) )) + __pyx_t_14)) ))) != 0);
          __pyx_t_8 = __pyx_t_7;
          __pyx_L15_bool_binop_done:;
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":875
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:
 *                 continue             # <<<<<<<<<<<<<<
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0
 */
            goto __pyx_L12_continue;

            /* "pyFAI/ext/_distortion.pyx":874
 *             i = idx // shape_in1
 *             j = idx % shape_in1
 *             if do_mask and mask[i, j]:             # <<<<<<<<<<<<<<
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 */
          }

          /* "pyFAI/ext/_distortion.pyx":876
 *             if do_mask and mask[i, j]:
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index             # <<<<<<<<<<<<<<
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]
 */
          __pyx_v_idx = ((__pyx_v_i * __pyx_v_shape_in1) + __pyx_v_j);

          /* "pyFAI/ext/_distortion.pyx":877
 *                 continue
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0             # <<<<<<<<<<<<<<
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]
 */
          {
              __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_temp_scalar = 0.0;
              {
                  Py_ssize_t __pyx_temp_extent = __pyx_v_buffer.shape[0] * __pyx_v_buffer.shape[1];
                  Py_ssize_t __pyx_temp_idx;
                  __pyx_t_5pyFAI_3ext_12shared_types_float32_t *__pyx_temp_pointer = (__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_buffer.data;
                  for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
                    *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
                    __pyx_temp_pointer += 1;
                  }
              }
          }

          /* "pyFAI/ext/_distortion.pyx":878
 *             idx = i * shape_in1 + j  # pixel index
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]             # <<<<<<<<<<<<<<
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]
 */
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_13 = __pyx_v_j;
          __pyx_t_15 = 0;
          __pyx_t_16 = 0;
          __pyx_v_A0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":879
 *             buffer[:, :] = 0.0
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]             # <<<<<<<<<<<<<<
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_13 = 0;
          __pyx_t_14 = 1;
          __pyx_v_A1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

          /* "pyFAI/ext/_distortion.pyx":880
 *             A0 = pos[i, j, 0, 0]
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]             # <<<<<<<<<<<<<<
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]
 */
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_13 = __pyx_v_j;
          __pyx_t_15 = 1;
          __pyx_t_16 = 0;
          __pyx_v_B0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":881
 *             A1 = pos[i, j, 0, 1]
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]             # <<<<<<<<<<<<<<
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_13 = 1;
          __pyx_t_14 = 1;
          __pyx_v_B1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

          /* "pyFAI/ext/_distortion.pyx":882
 *             B0 = pos[i, j, 1, 0]
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]             # <<<<<<<<<<<<<<
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]
 */
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_13 = __pyx_v_j;
          __pyx_t_15 = 2;
          __pyx_t_16 = 0;
          __pyx_v_C0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":883
 *             B1 = pos[i, j, 1, 1]
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]             # <<<<<<<<<<<<<<
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_13 = 2;
          __pyx_t_14 = 1;
          __pyx_v_C1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

          /* "pyFAI/ext/_distortion.pyx":884
 *             C0 = pos[i, j, 2, 0]
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]             # <<<<<<<<<<<<<<
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 */
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_13 = __pyx_v_j;
          __pyx_t_15 = 3;
          __pyx_t_16 = 0;
          __pyx_v_D0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_14 * __pyx_v_pos.strides[0]) ) + __pyx_t_13 * __pyx_v_pos.strides[1]) ) + __pyx_t_15 * __pyx_v_pos.strides[2]) )) + __pyx_t_16)) )));

          /* "pyFAI/ext/_distortion.pyx":885
 *             C1 = pos[i, j, 2, 1]
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]             # <<<<<<<<<<<<<<
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 */
          __pyx_t_16 = __pyx_v_i;
          __pyx_t_15 = __pyx_v_j;
          __pyx_t_13 = 3;
          __pyx_t_14 = 1;
          __pyx_v_D1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_16 * __pyx_v_pos.strides[0]) ) + __pyx_t_15 * __pyx_v_pos.strides[1]) ) + __pyx_t_13 * __pyx_v_pos.strides[2]) )) + __pyx_t_14)) )));

          /* "pyFAI/ext/_distortion.pyx":886
 *             D0 = pos[i, j, 3, 0]
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)             # <<<<<<<<<<<<<<
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0
 */
          __pyx_v_foffset0 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0);

          /* "pyFAI/ext/_distortion.pyx":887
 *             D1 = pos[i, j, 3, 1]
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)             # <<<<<<<<<<<<<<
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1
 */
          __pyx_v_foffset1 = __pyx_f_5pyFAI_3ext_11_distortion__floor_min4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1);

          /* "pyFAI/ext/_distortion.pyx":888
 *             foffset0 = _floor_min4(A0, B0, C0, D0)
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0             # <<<<<<<<<<<<<<
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 */
          __pyx_v_offset0 = ((int)__pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":889
 *             foffset1 = _floor_min4(A1, B1, C1, D1)
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1             # <<<<<<<<<<<<<<
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 */
          __pyx_v_offset1 = ((int)__pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":890
 *             offset0 = <int> foffset0
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0             # <<<<<<<<<<<<<<
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 */
          __pyx_v_box_size0 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A0, __pyx_v_B0, __pyx_v_C0, __pyx_v_D0)) - __pyx_v_offset0);

          /* "pyFAI/ext/_distortion.pyx":891
 *             offset1 = <int> foffset1
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1             # <<<<<<<<<<<<<<
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 *                 # Increase size of the buffer
 */
          __pyx_v_box_size1 = (((int)__pyx_f_5pyFAI_3ext_11_distortion__ceil_max4(__pyx_v_A1, __pyx_v_B1, __pyx_v_C1, __pyx_v_D1)) - __pyx_v_offset1);

          /* "pyFAI/ext/_distortion.pyx":892
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 */
          __pyx_t_7 = ((__pyx_v_box_size0 > __pyx_v_delta0) != 0);
          if (!__pyx_t_7) {
          } else {
            __pyx_t_8 = __pyx_t_7;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_7 = ((__pyx_v_box_size1 > __pyx_v_delta1) != 0);
          __pyx_t_8 = __pyx_t_7;
          __pyx_L18_bool_binop_done:;
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":894
 *             if (box_size0 > delta0) or (box_size1 > delta1):
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0             # <<<<<<<<<<<<<<
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:
 */
            if (((__pyx_v_offset0 > __pyx_v_delta0) != 0)) {
              __pyx_t_17 = __pyx_v_offset0;
            } else {
              __pyx_t_17 = __pyx_v_delta0;
            }
            __pyx_v_delta0 = __pyx_t_17;

            /* "pyFAI/ext/_distortion.pyx":895
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 */
            if (((__pyx_v_offset1 > __pyx_v_delta1) != 0)) {
              __pyx_t_17 = __pyx_v_offset1;
            } else {
              __pyx_t_17 = __pyx_v_delta1;
            }
            __pyx_v_delta1 = __pyx_t_17;

            /* "pyFAI/ext/_distortion.pyx":896
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
            {
                #ifdef WITH_THREAD
                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                #endif
                /*try:*/ {

                  /* "pyFAI/ext/_distortion.pyx":897
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 * 
 *             A0 = A0 - foffset0
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_10);
                  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_delta0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_10);
                  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_delta1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_10);
                  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_10);
                  __Pyx_GIVEREF(__pyx_t_1);
                  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
                  __pyx_t_10 = 0;
                  __pyx_t_1 = 0;
                  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_10);
                  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_float32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 897, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
                  __pyx_v_buffer = __pyx_t_11;
                  __pyx_t_11.memview = NULL;
                  __pyx_t_11.data = NULL;
                }

                /* "pyFAI/ext/_distortion.pyx":896
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 *                 delta1 = offset1 if offset1 > delta1 else delta1
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L24;
                  }
                  __pyx_L23_error: {
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    goto __pyx_L10_error;
                  }
                  __pyx_L24:;
                }
            }

            /* "pyFAI/ext/_distortion.pyx":892
 *             box_size0 = (<int> _ceil_max4(A0, B0, C0, D0)) - offset0
 *             box_size1 = (<int> _ceil_max4(A1, B1, C1, D1)) - offset1
 *             if (box_size0 > delta0) or (box_size1 > delta1):             # <<<<<<<<<<<<<<
 *                 # Increase size of the buffer
 *                 delta0 = offset0 if offset0 > delta0 else delta0
 */
          }

          /* "pyFAI/ext/_distortion.pyx":899
 *                     buffer = numpy.zeros((delta0, delta1), dtype=numpy.float32)
 * 
 *             A0 = A0 - foffset0             # <<<<<<<<<<<<<<
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0
 */
          __pyx_v_A0 = (__pyx_v_A0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":900
 * 
 *             A0 = A0 - foffset0
 *             A1 = A1 - foffset1             # <<<<<<<<<<<<<<
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1
 */
          __pyx_v_A1 = (__pyx_v_A1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":901
 *             A0 = A0 - foffset0
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0             # <<<<<<<<<<<<<<
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0
 */
          __pyx_v_B0 = (__pyx_v_B0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":902
 *             A1 = A1 - foffset1
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1             # <<<<<<<<<<<<<<
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1
 */
          __pyx_v_B1 = (__pyx_v_B1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":903
 *             B0 = B0 - foffset0
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0             # <<<<<<<<<<<<<<
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0
 */
          __pyx_v_C0 = (__pyx_v_C0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":904
 *             B1 = B1 - foffset1
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1             # <<<<<<<<<<<<<<
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 */
          __pyx_v_C1 = (__pyx_v_C1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":905
 *             C0 = C0 - foffset0
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0             # <<<<<<<<<<<<<<
 *             D1 = D1 - foffset1
 *             if B0 != A0:
 */
          __pyx_v_D0 = (__pyx_v_D0 - __pyx_v_foffset0);

          /* "pyFAI/ext/_distortion.pyx":906
 *             C1 = C1 - foffset1
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1             # <<<<<<<<<<<<<<
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)
 */
          __pyx_v_D1 = (__pyx_v_D1 - __pyx_v_foffset1);

          /* "pyFAI/ext/_distortion.pyx":907
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 *             if B0 != A0:             # <<<<<<<<<<<<<<
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0
 */
          __pyx_t_8 = ((__pyx_v_B0 != __pyx_v_A0) != 0);
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":908
 *             D1 = D1 - foffset1
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)             # <<<<<<<<<<<<<<
 *                 cAB = A1 - pAB * A0
 *             else:
 */
            __pyx_v_pAB = ((__pyx_v_B1 - __pyx_v_A1) / (__pyx_v_B0 - __pyx_v_A0));

            /* "pyFAI/ext/_distortion.pyx":909
 *             if B0 != A0:
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0             # <<<<<<<<<<<<<<
 *             else:
 *                 pAB = cAB = 0.0
 */
            __pyx_v_cAB = (__pyx_v_A1 - (__pyx_v_pAB * __pyx_v_A0));

            /* "pyFAI/ext/_distortion.pyx":907
 *             D0 = D0 - foffset0
 *             D1 = D1 - foffset1
 *             if B0 != A0:             # <<<<<<<<<<<<<<
 *                 pAB = (B1 - A1) / (B0 - A0)
 *                 cAB = A1 - pAB * A0
 */
            goto __pyx_L25;
          }

          /* "pyFAI/ext/_distortion.pyx":911
 *                 cAB = A1 - pAB * A0
 *             else:
 *                 pAB = cAB = 0.0             # <<<<<<<<<<<<<<
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)
 */
          /*else*/ {
            __pyx_v_pAB = 0.0;
            __pyx_v_cAB = 0.0;
          }
          __pyx_L25:;

          /* "pyFAI/ext/_distortion.pyx":912
 *             else:
 *                 pAB = cAB = 0.0
 *             if C0 != B0:             # <<<<<<<<<<<<<<
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0
 */
          __pyx_t_8 = ((__pyx_v_C0 != __pyx_v_B0) != 0);
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":913
 *                 pAB = cAB = 0.0
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)             # <<<<<<<<<<<<<<
 *                 cBC = B1 - pBC * B0
 *             else:
 */
            __pyx_v_pBC = ((__pyx_v_C1 - __pyx_v_B1) / (__pyx_v_C0 - __pyx_v_B0));

            /* "pyFAI/ext/_distortion.pyx":914
 *             if C0 != B0:
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0             # <<<<<<<<<<<<<<
 *             else:
 *                 pBC = cBC = 0.0
 */
            __pyx_v_cBC = (__pyx_v_B1 - (__pyx_v_pBC * __pyx_v_B0));

            /* "pyFAI/ext/_distortion.pyx":912
 *             else:
 *                 pAB = cAB = 0.0
 *             if C0 != B0:             # <<<<<<<<<<<<<<
 *                 pBC = (C1 - B1) / (C0 - B0)
 *                 cBC = B1 - pBC * B0
 */
            goto __pyx_L26;
          }

          /* "pyFAI/ext/_distortion.pyx":916
 *                 cBC = B1 - pBC * B0
 *             else:
 *                 pBC = cBC = 0.0             # <<<<<<<<<<<<<<
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)
 */
          /*else*/ {
            __pyx_v_pBC = 0.0;
            __pyx_v_cBC = 0.0;
          }
          __pyx_L26:;

          /* "pyFAI/ext/_distortion.pyx":917
 *             else:
 *                 pBC = cBC = 0.0
 *             if D0 != C0:             # <<<<<<<<<<<<<<
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0
 */
          __pyx_t_8 = ((__pyx_v_D0 != __pyx_v_C0) != 0);
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":918
 *                 pBC = cBC = 0.0
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)             # <<<<<<<<<<<<<<
 *                 cCD = C1 - pCD * C0
 *             else:
 */
            __pyx_v_pCD = ((__pyx_v_D1 - __pyx_v_C1) / (__pyx_v_D0 - __pyx_v_C0));

            /* "pyFAI/ext/_distortion.pyx":919
 *             if D0 != C0:
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0             # <<<<<<<<<<<<<<
 *             else:
 *                 pCD = cCD = 0.0
 */
            __pyx_v_cCD = (__pyx_v_C1 - (__pyx_v_pCD * __pyx_v_C0));

            /* "pyFAI/ext/_distortion.pyx":917
 *             else:
 *                 pBC = cBC = 0.0
 *             if D0 != C0:             # <<<<<<<<<<<<<<
 *                 pCD = (D1 - C1) / (D0 - C0)
 *                 cCD = C1 - pCD * C0
 */
            goto __pyx_L27;
          }

          /* "pyFAI/ext/_distortion.pyx":921
 *                 cCD = C1 - pCD * C0
 *             else:
 *                 pCD = cCD = 0.0             # <<<<<<<<<<<<<<
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)
 */
          /*else*/ {
            __pyx_v_pCD = 0.0;
            __pyx_v_cCD = 0.0;
          }
          __pyx_L27:;

          /* "pyFAI/ext/_distortion.pyx":922
 *             else:
 *                 pCD = cCD = 0.0
 *             if A0 != D0:             # <<<<<<<<<<<<<<
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0
 */
          __pyx_t_8 = ((__pyx_v_A0 != __pyx_v_D0) != 0);
          if (__pyx_t_8) {

            /* "pyFAI/ext/_distortion.pyx":923
 *                 pCD = cCD = 0.0
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)             # <<<<<<<<<<<<<<
 *                 cDA = D1 - pDA * D0
 *             else:
 */
            __pyx_v_pDA = ((__pyx_v_A1 - __pyx_v_D1) / (__pyx_v_A0 - __pyx_v_D0));

            /* "pyFAI/ext/_distortion.pyx":924
 *             if A0 != D0:
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0             # <<<<<<<<<<<<<<
 *             else:
 *                 pDA = cDA = 0.0
 */
            __pyx_v_cDA = (__pyx_v_D1 - (__pyx_v_pDA * __pyx_v_D0));

            /* "pyFAI/ext/_distortion.pyx":922
 *             else:
 *                 pCD = cCD = 0.0
 *             if A0 != D0:             # <<<<<<<<<<<<<<
 *                 pDA = (A1 - D1) / (A0 - D0)
 *                 cDA = D1 - pDA * D0
 */
            goto __pyx_L28;
          }

          /* "pyFAI/ext/_distortion.pyx":926
 *                 cDA = D1 - pDA * D0
 *             else:
 *                 pDA = cDA = 0.0             # <<<<<<<<<<<<<<
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)
 */
          /*else*/ {
            __pyx_v_pDA = 0.0;
            __pyx_v_cDA = 0.0;
          }
          __pyx_L28:;

          /* "pyFAI/ext/_distortion.pyx":928
 *                 pDA = cDA = 0.0
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)             # <<<<<<<<<<<<<<
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_B0, __pyx_v_A0, __pyx_v_pAB, __pyx_v_cAB);

          /* "pyFAI/ext/_distortion.pyx":929
 * 
 *             integrate(buffer, B0, A0, pAB, cAB)
 *             integrate(buffer, A0, D0, pDA, cDA)             # <<<<<<<<<<<<<<
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_A0, __pyx_v_D0, __pyx_v_pDA, __pyx_v_cDA);

          /* "pyFAI/ext/_distortion.pyx":930
 *             integrate(buffer, B0, A0, pAB, cAB)
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)             # <<<<<<<<<<<<<<
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_D0, __pyx_v_C0, __pyx_v_pCD, __pyx_v_cCD);

          /* "pyFAI/ext/_distortion.pyx":931
 *             integrate(buffer, A0, D0, pDA, cDA)
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)             # <<<<<<<<<<<<<<
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):
 */
          __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_C0, __pyx_v_B0, __pyx_v_pBC, __pyx_v_cBC);

          /* "pyFAI/ext/_distortion.pyx":932
 *             integrate(buffer, D0, C0, pCD, cCD)
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))             # <<<<<<<<<<<<<<
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 */
          __pyx_v_area = (0.5 * (((__pyx_v_C0 - __pyx_v_A0) * (__pyx_v_D1 - __pyx_v_B1)) - ((__pyx_v_C1 - __pyx_v_A1) * (__pyx_v_D0 - __pyx_v_B0))));

          /* "pyFAI/ext/_distortion.pyx":933
 *             integrate(buffer, C0, B0, pBC, cBC)
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):             # <<<<<<<<<<<<<<
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:
 */
          __pyx_t_17 = __pyx_v_box_size0;
          __pyx_t_18 = __pyx_t_17;
          for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
            __pyx_v_ms = __pyx_t_19;

            /* "pyFAI/ext/_distortion.pyx":934
 *             area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *             for ms in range(box_size0):
 *                 ml = ms + offset0             # <<<<<<<<<<<<<<
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue
 */
            __pyx_v_ml = (__pyx_v_ms + __pyx_v_offset0);

            /* "pyFAI/ext/_distortion.pyx":935
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for ns in range(box_size1):
 */
            __pyx_t_7 = ((__pyx_v_ml < 0) != 0);
            if (!__pyx_t_7) {
            } else {
              __pyx_t_8 = __pyx_t_7;
              goto __pyx_L32_bool_binop_done;
            }
            __pyx_t_7 = ((__pyx_v_ml >= __pyx_v_shape_out0) != 0);
            __pyx_t_8 = __pyx_t_7;
            __pyx_L32_bool_binop_done:;
            if (__pyx_t_8) {

              /* "pyFAI/ext/_distortion.pyx":936
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 for ns in range(box_size1):
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 */
              goto __pyx_L29_continue;

              /* "pyFAI/ext/_distortion.pyx":935
 *             for ms in range(box_size0):
 *                 ml = ms + offset0
 *                 if ml < 0 or ml >= shape_out0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for ns in range(box_size1):
 */
            }

            /* "pyFAI/ext/_distortion.pyx":937
 *                 if ml < 0 or ml >= shape_out0:
 *                     continue
 *                 for ns in range(box_size1):             # <<<<<<<<<<<<<<
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 */
            __pyx_t_20 = __pyx_v_box_size1;
            __pyx_t_21 = __pyx_t_20;
            for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
              __pyx_v_ns = __pyx_t_22;

              /* "pyFAI/ext/_distortion.pyx":939
 *                 for ns in range(box_size1):
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1             # <<<<<<<<<<<<<<
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue
 */
              __pyx_v_nl = (__pyx_v_ns + __pyx_v_offset1);

              /* "pyFAI/ext/_distortion.pyx":940
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:             # <<<<<<<<<<<<<<
 *                         continue
 *                     value = buffer[ms, ns] / area
 */
              __pyx_t_7 = ((__pyx_v_nl < 0) != 0);
              if (!__pyx_t_7) {
              } else {
                __pyx_t_8 = __pyx_t_7;
                goto __pyx_L37_bool_binop_done;
              }
              __pyx_t_7 = ((__pyx_v_nl >= __pyx_v_shape_out1) != 0);
              __pyx_t_8 = __pyx_t_7;
              __pyx_L37_bool_binop_done:;
              if (__pyx_t_8) {

                /* "pyFAI/ext/_distortion.pyx":941
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue             # <<<<<<<<<<<<<<
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:
 */
                goto __pyx_L34_continue;

                /* "pyFAI/ext/_distortion.pyx":940
 *                     # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                     nl = ns + offset1
 *                     if nl < 0 or nl >= shape_out1:             # <<<<<<<<<<<<<<
 *                         continue
 *                     value = buffer[ms, ns] / area
 */
              }

              /* "pyFAI/ext/_distortion.pyx":942
 *                     if nl < 0 or nl >= shape_out1:
 *                         continue
 *                     value = buffer[ms, ns] / area             # <<<<<<<<<<<<<<
 *                     if value == 0.0:
 *                         continue
 */
              __pyx_t_14 = __pyx_v_ms;
              __pyx_t_13 = __pyx_v_ns;
              __pyx_v_value = ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_13)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_area));

              /* "pyFAI/ext/_distortion.pyx":943
 *                         continue
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if value < 0.0 or value > 1.0001:
 */
              __pyx_t_8 = ((__pyx_v_value == 0.0) != 0);
              if (__pyx_t_8) {

                /* "pyFAI/ext/_distortion.pyx":944
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 */
                goto __pyx_L34_continue;

                /* "pyFAI/ext/_distortion.pyx":943
 *                         continue
 *                     value = buffer[ms, ns] / area
 *                     if value == 0.0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if value < 0.0 or value > 1.0001:
 */
              }

              /* "pyFAI/ext/_distortion.pyx":945
 *                     if value == 0.0:
 *                         continue
 *                     if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 */
              __pyx_t_7 = ((__pyx_v_value < 0.0) != 0);
              if (!__pyx_t_7) {
              } else {
                __pyx_t_8 = __pyx_t_7;
                goto __pyx_L41_bool_binop_done;
              }
              __pyx_t_7 = ((__pyx_v_value > 1.0001) != 0);
              __pyx_t_8 = __pyx_t_7;
              __pyx_L41_bool_binop_done:;
              if (__pyx_t_8) {

                /* "pyFAI/ext/_distortion.pyx":947
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                __pyx_t_8 = ((__pyx_v_err_cnt < 0x3E8) != 0);
                if (__pyx_t_8) {

                  /* "pyFAI/ext/_distortion.pyx":948
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 *                             with gil:             # <<<<<<<<<<<<<<
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      /*try:*/ {

                        /* "pyFAI/ext/_distortion.pyx":949
 *                         if err_cnt < 1000:
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])             # <<<<<<<<<<<<<<
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1
 */
                        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_2);
                        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_3);
                        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ms); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_1);
                        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_box_size0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_9);
                        __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_ns); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_10);
                        __pyx_t_23 = __Pyx_PyInt_From_int(__pyx_v_box_size1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_23);
                        __pyx_t_13 = __pyx_v_ms;
                        __pyx_t_14 = __pyx_v_ns;
                        __pyx_t_24 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_13 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        __pyx_t_25 = PyFloat_FromDouble(__pyx_v_area); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        __pyx_t_26 = PyFloat_FromDouble(__pyx_v_value); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_26);
                        __pyx_t_14 = 0;
                        __pyx_t_13 = 0;
                        __pyx_t_27 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_13)) )))); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_27);
                        __pyx_t_13 = 0;
                        __pyx_t_14 = 1;
                        __pyx_t_28 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_13 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_28);
                        __pyx_t_14 = 1;
                        __pyx_t_13 = 0;
                        __pyx_t_29 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_14 * __pyx_v_buffer.strides[0]) )) + __pyx_t_13)) )))); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_13 = 1;
                        __pyx_t_14 = 1;
                        __pyx_t_30 = PyFloat_FromDouble((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_13 * __pyx_v_buffer.strides[0]) )) + __pyx_t_14)) )))); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_30);
                        __pyx_t_31 = PyTuple_New(13); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_2);
                        PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_t_2);
                        __Pyx_GIVEREF(__pyx_t_3);
                        PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_3);
                        __Pyx_GIVEREF(__pyx_t_1);
                        PyTuple_SET_ITEM(__pyx_t_31, 2, __pyx_t_1);
                        __Pyx_GIVEREF(__pyx_t_9);
                        PyTuple_SET_ITEM(__pyx_t_31, 3, __pyx_t_9);
                        __Pyx_GIVEREF(__pyx_t_10);
                        PyTuple_SET_ITEM(__pyx_t_31, 4, __pyx_t_10);
                        __Pyx_GIVEREF(__pyx_t_23);
                        PyTuple_SET_ITEM(__pyx_t_31, 5, __pyx_t_23);
                        __Pyx_GIVEREF(__pyx_t_24);
                        PyTuple_SET_ITEM(__pyx_t_31, 6, __pyx_t_24);
                        __Pyx_GIVEREF(__pyx_t_25);
                        PyTuple_SET_ITEM(__pyx_t_31, 7, __pyx_t_25);
                        __Pyx_GIVEREF(__pyx_t_26);
                        PyTuple_SET_ITEM(__pyx_t_31, 8, __pyx_t_26);
                        __Pyx_GIVEREF(__pyx_t_27);
                        PyTuple_SET_ITEM(__pyx_t_31, 9, __pyx_t_27);
                        __Pyx_GIVEREF(__pyx_t_28);
                        PyTuple_SET_ITEM(__pyx_t_31, 10, __pyx_t_28);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_31, 11, __pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_30);
                        PyTuple_SET_ITEM(__pyx_t_31, 12, __pyx_t_30);
                        __pyx_t_2 = 0;
                        __pyx_t_3 = 0;
                        __pyx_t_1 = 0;
                        __pyx_t_9 = 0;
                        __pyx_t_10 = 0;
                        __pyx_t_23 = 0;
                        __pyx_t_24 = 0;
                        __pyx_t_25 = 0;
                        __pyx_t_26 = 0;
                        __pyx_t_27 = 0;
                        __pyx_t_28 = 0;
                        __pyx_t_29 = 0;
                        __pyx_t_30 = 0;
                        __pyx_t_30 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_31, NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 949, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_30);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

                        /* "pyFAI/ext/_distortion.pyx":950
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))             # <<<<<<<<<<<<<<
 *                             err_cnt += 1
 *                         continue
 */
                        __pyx_t_30 = PyTuple_New(16); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_30);
                        __pyx_t_32 = 0;
                        __pyx_t_33 = 127;
                        __Pyx_INCREF(__pyx_kp_u_A0);
                        __pyx_t_32 += 4;
                        __Pyx_GIVEREF(__pyx_kp_u_A0);
                        PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_kp_u_A0);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_A0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_29 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_29);
                        __pyx_t_29 = 0;
                        __Pyx_INCREF(__pyx_kp_u_A1);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_A1);
                        PyTuple_SET_ITEM(__pyx_t_30, 2, __pyx_kp_u_A1);
                        __pyx_t_29 = PyFloat_FromDouble(__pyx_v_A1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_29), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_30, 3, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_B0);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_B0);
                        PyTuple_SET_ITEM(__pyx_t_30, 4, __pyx_kp_u_B0);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_B0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_29 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_30, 5, __pyx_t_29);
                        __pyx_t_29 = 0;
                        __Pyx_INCREF(__pyx_kp_u_B1);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_B1);
                        PyTuple_SET_ITEM(__pyx_t_30, 6, __pyx_kp_u_B1);
                        __pyx_t_29 = PyFloat_FromDouble(__pyx_v_B1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_29), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_30, 7, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_C0);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_C0);
                        PyTuple_SET_ITEM(__pyx_t_30, 8, __pyx_kp_u_C0);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_C0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_29 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_30, 9, __pyx_t_29);
                        __pyx_t_29 = 0;
                        __Pyx_INCREF(__pyx_kp_u_C1);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_C1);
                        PyTuple_SET_ITEM(__pyx_t_30, 10, __pyx_kp_u_C1);
                        __pyx_t_29 = PyFloat_FromDouble(__pyx_v_C1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_29), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_30, 11, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __Pyx_INCREF(__pyx_kp_u_D0);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_D0);
                        PyTuple_SET_ITEM(__pyx_t_30, 12, __pyx_kp_u_D0);
                        __pyx_t_31 = PyFloat_FromDouble(__pyx_v_D0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __pyx_t_29 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_31), __pyx_empty_unicode); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_29) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_29);
                        __Pyx_GIVEREF(__pyx_t_29);
                        PyTuple_SET_ITEM(__pyx_t_30, 13, __pyx_t_29);
                        __pyx_t_29 = 0;
                        __Pyx_INCREF(__pyx_kp_u_D1);
                        __pyx_t_32 += 5;
                        __Pyx_GIVEREF(__pyx_kp_u_D1);
                        PyTuple_SET_ITEM(__pyx_t_30, 14, __pyx_kp_u_D1);
                        __pyx_t_29 = PyFloat_FromDouble(__pyx_v_D1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_29);
                        __pyx_t_31 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_29), __pyx_empty_unicode); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
                        __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_31) : __pyx_t_33;
                        __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_31);
                        __Pyx_GIVEREF(__pyx_t_31);
                        PyTuple_SET_ITEM(__pyx_t_30, 15, __pyx_t_31);
                        __pyx_t_31 = 0;
                        __pyx_t_31 = __Pyx_PyUnicode_Join(__pyx_t_30, 16, __pyx_t_32, __pyx_t_33); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_31);
                        __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
                        __pyx_t_30 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_31); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 950, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_30);
                        __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                        __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
                      }

                      /* "pyFAI/ext/_distortion.pyx":948
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 *                             with gil:             # <<<<<<<<<<<<<<
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 */
                      /*finally:*/ {
                        /*normal exit:*/{
                          #ifdef WITH_THREAD
                          __Pyx_PyGILState_Release(__pyx_gilstate_save);
                          #endif
                          goto __pyx_L48;
                        }
                        __pyx_L47_error: {
                          #ifdef WITH_THREAD
                          __Pyx_PyGILState_Release(__pyx_gilstate_save);
                          #endif
                          goto __pyx_L10_error;
                        }
                        __pyx_L48:;
                      }
                  }

                  /* "pyFAI/ext/_distortion.pyx":951
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
                  __pyx_v_err_cnt = (__pyx_v_err_cnt + 1);

                  /* "pyFAI/ext/_distortion.pyx":947
 *                     if value < 0.0 or value > 1.0001:
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:             # <<<<<<<<<<<<<<
 *                             with gil:
 *                                 print(i, j, ms, box_size0, ns, box_size1, buffer[ms, ns], area, value, buffer[0, 0], buffer[0, 1], buffer[1, 0], buffer[1, 1])
 */
                }

                /* "pyFAI/ext/_distortion.pyx":952
 *                                 print(" A0=%s; A1=%s; B0=%s; B1=%s; C0=%s; C1=%s; D0=%s; D1=%s" % (A0, A1, B0, B1, C0, C1, D0, D1))
 *                             err_cnt += 1
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *                     bin_number = ml * shape_out1 + nl
 */
                goto __pyx_L34_continue;

                /* "pyFAI/ext/_distortion.pyx":945
 *                     if value == 0.0:
 *                         continue
 *                     if value < 0.0 or value > 1.0001:             # <<<<<<<<<<<<<<
 *                         # here we print pathological cases for debugging
 *                         if err_cnt < 1000:
 */
              }

              /* "pyFAI/ext/_distortion.pyx":954
 *                         continue
 * 
 *                     bin_number = ml * shape_out1 + nl             # <<<<<<<<<<<<<<
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1
 */
              __pyx_v_bin_number = ((__pyx_v_ml * __pyx_v_shape_out1) + __pyx_v_nl);

              /* "pyFAI/ext/_distortion.pyx":956
 *                     bin_number = ml * shape_out1 + nl
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1             # <<<<<<<<<<<<<<
 *                     if counter >= large_size:
 *                         with gil:
 */
              __pyx_v_counter = (__pyx_v_counter + 1);

              /* "pyFAI/ext/_distortion.pyx":957
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1
 *                     if counter >= large_size:             # <<<<<<<<<<<<<<
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 */
              __pyx_t_8 = ((__pyx_v_counter >= __pyx_v_large_size) != 0);
              if (__pyx_t_8) {

                /* "pyFAI/ext/_distortion.pyx":958
 *                     counter += 1
 *                     if counter >= large_size:
 *                         with gil:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 */
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    /*try:*/ {

                      /* "pyFAI/ext/_distortion.pyx":960
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +             # <<<<<<<<<<<<<<
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     builder.cinsert(bin_number, idx, value)
 */
                      __pyx_t_30 = __Pyx_PyInt_From_int(__pyx_v_bins_per_pixel); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 960, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_30);
                      __pyx_t_31 = PyUnicode_Format(__pyx_kp_u_Please_increase_bins_per_pixel_s, __pyx_t_30); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 960, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_31);
                      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

                      /* "pyFAI/ext/_distortion.pyx":959
 *                     if counter >= large_size:
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +             # <<<<<<<<<<<<<<
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 */
                      __pyx_t_30 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Provided_temporary_space_for_sto, __pyx_t_31); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 959, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_30);
                      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

                      /* "pyFAI/ext/_distortion.pyx":961
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))             # <<<<<<<<<<<<<<
 *                     builder.cinsert(bin_number, idx, value)
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 */
                      __pyx_t_31 = PyFloat_FromDouble(ceil((((1.1 * __pyx_v_bins_per_pixel) * __pyx_v_size_in) / ((double)__pyx_v_idx)))); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 961, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_31);
                      __pyx_t_29 = PyUnicode_Format(__pyx_kp_u_The_suggested_value_is_i_or_grea, __pyx_t_31); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 961, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_29);
                      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

                      /* "pyFAI/ext/_distortion.pyx":960
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +             # <<<<<<<<<<<<<<
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     builder.cinsert(bin_number, idx, value)
 */
                      __pyx_t_31 = __Pyx_PyUnicode_Concat(__pyx_t_30, __pyx_t_29); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 960, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_31);
                      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
                      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

                      /* "pyFAI/ext/_distortion.pyx":959
 *                     if counter >= large_size:
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +             # <<<<<<<<<<<<<<
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 */
                      __pyx_t_29 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_31); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 959, __pyx_L53_error)
                      __Pyx_GOTREF(__pyx_t_29);
                      __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
                      __Pyx_Raise(__pyx_t_29, 0, 0, 0);
                      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
                      __PYX_ERR(0, 959, __pyx_L53_error)
                    }

                    /* "pyFAI/ext/_distortion.pyx":958
 *                     counter += 1
 *                     if counter >= large_size:
 *                         with gil:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 */
                    /*finally:*/ {
                      __pyx_L53_error: {
                        #ifdef WITH_THREAD
                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                        #endif
                        goto __pyx_L10_error;
                      }
                    }
                }

                /* "pyFAI/ext/_distortion.pyx":957
 *                     # with gil: #Use the gil to perform an atomic operation
 *                     counter += 1
 *                     if counter >= large_size:             # <<<<<<<<<<<<<<
 *                         with gil:
 *                             raise RuntimeError("Provided temporary space for storage is not enough. " +
 */
              }

              /* "pyFAI/ext/_distortion.pyx":962
 *                                                "Please increase bins_per_pixel=%s. " % bins_per_pixel +
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     builder.cinsert(bin_number, idx, value)             # <<<<<<<<<<<<<<
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 *                 counter, counter / size_in, bins_per_pixel)
 */
              ((struct __pyx_vtabstruct_5pyFAI_3ext_14sparse_builder_SparseBuilder *)__pyx_v_builder->__pyx_vtab)->cinsert(__pyx_v_builder, __pyx_v_bin_number, __pyx_v_idx, __pyx_v_value);
              __pyx_L34_continue:;
            }
            __pyx_L29_continue:;
          }
          __pyx_L12_continue:;
        }
      }

      /* "pyFAI/ext/_distortion.pyx":869
 *     buffer = numpy.empty((delta0, delta1), dtype=numpy.float32)
 *     counter = -1  # bin index
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # i, j, idx are indices of the raw image uncorrected
 *         for idx in range(size_in):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L11;
        }
        __pyx_L10_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L11:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":963
 *                                                "The suggested value is %i or greater." % ceil(1.1 * bins_per_pixel * size_in / idx))
 *                     builder.cinsert(bin_number, idx, value)
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",             # <<<<<<<<<<<<<<
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_n_s_logger); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_n_s_info); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pyFAI/ext/_distortion.pyx":964
 *                     builder.cinsert(bin_number, idx, value)
 *     logger.info("number of elements: %s, average per bin %.3f allocated max: %s",
 *                 counter, counter / size_in, bins_per_pixel)             # <<<<<<<<<<<<<<
 * 
 *     if format == "csr":
 */
  __pyx_t_31 = __Pyx_PyInt_From_int(__pyx_v_counter); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_28 = __Pyx_PyInt_From_int((__pyx_v_counter / __pyx_v_size_in)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_27 = __Pyx_PyInt_From_int(__pyx_v_bins_per_pixel); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_26 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_30))) {
    __pyx_t_26 = PyMethod_GET_SELF(__pyx_t_30);
    if (likely(__pyx_t_26)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_30);
      __Pyx_INCREF(__pyx_t_26);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_30, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_30)) {
    PyObject *__pyx_temp[5] = {__pyx_t_26, __pyx_kp_u_number_of_elements_s_average_per, __pyx_t_31, __pyx_t_28, __pyx_t_27};
    __pyx_t_29 = __Pyx_PyFunction_FastCall(__pyx_t_30, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_30)) {
    PyObject *__pyx_temp[5] = {__pyx_t_26, __pyx_kp_u_number_of_elements_s_average_per, __pyx_t_31, __pyx_t_28, __pyx_t_27};
    __pyx_t_29 = __Pyx_PyCFunction_FastCall(__pyx_t_30, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  } else
  #endif
  {
    __pyx_t_25 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    if (__pyx_t_26) {
      __Pyx_GIVEREF(__pyx_t_26); PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_26); __pyx_t_26 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_number_of_elements_s_average_per);
    __Pyx_GIVEREF(__pyx_kp_u_number_of_elements_s_average_per);
    PyTuple_SET_ITEM(__pyx_t_25, 0+__pyx_t_5, __pyx_kp_u_number_of_elements_s_average_per);
    __Pyx_GIVEREF(__pyx_t_31);
    PyTuple_SET_ITEM(__pyx_t_25, 1+__pyx_t_5, __pyx_t_31);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_25, 2+__pyx_t_5, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_25, 3+__pyx_t_5, __pyx_t_27);
    __pyx_t_31 = 0;
    __pyx_t_28 = 0;
    __pyx_t_27 = 0;
    __pyx_t_29 = __Pyx_PyObject_Call(__pyx_t_30, __pyx_t_25, NULL); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  }
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pyFAI/ext/_distortion.pyx":966
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 *     if format == "csr":             # <<<<<<<<<<<<<<
 *         res = builder.to_csr()
 *     elif format == "lut":
 */
  __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_format, __pyx_n_u_csr, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 966, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "pyFAI/ext/_distortion.pyx":967
 * 
 *     if format == "csr":
 *         res = builder.to_csr()             # <<<<<<<<<<<<<<
 *     elif format == "lut":
 *         raise NotImplementedError("")
 */
    __pyx_t_30 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_builder), __pyx_n_s_to_csr); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_25 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_30))) {
      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_30);
      if (likely(__pyx_t_25)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_30);
        __Pyx_INCREF(__pyx_t_25);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_30, function);
      }
    }
    __pyx_t_29 = (__pyx_t_25) ? __Pyx_PyObject_CallOneArg(__pyx_t_30, __pyx_t_25) : __Pyx_PyObject_CallNoArg(__pyx_t_30);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_v_res = __pyx_t_29;
    __pyx_t_29 = 0;

    /* "pyFAI/ext/_distortion.pyx":966
 *                 counter, counter / size_in, bins_per_pixel)
 * 
 *     if format == "csr":             # <<<<<<<<<<<<<<
 *         res = builder.to_csr()
 *     elif format == "lut":
 */
    goto __pyx_L55;
  }

  /* "pyFAI/ext/_distortion.pyx":968
 *     if format == "csr":
 *         res = builder.to_csr()
 *     elif format == "lut":             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("")
 *     else:
 */
  __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_format, __pyx_n_u_lut, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
  if (unlikely(__pyx_t_8)) {

    /* "pyFAI/ext/_distortion.pyx":969
 *         res = builder.to_csr()
 *     elif format == "lut":
 *         raise NotImplementedError("")             # <<<<<<<<<<<<<<
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)
 */
    __pyx_t_29 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_Raise(__pyx_t_29, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __PYX_ERR(0, 969, __pyx_L1_error)

    /* "pyFAI/ext/_distortion.pyx":968
 *     if format == "csr":
 *         res = builder.to_csr()
 *     elif format == "lut":             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("")
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":971
 *         raise NotImplementedError("")
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  /*else*/ {
    __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_INCREF(__pyx_kp_u_Unimplemented_sparse_matrix_form);
    __Pyx_GIVEREF(__pyx_kp_u_Unimplemented_sparse_matrix_form);
    PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_kp_u_Unimplemented_sparse_matrix_form);
    __Pyx_INCREF(__pyx_v_format);
    __Pyx_GIVEREF(__pyx_v_format);
    PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_v_format);
    __pyx_t_30 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_t_29, NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __Pyx_Raise(__pyx_t_30, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __PYX_ERR(0, 971, __pyx_L1_error)
  }
  __pyx_L55:;

  /* "pyFAI/ext/_distortion.pyx":972
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":823
 * 
 * 
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                    shape,
 *                    max_pixel_size=(8, 8),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_AddTraceback("pyFAI.ext._distortion.calc_sparse_v2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_builder);
  __Pyx_XDECREF(__pyx_v_res);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":975
 * 
 * 
 * def resize_image_2D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_17resize_image_2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_16resize_image_2D[] = "resize_image_2D(image, shape=None)\n\n    Reshape the image in such a way it has the required shape\n\n    :param image: 2D-array with the image\n    :param shape: expected shape of input image\n    :return: 2D image with the proper shape\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_17resize_image_2D = {"resize_image_2D", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_17resize_image_2D, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_16resize_image_2D};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_17resize_image_2D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resize_image_2D (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape,0};
    PyObject* values[2] = {0,0};

    /* "pyFAI/ext/_distortion.pyx":976
 * 
 * def resize_image_2D(image not None,
 *                     shape=None):             # <<<<<<<<<<<<<<
 *     """
 *     Reshape the image in such a way it has the required shape
 */
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "resize_image_2D") < 0)) __PYX_ERR(0, 975, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resize_image_2D", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 975, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.resize_image_2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_image) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "image"); __PYX_ERR(0, 975, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_16resize_image_2D(__pyx_self, __pyx_v_image, __pyx_v_shape);

  /* "pyFAI/ext/_distortion.pyx":975
 * 
 * 
 * def resize_image_2D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_16resize_image_2D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape) {
  PyObject *__pyx_v_shape_in0 = NULL;
  PyObject *__pyx_v_shape_in1 = NULL;
  PyObject *__pyx_v_shape_img0 = NULL;
  PyObject *__pyx_v_shape_img1 = NULL;
  PyObject *__pyx_v_new_image = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resize_image_2D", 0);

  /* "pyFAI/ext/_distortion.pyx":984
 *     :return: 2D image with the proper shape
 *     """
 *     if shape is None:             # <<<<<<<<<<<<<<
 *         return image
 *     assert image.ndim == 2, "image is 2D"
 */
  __pyx_t_1 = (__pyx_v_shape == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":985
 *     """
 *     if shape is None:
 *         return image             # <<<<<<<<<<<<<<
 *     assert image.ndim == 2, "image is 2D"
 *     shape_in0, shape_in1 = shape
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_image);
    __pyx_r = __pyx_v_image;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":984
 *     :return: 2D image with the proper shape
 *     """
 *     if shape is None:             # <<<<<<<<<<<<<<
 *         return image
 *     assert image.ndim == 2, "image is 2D"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":986
 *     if shape is None:
 *         return image
 *     assert image.ndim == 2, "image is 2D"             # <<<<<<<<<<<<<<
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1 = image.shape
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_image_is_2D);
      __PYX_ERR(0, 986, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":987
 *         return image
 *     assert image.ndim == 2, "image is 2D"
 *     shape_in0, shape_in1 = shape             # <<<<<<<<<<<<<<
 *     shape_img0, shape_img1 = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 987, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 987, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 987, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_v_shape_in0 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_shape_in1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":988
 *     assert image.ndim == 2, "image is 2D"
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1 = image.shape             # <<<<<<<<<<<<<<
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 *         return image
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 988, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_7)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_6(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_7), 2) < 0) __PYX_ERR(0, 988, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 988, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_shape_img0 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_shape_img1 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":989
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1 = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):             # <<<<<<<<<<<<<<
 *         return image
 * 
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_shape_img0, __pyx_v_shape_in0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 989, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 989, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":990
 *     shape_img0, shape_img1 = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 *         return image             # <<<<<<<<<<<<<<
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_image);
    __pyx_r = __pyx_v_image;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":989
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1 = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):             # <<<<<<<<<<<<<<
 *         return image
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":992
 *         return image
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_shape_in0);
  __Pyx_GIVEREF(__pyx_v_shape_in0);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_shape_in0);
  __Pyx_INCREF(__pyx_v_shape_in1);
  __Pyx_GIVEREF(__pyx_v_shape_in1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_shape_in1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_new_image = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":993
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:             # <<<<<<<<<<<<<<
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1] = image
 */
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_shape_img0, __pyx_v_shape_in0, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 993, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":994
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:shape_img0, :shape_img1] = image
 *         else:
 */
    __pyx_t_8 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 994, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_2) {

      /* "pyFAI/ext/_distortion.pyx":995
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1] = image             # <<<<<<<<<<<<<<
 *         else:
 *             new_image[:shape_img0, :] = image[:, :shape_in1]
 */
      __pyx_t_8 = PySlice_New(Py_None, __pyx_v_shape_img0, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_img1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __pyx_t_8 = 0;
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_4, __pyx_v_image) < 0)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyFAI/ext/_distortion.pyx":994
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:shape_img0, :shape_img1] = image
 *         else:
 */
      goto __pyx_L12;
    }

    /* "pyFAI/ext/_distortion.pyx":997
 *             new_image[:shape_img0, :shape_img1] = image
 *         else:
 *             new_image[:shape_img0, :] = image[:, :shape_in1]             # <<<<<<<<<<<<<<
 *     else:
 *         if shape_img1 < shape_in1:
 */
    /*else*/ {
      __pyx_t_4 = PySlice_New(Py_None, __pyx_v_shape_in1, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__20);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_img0, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_slice__20);
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_8, __pyx_t_4) < 0)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L12:;

    /* "pyFAI/ext/_distortion.pyx":993
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:             # <<<<<<<<<<<<<<
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1] = image
 */
    goto __pyx_L11;
  }

  /* "pyFAI/ext/_distortion.pyx":999
 *             new_image[:shape_img0, :] = image[:, :shape_in1]
 *     else:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:, :shape_img1] = image[:shape_in0, :]
 *         else:
 */
  /*else*/ {
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 999, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "pyFAI/ext/_distortion.pyx":1000
 *     else:
 *         if shape_img1 < shape_in1:
 *             new_image[:, :shape_img1] = image[:shape_in0, :]             # <<<<<<<<<<<<<<
 *         else:
 *             new_image[:, :] = image[:shape_in0, :shape_in1]
 */
      __pyx_t_4 = PySlice_New(Py_None, __pyx_v_shape_in0, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_slice__20);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PySlice_New(Py_None, __pyx_v_shape_img1, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__20);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
      __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_3, __pyx_t_4) < 0)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyFAI/ext/_distortion.pyx":999
 *             new_image[:shape_img0, :] = image[:, :shape_in1]
 *     else:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:, :shape_img1] = image[:shape_in0, :]
 *         else:
 */
      goto __pyx_L13;
    }

    /* "pyFAI/ext/_distortion.pyx":1002
 *             new_image[:, :shape_img1] = image[:shape_in0, :]
 *         else:
 *             new_image[:, :] = image[:shape_in0, :shape_in1]             # <<<<<<<<<<<<<<
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 */
    /*else*/ {
      __pyx_t_4 = PySlice_New(Py_None, __pyx_v_shape_in0, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_in1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_tuple__21, __pyx_t_3) < 0)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L13:;
  }
  __pyx_L11:;

  /* "pyFAI/ext/_distortion.pyx":1003
 *         else:
 *             new_image[:, :] = image[:shape_in0, :shape_in1]
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",             # <<<<<<<<<<<<<<
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 *     return new_image
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_logger); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":1004
 *             new_image[:, :] = image[:shape_in0, :shape_in1]
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)             # <<<<<<<<<<<<<<
 *     return new_image
 * 
 */
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_kp_u_Patching_image_of_shape_ix_i_on, __pyx_v_shape_img1, __pyx_v_shape_img0, __pyx_v_shape_in1, __pyx_v_shape_in0};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_kp_u_Patching_image_of_shape_ix_i_on, __pyx_v_shape_img1, __pyx_v_shape_img0, __pyx_v_shape_in1, __pyx_v_shape_in0};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_Patching_image_of_shape_ix_i_on);
    __Pyx_GIVEREF(__pyx_kp_u_Patching_image_of_shape_ix_i_on);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_9, __pyx_kp_u_Patching_image_of_shape_ix_i_on);
    __Pyx_INCREF(__pyx_v_shape_img1);
    __Pyx_GIVEREF(__pyx_v_shape_img1);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_9, __pyx_v_shape_img1);
    __Pyx_INCREF(__pyx_v_shape_img0);
    __Pyx_GIVEREF(__pyx_v_shape_img0);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_9, __pyx_v_shape_img0);
    __Pyx_INCREF(__pyx_v_shape_in1);
    __Pyx_GIVEREF(__pyx_v_shape_in1);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_9, __pyx_v_shape_in1);
    __Pyx_INCREF(__pyx_v_shape_in0);
    __Pyx_GIVEREF(__pyx_v_shape_in0);
    PyTuple_SET_ITEM(__pyx_t_5, 4+__pyx_t_9, __pyx_v_shape_in0);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1005
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 *     return new_image             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_new_image);
  __pyx_r = __pyx_v_new_image;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":975
 * 
 * 
 * def resize_image_2D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyFAI.ext._distortion.resize_image_2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape_in0);
  __Pyx_XDECREF(__pyx_v_shape_in1);
  __Pyx_XDECREF(__pyx_v_shape_img0);
  __Pyx_XDECREF(__pyx_v_shape_img1);
  __Pyx_XDECREF(__pyx_v_new_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1008
 * 
 * 
 * def resize_image_3D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_19resize_image_3D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_18resize_image_3D[] = "resize_image_3D(image, shape=None)\n\n    Reshape the image in such a way it has the required shape\n    This version is optimized for n-channel images used after preprocesing like:\n    nlines * ncolumn * (value, variance, normalization)\n\n    :param image: 3D-array with the preprocessed image\n    :param shape: expected shape of input image (2D only)\n    :return: 3D image with the proper shape\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_19resize_image_3D = {"resize_image_3D", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_19resize_image_3D, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_18resize_image_3D};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_19resize_image_3D(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resize_image_3D (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape,0};
    PyObject* values[2] = {0,0};

    /* "pyFAI/ext/_distortion.pyx":1009
 * 
 * def resize_image_3D(image not None,
 *                     shape=None):             # <<<<<<<<<<<<<<
 *     """
 *     Reshape the image in such a way it has the required shape
 */
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "resize_image_3D") < 0)) __PYX_ERR(0, 1008, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resize_image_3D", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1008, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.resize_image_3D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_image) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "image"); __PYX_ERR(0, 1008, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_18resize_image_3D(__pyx_self, __pyx_v_image, __pyx_v_shape);

  /* "pyFAI/ext/_distortion.pyx":1008
 * 
 * 
 * def resize_image_3D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_18resize_image_3D(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape) {
  PyObject *__pyx_v_shape_in0 = NULL;
  PyObject *__pyx_v_shape_in1 = NULL;
  PyObject *__pyx_v_shape_img0 = NULL;
  PyObject *__pyx_v_shape_img1 = NULL;
  PyObject *__pyx_v_nchan = NULL;
  PyObject *__pyx_v_new_image = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resize_image_3D", 0);

  /* "pyFAI/ext/_distortion.pyx":1019
 *     :return: 3D image with the proper shape
 *     """
 *     if shape is None:             # <<<<<<<<<<<<<<
 *         return image
 *     assert image.ndim == 3, "image is 3D"
 */
  __pyx_t_1 = (__pyx_v_shape == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1020
 *     """
 *     if shape is None:
 *         return image             # <<<<<<<<<<<<<<
 *     assert image.ndim == 3, "image is 3D"
 *     shape_in0, shape_in1 = shape
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_image);
    __pyx_r = __pyx_v_image;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1019
 *     :return: 3D image with the proper shape
 *     """
 *     if shape is None:             # <<<<<<<<<<<<<<
 *         return image
 *     assert image.ndim == 3, "image is 3D"
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1021
 *     if shape is None:
 *         return image
 *     assert image.ndim == 3, "image is 3D"             # <<<<<<<<<<<<<<
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1, nchan = image.shape
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_image_is_3D);
      __PYX_ERR(0, 1021, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1022
 *         return image
 *     assert image.ndim == 3, "image is 3D"
 *     shape_in0, shape_in1 = shape             # <<<<<<<<<<<<<<
 *     shape_img0, shape_img1, nchan = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
    PyObject* sequence = __pyx_v_shape;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1022, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1022, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_v_shape_in0 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_shape_in1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1023
 *     assert image.ndim == 3, "image is 3D"
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1, nchan = image.shape             # <<<<<<<<<<<<<<
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 *         return image
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1023, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_7);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_8 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_8)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_8); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_6(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 2; __pyx_t_7 = __pyx_t_6(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_8), 3) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1023, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_shape_img0 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_shape_img1 = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_nchan = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyFAI/ext/_distortion.pyx":1024
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1, nchan = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):             # <<<<<<<<<<<<<<
 *         return image
 * 
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_shape_img0, __pyx_v_shape_in0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1025
 *     shape_img0, shape_img1, nchan = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):
 *         return image             # <<<<<<<<<<<<<<
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_image);
    __pyx_r = __pyx_v_image;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1024
 *     shape_in0, shape_in1 = shape
 *     shape_img0, shape_img1, nchan = image.shape
 *     if (shape_img0 == shape_in0) and (shape_img1 == shape_in1):             # <<<<<<<<<<<<<<
 *         return image
 * 
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1027
 *         return image
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_shape_in0);
  __Pyx_GIVEREF(__pyx_v_shape_in0);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_shape_in0);
  __Pyx_INCREF(__pyx_v_shape_in1);
  __Pyx_GIVEREF(__pyx_v_shape_in1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_shape_in1);
  __Pyx_INCREF(__pyx_v_nchan);
  __Pyx_GIVEREF(__pyx_v_nchan);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_nchan);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_new_image = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":1028
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:             # <<<<<<<<<<<<<<
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1, :] = image
 */
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_shape_img0, __pyx_v_shape_in0, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1029
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:shape_img0, :shape_img1, :] = image
 *         else:
 */
    __pyx_t_8 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_2) {

      /* "pyFAI/ext/_distortion.pyx":1030
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1, :] = image             # <<<<<<<<<<<<<<
 *         else:
 *             new_image[:shape_img0, :, :] = image[:, :shape_in1, :]
 */
      __pyx_t_8 = PySlice_New(Py_None, __pyx_v_shape_img0, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1030, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_img1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1030, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_slice__20);
      __pyx_t_8 = 0;
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_5, __pyx_v_image) < 0)) __PYX_ERR(0, 1030, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":1029
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:shape_img0, :shape_img1, :] = image
 *         else:
 */
      goto __pyx_L12;
    }

    /* "pyFAI/ext/_distortion.pyx":1032
 *             new_image[:shape_img0, :shape_img1, :] = image
 *         else:
 *             new_image[:shape_img0, :, :] = image[:, :shape_in1, :]             # <<<<<<<<<<<<<<
 *     else:
 *         if shape_img1 < shape_in1:
 */
    /*else*/ {
      __pyx_t_5 = PySlice_New(Py_None, __pyx_v_shape_in1, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__20);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_slice__20);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_img0, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_slice__20);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_slice__20);
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_8, __pyx_t_5) < 0)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_L12:;

    /* "pyFAI/ext/_distortion.pyx":1028
 * 
 *     new_image = numpy.zeros((shape_in0, shape_in1, nchan), dtype=numpy.float32)
 *     if shape_img0 < shape_in0:             # <<<<<<<<<<<<<<
 *         if shape_img1 < shape_in1:
 *             new_image[:shape_img0, :shape_img1, :] = image
 */
    goto __pyx_L11;
  }

  /* "pyFAI/ext/_distortion.pyx":1034
 *             new_image[:shape_img0, :, :] = image[:, :shape_in1, :]
 *     else:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:, :shape_img1, :] = image[:shape_in0, :, :]
 *         else:
 */
  /*else*/ {
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_shape_img1, __pyx_v_shape_in1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "pyFAI/ext/_distortion.pyx":1035
 *     else:
 *         if shape_img1 < shape_in1:
 *             new_image[:, :shape_img1, :] = image[:shape_in0, :, :]             # <<<<<<<<<<<<<<
 *         else:
 *             new_image[:, :, :] = image[:shape_in0, :shape_in1, :]
 */
      __pyx_t_5 = PySlice_New(Py_None, __pyx_v_shape_in0, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_slice__20);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_slice__20);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PySlice_New(Py_None, __pyx_v_shape_img1, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__20);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_slice__20);
      __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyFAI/ext/_distortion.pyx":1034
 *             new_image[:shape_img0, :, :] = image[:, :shape_in1, :]
 *     else:
 *         if shape_img1 < shape_in1:             # <<<<<<<<<<<<<<
 *             new_image[:, :shape_img1, :] = image[:shape_in0, :, :]
 *         else:
 */
      goto __pyx_L13;
    }

    /* "pyFAI/ext/_distortion.pyx":1037
 *             new_image[:, :shape_img1, :] = image[:shape_in0, :, :]
 *         else:
 *             new_image[:, :, :] = image[:shape_in0, :shape_in1, :]             # <<<<<<<<<<<<<<
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 */
    /*else*/ {
      __pyx_t_5 = PySlice_New(Py_None, __pyx_v_shape_in0, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PySlice_New(Py_None, __pyx_v_shape_in1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
      __Pyx_INCREF(__pyx_slice__20);
      __Pyx_GIVEREF(__pyx_slice__20);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_slice__20);
      __pyx_t_5 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_image, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_tuple__22, __pyx_t_3) < 0)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L13:;
  }
  __pyx_L11:;

  /* "pyFAI/ext/_distortion.pyx":1038
 *         else:
 *             new_image[:, :, :] = image[:shape_in0, :shape_in1, :]
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",             # <<<<<<<<<<<<<<
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 *     return new_image
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_logger); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":1039
 *             new_image[:, :, :] = image[:shape_in0, :shape_in1, :]
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)             # <<<<<<<<<<<<<<
 *     return new_image
 * 
 */
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_kp_u_Patching_image_of_shape_ix_i_on, __pyx_v_shape_img1, __pyx_v_shape_img0, __pyx_v_shape_in1, __pyx_v_shape_in0};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[6] = {__pyx_t_8, __pyx_kp_u_Patching_image_of_shape_ix_i_on, __pyx_v_shape_img1, __pyx_v_shape_img0, __pyx_v_shape_in1, __pyx_v_shape_in0};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_Patching_image_of_shape_ix_i_on);
    __Pyx_GIVEREF(__pyx_kp_u_Patching_image_of_shape_ix_i_on);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_kp_u_Patching_image_of_shape_ix_i_on);
    __Pyx_INCREF(__pyx_v_shape_img1);
    __Pyx_GIVEREF(__pyx_v_shape_img1);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_shape_img1);
    __Pyx_INCREF(__pyx_v_shape_img0);
    __Pyx_GIVEREF(__pyx_v_shape_img0);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_9, __pyx_v_shape_img0);
    __Pyx_INCREF(__pyx_v_shape_in1);
    __Pyx_GIVEREF(__pyx_v_shape_in1);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_9, __pyx_v_shape_in1);
    __Pyx_INCREF(__pyx_v_shape_in0);
    __Pyx_GIVEREF(__pyx_v_shape_in0);
    PyTuple_SET_ITEM(__pyx_t_7, 4+__pyx_t_9, __pyx_v_shape_in0);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1040
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 *     return new_image             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_new_image);
  __pyx_r = __pyx_v_new_image;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1008
 * 
 * 
 * def resize_image_3D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyFAI.ext._distortion.resize_image_3D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape_in0);
  __Pyx_XDECREF(__pyx_v_shape_in1);
  __Pyx_XDECREF(__pyx_v_shape_img0);
  __Pyx_XDECREF(__pyx_v_shape_img1);
  __Pyx_XDECREF(__pyx_v_nchan);
  __Pyx_XDECREF(__pyx_v_new_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1043
 * 
 * 
 * def correct(image, shape_in, shape_out, LUT not None, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *             method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_21correct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_20correct[] = "correct(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None, method=u'double')\nCorrect an image based on the look-up table calculated ...\n    dispatch according to LUT type\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :param method: integration method: can be \"kahan\" using single precision\n            compensated for error or \"double\" in double precision (64 bits)\n\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_21correct = {"correct", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_21correct, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_20correct};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_21correct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_in = 0;
  PyObject *__pyx_v_shape_out = 0;
  PyObject *__pyx_v_LUT = 0;
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  PyObject *__pyx_v_method = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_in,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,&__pyx_n_s_method,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)__pyx_n_u_double);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_in)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct", 0, 4, 7, 1); __PYX_ERR(0, 1043, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct", 0, 4, 7, 2); __PYX_ERR(0, 1043, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct", 0, 4, 7, 3); __PYX_ERR(0, 1043, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct") < 0)) __PYX_ERR(0, 1043, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_in = values[1];
    __pyx_v_shape_out = values[2];
    __pyx_v_LUT = values[3];
    __pyx_v_dummy = values[4];
    __pyx_v_delta_dummy = values[5];
    __pyx_v_method = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1043, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1043, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_20correct(__pyx_self, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_20correct(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_method) {
  int __pyx_v_preprocessed_data;
  PyObject *__pyx_v_shape_out0 = NULL;
  PyObject *__pyx_v_shape_out1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct", 0);
  __Pyx_INCREF(__pyx_v_image);

  /* "pyFAI/ext/_distortion.pyx":1059
 *     :return: corrected 2D image
 *     """
 *     if (image.ndim == 3):             # <<<<<<<<<<<<<<
 *         # new generation of processing with (signal, variance, normalization)
 *         preprocessed_data = True
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1061
 *     if (image.ndim == 3):
 *         # new generation of processing with (signal, variance, normalization)
 *         preprocessed_data = True             # <<<<<<<<<<<<<<
 *         image = resize_image_3D(image, shape_in)
 *     else:
 */
    __pyx_v_preprocessed_data = 1;

    /* "pyFAI/ext/_distortion.pyx":1062
 *         # new generation of processing with (signal, variance, normalization)
 *         preprocessed_data = True
 *         image = resize_image_3D(image, shape_in)             # <<<<<<<<<<<<<<
 *     else:
 *         preprocessed_data = False
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_resize_image_3D); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_image, __pyx_v_shape_in};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_image, __pyx_v_shape_in};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_in);
      __Pyx_GIVEREF(__pyx_v_shape_in);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_shape_in);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_image, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyFAI/ext/_distortion.pyx":1059
 *     :return: corrected 2D image
 *     """
 *     if (image.ndim == 3):             # <<<<<<<<<<<<<<
 *         # new generation of processing with (signal, variance, normalization)
 *         preprocessed_data = True
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1064
 *         image = resize_image_3D(image, shape_in)
 *     else:
 *         preprocessed_data = False             # <<<<<<<<<<<<<<
 *         image = resize_image_2D(image, shape_in)
 * 
 */
  /*else*/ {
    __pyx_v_preprocessed_data = 0;

    /* "pyFAI/ext/_distortion.pyx":1065
 *     else:
 *         preprocessed_data = False
 *         image = resize_image_2D(image, shape_in)             # <<<<<<<<<<<<<<
 * 
 *     if len(LUT) == 3:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_resize_image_2D); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_in);
      __Pyx_GIVEREF(__pyx_v_shape_in);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_shape_in);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_image, __pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1067
 *         image = resize_image_2D(image, shape_in)
 * 
 *     if len(LUT) == 3:             # <<<<<<<<<<<<<<
 *         # CSR format:
 *         if preprocessed_data:
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_LUT); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1067, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_7 == 3) != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1069
 *     if len(LUT) == 3:
 *         # CSR format:
 *         if preprocessed_data:             # <<<<<<<<<<<<<<
 *             return correct_CSR_preproc_double(image, shape_out, LUT, dummy, delta_dummy)
 *         else:
 */
    __pyx_t_3 = (__pyx_v_preprocessed_data != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1070
 *         # CSR format:
 *         if preprocessed_data:
 *             return correct_CSR_preproc_double(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 *         else:
 *             return correct_CSR(image, shape_in, shape_out, LUT, dummy, delta_dummy, method)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_correct_CSR_preproc_double); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[6] = {__pyx_t_4, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[6] = {__pyx_t_4, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(5+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_image);
        __Pyx_GIVEREF(__pyx_v_image);
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_image);
        __Pyx_INCREF(__pyx_v_shape_out);
        __Pyx_GIVEREF(__pyx_v_shape_out);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_shape_out);
        __Pyx_INCREF(__pyx_v_LUT);
        __Pyx_GIVEREF(__pyx_v_LUT);
        PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_LUT);
        __Pyx_INCREF(__pyx_v_dummy);
        __Pyx_GIVEREF(__pyx_v_dummy);
        PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, __pyx_v_dummy);
        __Pyx_INCREF(__pyx_v_delta_dummy);
        __Pyx_GIVEREF(__pyx_v_delta_dummy);
        PyTuple_SET_ITEM(__pyx_t_6, 4+__pyx_t_5, __pyx_v_delta_dummy);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "pyFAI/ext/_distortion.pyx":1069
 *     if len(LUT) == 3:
 *         # CSR format:
 *         if preprocessed_data:             # <<<<<<<<<<<<<<
 *             return correct_CSR_preproc_double(image, shape_out, LUT, dummy, delta_dummy)
 *         else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1072
 *             return correct_CSR_preproc_double(image, shape_out, LUT, dummy, delta_dummy)
 *         else:
 *             return correct_CSR(image, shape_in, shape_out, LUT, dummy, delta_dummy, method)             # <<<<<<<<<<<<<<
 *     else:
 *         # LUT format
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_correct_CSR); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[8] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 7+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[8] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 7+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(7+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_image);
        __Pyx_GIVEREF(__pyx_v_image);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_image);
        __Pyx_INCREF(__pyx_v_shape_in);
        __Pyx_GIVEREF(__pyx_v_shape_in);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_shape_in);
        __Pyx_INCREF(__pyx_v_shape_out);
        __Pyx_GIVEREF(__pyx_v_shape_out);
        PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_v_shape_out);
        __Pyx_INCREF(__pyx_v_LUT);
        __Pyx_GIVEREF(__pyx_v_LUT);
        PyTuple_SET_ITEM(__pyx_t_4, 3+__pyx_t_5, __pyx_v_LUT);
        __Pyx_INCREF(__pyx_v_dummy);
        __Pyx_GIVEREF(__pyx_v_dummy);
        PyTuple_SET_ITEM(__pyx_t_4, 4+__pyx_t_5, __pyx_v_dummy);
        __Pyx_INCREF(__pyx_v_delta_dummy);
        __Pyx_GIVEREF(__pyx_v_delta_dummy);
        PyTuple_SET_ITEM(__pyx_t_4, 5+__pyx_t_5, __pyx_v_delta_dummy);
        __Pyx_INCREF(__pyx_v_method);
        __Pyx_GIVEREF(__pyx_v_method);
        PyTuple_SET_ITEM(__pyx_t_4, 6+__pyx_t_5, __pyx_v_method);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "pyFAI/ext/_distortion.pyx":1067
 *         image = resize_image_2D(image, shape_in)
 * 
 *     if len(LUT) == 3:             # <<<<<<<<<<<<<<
 *         # CSR format:
 *         if preprocessed_data:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1075
 *     else:
 *         # LUT format
 *         if preprocessed_data:             # <<<<<<<<<<<<<<
 *             shape_out0, shape_out1 = shape_out
 *             assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_preprocessed_data != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1076
 *         # LUT format
 *         if preprocessed_data:
 *             shape_out0, shape_out1 = shape_out             # <<<<<<<<<<<<<<
 *             assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * #             if method == "kahan":
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_shape_out))) || (PyList_CheckExact(__pyx_v_shape_out))) {
        PyObject* sequence = __pyx_v_shape_out;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1076, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_4 = PyObject_GetIter(__pyx_v_shape_out); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_1 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1076, __pyx_L1_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1076, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_v_shape_out0 = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_shape_out1 = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "pyFAI/ext/_distortion.pyx":1077
 *         if preprocessed_data:
 *             shape_out0, shape_out1 = shape_out
 *             assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"             # <<<<<<<<<<<<<<
 * #             if method == "kahan":
 * #                 return correct_LUT_preproc_kahan(image, shape_out, LUT, dummy, delta_dummy)
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_1 = PyNumber_Multiply(__pyx_v_shape_out0, __pyx_v_shape_out1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_LUT, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_3)) {
          PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_LUT_shape);
          __PYX_ERR(0, 1077, __pyx_L1_error)
        }
      }
      #endif

      /* "pyFAI/ext/_distortion.pyx":1082
 * #             else:
 * #                 return correct_LUT_preproc_double(image, shape_out, LUT, dummy, delta_dummy)
 *             return correct_LUT_preproc_double(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 *         else:
 *             return correct_LUT(image, shape_in, shape_out, LUT, dummy, delta_dummy, method)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_correct_LUT_preproc_double); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1082, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[6] = {__pyx_t_1, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[6] = {__pyx_t_1, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 5+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(5+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_image);
        __Pyx_GIVEREF(__pyx_v_image);
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_image);
        __Pyx_INCREF(__pyx_v_shape_out);
        __Pyx_GIVEREF(__pyx_v_shape_out);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_shape_out);
        __Pyx_INCREF(__pyx_v_LUT);
        __Pyx_GIVEREF(__pyx_v_LUT);
        PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_LUT);
        __Pyx_INCREF(__pyx_v_dummy);
        __Pyx_GIVEREF(__pyx_v_dummy);
        PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, __pyx_v_dummy);
        __Pyx_INCREF(__pyx_v_delta_dummy);
        __Pyx_GIVEREF(__pyx_v_delta_dummy);
        PyTuple_SET_ITEM(__pyx_t_6, 4+__pyx_t_5, __pyx_v_delta_dummy);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "pyFAI/ext/_distortion.pyx":1075
 *     else:
 *         # LUT format
 *         if preprocessed_data:             # <<<<<<<<<<<<<<
 *             shape_out0, shape_out1 = shape_out
 *             assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1084
 *             return correct_LUT_preproc_double(image, shape_out, LUT, dummy, delta_dummy)
 *         else:
 *             return correct_LUT(image, shape_in, shape_out, LUT, dummy, delta_dummy, method)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_correct_LUT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[8] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 7+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[8] = {__pyx_t_6, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 7+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(7+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_image);
        __Pyx_GIVEREF(__pyx_v_image);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_image);
        __Pyx_INCREF(__pyx_v_shape_in);
        __Pyx_GIVEREF(__pyx_v_shape_in);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_v_shape_in);
        __Pyx_INCREF(__pyx_v_shape_out);
        __Pyx_GIVEREF(__pyx_v_shape_out);
        PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_5, __pyx_v_shape_out);
        __Pyx_INCREF(__pyx_v_LUT);
        __Pyx_GIVEREF(__pyx_v_LUT);
        PyTuple_SET_ITEM(__pyx_t_1, 3+__pyx_t_5, __pyx_v_LUT);
        __Pyx_INCREF(__pyx_v_dummy);
        __Pyx_GIVEREF(__pyx_v_dummy);
        PyTuple_SET_ITEM(__pyx_t_1, 4+__pyx_t_5, __pyx_v_dummy);
        __Pyx_INCREF(__pyx_v_delta_dummy);
        __Pyx_GIVEREF(__pyx_v_delta_dummy);
        PyTuple_SET_ITEM(__pyx_t_1, 5+__pyx_t_5, __pyx_v_delta_dummy);
        __Pyx_INCREF(__pyx_v_method);
        __Pyx_GIVEREF(__pyx_v_method);
        PyTuple_SET_ITEM(__pyx_t_1, 6+__pyx_t_5, __pyx_v_method);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }

  /* "pyFAI/ext/_distortion.pyx":1043
 * 
 * 
 * def correct(image, shape_in, shape_out, LUT not None, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *             method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape_out0);
  __Pyx_XDECREF(__pyx_v_shape_out1);
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1087
 * 
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                 dummy=None, delta_dummy=None, method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_23correct_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_22correct_LUT[] = "correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT, dummy=None, delta_dummy=None, method=u'double')\nCorrect an image based on the look-up table calculated ...\n    dispatch between kahan and double\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :param method: integration method: can be \"kahan\" using single precision\n            compensated for error or \"double\" in double precision (64 bits)\n\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_23correct_LUT = {"correct_LUT", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_23correct_LUT, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_22correct_LUT};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_23correct_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_in = 0;
  PyObject *__pyx_v_shape_out = 0;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  PyObject *__pyx_v_method = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_LUT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_in,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,&__pyx_n_s_method,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":1088
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,
 *                 dummy=None, delta_dummy=None, method="double"):             # <<<<<<<<<<<<<<
 *     """Correct an image based on the look-up table calculated ...
 *     dispatch between kahan and double
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)__pyx_n_u_double);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_in)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT", 0, 4, 7, 1); __PYX_ERR(0, 1087, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT", 0, 4, 7, 2); __PYX_ERR(0, 1087, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT", 0, 4, 7, 3); __PYX_ERR(0, 1087, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_LUT") < 0)) __PYX_ERR(0, 1087, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_in = values[1];
    __pyx_v_shape_out = values[2];
    __pyx_v_LUT = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_LUT.memview)) __PYX_ERR(0, 1087, __pyx_L3_error)
    __pyx_v_dummy = values[4];
    __pyx_v_delta_dummy = values[5];
    __pyx_v_method = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_LUT", 0, 4, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1087, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1087, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_22correct_LUT(__pyx_self, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_method);

  /* "pyFAI/ext/_distortion.pyx":1087
 * 
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                 dummy=None, delta_dummy=None, method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_22correct_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_method) {
  PyObject *__pyx_v_shape_out0 = NULL;
  PyObject *__pyx_v_shape_out1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_LUT", 0);
  __Pyx_INCREF(__pyx_v_image);

  /* "pyFAI/ext/_distortion.pyx":1103
 *     :return: corrected 2D image
 *     """
 *     shape_out0, shape_out1 = shape_out             # <<<<<<<<<<<<<<
 *     assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 *     image = resize_image_2D(image, shape_in)
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape_out))) || (PyList_CheckExact(__pyx_v_shape_out))) {
    PyObject* sequence = __pyx_v_shape_out;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1103, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_shape_out); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1103, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1103, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_shape_out0 = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_shape_out1 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyFAI/ext/_distortion.pyx":1104
 *     """
 *     shape_out0, shape_out1 = shape_out
 *     assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"             # <<<<<<<<<<<<<<
 *     image = resize_image_2D(image, shape_in)
 *     if method == "kahan":
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_shape_out0, __pyx_v_shape_out1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_LUT.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_LUT_shape);
      __PYX_ERR(0, 1104, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1105
 *     shape_out0, shape_out1 = shape_out
 *     assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 *     image = resize_image_2D(image, shape_in)             # <<<<<<<<<<<<<<
 *     if method == "kahan":
 *         return correct_LUT_kahan(image, shape_out, LUT, dummy, delta_dummy)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_resize_image_2D); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_image, __pyx_v_shape_in};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_image, __pyx_v_shape_in};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_image);
    __Pyx_GIVEREF(__pyx_v_image);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_image);
    __Pyx_INCREF(__pyx_v_shape_in);
    __Pyx_GIVEREF(__pyx_v_shape_in);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_shape_in);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_image, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1106
 *     assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 *     image = resize_image_2D(image, shape_in)
 *     if method == "kahan":             # <<<<<<<<<<<<<<
 *         return correct_LUT_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_kahan, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyFAI/ext/_distortion.pyx":1107
 *     image = resize_image_2D(image, shape_in)
 *     if method == "kahan":
 *         return correct_LUT_kahan(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 *     else:
 *         return correct_LUT_double(image, shape_out, LUT, dummy, delta_dummy)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_correct_LUT_kahan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_LUT, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_2, __pyx_v_image, __pyx_v_shape_out, __pyx_t_7, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_2, __pyx_v_image, __pyx_v_shape_out, __pyx_t_7, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(5+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_v_shape_out);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_6, __pyx_t_7);
      __Pyx_INCREF(__pyx_v_dummy);
      __Pyx_GIVEREF(__pyx_v_dummy);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_6, __pyx_v_dummy);
      __Pyx_INCREF(__pyx_v_delta_dummy);
      __Pyx_GIVEREF(__pyx_v_delta_dummy);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_6, __pyx_v_delta_dummy);
      __pyx_t_7 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1106
 *     assert shape_out0 * shape_out1 == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 *     image = resize_image_2D(image, shape_in)
 *     if method == "kahan":             # <<<<<<<<<<<<<<
 *         return correct_LUT_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1109
 *         return correct_LUT_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 *         return correct_LUT_double(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_correct_LUT_double); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_LUT, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_image, __pyx_v_shape_out, __pyx_t_8, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_image, __pyx_v_shape_out, __pyx_t_8, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(5+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_v_shape_out);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_8);
      __Pyx_INCREF(__pyx_v_dummy);
      __Pyx_GIVEREF(__pyx_v_dummy);
      PyTuple_SET_ITEM(__pyx_t_2, 3+__pyx_t_6, __pyx_v_dummy);
      __Pyx_INCREF(__pyx_v_delta_dummy);
      __Pyx_GIVEREF(__pyx_v_delta_dummy);
      PyTuple_SET_ITEM(__pyx_t_2, 4+__pyx_t_6, __pyx_v_delta_dummy);
      __pyx_t_8 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pyFAI/ext/_distortion.pyx":1087
 * 
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                 dummy=None, delta_dummy=None, method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape_out0);
  __Pyx_XDECREF(__pyx_v_shape_out1);
  __Pyx_XDECREF(__pyx_v_image);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1112
 * 
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                       dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_25correct_LUT_kahan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_24correct_LUT_kahan[] = "correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT, dummy=None, delta_dummy=None)\nCorrect an image based on the look-up table calculated ...\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_25correct_LUT_kahan = {"correct_LUT_kahan", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_25correct_LUT_kahan, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_24correct_LUT_kahan};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_25correct_LUT_kahan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_LUT_kahan (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":1113
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,
 *                       dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """Correct an image based on the look-up table calculated ...
 * 
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_kahan", 0, 3, 5, 1); __PYX_ERR(0, 1112, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_kahan", 0, 3, 5, 2); __PYX_ERR(0, 1112, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_LUT_kahan") < 0)) __PYX_ERR(0, 1112, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_LUT.memview)) __PYX_ERR(0, 1112, __pyx_L3_error)
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_LUT_kahan", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1112, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_kahan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1112, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_24correct_LUT_kahan(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy);

  /* "pyFAI/ext/_distortion.pyx":1112
 * 
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                       dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_24correct_LUT_kahan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  float __pyx_v_coef;
  float __pyx_v_sum;
  float __pyx_v_error;
  float __pyx_v_t;
  float __pyx_v_y;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  __pyx_t_5pyFAI_3ext_11_distortion_index_t __pyx_t_18;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_LUT_kahan", 0);

  /* "pyFAI/ext/_distortion.pyx":1128
 *         float coef, sum, error, t, y, value, cdummy, cdelta_dummy
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 *     if do_dummy:
 *         cdummy = dummy
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1129
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1130
 *         bint do_dummy = dummy is not None
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1131
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1132
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = cdelta_dummy = numpy.Nan
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1131
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1129
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1134
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = cdelta_dummy = numpy.Nan             # <<<<<<<<<<<<<<
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Nan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cdummy = __pyx_t_2;
    __pyx_v_cdelta_dummy = __pyx_t_2;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1135
 *     else:
 *         cdummy = cdelta_dummy = numpy.Nan
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"             # <<<<<<<<<<<<<<
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_prod); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_shape_out) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_shape_out);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_LUT.shape[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_LUT_shape);
      __PYX_ERR(0, 1135, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1137
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_shape_out);
  __Pyx_GIVEREF(__pyx_v_shape_out);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape_out);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_out = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":1138
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_lout = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1139
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1140
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size             # <<<<<<<<<<<<<<
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0
 */
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_lin, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_size = __pyx_t_10;

  /* "pyFAI/ext/_distortion.pyx":1141
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0
 *         error = 0.0  # Implement Kahan summation
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_11 = (__pyx_v_LUT.shape[0]);
        if ((1 == 0)) abort();
        {
            float __pyx_parallel_temp0 = ((float)__PYX_NAN());
            float __pyx_parallel_temp1 = ((float)__PYX_NAN());
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            int __pyx_parallel_temp4 = ((int)0xbad0bad0);
            float __pyx_parallel_temp5 = ((float)__PYX_NAN());
            float __pyx_parallel_temp6 = ((float)__PYX_NAN());
            float __pyx_parallel_temp7 = ((float)__PYX_NAN());
            float __pyx_parallel_temp8 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_13 = (__pyx_t_11 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_13 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_10, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_3) firstprivate(__pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) lastprivate(__pyx_v_error) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum) lastprivate(__pyx_v_t) lastprivate(__pyx_v_value) lastprivate(__pyx_v_y) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_13; __pyx_t_12++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_12);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((float)__PYX_NAN());
                            __pyx_v_error = ((float)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum = ((float)__PYX_NAN());
                            __pyx_v_t = ((float)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());
                            __pyx_v_y = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1142
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0             # <<<<<<<<<<<<<<
 *         error = 0.0  # Implement Kahan summation
 *         for j in range(LUT.shape[1]):
 */
                            __pyx_v_sum = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1143
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0
 *         error = 0.0  # Implement Kahan summation             # <<<<<<<<<<<<<<
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 */
                            __pyx_v_error = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1144
 *         sum = 0.0
 *         error = 0.0  # Implement Kahan summation
 *         for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 */
                            __pyx_t_14 = (__pyx_v_LUT.shape[1]);
                            __pyx_t_15 = __pyx_t_14;
                            for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_15; __pyx_t_10+=1) {
                              __pyx_v_j = __pyx_t_10;

                              /* "pyFAI/ext/_distortion.pyx":1145
 *         error = 0.0  # Implement Kahan summation
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx             # <<<<<<<<<<<<<<
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 */
                              __pyx_t_16 = __pyx_v_i;
                              __pyx_t_17 = __pyx_v_j;
                              __pyx_t_18 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_16 * __pyx_v_LUT.strides[0]) )) + __pyx_t_17)) ))).idx;
                              __pyx_v_idx = __pyx_t_18;

                              /* "pyFAI/ext/_distortion.pyx":1146
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef             # <<<<<<<<<<<<<<
 *             if coef <= 0:
 *                 continue
 */
                              __pyx_t_17 = __pyx_v_i;
                              __pyx_t_16 = __pyx_v_j;
                              __pyx_t_19 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_17 * __pyx_v_LUT.strides[0]) )) + __pyx_t_16)) ))).coef;
                              __pyx_v_coef = __pyx_t_19;

                              /* "pyFAI/ext/_distortion.pyx":1147
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1148
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L12_continue;

                                /* "pyFAI/ext/_distortion.pyx":1147
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1149
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1150
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1151
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 *             value = lin[idx]
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                                      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_GIVEREF(__pyx_t_7);
                                      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
                                      __Pyx_GIVEREF(__pyx_t_5);
                                      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
                                      __pyx_t_7 = 0;
                                      __pyx_t_5 = 0;
                                      __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      __pyx_t_6 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                                        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
                                        if (likely(__pyx_t_6)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                                          __Pyx_INCREF(__pyx_t_6);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_4, function);
                                        }
                                      }
                                      __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
                                      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1151, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_8);
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1152
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 */
                                      goto __pyx_L16_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1150
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L16_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L12_continue;
                                      }
                                      __pyx_L19_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L10_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1149
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1153
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 *             value = lin[idx]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_16 = __pyx_v_idx;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_16)) )));

                              /* "pyFAI/ext/_distortion.pyx":1154
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             y = value * coef - error
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L22_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L22_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1155
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             y = value * coef - error
 *             t = sum + y
 */
                                goto __pyx_L12_continue;

                                /* "pyFAI/ext/_distortion.pyx":1154
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             y = value * coef - error
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1156
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             y = value * coef - error             # <<<<<<<<<<<<<<
 *             t = sum + y
 *             error = (t - sum) - y
 */
                              __pyx_v_y = ((__pyx_v_value * __pyx_v_coef) - __pyx_v_error);

                              /* "pyFAI/ext/_distortion.pyx":1157
 *                 continue
 *             y = value * coef - error
 *             t = sum + y             # <<<<<<<<<<<<<<
 *             error = (t - sum) - y
 *             sum = t
 */
                              __pyx_v_t = (__pyx_v_sum + __pyx_v_y);

                              /* "pyFAI/ext/_distortion.pyx":1158
 *             y = value * coef - error
 *             t = sum + y
 *             error = (t - sum) - y             # <<<<<<<<<<<<<<
 *             sum = t
 *         if do_dummy and (sum == 0.0):
 */
                              __pyx_v_error = ((__pyx_v_t - __pyx_v_sum) - __pyx_v_y);

                              /* "pyFAI/ext/_distortion.pyx":1159
 *             t = sum + y
 *             error = (t - sum) - y
 *             sum = t             # <<<<<<<<<<<<<<
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 */
                              __pyx_v_sum = __pyx_v_t;
                              __pyx_L12_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1160
 *             error = (t - sum) - y
 *             sum = t
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            __pyx_t_1 = (__pyx_v_do_dummy != 0);
                            if (__pyx_t_1) {
                            } else {
                              __pyx_t_3 = __pyx_t_1;
                              goto __pyx_L25_bool_binop_done;
                            }
                            __pyx_t_1 = ((__pyx_v_sum == 0.0) != 0);
                            __pyx_t_3 = __pyx_t_1;
                            __pyx_L25_bool_binop_done:;
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1161
 *             sum = t
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy             # <<<<<<<<<<<<<<
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out
 */
                              __pyx_v_sum = __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1160
 *             error = (t - sum) - y
 *             sum = t
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1162
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
                            __pyx_t_16 = __pyx_v_i;
                            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_16)) )) += __pyx_v_sum;
                            goto __pyx_L28;
                            __pyx_L10_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L27;
                            __pyx_L27:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates0)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_error;
                                __pyx_parallel_temp2 = __pyx_v_i;
                                __pyx_parallel_temp3 = __pyx_v_idx;
                                __pyx_parallel_temp4 = __pyx_v_j;
                                __pyx_parallel_temp5 = __pyx_v_sum;
                                __pyx_parallel_temp6 = __pyx_v_t;
                                __pyx_parallel_temp7 = __pyx_v_value;
                                __pyx_parallel_temp8 = __pyx_v_y;
                            }
                            __pyx_L28:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_5);
                    __pyx_t_5 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    __Pyx_XDECREF(__pyx_t_8);
                    __pyx_t_8 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_error = __pyx_parallel_temp1;
              __pyx_v_i = __pyx_parallel_temp2;
              __pyx_v_idx = __pyx_parallel_temp3;
              __pyx_v_j = __pyx_parallel_temp4;
              __pyx_v_sum = __pyx_parallel_temp5;
              __pyx_v_t = __pyx_parallel_temp6;
              __pyx_v_value = __pyx_parallel_temp7;
              __pyx_v_y = __pyx_parallel_temp8;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L6_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1141
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0
 *         error = 0.0  # Implement Kahan summation
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L6_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L7:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1163
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1112
 * 
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                       dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_kahan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1166
 * 
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                        dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_27correct_LUT_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_26correct_LUT_double[] = "correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT, dummy=None, delta_dummy=None)\nCorrect an image based on the look-up table calculated ...\n    double precision accumulated\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_27correct_LUT_double = {"correct_LUT_double", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_27correct_LUT_double, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_26correct_LUT_double};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_27correct_LUT_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_LUT_double (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":1167
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,
 *                        dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """Correct an image based on the look-up table calculated ...
 *     double precision accumulated
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_double", 0, 3, 5, 1); __PYX_ERR(0, 1166, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_double", 0, 3, 5, 2); __PYX_ERR(0, 1166, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_LUT_double") < 0)) __PYX_ERR(0, 1166, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_LUT.memview)) __PYX_ERR(0, 1166, __pyx_L3_error)
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_LUT_double", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1166, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1166, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_26correct_LUT_double(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy);

  /* "pyFAI/ext/_distortion.pyx":1166
 * 
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                        dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_26correct_LUT_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  double __pyx_v_sum;
  double __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  __pyx_t_5pyFAI_3ext_11_distortion_index_t __pyx_t_18;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_LUT_double", 0);

  /* "pyFAI/ext/_distortion.pyx":1184
 *         double sum, coef
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 *     if do_dummy:
 *         cdummy = dummy
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1185
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1186
 *         bint do_dummy = dummy is not None
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1186, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1187
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1188
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = numpy.NaN
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1187
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1185
 *         float32_t[::1] lout, lin
 *         bint do_dummy = dummy is not None
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1190
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = numpy.NaN             # <<<<<<<<<<<<<<
 *         cdelta_dummy = 0.0
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_NaN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1191
 *     else:
 *         cdummy = numpy.NaN
 *         cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 * 
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 */
    __pyx_v_cdelta_dummy = 0.0;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1193
 *         cdelta_dummy = 0.0
 * 
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"             # <<<<<<<<<<<<<<
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_prod); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_shape_out) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_shape_out);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_LUT.shape[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_LUT_shape);
      __PYX_ERR(0, 1193, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1195
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_shape_out);
  __Pyx_GIVEREF(__pyx_v_shape_out);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape_out);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_out = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyFAI/ext/_distortion.pyx":1196
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_8 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_lout = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1197
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1198
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size             # <<<<<<<<<<<<<<
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0
 */
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_lin, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_size = __pyx_t_10;

  /* "pyFAI/ext/_distortion.pyx":1199
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0
 *         for j in range(LUT.shape[1]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_11 = (__pyx_v_LUT.shape[0]);
        if ((1 == 0)) abort();
        {
            double __pyx_parallel_temp0 = ((double)__PYX_NAN());
            int __pyx_parallel_temp1 = ((int)0xbad0bad0);
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            double __pyx_parallel_temp4 = ((double)__PYX_NAN());
            float __pyx_parallel_temp5 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_13 = (__pyx_t_11 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_13 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_10, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_3) firstprivate(__pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum) lastprivate(__pyx_v_value) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_13; __pyx_t_12++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_12);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((double)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum = ((double)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1200
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0             # <<<<<<<<<<<<<<
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 */
                            __pyx_v_sum = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1201
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum = 0.0
 *         for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 */
                            __pyx_t_14 = (__pyx_v_LUT.shape[1]);
                            __pyx_t_15 = __pyx_t_14;
                            for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_15; __pyx_t_10+=1) {
                              __pyx_v_j = __pyx_t_10;

                              /* "pyFAI/ext/_distortion.pyx":1202
 *         sum = 0.0
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx             # <<<<<<<<<<<<<<
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 */
                              __pyx_t_16 = __pyx_v_i;
                              __pyx_t_17 = __pyx_v_j;
                              __pyx_t_18 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_16 * __pyx_v_LUT.strides[0]) )) + __pyx_t_17)) ))).idx;
                              __pyx_v_idx = __pyx_t_18;

                              /* "pyFAI/ext/_distortion.pyx":1203
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef             # <<<<<<<<<<<<<<
 *             if coef <= 0:
 *                 continue
 */
                              __pyx_t_17 = __pyx_v_i;
                              __pyx_t_16 = __pyx_v_j;
                              __pyx_t_19 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_17 * __pyx_v_LUT.strides[0]) )) + __pyx_t_16)) ))).coef;
                              __pyx_v_coef = __pyx_t_19;

                              /* "pyFAI/ext/_distortion.pyx":1204
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1205
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L12_continue;

                                /* "pyFAI/ext/_distortion.pyx":1204
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1206
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1207
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1208
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 *             value = lin[idx]
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                                      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_GIVEREF(__pyx_t_7);
                                      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
                                      __Pyx_GIVEREF(__pyx_t_5);
                                      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
                                      __pyx_t_7 = 0;
                                      __pyx_t_5 = 0;
                                      __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      __pyx_t_6 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                                        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
                                        if (likely(__pyx_t_6)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                                          __Pyx_INCREF(__pyx_t_6);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_4, function);
                                        }
                                      }
                                      __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
                                      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1208, __pyx_L19_error)
                                      __Pyx_GOTREF(__pyx_t_8);
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1209
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 */
                                      goto __pyx_L16_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1207
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L16_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L12_continue;
                                      }
                                      __pyx_L19_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L10_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1206
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1210
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 *             value = lin[idx]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_16 = __pyx_v_idx;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_16)) )));

                              /* "pyFAI/ext/_distortion.pyx":1211
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum = value * coef + sum
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L22_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L22_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1212
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             sum = value * coef + sum
 *         if do_dummy and (sum == 0.0):
 */
                                goto __pyx_L12_continue;

                                /* "pyFAI/ext/_distortion.pyx":1211
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum = value * coef + sum
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1213
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             sum = value * coef + sum             # <<<<<<<<<<<<<<
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 */
                              __pyx_v_sum = ((__pyx_v_value * __pyx_v_coef) + __pyx_v_sum);
                              __pyx_L12_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1214
 *                 continue
 *             sum = value * coef + sum
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            __pyx_t_1 = (__pyx_v_do_dummy != 0);
                            if (__pyx_t_1) {
                            } else {
                              __pyx_t_3 = __pyx_t_1;
                              goto __pyx_L25_bool_binop_done;
                            }
                            __pyx_t_1 = ((__pyx_v_sum == 0.0) != 0);
                            __pyx_t_3 = __pyx_t_1;
                            __pyx_L25_bool_binop_done:;
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1215
 *             sum = value * coef + sum
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy             # <<<<<<<<<<<<<<
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out
 */
                              __pyx_v_sum = __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1214
 *                 continue
 *             sum = value * coef + sum
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1216
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
                            __pyx_t_16 = __pyx_v_i;
                            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_16)) )) += __pyx_v_sum;
                            goto __pyx_L28;
                            __pyx_L10_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L27;
                            __pyx_L27:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates1)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_i;
                                __pyx_parallel_temp2 = __pyx_v_idx;
                                __pyx_parallel_temp3 = __pyx_v_j;
                                __pyx_parallel_temp4 = __pyx_v_sum;
                                __pyx_parallel_temp5 = __pyx_v_value;
                            }
                            __pyx_L28:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_5);
                    __pyx_t_5 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    __Pyx_XDECREF(__pyx_t_8);
                    __pyx_t_8 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_i = __pyx_parallel_temp1;
              __pyx_v_idx = __pyx_parallel_temp2;
              __pyx_v_j = __pyx_parallel_temp3;
              __pyx_v_sum = __pyx_parallel_temp4;
              __pyx_v_value = __pyx_parallel_temp5;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L6_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1199
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = lin.size
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0
 *         for j in range(LUT.shape[1]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L6_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L7:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1217
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1166
 * 
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                        dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1220
 * 
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                 variance=None, method="double"):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_29correct_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_28correct_CSR[] = "correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None, variance=None, method=u'double')\n\n    Correct an image based on the look-up table calculated ...\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 3-tuple array of ndarray\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :param variance: unused for now ... TODO: propagate variance.\n    :param method: integration method: can be \"kahan\" using single precision compensated for error or \"double\" in double precision (64 bits)\n    :return: corrected 2D image\n\n    Nota: patch image on proper buffer size if needed.\n\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_29correct_CSR = {"correct_CSR", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_29correct_CSR, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_28correct_CSR};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_29correct_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_in = 0;
  PyObject *__pyx_v_shape_out = 0;
  PyObject *__pyx_v_LUT = 0;
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  CYTHON_UNUSED PyObject *__pyx_v_variance = 0;
  PyObject *__pyx_v_method = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_CSR (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_in,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,&__pyx_n_s_variance,&__pyx_n_s_method,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);

    /* "pyFAI/ext/_distortion.pyx":1221
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,
 *                 variance=None, method="double"):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)__pyx_n_u_double);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_in)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR", 0, 4, 8, 1); __PYX_ERR(0, 1220, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR", 0, 4, 8, 2); __PYX_ERR(0, 1220, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR", 0, 4, 8, 3); __PYX_ERR(0, 1220, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_variance);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_method);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_CSR") < 0)) __PYX_ERR(0, 1220, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_in = values[1];
    __pyx_v_shape_out = values[2];
    __pyx_v_LUT = values[3];
    __pyx_v_dummy = values[4];
    __pyx_v_delta_dummy = values[5];
    __pyx_v_variance = values[6];
    __pyx_v_method = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_CSR", 0, 4, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1220, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_28correct_CSR(__pyx_self, __pyx_v_image, __pyx_v_shape_in, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_variance, __pyx_v_method);

  /* "pyFAI/ext/_distortion.pyx":1220
 * 
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                 variance=None, method="double"):
 *     """
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_28correct_CSR(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_in, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, CYTHON_UNUSED PyObject *__pyx_v_variance, PyObject *__pyx_v_method) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_CSR", 0);
  __Pyx_INCREF(__pyx_v_image);

  /* "pyFAI/ext/_distortion.pyx":1238
 * 
 *     """
 *     image = resize_image_2D(image, shape_in)             # <<<<<<<<<<<<<<
 * 
 *     if method == "kahan":
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_resize_image_2D); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_image, __pyx_v_shape_in};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_image, __pyx_v_shape_in};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_image);
    __Pyx_GIVEREF(__pyx_v_image);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_image);
    __Pyx_INCREF(__pyx_v_shape_in);
    __Pyx_GIVEREF(__pyx_v_shape_in);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_shape_in);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_image, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1240
 *     image = resize_image_2D(image, shape_in)
 * 
 *     if method == "kahan":             # <<<<<<<<<<<<<<
 *         return correct_CSR_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 */
  __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_kahan, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1240, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "pyFAI/ext/_distortion.pyx":1241
 * 
 *     if method == "kahan":
 *         return correct_CSR_kahan(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 *     else:
 *         return correct_CSR_double(image, shape_out, LUT, dummy, delta_dummy)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_correct_CSR_kahan); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[6] = {__pyx_t_5, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 5+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[6] = {__pyx_t_5, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 5+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(5+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_v_shape_out);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_shape_out);
      __Pyx_INCREF(__pyx_v_LUT);
      __Pyx_GIVEREF(__pyx_v_LUT);
      PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_4, __pyx_v_LUT);
      __Pyx_INCREF(__pyx_v_dummy);
      __Pyx_GIVEREF(__pyx_v_dummy);
      PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_4, __pyx_v_dummy);
      __Pyx_INCREF(__pyx_v_delta_dummy);
      __Pyx_GIVEREF(__pyx_v_delta_dummy);
      PyTuple_SET_ITEM(__pyx_t_3, 4+__pyx_t_4, __pyx_v_delta_dummy);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1240
 *     image = resize_image_2D(image, shape_in)
 * 
 *     if method == "kahan":             # <<<<<<<<<<<<<<
 *         return correct_CSR_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1243
 *         return correct_CSR_kahan(image, shape_out, LUT, dummy, delta_dummy)
 *     else:
 *         return correct_CSR_double(image, shape_out, LUT, dummy, delta_dummy)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_correct_CSR_double); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[6] = {__pyx_t_3, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 5+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[6] = {__pyx_t_3, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 5+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(5+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_image);
      __Pyx_INCREF(__pyx_v_shape_out);
      __Pyx_GIVEREF(__pyx_v_shape_out);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_shape_out);
      __Pyx_INCREF(__pyx_v_LUT);
      __Pyx_GIVEREF(__pyx_v_LUT);
      PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_LUT);
      __Pyx_INCREF(__pyx_v_dummy);
      __Pyx_GIVEREF(__pyx_v_dummy);
      PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_dummy);
      __Pyx_INCREF(__pyx_v_delta_dummy);
      __Pyx_GIVEREF(__pyx_v_delta_dummy);
      PyTuple_SET_ITEM(__pyx_t_5, 4+__pyx_t_4, __pyx_v_delta_dummy);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "pyFAI/ext/_distortion.pyx":1220
 * 
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                 variance=None, method="double"):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1246
 * 
 * 
 * def correct_CSR_kahan(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_31correct_CSR_kahan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_30correct_CSR_kahan[] = "correct_CSR_kahan(image, shape_out, LUT, dummy=None, delta_dummy=None)\n\n    Correct an image based on the look-up table calculated ...\n    using kahan's error compensated algorithm\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 3-tuple array of ndarray\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_31correct_CSR_kahan = {"correct_CSR_kahan", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_31correct_CSR_kahan, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_30correct_CSR_kahan};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_31correct_CSR_kahan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  PyObject *__pyx_v_LUT = 0;
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_CSR_kahan (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_kahan", 0, 3, 5, 1); __PYX_ERR(0, 1246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_kahan", 0, 3, 5, 2); __PYX_ERR(0, 1246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_CSR_kahan") < 0)) __PYX_ERR(0, 1246, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = values[2];
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_CSR_kahan", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1246, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_kahan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_30correct_CSR_kahan(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_30correct_CSR_kahan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  CYTHON_UNUSED int __pyx_v_bins;
  float __pyx_v_coef;
  float __pyx_v_error;
  float __pyx_v_sum;
  float __pyx_v_y;
  float __pyx_v_t;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_CSR_kahan", 0);

  /* "pyFAI/ext/_distortion.pyx":1264
 *         float32_t[::1] lout, lin, data
 *         int[::1] indices, indptr
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 * 
 *     if do_dummy:
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1266
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1267
 * 
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1267, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1268
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1269
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = numpy.NaN
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1268
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1266
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1271
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = numpy.NaN             # <<<<<<<<<<<<<<
 *         cdelta_dummy = 0.0
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_NaN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1271, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1272
 *     else:
 *         cdummy = numpy.NaN
 *         cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 * 
 *     data, indices, indptr = LUT
 */
    __pyx_v_cdelta_dummy = 0.0;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1274
 *         cdelta_dummy = 0.0
 * 
 *     data, indices, indptr = LUT             # <<<<<<<<<<<<<<
 *     bins = indptr.shape[0] - 1
 * 
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_LUT))) || (PyList_CheckExact(__pyx_v_LUT))) {
    PyObject* sequence = __pyx_v_LUT;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1274, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_v_LUT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 2; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 3) < 0) __PYX_ERR(0, 1274, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1274, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_data = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;
  __pyx_v_indices = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;
  __pyx_v_indptr = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1275
 * 
 *     data, indices, indptr = LUT
 *     bins = indptr.shape[0] - 1             # <<<<<<<<<<<<<<
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 */
  __pyx_v_bins = ((__pyx_v_indptr.shape[0]) - 1);

  /* "pyFAI/ext/_distortion.pyx":1277
 *     bins = indptr.shape[0] - 1
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_shape_out);
  __Pyx_GIVEREF(__pyx_v_shape_out);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_shape_out);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_12) < 0) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_out = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "pyFAI/ext/_distortion.pyx":1278
 * 
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = image.size
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_12 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_12, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_v_lout = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1279
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     size = image.size
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_12 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_12);
  __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1280
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = image.size             # <<<<<<<<<<<<<<
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_size = __pyx_t_13;

  /* "pyFAI/ext/_distortion.pyx":1282
 *     size = image.size
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0    # Implement Kahan summation
 *         error = 0.0
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_13 = __pyx_v_bins;
        if ((1 == 0)) abort();
        {
            float __pyx_parallel_temp0 = ((float)__PYX_NAN());
            float __pyx_parallel_temp1 = ((float)__PYX_NAN());
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            int __pyx_parallel_temp4 = ((int)0xbad0bad0);
            float __pyx_parallel_temp5 = ((float)__PYX_NAN());
            float __pyx_parallel_temp6 = ((float)__PYX_NAN());
            float __pyx_parallel_temp7 = ((float)__PYX_NAN());
            float __pyx_parallel_temp8 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_15 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_3) firstprivate(__pyx_t_12, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) lastprivate(__pyx_v_error) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum) lastprivate(__pyx_v_t) lastprivate(__pyx_v_value) lastprivate(__pyx_v_y) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_14);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((float)__PYX_NAN());
                            __pyx_v_error = ((float)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum = ((float)__PYX_NAN());
                            __pyx_v_t = ((float)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());
                            __pyx_v_y = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1283
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum = 0.0    # Implement Kahan summation             # <<<<<<<<<<<<<<
 *         error = 0.0
 *         for j in range(indptr[i], indptr[i + 1]):
 */
                            __pyx_v_sum = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1284
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum = 0.0    # Implement Kahan summation
 *         error = 0.0             # <<<<<<<<<<<<<<
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]
 */
                            __pyx_v_error = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1285
 *         sum = 0.0    # Implement Kahan summation
 *         error = 0.0
 *         for j in range(indptr[i], indptr[i + 1]):             # <<<<<<<<<<<<<<
 *             idx = indices[j]
 *             coef = data[j]
 */
                            __pyx_t_16 = (__pyx_v_i + 1);
                            __pyx_t_17 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_16)) )));
                            __pyx_t_16 = __pyx_v_i;
                            __pyx_t_18 = __pyx_t_17;
                            for (__pyx_t_19 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_16)) ))); __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
                              __pyx_v_j = __pyx_t_19;

                              /* "pyFAI/ext/_distortion.pyx":1286
 *         error = 0.0
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]             # <<<<<<<<<<<<<<
 *             coef = data[j]
 *             if coef <= 0:
 */
                              __pyx_t_20 = __pyx_v_j;
                              __pyx_v_idx = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indices.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1287
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]
 *             coef = data[j]             # <<<<<<<<<<<<<<
 *             if coef <= 0:
 *                 continue
 */
                              __pyx_t_20 = __pyx_v_j;
                              __pyx_v_coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1288
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1289
 *             coef = data[j]
 *             if coef <= 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L14_continue;

                                /* "pyFAI/ext/_distortion.pyx":1288
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1290
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1291
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1292
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 *             value = lin[idx]
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_logger); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_12);
                                      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_warning); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                                      __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_12);
                                      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_GIVEREF(__pyx_t_12);
                                      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_12);
                                      __Pyx_GIVEREF(__pyx_t_5);
                                      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
                                      __pyx_t_12 = 0;
                                      __pyx_t_5 = 0;
                                      __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      __pyx_t_4 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                                        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
                                        if (likely(__pyx_t_4)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                                          __Pyx_INCREF(__pyx_t_4);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_6, function);
                                        }
                                      }
                                      __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
                                      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1292, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1293
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 */
                                      goto __pyx_L18_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1291
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L18_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L14_continue;
                                      }
                                      __pyx_L21_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L12_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1290
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1294
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 *             value = lin[idx]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_20 = __pyx_v_idx;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1295
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             y = value * coef - error
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L24_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L24_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1296
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             y = value * coef - error
 *             t = sum + y
 */
                                goto __pyx_L14_continue;

                                /* "pyFAI/ext/_distortion.pyx":1295
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             y = value * coef - error
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1297
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             y = value * coef - error             # <<<<<<<<<<<<<<
 *             t = sum + y
 *             error = (t - sum) - y
 */
                              __pyx_v_y = ((__pyx_v_value * __pyx_v_coef) - __pyx_v_error);

                              /* "pyFAI/ext/_distortion.pyx":1298
 *                 continue
 *             y = value * coef - error
 *             t = sum + y             # <<<<<<<<<<<<<<
 *             error = (t - sum) - y
 *             sum = t
 */
                              __pyx_v_t = (__pyx_v_sum + __pyx_v_y);

                              /* "pyFAI/ext/_distortion.pyx":1299
 *             y = value * coef - error
 *             t = sum + y
 *             error = (t - sum) - y             # <<<<<<<<<<<<<<
 *             sum = t
 *         if do_dummy and (sum == 0.0):
 */
                              __pyx_v_error = ((__pyx_v_t - __pyx_v_sum) - __pyx_v_y);

                              /* "pyFAI/ext/_distortion.pyx":1300
 *             t = sum + y
 *             error = (t - sum) - y
 *             sum = t             # <<<<<<<<<<<<<<
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 */
                              __pyx_v_sum = __pyx_v_t;
                              __pyx_L14_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1301
 *             error = (t - sum) - y
 *             sum = t
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            __pyx_t_1 = (__pyx_v_do_dummy != 0);
                            if (__pyx_t_1) {
                            } else {
                              __pyx_t_3 = __pyx_t_1;
                              goto __pyx_L27_bool_binop_done;
                            }
                            __pyx_t_1 = ((__pyx_v_sum == 0.0) != 0);
                            __pyx_t_3 = __pyx_t_1;
                            __pyx_L27_bool_binop_done:;
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1302
 *             sum = t
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy             # <<<<<<<<<<<<<<
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out
 */
                              __pyx_v_sum = __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1301
 *             error = (t - sum) - y
 *             sum = t
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1303
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
                            __pyx_t_16 = __pyx_v_i;
                            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_16)) )) += __pyx_v_sum;
                            goto __pyx_L30;
                            __pyx_L12_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L29;
                            __pyx_L29:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates2)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_error;
                                __pyx_parallel_temp2 = __pyx_v_i;
                                __pyx_parallel_temp3 = __pyx_v_idx;
                                __pyx_parallel_temp4 = __pyx_v_j;
                                __pyx_parallel_temp5 = __pyx_v_sum;
                                __pyx_parallel_temp6 = __pyx_v_t;
                                __pyx_parallel_temp7 = __pyx_v_value;
                                __pyx_parallel_temp8 = __pyx_v_y;
                            }
                            __pyx_L30:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_12);
                    __pyx_t_12 = NULL;
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_5);
                    __pyx_t_5 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_error = __pyx_parallel_temp1;
              __pyx_v_i = __pyx_parallel_temp2;
              __pyx_v_idx = __pyx_parallel_temp3;
              __pyx_v_j = __pyx_parallel_temp4;
              __pyx_v_sum = __pyx_parallel_temp5;
              __pyx_v_t = __pyx_parallel_temp6;
              __pyx_v_value = __pyx_parallel_temp7;
              __pyx_v_y = __pyx_parallel_temp8;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L8_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1282
 *     size = image.size
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0    # Implement Kahan summation
 *         error = 0.0
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L8_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L9:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1304
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1246
 * 
 * 
 * def correct_CSR_kahan(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_kahan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1307
 * 
 * 
 * def correct_CSR_double(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_33correct_CSR_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_32correct_CSR_double[] = "correct_CSR_double(image, shape_out, LUT, dummy=None, delta_dummy=None)\n\n    Correct an image based on the look-up table calculated ...\n    using double precision accumulator\n\n    :param image: 2D-array with the image\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 3-tuple array of ndarray\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :return: corrected 2D image\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_33correct_CSR_double = {"correct_CSR_double", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_33correct_CSR_double, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_32correct_CSR_double};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_33correct_CSR_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  PyObject *__pyx_v_LUT = 0;
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_CSR_double (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_double", 0, 3, 5, 1); __PYX_ERR(0, 1307, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_double", 0, 3, 5, 2); __PYX_ERR(0, 1307, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_CSR_double") < 0)) __PYX_ERR(0, 1307, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = values[2];
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_CSR_double", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1307, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_32correct_CSR_double(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_32correct_CSR_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  int __pyx_v_bins;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  double __pyx_v_coef;
  double __pyx_v_sum;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_CSR_double", 0);

  /* "pyFAI/ext/_distortion.pyx":1326
 *         float32_t[::1] lout, lin, data
 *         int[::1] indices, indptr
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 * 
 *     if do_dummy:
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1328
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1329
 * 
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1329, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1330
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1331
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = numpy.NaN
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1330
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1328
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1333
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = numpy.NaN             # <<<<<<<<<<<<<<
 *         cdelta_dummy = 0.0
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_NaN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1334
 *     else:
 *         cdummy = numpy.NaN
 *         cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 * 
 *     data, indices, indptr = LUT
 */
    __pyx_v_cdelta_dummy = 0.0;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1336
 *         cdelta_dummy = 0.0
 * 
 *     data, indices, indptr = LUT             # <<<<<<<<<<<<<<
 *     bins = indptr.size - 1
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_LUT))) || (PyList_CheckExact(__pyx_v_LUT))) {
    PyObject* sequence = __pyx_v_LUT;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1336, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_v_LUT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 2; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 3) < 0) __PYX_ERR(0, 1336, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1336, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_data = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;
  __pyx_v_indices = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;
  __pyx_v_indptr = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1337
 * 
 *     data, indices, indptr = LUT
 *     bins = indptr.size - 1             # <<<<<<<<<<<<<<
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 */
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_int, (int (*)(char *, PyObject *)) __pyx_memview_set_int, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_bins = __pyx_t_12;

  /* "pyFAI/ext/_distortion.pyx":1338
 *     data, indices, indptr = LUT
 *     bins = indptr.size - 1
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"             # <<<<<<<<<<<<<<
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_prod); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_shape_out) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_shape_out);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_indptr_siz);
      __PYX_ERR(0, 1338, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1339
 *     bins = indptr.size - 1
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_shape_out);
  __Pyx_GIVEREF(__pyx_v_shape_out);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape_out);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_13) < 0) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_out = __pyx_t_13;
  __pyx_t_13 = 0;

  /* "pyFAI/ext/_distortion.pyx":1340
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = image.size
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_13 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_13, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_lout = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1341
 *     out = numpy.zeros(shape_out, dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     size = image.size
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_numpy); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_13 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_13);
  __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1342
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *     size = image.size             # <<<<<<<<<<<<<<
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_size = __pyx_t_12;

  /* "pyFAI/ext/_distortion.pyx":1344
 *     size = image.size
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0    # double precision
 *         for j in range(indptr[i], indptr[i + 1]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_12 = __pyx_v_bins;
        if ((1 == 0)) abort();
        {
            double __pyx_parallel_temp0 = ((double)__PYX_NAN());
            int __pyx_parallel_temp1 = ((int)0xbad0bad0);
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            double __pyx_parallel_temp4 = ((double)__PYX_NAN());
            float __pyx_parallel_temp5 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_15 = (__pyx_t_12 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_15 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_3) firstprivate(__pyx_t_13, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum) lastprivate(__pyx_v_value) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_14);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((double)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum = ((double)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1345
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum = 0.0    # double precision             # <<<<<<<<<<<<<<
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]
 */
                            __pyx_v_sum = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1346
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum = 0.0    # double precision
 *         for j in range(indptr[i], indptr[i + 1]):             # <<<<<<<<<<<<<<
 *             idx = indices[j]
 *             coef = data[j]
 */
                            __pyx_t_16 = (__pyx_v_i + 1);
                            __pyx_t_17 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_16)) )));
                            __pyx_t_16 = __pyx_v_i;
                            __pyx_t_18 = __pyx_t_17;
                            for (__pyx_t_19 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_16)) ))); __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
                              __pyx_v_j = __pyx_t_19;

                              /* "pyFAI/ext/_distortion.pyx":1347
 *         sum = 0.0    # double precision
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]             # <<<<<<<<<<<<<<
 *             coef = data[j]
 *             if coef <= 0.0:
 */
                              __pyx_t_20 = __pyx_v_j;
                              __pyx_v_idx = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indices.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1348
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]
 *             coef = data[j]             # <<<<<<<<<<<<<<
 *             if coef <= 0.0:
 *                 continue
 */
                              __pyx_t_20 = __pyx_v_j;
                              __pyx_v_coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1349
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0.0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1350
 *             coef = data[j]
 *             if coef <= 0.0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L14_continue;

                                /* "pyFAI/ext/_distortion.pyx":1349
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0.0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1351
 *             if coef <= 0.0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1352
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1353
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 *             value = lin[idx]
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_logger); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_13);
                                      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                                      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_13);
                                      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __Pyx_GIVEREF(__pyx_t_13);
                                      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_13);
                                      __Pyx_GIVEREF(__pyx_t_6);
                                      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
                                      __pyx_t_13 = 0;
                                      __pyx_t_6 = 0;
                                      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      __pyx_t_5 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                                        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                                        if (likely(__pyx_t_5)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                                          __Pyx_INCREF(__pyx_t_5);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_4, function);
                                        }
                                      }
                                      __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
                                      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1353, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1354
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 */
                                      goto __pyx_L18_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1352
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L18_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L14_continue;
                                      }
                                      __pyx_L21_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L12_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1351
 *             if coef <= 0.0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1355
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 *             value = lin[idx]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_20 = __pyx_v_idx;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1356
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum = sum + value * coef  # += operator not allowed in // sections
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L24_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L24_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1357
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             sum = sum + value * coef  # += operator not allowed in // sections
 *         if do_dummy and (sum == 0.0):
 */
                                goto __pyx_L14_continue;

                                /* "pyFAI/ext/_distortion.pyx":1356
 *                     continue
 *             value = lin[idx]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum = sum + value * coef  # += operator not allowed in // sections
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1358
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             sum = sum + value * coef  # += operator not allowed in // sections             # <<<<<<<<<<<<<<
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 */
                              __pyx_v_sum = (__pyx_v_sum + (__pyx_v_value * __pyx_v_coef));
                              __pyx_L14_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1359
 *                 continue
 *             sum = sum + value * coef  # += operator not allowed in // sections
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            __pyx_t_1 = (__pyx_v_do_dummy != 0);
                            if (__pyx_t_1) {
                            } else {
                              __pyx_t_3 = __pyx_t_1;
                              goto __pyx_L27_bool_binop_done;
                            }
                            __pyx_t_1 = ((__pyx_v_sum == 0.0) != 0);
                            __pyx_t_3 = __pyx_t_1;
                            __pyx_L27_bool_binop_done:;
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1360
 *             sum = sum + value * coef  # += operator not allowed in // sections
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy             # <<<<<<<<<<<<<<
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out
 */
                              __pyx_v_sum = __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1359
 *                 continue
 *             sum = sum + value * coef  # += operator not allowed in // sections
 *         if do_dummy and (sum == 0.0):             # <<<<<<<<<<<<<<
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1361
 *         if do_dummy and (sum == 0.0):
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
                            __pyx_t_16 = __pyx_v_i;
                            *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_16)) )) += __pyx_v_sum;
                            goto __pyx_L30;
                            __pyx_L12_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L29;
                            __pyx_L29:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates3)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_i;
                                __pyx_parallel_temp2 = __pyx_v_idx;
                                __pyx_parallel_temp3 = __pyx_v_j;
                                __pyx_parallel_temp4 = __pyx_v_sum;
                                __pyx_parallel_temp5 = __pyx_v_value;
                            }
                            __pyx_L30:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_13);
                    __pyx_t_13 = NULL;
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_5);
                    __pyx_t_5 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_i = __pyx_parallel_temp1;
              __pyx_v_idx = __pyx_parallel_temp2;
              __pyx_v_j = __pyx_parallel_temp3;
              __pyx_v_sum = __pyx_parallel_temp4;
              __pyx_v_value = __pyx_parallel_temp5;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L8_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1344
 *     size = image.size
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum = 0.0    # double precision
 *         for j in range(indptr[i], indptr[i + 1]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L8_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L9:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1362
 *             sum = cdummy
 *         lout[i] += sum  # this += is for Cython's reduction
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1307
 * 
 * 
 * def correct_CSR_double(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1365
 * 
 * 
 * def correct_LUT_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_35correct_LUT_preproc_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_34correct_LUT_preproc_double[] = "correct_LUT_preproc_double(image, shape_out, lut_t[:, ::1] LUT, dummy=None, delta_dummy=None, empty=numpy.NaN)\nCorrect an image based on the look-up table calculated ...\n    implementation using double precision accumulator\n\n    :param image: 2D-array with the image (signal, variance, normalization)\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :param empty: numerical value for empty pixels (if dummy is not provided)\n    :param method: integration method: can be \"kahan\" using single precision\n            compensated for error or \"double\" in double precision (64 bits)\n\n    :return: corrected 2D image + array with (signal, variance, norm)\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_35correct_LUT_preproc_double = {"correct_LUT_preproc_double", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_35correct_LUT_preproc_double, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_34correct_LUT_preproc_double};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_35correct_LUT_preproc_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  PyObject *__pyx_v_empty = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_LUT_preproc_double (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,&__pyx_n_s_empty,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":1367
 * def correct_LUT_preproc_double(image, shape_out,
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                                empty=numpy.NaN):
 *     """Correct an image based on the look-up table calculated ...
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    values[5] = __pyx_k__23;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_preproc_double", 0, 3, 6, 1); __PYX_ERR(0, 1365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_LUT_preproc_double", 0, 3, 6, 2); __PYX_ERR(0, 1365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_empty);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_LUT_preproc_double") < 0)) __PYX_ERR(0, 1365, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_LUT.memview)) __PYX_ERR(0, 1366, __pyx_L3_error)
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
    __pyx_v_empty = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_LUT_preproc_double", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1365, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_preproc_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT.memview) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1366, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_34correct_LUT_preproc_double(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_empty);

  /* "pyFAI/ext/_distortion.pyx":1365
 * 
 * 
 * def correct_LUT_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_34correct_LUT_preproc_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, __Pyx_memviewslice __pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_empty) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  int __pyx_v_nchan;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  double __pyx_v_sum_sig;
  double __pyx_v_sum_var;
  double __pyx_v_sum_norm;
  double __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lerr = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lprop = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_shape_out0 = NULL;
  PyObject *__pyx_v_shape_out1 = NULL;
  PyObject *__pyx_v_prop = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __pyx_t_5pyFAI_3ext_11_distortion_index_t __pyx_t_21;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_LUT_preproc_double", 0);

  /* "pyFAI/ext/_distortion.pyx":1391
 *         float32_t[::1]  lout, lerr
 *         float32_t[:, ::1] lin, lprop
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 * 
 *     if do_dummy:
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1393
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1394
 * 
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1394, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1395
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1396
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = empty
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1395
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1393
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1398
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = empty             # <<<<<<<<<<<<<<
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * 
 */
  /*else*/ {
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_empty); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1399
 *     else:
 *         cdummy = empty
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"             # <<<<<<<<<<<<<<
 * 
 *     nchan = image.shape[2]
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_prod); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_v_shape_out) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_shape_out);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_LUT.shape[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1399, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_LUT_shape);
      __PYX_ERR(0, 1399, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1401
 *     assert numpy.prod(shape_out) == LUT.shape[0], "shape_out0 * shape_out1 == LUT.shape[0]"
 * 
 *     nchan = image.shape[2]             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape_out
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_nchan = __pyx_t_7;

  /* "pyFAI/ext/_distortion.pyx":1402
 * 
 *     nchan = image.shape[2]
 *     shape_out0, shape_out1 = shape_out             # <<<<<<<<<<<<<<
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape_out))) || (PyList_CheckExact(__pyx_v_shape_out))) {
    PyObject* sequence = __pyx_v_shape_out;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1402, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_shape_out); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1402, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1402, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_v_shape_out0 = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_shape_out1 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1404
 *     shape_out0, shape_out1 = shape_out
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_shape_out0);
  __Pyx_GIVEREF(__pyx_v_shape_out0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape_out0);
  __Pyx_INCREF(__pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_v_shape_out1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_float32); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_prop = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "pyFAI/ext/_distortion.pyx":1405
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 *     lprop = prop.reshape((-1, nchan))             # <<<<<<<<<<<<<<
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_prop, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_lprop = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1406
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_shape_out0);
  __Pyx_GIVEREF(__pyx_v_shape_out0);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_shape_out0);
  __Pyx_INCREF(__pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_v_shape_out1);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_shape_out1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10);
  __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_out = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":1407
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lout = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1408
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))             # <<<<<<<<<<<<<<
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_image);
  __Pyx_GIVEREF(__pyx_v_image);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_image);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_13) < 0) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_13);
  __pyx_t_13 = 0;
  __pyx_t_13 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_9 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_13, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lin = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1409
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 */
  __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1410
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *         lerr = err.ravel()
 *     size = lin.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_shape_out0);
    __Pyx_GIVEREF(__pyx_v_shape_out0);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_shape_out0);
    __Pyx_INCREF(__pyx_v_shape_out1);
    __Pyx_GIVEREF(__pyx_v_shape_out1);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_shape_out1);
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_numpy); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_float32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_err = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyFAI/ext/_distortion.pyx":1411
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()             # <<<<<<<<<<<<<<
 *     size = lin.shape[0]
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_ravel); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_lerr = __pyx_t_12;
    __pyx_t_12.memview = NULL;
    __pyx_t_12.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":1409
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1412
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 *     size = lin.shape[0]             # <<<<<<<<<<<<<<
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum_sig = 0.0
 */
  __pyx_v_size = (__pyx_v_lin.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":1413
 *         lerr = err.ravel()
 *     size = lin.shape[0]
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum_sig = 0.0
 *         sum_var = 0.0
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_14 = (__pyx_v_LUT.shape[0]);
        if ((1 == 0)) abort();
        {
            double __pyx_parallel_temp0 = ((double)__PYX_NAN());
            int __pyx_parallel_temp1 = ((int)0xbad0bad0);
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            double __pyx_parallel_temp4 = ((double)__PYX_NAN());
            double __pyx_parallel_temp5 = ((double)__PYX_NAN());
            double __pyx_parallel_temp6 = ((double)__PYX_NAN());
            float __pyx_parallel_temp7 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_16 = (__pyx_t_14 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_16 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_3, __pyx_t_7) firstprivate(__pyx_t_10, __pyx_t_13, __pyx_t_4, __pyx_t_6, __pyx_t_9) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum_norm) lastprivate(__pyx_v_sum_sig) lastprivate(__pyx_v_sum_var) lastprivate(__pyx_v_value) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_16; __pyx_t_15++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_15);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((double)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum_norm = ((double)__PYX_NAN());
                            __pyx_v_sum_sig = ((double)__PYX_NAN());
                            __pyx_v_sum_var = ((double)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1414
 *     size = lin.shape[0]
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum_sig = 0.0             # <<<<<<<<<<<<<<
 *         sum_var = 0.0
 *         sum_norm = 0.0
 */
                            __pyx_v_sum_sig = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1415
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *         sum_sig = 0.0
 *         sum_var = 0.0             # <<<<<<<<<<<<<<
 *         sum_norm = 0.0
 *         for j in range(LUT.shape[1]):
 */
                            __pyx_v_sum_var = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1416
 *         sum_sig = 0.0
 *         sum_var = 0.0
 *         sum_norm = 0.0             # <<<<<<<<<<<<<<
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 */
                            __pyx_v_sum_norm = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1417
 *         sum_var = 0.0
 *         sum_norm = 0.0
 *         for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 */
                            __pyx_t_17 = (__pyx_v_LUT.shape[1]);
                            __pyx_t_18 = __pyx_t_17;
                            for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_18; __pyx_t_7+=1) {
                              __pyx_v_j = __pyx_t_7;

                              /* "pyFAI/ext/_distortion.pyx":1418
 *         sum_norm = 0.0
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx             # <<<<<<<<<<<<<<
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 */
                              __pyx_t_19 = __pyx_v_i;
                              __pyx_t_20 = __pyx_v_j;
                              __pyx_t_21 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_19 * __pyx_v_LUT.strides[0]) )) + __pyx_t_20)) ))).idx;
                              __pyx_v_idx = __pyx_t_21;

                              /* "pyFAI/ext/_distortion.pyx":1419
 *         for j in range(LUT.shape[1]):
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef             # <<<<<<<<<<<<<<
 *             if coef <= 0:
 *                 continue
 */
                              __pyx_t_20 = __pyx_v_i;
                              __pyx_t_19 = __pyx_v_j;
                              __pyx_t_22 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_20 * __pyx_v_LUT.strides[0]) )) + __pyx_t_19)) ))).coef;
                              __pyx_v_coef = __pyx_t_22;

                              /* "pyFAI/ext/_distortion.pyx":1420
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1421
 *             coef = LUT[i, j].coef
 *             if coef <= 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L15_continue;

                                /* "pyFAI/ext/_distortion.pyx":1420
 *             idx = LUT[i, j].idx
 *             coef = LUT[i, j].coef
 *             if coef <= 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1422
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1423
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1424
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 *             value = lin[idx, 0]
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_logger); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_9);
                                      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_warning); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_10);
                                      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                                      __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_9);
                                      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_13);
                                      __Pyx_GIVEREF(__pyx_t_9);
                                      PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_9);
                                      __Pyx_GIVEREF(__pyx_t_4);
                                      PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_4);
                                      __pyx_t_9 = 0;
                                      __pyx_t_4 = 0;
                                      __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                                      __pyx_t_13 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
                                        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_10);
                                        if (likely(__pyx_t_13)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                                          __Pyx_INCREF(__pyx_t_13);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_10, function);
                                        }
                                      }
                                      __pyx_t_6 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_13, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_4);
                                      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                                      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1424, __pyx_L22_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1425
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 */
                                      goto __pyx_L19_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1423
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L19_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L15_continue;
                                      }
                                      __pyx_L22_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L13_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1422
 *             if coef <= 0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1426
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 *             value = lin[idx, 0]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_19 = __pyx_v_idx;
                              __pyx_t_20 = 0;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_19 * __pyx_v_lin.strides[0]) )) + __pyx_t_20)) )));

                              /* "pyFAI/ext/_distortion.pyx":1427
 *                     continue
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum_sig = value * coef + sum_sig
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L25_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L25_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1428
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:
 */
                                goto __pyx_L15_continue;

                                /* "pyFAI/ext/_distortion.pyx":1427
 *                     continue
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum_sig = value * coef + sum_sig
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1429
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             sum_sig = value * coef + sum_sig             # <<<<<<<<<<<<<<
 *             if nchan == 2:
 *                 # case (signal, norm)
 */
                              __pyx_v_sum_sig = ((__pyx_v_value * __pyx_v_coef) + __pyx_v_sum_sig);

                              /* "pyFAI/ext/_distortion.pyx":1430
 *                 continue
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 */
                              switch (__pyx_v_nchan) {
                                case 2:

                                /* "pyFAI/ext/_distortion.pyx":1432
 *             if nchan == 2:
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm             # <<<<<<<<<<<<<<
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 */
                                __pyx_t_20 = __pyx_v_idx;
                                __pyx_t_19 = 1;
                                __pyx_v_sum_norm = ((__pyx_v_coef * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_20 * __pyx_v_lin.strides[0]) )) + __pyx_t_19)) )))) + __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1430
 *                 continue
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 */
                                break;
                                case 3:

                                /* "pyFAI/ext/_distortion.pyx":1435
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var             # <<<<<<<<<<<<<<
 *                 sum_norm = coef * lin[idx, 2] + sum_norm
 *             else:
 */
                                __pyx_t_19 = __pyx_v_idx;
                                __pyx_t_20 = 1;
                                __pyx_v_sum_var = (((__pyx_v_coef * __pyx_v_coef) * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_19 * __pyx_v_lin.strides[0]) )) + __pyx_t_20)) )))) + __pyx_v_sum_var);

                                /* "pyFAI/ext/_distortion.pyx":1436
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var
 *                 sum_norm = coef * lin[idx, 2] + sum_norm             # <<<<<<<<<<<<<<
 *             else:
 *                 sum_norm = sum_norm + coef
 */
                                __pyx_t_20 = __pyx_v_idx;
                                __pyx_t_19 = 2;
                                __pyx_v_sum_norm = ((__pyx_v_coef * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_20 * __pyx_v_lin.strides[0]) )) + __pyx_t_19)) )))) + __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1433
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 *             elif nchan == 3:             # <<<<<<<<<<<<<<
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var
 */
                                break;
                                default:

                                /* "pyFAI/ext/_distortion.pyx":1438
 *                 sum_norm = coef * lin[idx, 2] + sum_norm
 *             else:
 *                 sum_norm = sum_norm + coef             # <<<<<<<<<<<<<<
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 */
                                __pyx_v_sum_norm = (__pyx_v_sum_norm + __pyx_v_coef);
                                break;
                              }
                              __pyx_L15_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1440
 *                 sum_norm = sum_norm + coef
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel             # <<<<<<<<<<<<<<
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 */
                            __pyx_t_3 = ((__pyx_v_sum_norm == 0.0) != 0);
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1441
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *             if nchan == 3:
 *                 lerr[i] += cdummy
 */
                              __pyx_t_19 = __pyx_v_i;
                              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_19)) )) += __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1442
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:             # <<<<<<<<<<<<<<
 *                 lerr[i] += cdummy
 *         else:
 */
                              __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1443
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 *                 lerr[i] += cdummy             # <<<<<<<<<<<<<<
 *         else:
 *             lprop[i, 0] += sum_sig
 */
                                __pyx_t_19 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lerr.data) + __pyx_t_19)) )) += __pyx_v_cdummy;

                                /* "pyFAI/ext/_distortion.pyx":1442
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:             # <<<<<<<<<<<<<<
 *                 lerr[i] += cdummy
 *         else:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1440
 *                 sum_norm = sum_norm + coef
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel             # <<<<<<<<<<<<<<
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 */
                              goto __pyx_L27;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1445
 *                 lerr[i] += cdummy
 *         else:
 *             lprop[i, 0] += sum_sig             # <<<<<<<<<<<<<<
 *             if nchan == 2:
 *                 # case (signal, norm)
 */
                            /*else*/ {
                              __pyx_t_19 = __pyx_v_i;
                              __pyx_t_20 = 0;
                              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_19 * __pyx_v_lprop.strides[0]) )) + __pyx_t_20)) )) += __pyx_v_sum_sig;

                              /* "pyFAI/ext/_distortion.pyx":1446
 *         else:
 *             lprop[i, 0] += sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 lprop[i, 1] += sum_norm
 */
                              switch (__pyx_v_nchan) {
                                case 2:

                                /* "pyFAI/ext/_distortion.pyx":1448
 *             if nchan == 2:
 *                 # case (signal, norm)
 *                 lprop[i, 1] += sum_norm             # <<<<<<<<<<<<<<
 *                 lout[i] += sum_sig / sum_norm
 *             elif nchan == 3:
 */
                                __pyx_t_20 = __pyx_v_i;
                                __pyx_t_19 = 1;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_20 * __pyx_v_lprop.strides[0]) )) + __pyx_t_19)) )) += __pyx_v_sum_norm;

                                /* "pyFAI/ext/_distortion.pyx":1449
 *                 # case (signal, norm)
 *                 lprop[i, 1] += sum_norm
 *                 lout[i] += sum_sig / sum_norm             # <<<<<<<<<<<<<<
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 */
                                __pyx_t_19 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_19)) )) += (__pyx_v_sum_sig / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1446
 *         else:
 *             lprop[i, 0] += sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 lprop[i, 1] += sum_norm
 */
                                break;
                                case 3:

                                /* "pyFAI/ext/_distortion.pyx":1452
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var             # <<<<<<<<<<<<<<
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm
 */
                                __pyx_t_19 = __pyx_v_i;
                                __pyx_t_20 = 1;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_19 * __pyx_v_lprop.strides[0]) )) + __pyx_t_20)) )) += __pyx_v_sum_var;

                                /* "pyFAI/ext/_distortion.pyx":1453
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var
 *                 lprop[i, 2] += sum_norm             # <<<<<<<<<<<<<<
 *                 lout[i] += sum_sig / sum_norm
 *                 lerr[i] += sqrt(sum_var) / sum_norm
 */
                                __pyx_t_20 = __pyx_v_i;
                                __pyx_t_19 = 2;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_20 * __pyx_v_lprop.strides[0]) )) + __pyx_t_19)) )) += __pyx_v_sum_norm;

                                /* "pyFAI/ext/_distortion.pyx":1454
 *                 lprop[i, 1] += sum_var
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm             # <<<<<<<<<<<<<<
 *                 lerr[i] += sqrt(sum_var) / sum_norm
 *             else:
 */
                                __pyx_t_19 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_19)) )) += (__pyx_v_sum_sig / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1455
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm
 *                 lerr[i] += sqrt(sum_var) / sum_norm             # <<<<<<<<<<<<<<
 *             else:
 *                 # Case signal only. No normalization to behave like FIT2D does
 */
                                __pyx_t_19 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lerr.data) + __pyx_t_19)) )) += (sqrt(__pyx_v_sum_var) / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1450
 *                 lprop[i, 1] += sum_norm
 *                 lout[i] += sum_sig / sum_norm
 *             elif nchan == 3:             # <<<<<<<<<<<<<<
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var
 */
                                break;
                                default:

                                /* "pyFAI/ext/_distortion.pyx":1458
 *             else:
 *                 # Case signal only. No normalization to behave like FIT2D does
 *                 lout[i] += sum_sig             # <<<<<<<<<<<<<<
 * 
 *     if nchan == 3:
 */
                                __pyx_t_19 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_19)) )) += __pyx_v_sum_sig;
                                break;
                              }
                            }
                            __pyx_L27:;
                            goto __pyx_L30;
                            __pyx_L13_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L29;
                            __pyx_L29:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates4)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_i;
                                __pyx_parallel_temp2 = __pyx_v_idx;
                                __pyx_parallel_temp3 = __pyx_v_j;
                                __pyx_parallel_temp4 = __pyx_v_sum_norm;
                                __pyx_parallel_temp5 = __pyx_v_sum_sig;
                                __pyx_parallel_temp6 = __pyx_v_sum_var;
                                __pyx_parallel_temp7 = __pyx_v_value;
                            }
                            __pyx_L30:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_10);
                    __pyx_t_10 = NULL;
                    __Pyx_XDECREF(__pyx_t_13);
                    __pyx_t_13 = NULL;
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_9);
                    __pyx_t_9 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_i = __pyx_parallel_temp1;
              __pyx_v_idx = __pyx_parallel_temp2;
              __pyx_v_j = __pyx_parallel_temp3;
              __pyx_v_sum_norm = __pyx_parallel_temp4;
              __pyx_v_sum_sig = __pyx_parallel_temp5;
              __pyx_v_sum_var = __pyx_parallel_temp6;
              __pyx_v_value = __pyx_parallel_temp7;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L9_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1413
 *         lerr = err.ravel()
 *     size = lin.shape[0]
 *     for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum_sig = 0.0
 *         sum_var = 0.0
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L10;
        }
        __pyx_L9_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L10:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1460
 *                 lout[i] += sum_sig
 * 
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         return out, err, prop
 *     else:
 */
  __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1461
 * 
 *     if nchan == 3:
 *         return out, err, prop             # <<<<<<<<<<<<<<
 *     else:
 *         return out, prop
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_err)) { __Pyx_RaiseUnboundLocalError("err"); __PYX_ERR(0, 1461, __pyx_L1_error) }
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_out);
    __Pyx_GIVEREF(__pyx_v_out);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_out);
    __Pyx_INCREF(__pyx_v_err);
    __Pyx_GIVEREF(__pyx_v_err);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_err);
    __Pyx_INCREF(__pyx_v_prop);
    __Pyx_GIVEREF(__pyx_v_prop);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_prop);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1460
 *                 lout[i] += sum_sig
 * 
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         return out, err, prop
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1463
 *         return out, err, prop
 *     else:
 *         return out, prop             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_out);
    __Pyx_GIVEREF(__pyx_v_out);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_out);
    __Pyx_INCREF(__pyx_v_prop);
    __Pyx_GIVEREF(__pyx_v_prop);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_prop);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "pyFAI/ext/_distortion.pyx":1365
 * 
 * 
 * def correct_LUT_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_LUT_preproc_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lerr, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lprop, 1);
  __Pyx_XDECREF(__pyx_v_shape_out0);
  __Pyx_XDECREF(__pyx_v_shape_out1);
  __Pyx_XDECREF(__pyx_v_prop);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_err);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1466
 * 
 * 
 * def correct_CSR_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_37correct_CSR_preproc_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_36correct_CSR_preproc_double[] = "correct_CSR_preproc_double(image, shape_out, LUT, dummy=None, delta_dummy=None, empty=numpy.NaN)\nCorrect an image based on the look-up table calculated ...\n    implementation using double precision accumulator\n\n    :param image: 2D-array with the image (signal, variance, normalization)\n    :param shape_in: shape of input image\n    :param shape_out: shape of output image\n    :param LUT: Look up table, here a 3-tuple array of ndarray\n    :param dummy: value for invalid pixels\n    :param delta_dummy: precision for invalid pixels\n    :param empty: numerical value for empty pixels (if dummy is not provided)\n    :param method: integration method: can be \"kahan\" using single precision\n            compensated for error or \"double\" in double precision (64 bits)\n\n    :return: corrected 2D image + array with (signal, variance, norm)\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_37correct_CSR_preproc_double = {"correct_CSR_preproc_double", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_37correct_CSR_preproc_double, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_36correct_CSR_preproc_double};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_37correct_CSR_preproc_double(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape_out = 0;
  PyObject *__pyx_v_LUT = 0;
  PyObject *__pyx_v_dummy = 0;
  PyObject *__pyx_v_delta_dummy = 0;
  PyObject *__pyx_v_empty = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct_CSR_preproc_double (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape_out,&__pyx_n_s_LUT,&__pyx_n_s_dummy,&__pyx_n_s_delta_dummy,&__pyx_n_s_empty,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "pyFAI/ext/_distortion.pyx":1468
 * def correct_CSR_preproc_double(image, shape_out,
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                                empty=numpy.NaN):
 *     """Correct an image based on the look-up table calculated ...
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    values[5] = __pyx_k__24;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_preproc_double", 0, 3, 6, 1); __PYX_ERR(0, 1466, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct_CSR_preproc_double", 0, 3, 6, 2); __PYX_ERR(0, 1466, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dummy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_dummy);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_empty);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct_CSR_preproc_double") < 0)) __PYX_ERR(0, 1466, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_shape_out = values[1];
    __pyx_v_LUT = values[2];
    __pyx_v_dummy = values[3];
    __pyx_v_delta_dummy = values[4];
    __pyx_v_empty = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct_CSR_preproc_double", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1466, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_preproc_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_LUT) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "LUT"); __PYX_ERR(0, 1467, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_36correct_CSR_preproc_double(__pyx_self, __pyx_v_image, __pyx_v_shape_out, __pyx_v_LUT, __pyx_v_dummy, __pyx_v_delta_dummy, __pyx_v_empty);

  /* "pyFAI/ext/_distortion.pyx":1466
 * 
 * 
 * def correct_CSR_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_36correct_CSR_preproc_double(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape_out, PyObject *__pyx_v_LUT, PyObject *__pyx_v_dummy, PyObject *__pyx_v_delta_dummy, PyObject *__pyx_v_empty) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  int __pyx_v_bins;
  int __pyx_v_nchan;
  float __pyx_v_value;
  float __pyx_v_cdummy;
  float __pyx_v_cdelta_dummy;
  double __pyx_v_sum_sig;
  double __pyx_v_sum_var;
  double __pyx_v_sum_norm;
  double __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lerr = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lprop = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_do_dummy;
  PyObject *__pyx_v_shape_out0 = NULL;
  PyObject *__pyx_v_shape_out1 = NULL;
  PyObject *__pyx_v_prop = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  float __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct_CSR_preproc_double", 0);

  /* "pyFAI/ext/_distortion.pyx":1493
 *         float32_t[:, ::1] lin, lprop
 *         int[::1] indices, indptr
 *         bint do_dummy = dummy is not None             # <<<<<<<<<<<<<<
 * 
 *     if do_dummy:
 */
  __pyx_t_1 = (__pyx_v_dummy != Py_None);
  __pyx_v_do_dummy = __pyx_t_1;

  /* "pyFAI/ext/_distortion.pyx":1495
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
  __pyx_t_1 = (__pyx_v_do_dummy != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1496
 * 
 *     if do_dummy:
 *         cdummy = dummy             # <<<<<<<<<<<<<<
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0
 */
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_dummy); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1496, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;

    /* "pyFAI/ext/_distortion.pyx":1497
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    __pyx_t_1 = (__pyx_v_delta_dummy == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1498
 *         cdummy = dummy
 *         if delta_dummy is None:
 *             cdelta_dummy = 0.0             # <<<<<<<<<<<<<<
 *     else:
 *         cdummy = empty
 */
      __pyx_v_cdelta_dummy = 0.0;

      /* "pyFAI/ext/_distortion.pyx":1497
 *     if do_dummy:
 *         cdummy = dummy
 *         if delta_dummy is None:             # <<<<<<<<<<<<<<
 *             cdelta_dummy = 0.0
 *     else:
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1495
 *         bint do_dummy = dummy is not None
 * 
 *     if do_dummy:             # <<<<<<<<<<<<<<
 *         cdummy = dummy
 *         if delta_dummy is None:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1500
 *             cdelta_dummy = 0.0
 *     else:
 *         cdummy = empty             # <<<<<<<<<<<<<<
 *     data, indices, indptr = LUT
 *     bins = indptr.size - 1
 */
  /*else*/ {
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_v_empty); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L1_error)
    __pyx_v_cdummy = __pyx_t_2;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1501
 *     else:
 *         cdummy = empty
 *     data, indices, indptr = LUT             # <<<<<<<<<<<<<<
 *     bins = indptr.size - 1
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_LUT))) || (PyList_CheckExact(__pyx_v_LUT))) {
    PyObject* sequence = __pyx_v_LUT;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1501, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_v_LUT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 2; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 3) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1501, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_int(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_data = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;
  __pyx_v_indices = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;
  __pyx_v_indptr = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1502
 *         cdummy = empty
 *     data, indices, indptr = LUT
 *     bins = indptr.size - 1             # <<<<<<<<<<<<<<
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 * 
 */
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_int, (int (*)(char *, PyObject *)) __pyx_memview_set_int, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_bins = __pyx_t_12;

  /* "pyFAI/ext/_distortion.pyx":1503
 *     data, indices, indptr = LUT
 *     bins = indptr.size - 1
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"             # <<<<<<<<<<<<<<
 * 
 *     nchan = image.shape[2]
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_prod); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_shape_out) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_shape_out);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_bins); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_u_shape_out0_shape_out1_indptr_siz);
      __PYX_ERR(0, 1503, __pyx_L1_error)
    }
  }
  #endif

  /* "pyFAI/ext/_distortion.pyx":1505
 *     assert numpy.prod(shape_out) == bins, "shape_out0*shape_out1 == indptr.size-1"
 * 
 *     nchan = image.shape[2]             # <<<<<<<<<<<<<<
 *     shape_out0, shape_out1 = shape_out
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_nchan = __pyx_t_12;

  /* "pyFAI/ext/_distortion.pyx":1506
 * 
 *     nchan = image.shape[2]
 *     shape_out0, shape_out1 = shape_out             # <<<<<<<<<<<<<<
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_shape_out))) || (PyList_CheckExact(__pyx_v_shape_out))) {
    PyObject* sequence = __pyx_v_shape_out;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1506, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_6 = PyObject_GetIter(__pyx_v_shape_out); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L7_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L7_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_6), 2) < 0) __PYX_ERR(0, 1506, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L8_unpacking_done;
    __pyx_L7_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1506, __pyx_L1_error)
    __pyx_L8_unpacking_done:;
  }
  __pyx_v_shape_out0 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_shape_out1 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1508
 *     shape_out0, shape_out1 = shape_out
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_shape_out0);
  __Pyx_GIVEREF(__pyx_v_shape_out0);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_shape_out0);
  __Pyx_INCREF(__pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_v_shape_out1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_13) < 0) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_prop = __pyx_t_13;
  __pyx_t_13 = 0;

  /* "pyFAI/ext/_distortion.pyx":1509
 * 
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 *     lprop = prop.reshape((-1, nchan))             # <<<<<<<<<<<<<<
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_prop, __pyx_n_s_reshape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_13 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_13, PyBUF_WRITABLE); if (unlikely(!__pyx_t_14.memview)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_lprop = __pyx_t_14;
  __pyx_t_14.memview = NULL;
  __pyx_t_14.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1510
 *     prop = numpy.zeros((shape_out0, shape_out1, nchan), dtype=numpy.float32)
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_numpy); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_shape_out0);
  __Pyx_GIVEREF(__pyx_v_shape_out0);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_shape_out0);
  __Pyx_INCREF(__pyx_v_shape_out1);
  __Pyx_GIVEREF(__pyx_v_shape_out1);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_shape_out1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_13);
  __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_out = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyFAI/ext/_distortion.pyx":1511
 *     lprop = prop.reshape((-1, nchan))
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:
 */
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_13, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lout = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1512
 *     out = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))             # <<<<<<<<<<<<<<
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_numpy); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_image);
  __Pyx_GIVEREF(__pyx_v_image);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_image);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_15) < 0) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_reshape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_nchan); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_7 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_15, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_14.memview)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_14;
  __pyx_t_14.memview = NULL;
  __pyx_t_14.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1513
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 */
  __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1514
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *         lerr = err.ravel()
 *     size = lin.shape[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_shape_out0);
    __Pyx_GIVEREF(__pyx_v_shape_out0);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_shape_out0);
    __Pyx_INCREF(__pyx_v_shape_out1);
    __Pyx_GIVEREF(__pyx_v_shape_out1);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_shape_out1);
    __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_numpy); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_float32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_13, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_err = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "pyFAI/ext/_distortion.pyx":1515
 *     if nchan == 3:
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()             # <<<<<<<<<<<<<<
 *     size = lin.shape[0]
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_ravel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_lerr = __pyx_t_9;
    __pyx_t_9.memview = NULL;
    __pyx_t_9.data = NULL;

    /* "pyFAI/ext/_distortion.pyx":1513
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).reshape((-1, nchan))
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1516
 *         err = numpy.zeros((shape_out0, shape_out1), dtype=numpy.float32)
 *         lerr = err.ravel()
 *     size = lin.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 */
  __pyx_v_size = (__pyx_v_lin.shape[0]);

  /* "pyFAI/ext/_distortion.pyx":1518
 *     size = lin.shape[0]
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum_sig = 0.0
 *         sum_var = 0.0
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_12 = __pyx_v_bins;
        if ((1 == 0)) abort();
        {
            double __pyx_parallel_temp0 = ((double)__PYX_NAN());
            int __pyx_parallel_temp1 = ((int)0xbad0bad0);
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            double __pyx_parallel_temp4 = ((double)__PYX_NAN());
            double __pyx_parallel_temp5 = ((double)__PYX_NAN());
            double __pyx_parallel_temp6 = ((double)__PYX_NAN());
            float __pyx_parallel_temp7 = ((float)__PYX_NAN());
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_17 = (__pyx_t_12 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_17 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_1, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_3) firstprivate(__pyx_t_13, __pyx_t_15, __pyx_t_4, __pyx_t_6, __pyx_t_7) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) lastprivate(__pyx_v_sum_norm) lastprivate(__pyx_v_sum_sig) lastprivate(__pyx_v_sum_var) lastprivate(__pyx_v_value) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_17; __pyx_t_16++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_16);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((double)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);
                            __pyx_v_sum_norm = ((double)__PYX_NAN());
                            __pyx_v_sum_sig = ((double)__PYX_NAN());
                            __pyx_v_sum_var = ((double)__PYX_NAN());
                            __pyx_v_value = ((float)__PYX_NAN());

                            /* "pyFAI/ext/_distortion.pyx":1519
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum_sig = 0.0             # <<<<<<<<<<<<<<
 *         sum_var = 0.0
 *         sum_norm = 0.0
 */
                            __pyx_v_sum_sig = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1520
 *     for i in prange(bins, nogil=True, schedule="static"):
 *         sum_sig = 0.0
 *         sum_var = 0.0             # <<<<<<<<<<<<<<
 *         sum_norm = 0.0
 * 
 */
                            __pyx_v_sum_var = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1521
 *         sum_sig = 0.0
 *         sum_var = 0.0
 *         sum_norm = 0.0             # <<<<<<<<<<<<<<
 * 
 *         for j in range(indptr[i], indptr[i + 1]):
 */
                            __pyx_v_sum_norm = 0.0;

                            /* "pyFAI/ext/_distortion.pyx":1523
 *         sum_norm = 0.0
 * 
 *         for j in range(indptr[i], indptr[i + 1]):             # <<<<<<<<<<<<<<
 *             idx = indices[j]
 *             coef = data[j]
 */
                            __pyx_t_18 = (__pyx_v_i + 1);
                            __pyx_t_19 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_18)) )));
                            __pyx_t_18 = __pyx_v_i;
                            __pyx_t_20 = __pyx_t_19;
                            for (__pyx_t_21 = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indptr.data) + __pyx_t_18)) ))); __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                              __pyx_v_j = __pyx_t_21;

                              /* "pyFAI/ext/_distortion.pyx":1524
 * 
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]             # <<<<<<<<<<<<<<
 *             coef = data[j]
 *             if coef <= 0.0:
 */
                              __pyx_t_22 = __pyx_v_j;
                              __pyx_v_idx = (*((int *) ( /* dim=0 */ ((char *) (((int *) __pyx_v_indices.data) + __pyx_t_22)) )));

                              /* "pyFAI/ext/_distortion.pyx":1525
 *         for j in range(indptr[i], indptr[i + 1]):
 *             idx = indices[j]
 *             coef = data[j]             # <<<<<<<<<<<<<<
 *             if coef <= 0.0:
 *                 continue
 */
                              __pyx_t_22 = __pyx_v_j;
                              __pyx_v_coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_22)) )));

                              /* "pyFAI/ext/_distortion.pyx":1526
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0.0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1527
 *             coef = data[j]
 *             if coef <= 0.0:
 *                 continue             # <<<<<<<<<<<<<<
 *             if idx >= size:
 *                 with gil:
 */
                                goto __pyx_L17_continue;

                                /* "pyFAI/ext/_distortion.pyx":1526
 *             idx = indices[j]
 *             coef = data[j]
 *             if coef <= 0.0:             # <<<<<<<<<<<<<<
 *                 continue
 *             if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1528
 *             if coef <= 0.0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1529
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1530
 *             if idx >= size:
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warning); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_13);
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                                      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_15);
                                      __Pyx_GIVEREF(__pyx_t_7);
                                      PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_7);
                                      __Pyx_GIVEREF(__pyx_t_6);
                                      PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_6);
                                      __pyx_t_7 = 0;
                                      __pyx_t_6 = 0;
                                      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_15); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_6);
                                      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                                      __pyx_t_15 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
                                        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
                                        if (likely(__pyx_t_15)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                                          __Pyx_INCREF(__pyx_t_15);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_13, function);
                                        }
                                      }
                                      __pyx_t_4 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_15, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_6);
                                      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                                      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1530, __pyx_L24_error)
                                      __Pyx_GOTREF(__pyx_t_4);
                                      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1531
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *             value = lin[idx, 0]
 */
                                      goto __pyx_L21_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1529
 *                 continue
 *             if idx >= size:
 *                 with gil:             # <<<<<<<<<<<<<<
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                     continue
 */
                                    /*finally:*/ {
                                      __pyx_L21_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L17_continue;
                                      }
                                      __pyx_L24_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L15_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1528
 *             if coef <= 0.0:
 *                 continue
 *             if idx >= size:             # <<<<<<<<<<<<<<
 *                 with gil:
 *                     logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1533
 *                     continue
 * 
 *             value = lin[idx, 0]             # <<<<<<<<<<<<<<
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 */
                              __pyx_t_22 = __pyx_v_idx;
                              __pyx_t_23 = 0;
                              __pyx_v_value = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_22 * __pyx_v_lin.strides[0]) )) + __pyx_t_23)) )));

                              /* "pyFAI/ext/_distortion.pyx":1534
 * 
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum_sig = value * coef + sum_sig
 */
                              __pyx_t_1 = (__pyx_v_do_dummy != 0);
                              if (__pyx_t_1) {
                              } else {
                                __pyx_t_3 = __pyx_t_1;
                                goto __pyx_L27_bool_binop_done;
                              }
                              __pyx_t_1 = ((fabs((__pyx_v_value - __pyx_v_cdummy)) <= __pyx_v_cdelta_dummy) != 0);
                              __pyx_t_3 = __pyx_t_1;
                              __pyx_L27_bool_binop_done:;
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1535
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue             # <<<<<<<<<<<<<<
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:
 */
                                goto __pyx_L17_continue;

                                /* "pyFAI/ext/_distortion.pyx":1534
 * 
 *             value = lin[idx, 0]
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:             # <<<<<<<<<<<<<<
 *                 continue
 *             sum_sig = value * coef + sum_sig
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1536
 *             if do_dummy and fabs(value - cdummy) <= cdelta_dummy:
 *                 continue
 *             sum_sig = value * coef + sum_sig             # <<<<<<<<<<<<<<
 *             if nchan == 2:
 *                 # case (signal, norm)
 */
                              __pyx_v_sum_sig = ((__pyx_v_value * __pyx_v_coef) + __pyx_v_sum_sig);

                              /* "pyFAI/ext/_distortion.pyx":1537
 *                 continue
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 */
                              switch (__pyx_v_nchan) {
                                case 2:

                                /* "pyFAI/ext/_distortion.pyx":1539
 *             if nchan == 2:
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm             # <<<<<<<<<<<<<<
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 */
                                __pyx_t_23 = __pyx_v_idx;
                                __pyx_t_22 = 1;
                                __pyx_v_sum_norm = ((__pyx_v_coef * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_23 * __pyx_v_lin.strides[0]) )) + __pyx_t_22)) )))) + __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1537
 *                 continue
 *             sum_sig = value * coef + sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 */
                                break;
                                case 3:

                                /* "pyFAI/ext/_distortion.pyx":1542
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var             # <<<<<<<<<<<<<<
 *                 sum_norm = coef * lin[idx, 2] + sum_norm
 *             else:
 */
                                __pyx_t_22 = __pyx_v_idx;
                                __pyx_t_23 = 1;
                                __pyx_v_sum_var = (((__pyx_v_coef * __pyx_v_coef) * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_22 * __pyx_v_lin.strides[0]) )) + __pyx_t_23)) )))) + __pyx_v_sum_var);

                                /* "pyFAI/ext/_distortion.pyx":1543
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var
 *                 sum_norm = coef * lin[idx, 2] + sum_norm             # <<<<<<<<<<<<<<
 *             else:
 *                 sum_norm = sum_norm + coef
 */
                                __pyx_t_23 = __pyx_v_idx;
                                __pyx_t_22 = 2;
                                __pyx_v_sum_norm = ((__pyx_v_coef * (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lin.data + __pyx_t_23 * __pyx_v_lin.strides[0]) )) + __pyx_t_22)) )))) + __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1540
 *                 # case (signal, norm)
 *                 sum_norm = coef * lin[idx, 1] + sum_norm
 *             elif nchan == 3:             # <<<<<<<<<<<<<<
 *                 # case (signal, variance,  normalization)
 *                 sum_var = coef * coef * lin[idx, 1] + sum_var
 */
                                break;
                                default:

                                /* "pyFAI/ext/_distortion.pyx":1545
 *                 sum_norm = coef * lin[idx, 2] + sum_norm
 *             else:
 *                 sum_norm = sum_norm + coef             # <<<<<<<<<<<<<<
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 */
                                __pyx_v_sum_norm = (__pyx_v_sum_norm + __pyx_v_coef);
                                break;
                              }
                              __pyx_L17_continue:;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1547
 *                 sum_norm = sum_norm + coef
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel             # <<<<<<<<<<<<<<
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 */
                            __pyx_t_3 = ((__pyx_v_sum_norm == 0.0) != 0);
                            if (__pyx_t_3) {

                              /* "pyFAI/ext/_distortion.pyx":1548
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction             # <<<<<<<<<<<<<<
 *             if nchan == 3:
 *                 lerr[i] += cdummy
 */
                              __pyx_t_18 = __pyx_v_i;
                              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_18)) )) += __pyx_v_cdummy;

                              /* "pyFAI/ext/_distortion.pyx":1549
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:             # <<<<<<<<<<<<<<
 *                 lerr[i] += cdummy
 *         else:
 */
                              __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1550
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 *                 lerr[i] += cdummy             # <<<<<<<<<<<<<<
 *         else:
 *             lprop[i, 0] += sum_sig
 */
                                __pyx_t_18 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lerr.data) + __pyx_t_18)) )) += __pyx_v_cdummy;

                                /* "pyFAI/ext/_distortion.pyx":1549
 *         if sum_norm == 0.0:  # No contribution to this output pixel
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:             # <<<<<<<<<<<<<<
 *                 lerr[i] += cdummy
 *         else:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1547
 *                 sum_norm = sum_norm + coef
 * 
 *         if sum_norm == 0.0:  # No contribution to this output pixel             # <<<<<<<<<<<<<<
 *             lout[i] += cdummy  # this += is for Cython's reduction
 *             if nchan == 3:
 */
                              goto __pyx_L29;
                            }

                            /* "pyFAI/ext/_distortion.pyx":1552
 *                 lerr[i] += cdummy
 *         else:
 *             lprop[i, 0] += sum_sig             # <<<<<<<<<<<<<<
 *             if nchan == 2:
 *                 # case (signal, norm)
 */
                            /*else*/ {
                              __pyx_t_18 = __pyx_v_i;
                              __pyx_t_22 = 0;
                              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_18 * __pyx_v_lprop.strides[0]) )) + __pyx_t_22)) )) += __pyx_v_sum_sig;

                              /* "pyFAI/ext/_distortion.pyx":1553
 *         else:
 *             lprop[i, 0] += sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 lout[i] += sum_sig / sum_norm
 */
                              switch (__pyx_v_nchan) {
                                case 2:

                                /* "pyFAI/ext/_distortion.pyx":1555
 *             if nchan == 2:
 *                 # case (signal, norm)
 *                 lout[i] += sum_sig / sum_norm             # <<<<<<<<<<<<<<
 *                 lprop[i, 1] += sum_norm
 *             elif nchan == 3:
 */
                                __pyx_t_22 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_22)) )) += (__pyx_v_sum_sig / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1556
 *                 # case (signal, norm)
 *                 lout[i] += sum_sig / sum_norm
 *                 lprop[i, 1] += sum_norm             # <<<<<<<<<<<<<<
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 */
                                __pyx_t_22 = __pyx_v_i;
                                __pyx_t_18 = 1;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_22 * __pyx_v_lprop.strides[0]) )) + __pyx_t_18)) )) += __pyx_v_sum_norm;

                                /* "pyFAI/ext/_distortion.pyx":1553
 *         else:
 *             lprop[i, 0] += sum_sig
 *             if nchan == 2:             # <<<<<<<<<<<<<<
 *                 # case (signal, norm)
 *                 lout[i] += sum_sig / sum_norm
 */
                                break;
                                case 3:

                                /* "pyFAI/ext/_distortion.pyx":1559
 *             elif nchan == 3:
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var             # <<<<<<<<<<<<<<
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm
 */
                                __pyx_t_18 = __pyx_v_i;
                                __pyx_t_22 = 1;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_18 * __pyx_v_lprop.strides[0]) )) + __pyx_t_22)) )) += __pyx_v_sum_var;

                                /* "pyFAI/ext/_distortion.pyx":1560
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var
 *                 lprop[i, 2] += sum_norm             # <<<<<<<<<<<<<<
 *                 lout[i] += sum_sig / sum_norm
 *                 lerr[i] += sqrt(sum_var) / sum_norm
 */
                                __pyx_t_22 = __pyx_v_i;
                                __pyx_t_18 = 2;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_lprop.data + __pyx_t_22 * __pyx_v_lprop.strides[0]) )) + __pyx_t_18)) )) += __pyx_v_sum_norm;

                                /* "pyFAI/ext/_distortion.pyx":1561
 *                 lprop[i, 1] += sum_var
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm             # <<<<<<<<<<<<<<
 *                 lerr[i] += sqrt(sum_var) / sum_norm
 *             else:
 */
                                __pyx_t_18 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_18)) )) += (__pyx_v_sum_sig / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1562
 *                 lprop[i, 2] += sum_norm
 *                 lout[i] += sum_sig / sum_norm
 *                 lerr[i] += sqrt(sum_var) / sum_norm             # <<<<<<<<<<<<<<
 *             else:
 *                 # Case signal only. No normalization to behave like FIT2D does
 */
                                __pyx_t_18 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lerr.data) + __pyx_t_18)) )) += (sqrt(__pyx_v_sum_var) / __pyx_v_sum_norm);

                                /* "pyFAI/ext/_distortion.pyx":1557
 *                 lout[i] += sum_sig / sum_norm
 *                 lprop[i, 1] += sum_norm
 *             elif nchan == 3:             # <<<<<<<<<<<<<<
 *                 # case (signal, variance,  normalization)
 *                 lprop[i, 1] += sum_var
 */
                                break;
                                default:

                                /* "pyFAI/ext/_distortion.pyx":1565
 *             else:
 *                 # Case signal only. No normalization to behave like FIT2D does
 *                 lout[i] += sum_sig             # <<<<<<<<<<<<<<
 * 
 *     if nchan == 3:
 */
                                __pyx_t_18 = __pyx_v_i;
                                *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_18)) )) += __pyx_v_sum_sig;
                                break;
                              }
                            }
                            __pyx_L29:;
                            goto __pyx_L32;
                            __pyx_L15_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L31;
                            __pyx_L31:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates5)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_i;
                                __pyx_parallel_temp2 = __pyx_v_idx;
                                __pyx_parallel_temp3 = __pyx_v_j;
                                __pyx_parallel_temp4 = __pyx_v_sum_norm;
                                __pyx_parallel_temp5 = __pyx_v_sum_sig;
                                __pyx_parallel_temp6 = __pyx_v_sum_var;
                                __pyx_parallel_temp7 = __pyx_v_value;
                            }
                            __pyx_L32:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_13);
                    __pyx_t_13 = NULL;
                    __Pyx_XDECREF(__pyx_t_15);
                    __pyx_t_15 = NULL;
                    __Pyx_XDECREF(__pyx_t_4);
                    __pyx_t_4 = NULL;
                    __Pyx_XDECREF(__pyx_t_6);
                    __pyx_t_6 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_i = __pyx_parallel_temp1;
              __pyx_v_idx = __pyx_parallel_temp2;
              __pyx_v_j = __pyx_parallel_temp3;
              __pyx_v_sum_norm = __pyx_parallel_temp4;
              __pyx_v_sum_sig = __pyx_parallel_temp5;
              __pyx_v_sum_var = __pyx_parallel_temp6;
              __pyx_v_value = __pyx_parallel_temp7;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L11_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1518
 *     size = lin.shape[0]
 * 
 *     for i in prange(bins, nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *         sum_sig = 0.0
 *         sum_var = 0.0
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L12;
        }
        __pyx_L11_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L12:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1567
 *                 lout[i] += sum_sig
 * 
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         return out, err, prop
 *     else:
 */
  __pyx_t_3 = ((__pyx_v_nchan == 3) != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1568
 * 
 *     if nchan == 3:
 *         return out, err, prop             # <<<<<<<<<<<<<<
 *     else:
 *         return out, prop
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_err)) { __Pyx_RaiseUnboundLocalError("err"); __PYX_ERR(0, 1568, __pyx_L1_error) }
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_out);
    __Pyx_GIVEREF(__pyx_v_out);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_out);
    __Pyx_INCREF(__pyx_v_err);
    __Pyx_GIVEREF(__pyx_v_err);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_err);
    __Pyx_INCREF(__pyx_v_prop);
    __Pyx_GIVEREF(__pyx_v_prop);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_prop);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "pyFAI/ext/_distortion.pyx":1567
 *                 lout[i] += sum_sig
 * 
 *     if nchan == 3:             # <<<<<<<<<<<<<<
 *         return out, err, prop
 *     else:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1570
 *         return out, err, prop
 *     else:
 *         return out, prop             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_out);
    __Pyx_GIVEREF(__pyx_v_out);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_out);
    __Pyx_INCREF(__pyx_v_prop);
    __Pyx_GIVEREF(__pyx_v_prop);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_prop);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "pyFAI/ext/_distortion.pyx":1466
 * 
 * 
 * def correct_CSR_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_13);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pyFAI.ext._distortion.correct_CSR_preproc_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lerr, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lprop, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __Pyx_XDECREF(__pyx_v_shape_out0);
  __Pyx_XDECREF(__pyx_v_shape_out1);
  __Pyx_XDECREF(__pyx_v_prop);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1573
 * 
 * 
 * def uncorrect_LUT(image, shape, lut_t[:, :]LUT):             # <<<<<<<<<<<<<<
 *     """
 *     Take an image which has been corrected and transform it into it's raw (with loss of information)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_39uncorrect_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_38uncorrect_LUT[] = "uncorrect_LUT(image, shape, lut_t[:, :] LUT)\n\n    Take an image which has been corrected and transform it into it's raw (with loss of information)\n\n    :param image: 2D-array with the image\n    :param shape: shape of output image\n    :param LUT: Look up table, here a 2D-array of struct\n    :return: uncorrected 2D image and a mask (pixels in raw image not existing)\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_39uncorrect_LUT = {"uncorrect_LUT", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_39uncorrect_LUT, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_38uncorrect_LUT};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_39uncorrect_LUT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape = 0;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uncorrect_LUT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape,&__pyx_n_s_LUT,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("uncorrect_LUT", 1, 3, 3, 1); __PYX_ERR(0, 1573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("uncorrect_LUT", 1, 3, 3, 2); __PYX_ERR(0, 1573, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "uncorrect_LUT") < 0)) __PYX_ERR(0, 1573, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_image = values[0];
    __pyx_v_shape = values[1];
    __pyx_v_LUT = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_LUT.memview)) __PYX_ERR(0, 1573, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uncorrect_LUT", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1573, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.uncorrect_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_38uncorrect_LUT(__pyx_self, __pyx_v_image, __pyx_v_shape, __pyx_v_LUT);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_38uncorrect_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape, __Pyx_memviewslice __pyx_v_LUT) {
  int __pyx_v_idx;
  int __pyx_v_j;
  float __pyx_v_total;
  float __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_lmask = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_17;
  int __pyx_t_18;
  __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uncorrect_LUT", 0);

  /* "pyFAI/ext/_distortion.pyx":1588
 *         float32_t[::1] lout, lin
 * 
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()             # <<<<<<<<<<<<<<
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_image);
  __Pyx_GIVEREF(__pyx_v_image);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_image);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ravel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_lin = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1589
 * 
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 *     out = numpy.zeros(shape, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_out = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1590
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     mask = numpy.zeros(shape, dtype=numpy.int8)             # <<<<<<<<<<<<<<
 *     lmask = mask.ravel()
 *     lout = out.ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_shape);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyFAI/ext/_distortion.pyx":1591
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     for idx in range(LUT.shape[0]):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_ravel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int8_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 1591, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lmask = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1592
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     for idx in range(LUT.shape[0]):
 *         total = 0.0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lout = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1593
 *     lmask = mask.ravel()
 *     lout = out.ravel()
 *     for idx in range(LUT.shape[0]):             # <<<<<<<<<<<<<<
 *         total = 0.0
 *         for j in range(LUT.shape[1]):
 */
  __pyx_t_9 = (__pyx_v_LUT.shape[0]);
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_idx = __pyx_t_11;

    /* "pyFAI/ext/_distortion.pyx":1594
 *     lout = out.ravel()
 *     for idx in range(LUT.shape[0]):
 *         total = 0.0             # <<<<<<<<<<<<<<
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 */
    __pyx_v_total = 0.0;

    /* "pyFAI/ext/_distortion.pyx":1595
 *     for idx in range(LUT.shape[0]):
 *         total = 0.0
 *         for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *             coef = LUT[idx, j].coef
 *             if coef > 0:
 */
    __pyx_t_12 = (__pyx_v_LUT.shape[1]);
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "pyFAI/ext/_distortion.pyx":1596
 *         total = 0.0
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef             # <<<<<<<<<<<<<<
 *             if coef > 0:
 *                 total += coef
 */
      __pyx_t_15 = __pyx_v_idx;
      __pyx_t_16 = __pyx_v_j;
      __pyx_t_17 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_15 * __pyx_v_LUT.strides[0]) ) + __pyx_t_16 * __pyx_v_LUT.strides[1]) ))).coef;
      __pyx_v_coef = __pyx_t_17;

      /* "pyFAI/ext/_distortion.pyx":1597
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 total += coef
 *         if total <= 0:
 */
      __pyx_t_18 = ((__pyx_v_coef > 0.0) != 0);
      if (__pyx_t_18) {

        /* "pyFAI/ext/_distortion.pyx":1598
 *             coef = LUT[idx, j].coef
 *             if coef > 0:
 *                 total += coef             # <<<<<<<<<<<<<<
 *         if total <= 0:
 *             lmask[idx] = 1
 */
        __pyx_v_total = (__pyx_v_total + __pyx_v_coef);

        /* "pyFAI/ext/_distortion.pyx":1597
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 total += coef
 *         if total <= 0:
 */
      }
    }

    /* "pyFAI/ext/_distortion.pyx":1599
 *             if coef > 0:
 *                 total += coef
 *         if total <= 0:             # <<<<<<<<<<<<<<
 *             lmask[idx] = 1
 *             continue
 */
    __pyx_t_18 = ((__pyx_v_total <= 0.0) != 0);
    if (__pyx_t_18) {

      /* "pyFAI/ext/_distortion.pyx":1600
 *                 total += coef
 *         if total <= 0:
 *             lmask[idx] = 1             # <<<<<<<<<<<<<<
 *             continue
 *         val = lin[idx] / total
 */
      __pyx_t_16 = __pyx_v_idx;
      *((int8_t *) ( /* dim=0 */ ((char *) (((int8_t *) __pyx_v_lmask.data) + __pyx_t_16)) )) = 1;

      /* "pyFAI/ext/_distortion.pyx":1601
 *         if total <= 0:
 *             lmask[idx] = 1
 *             continue             # <<<<<<<<<<<<<<
 *         val = lin[idx] / total
 *         for j in range(LUT.shape[1]):
 */
      goto __pyx_L3_continue;

      /* "pyFAI/ext/_distortion.pyx":1599
 *             if coef > 0:
 *                 total += coef
 *         if total <= 0:             # <<<<<<<<<<<<<<
 *             lmask[idx] = 1
 *             continue
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1602
 *             lmask[idx] = 1
 *             continue
 *         val = lin[idx] / total             # <<<<<<<<<<<<<<
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 */
    __pyx_t_16 = __pyx_v_idx;
    __pyx_t_2 = PyFloat_FromDouble(((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_16)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_total))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyFAI/ext/_distortion.pyx":1603
 *             continue
 *         val = lin[idx] / total
 *         for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *             coef = LUT[idx, j].coef
 *             if coef > 0:
 */
    __pyx_t_12 = (__pyx_v_LUT.shape[1]);
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "pyFAI/ext/_distortion.pyx":1604
 *         val = lin[idx] / total
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef             # <<<<<<<<<<<<<<
 *             if coef > 0:
 *                 lout[LUT[idx, j].idx] += val * coef
 */
      __pyx_t_16 = __pyx_v_idx;
      __pyx_t_15 = __pyx_v_j;
      __pyx_t_17 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_16 * __pyx_v_LUT.strides[0]) ) + __pyx_t_15 * __pyx_v_LUT.strides[1]) ))).coef;
      __pyx_v_coef = __pyx_t_17;

      /* "pyFAI/ext/_distortion.pyx":1605
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 lout[LUT[idx, j].idx] += val * coef
 *     return out, mask
 */
      __pyx_t_18 = ((__pyx_v_coef > 0.0) != 0);
      if (__pyx_t_18) {

        /* "pyFAI/ext/_distortion.pyx":1606
 *             coef = LUT[idx, j].coef
 *             if coef > 0:
 *                 lout[LUT[idx, j].idx] += val * coef             # <<<<<<<<<<<<<<
 *     return out, mask
 * 
 */
        __pyx_t_2 = PyFloat_FromDouble(__pyx_v_coef); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Multiply(__pyx_v_val, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_19 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_19 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1606, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_15 = __pyx_v_idx;
        __pyx_t_16 = __pyx_v_j;
        __pyx_t_20 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_15 * __pyx_v_LUT.strides[0]) ) + __pyx_t_16 * __pyx_v_LUT.strides[1]) ))).idx;
        *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_20)) )) += __pyx_t_19;

        /* "pyFAI/ext/_distortion.pyx":1605
 *         for j in range(LUT.shape[1]):
 *             coef = LUT[idx, j].coef
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 lout[LUT[idx, j].idx] += val * coef
 *     return out, mask
 */
      }
    }
    __pyx_L3_continue:;
  }

  /* "pyFAI/ext/_distortion.pyx":1607
 *             if coef > 0:
 *                 lout[LUT[idx, j].idx] += val * coef
 *     return out, mask             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_out);
  __Pyx_GIVEREF(__pyx_v_out);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_out);
  __Pyx_INCREF(__pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_mask);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1573
 * 
 * 
 * def uncorrect_LUT(image, shape, lut_t[:, :]LUT):             # <<<<<<<<<<<<<<
 *     """
 *     Take an image which has been corrected and transform it into it's raw (with loss of information)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.uncorrect_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lmask, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_val);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1610
 * 
 * 
 * def uncorrect_CSR(image, shape, LUT):             # <<<<<<<<<<<<<<
 *     """Take an image which has been corrected and transform it into it's raw (with loss of information)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_41uncorrect_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_40uncorrect_CSR[] = "uncorrect_CSR(image, shape, LUT)\nTake an image which has been corrected and transform it into it's raw (with loss of information)\n\n    :param image: 2D-array with the image\n    :param shape: shape of output image\n    :param LUT: Look up table, here a 3-tuple of ndarray\n    :return: uncorrected 2D image and a mask (pixels in raw image not existing)\n    ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_41uncorrect_CSR = {"uncorrect_CSR", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_41uncorrect_CSR, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_40uncorrect_CSR};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_41uncorrect_CSR(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_LUT = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uncorrect_CSR (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_shape,&__pyx_n_s_LUT,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("uncorrect_CSR", 1, 3, 3, 1); __PYX_ERR(0, 1610, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_LUT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("uncorrect_CSR", 1, 3, 3, 2); __PYX_ERR(0, 1610, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "uncorrect_CSR") < 0)) __PYX_ERR(0, 1610, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_image = values[0];
    __pyx_v_shape = values[1];
    __pyx_v_LUT = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uncorrect_CSR", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1610, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.uncorrect_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_40uncorrect_CSR(__pyx_self, __pyx_v_image, __pyx_v_shape, __pyx_v_LUT);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_40uncorrect_CSR(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_shape, PyObject *__pyx_v_LUT) {
  int __pyx_v_idx;
  int __pyx_v_j;
  int __pyx_v_nbins;
  float __pyx_v_total;
  float __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_lmask = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_indptr = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int32_t __pyx_t_14;
  int32_t __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uncorrect_CSR", 0);

  /* "pyFAI/ext/_distortion.pyx":1623
 *         int8_t[:] lmask
 *         float32_t[::1] lout, lin, data
 *         int32_t[::1] indices = LUT[1]             # <<<<<<<<<<<<<<
 *         int32_t[::1] indptr = LUT[2]
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_LUT, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1624
 *         float32_t[::1] lout, lin, data
 *         int32_t[::1] indices = LUT[1]
 *         int32_t[::1] indptr = LUT[2]             # <<<<<<<<<<<<<<
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     lout = out.ravel()
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_LUT, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_indptr = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1625
 *         int32_t[::1] indices = LUT[1]
 *         int32_t[::1] indptr = LUT[2]
 *     out = numpy.zeros(shape, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_out = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pyFAI/ext/_distortion.pyx":1626
 *         int32_t[::1] indptr = LUT[2]
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     lout = out.ravel()             # <<<<<<<<<<<<<<
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_lout = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1627
 *     out = numpy.zeros(shape, dtype=numpy.float32)
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()             # <<<<<<<<<<<<<<
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_image);
  __Pyx_GIVEREF(__pyx_v_image);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_image);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ravel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_lin = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1628
 *     lout = out.ravel()
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 *     mask = numpy.zeros(shape, dtype=numpy.int8)             # <<<<<<<<<<<<<<
 *     lmask = mask.ravel()
 *     data = LUT[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_shape);
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_mask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1629
 *     lin = numpy.ascontiguousarray(image, dtype=numpy.float32).ravel()
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()             # <<<<<<<<<<<<<<
 *     data = LUT[0]
 *     nbins = indptr.size - 1
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_ravel); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_int8_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_lmask = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1630
 *     mask = numpy.zeros(shape, dtype=numpy.int8)
 *     lmask = mask.ravel()
 *     data = LUT[0]             # <<<<<<<<<<<<<<
 *     nbins = indptr.size - 1
 *     for idx in range(nbins):
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_LUT, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_data = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1631
 *     lmask = mask.ravel()
 *     data = LUT[0]
 *     nbins = indptr.size - 1             # <<<<<<<<<<<<<<
 *     for idx in range(nbins):
 *         total = 0.0
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_indptr, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nbins = __pyx_t_10;

  /* "pyFAI/ext/_distortion.pyx":1632
 *     data = LUT[0]
 *     nbins = indptr.size - 1
 *     for idx in range(nbins):             # <<<<<<<<<<<<<<
 *         total = 0.0
 *         for j in range(indptr[idx], indptr[idx + 1]):
 */
  __pyx_t_10 = __pyx_v_nbins;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_idx = __pyx_t_12;

    /* "pyFAI/ext/_distortion.pyx":1633
 *     nbins = indptr.size - 1
 *     for idx in range(nbins):
 *         total = 0.0             # <<<<<<<<<<<<<<
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 */
    __pyx_v_total = 0.0;

    /* "pyFAI/ext/_distortion.pyx":1634
 *     for idx in range(nbins):
 *         total = 0.0
 *         for j in range(indptr[idx], indptr[idx + 1]):             # <<<<<<<<<<<<<<
 *             coef = data[j]
 *             if coef > 0:
 */
    __pyx_t_13 = (__pyx_v_idx + 1);
    __pyx_t_14 = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_13)) )));
    __pyx_t_13 = __pyx_v_idx;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_13)) ))); __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "pyFAI/ext/_distortion.pyx":1635
 *         total = 0.0
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]             # <<<<<<<<<<<<<<
 *             if coef > 0:
 *                 total += coef
 */
      __pyx_t_17 = __pyx_v_j;
      __pyx_v_coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_17)) )));

      /* "pyFAI/ext/_distortion.pyx":1636
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 total += coef
 *         if total <= 0:
 */
      __pyx_t_18 = ((__pyx_v_coef > 0.0) != 0);
      if (__pyx_t_18) {

        /* "pyFAI/ext/_distortion.pyx":1637
 *             coef = data[j]
 *             if coef > 0:
 *                 total += coef             # <<<<<<<<<<<<<<
 *         if total <= 0:
 *             lmask[idx] = 1
 */
        __pyx_v_total = (__pyx_v_total + __pyx_v_coef);

        /* "pyFAI/ext/_distortion.pyx":1636
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 total += coef
 *         if total <= 0:
 */
      }
    }

    /* "pyFAI/ext/_distortion.pyx":1638
 *             if coef > 0:
 *                 total += coef
 *         if total <= 0:             # <<<<<<<<<<<<<<
 *             lmask[idx] = 1
 *             continue
 */
    __pyx_t_18 = ((__pyx_v_total <= 0.0) != 0);
    if (__pyx_t_18) {

      /* "pyFAI/ext/_distortion.pyx":1639
 *                 total += coef
 *         if total <= 0:
 *             lmask[idx] = 1             # <<<<<<<<<<<<<<
 *             continue
 *         val = lin[idx] / total
 */
      __pyx_t_13 = __pyx_v_idx;
      *((int8_t *) ( /* dim=0 */ (__pyx_v_lmask.data + __pyx_t_13 * __pyx_v_lmask.strides[0]) )) = 1;

      /* "pyFAI/ext/_distortion.pyx":1640
 *         if total <= 0:
 *             lmask[idx] = 1
 *             continue             # <<<<<<<<<<<<<<
 *         val = lin[idx] / total
 *         for j in range(indptr[idx], indptr[idx + 1]):
 */
      goto __pyx_L3_continue;

      /* "pyFAI/ext/_distortion.pyx":1638
 *             if coef > 0:
 *                 total += coef
 *         if total <= 0:             # <<<<<<<<<<<<<<
 *             lmask[idx] = 1
 *             continue
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1641
 *             lmask[idx] = 1
 *             continue
 *         val = lin[idx] / total             # <<<<<<<<<<<<<<
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 */
    __pyx_t_13 = __pyx_v_idx;
    __pyx_t_1 = PyFloat_FromDouble(((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_13)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_total))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyFAI/ext/_distortion.pyx":1642
 *             continue
 *         val = lin[idx] / total
 *         for j in range(indptr[idx], indptr[idx + 1]):             # <<<<<<<<<<<<<<
 *             coef = data[j]
 *             if coef > 0:
 */
    __pyx_t_13 = (__pyx_v_idx + 1);
    __pyx_t_14 = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_13)) )));
    __pyx_t_13 = __pyx_v_idx;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indptr.data) + __pyx_t_13)) ))); __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "pyFAI/ext/_distortion.pyx":1643
 *         val = lin[idx] / total
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]             # <<<<<<<<<<<<<<
 *             if coef > 0:
 *                 lout[indices[j]] += val * coef
 */
      __pyx_t_17 = __pyx_v_j;
      __pyx_v_coef = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_data.data) + __pyx_t_17)) )));

      /* "pyFAI/ext/_distortion.pyx":1644
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 lout[indices[j]] += val * coef
 *     return out, mask
 */
      __pyx_t_18 = ((__pyx_v_coef > 0.0) != 0);
      if (__pyx_t_18) {

        /* "pyFAI/ext/_distortion.pyx":1645
 *             coef = data[j]
 *             if coef > 0:
 *                 lout[indices[j]] += val * coef             # <<<<<<<<<<<<<<
 *     return out, mask
 * 
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_coef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = PyNumber_Multiply(__pyx_v_val, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_19 = __pyx_PyFloat_AsFloat(__pyx_t_8); if (unlikely((__pyx_t_19 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_17 = __pyx_v_j;
        __pyx_t_20 = (*((int32_t *) ( /* dim=0 */ ((char *) (((int32_t *) __pyx_v_indices.data) + __pyx_t_17)) )));
        *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_20)) )) += __pyx_t_19;

        /* "pyFAI/ext/_distortion.pyx":1644
 *         for j in range(indptr[idx], indptr[idx + 1]):
 *             coef = data[j]
 *             if coef > 0:             # <<<<<<<<<<<<<<
 *                 lout[indices[j]] += val * coef
 *     return out, mask
 */
      }
    }
    __pyx_L3_continue:;
  }

  /* "pyFAI/ext/_distortion.pyx":1646
 *             if coef > 0:
 *                 lout[indices[j]] += val * coef
 *     return out, mask             # <<<<<<<<<<<<<<
 * 
 * ###########################################################################
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_out);
  __Pyx_GIVEREF(__pyx_v_out);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_out);
  __Pyx_INCREF(__pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_mask);
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1610
 * 
 * 
 * def uncorrect_CSR(image, shape, LUT):             # <<<<<<<<<<<<<<
 *     """Take an image which has been corrected and transform it into it's raw (with loss of information)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.uncorrect_CSR", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_lmask, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_indptr, 1);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1661
 * 
 *     """
 *     def __init__(self, detector="detector", shape=None):             # <<<<<<<<<<<<<<
 *         """
 *         :param detector: detector instance or detector name
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion___init__[] = "Distortion.__init__(self, detector=u'detector', shape=None)\n\n        :param detector: detector instance or detector name\n        ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion___init__};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_detector = 0;
  PyObject *__pyx_v_shape = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_detector,&__pyx_n_s_shape,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_n_u_detector));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_detector);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1661, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_detector = values[1];
    __pyx_v_shape = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1661, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion___init__(__pyx_self, __pyx_v_self, __pyx_v_detector, __pyx_v_shape);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_detector, PyObject *__pyx_v_shape) {
  PyObject *__pyx_8genexpr1__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyFAI/ext/_distortion.pyx":1665
 *         :param detector: detector instance or detector name
 *         """
 *         if isinstance(detector, str):             # <<<<<<<<<<<<<<
 *             self.detector = detector_factory(detector)
 *         else:  # we assume it is a Detector instance
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_detector); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1666
 *         """
 *         if isinstance(detector, str):
 *             self.detector = detector_factory(detector)             # <<<<<<<<<<<<<<
 *         else:  # we assume it is a Detector instance
 *             self.detector = detector
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_detector_factory); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_detector) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_detector);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_detector, __pyx_t_3) < 0) __PYX_ERR(0, 1666, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyFAI/ext/_distortion.pyx":1665
 *         :param detector: detector instance or detector name
 *         """
 *         if isinstance(detector, str):             # <<<<<<<<<<<<<<
 *             self.detector = detector_factory(detector)
 *         else:  # we assume it is a Detector instance
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1668
 *             self.detector = detector_factory(detector)
 *         else:  # we assume it is a Detector instance
 *             self.detector = detector             # <<<<<<<<<<<<<<
 *         if shape:
 *             self.shape = shape
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_detector, __pyx_v_detector) < 0) __PYX_ERR(0, 1668, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1669
 *         else:  # we assume it is a Detector instance
 *             self.detector = detector
 *         if shape:             # <<<<<<<<<<<<<<
 *             self.shape = shape
 *         elif "max_shape" in dir(self.detector):
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_shape); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1669, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1670
 *             self.detector = detector
 *         if shape:
 *             self.shape = shape             # <<<<<<<<<<<<<<
 *         elif "max_shape" in dir(self.detector):
 *             self.shape = self.detector.max_shape
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 1670, __pyx_L1_error)

    /* "pyFAI/ext/_distortion.pyx":1669
 *         else:  # we assume it is a Detector instance
 *             self.detector = detector
 *         if shape:             # <<<<<<<<<<<<<<
 *             self.shape = shape
 *         elif "max_shape" in dir(self.detector):
 */
    goto __pyx_L4;
  }

  /* "pyFAI/ext/_distortion.pyx":1671
 *         if shape:
 *             self.shape = shape
 *         elif "max_shape" in dir(self.detector):             # <<<<<<<<<<<<<<
 *             self.shape = self.detector.max_shape
 *         self.shape = tuple([int(i) for i in self.shape])
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Dir(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_max_shape, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "pyFAI/ext/_distortion.pyx":1672
 *             self.shape = shape
 *         elif "max_shape" in dir(self.detector):
 *             self.shape = self.detector.max_shape             # <<<<<<<<<<<<<<
 *         self.shape = tuple([int(i) for i in self.shape])
 *         self._sem = threading.Semaphore()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_max_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 1672, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyFAI/ext/_distortion.pyx":1671
 *         if shape:
 *             self.shape = shape
 *         elif "max_shape" in dir(self.detector):             # <<<<<<<<<<<<<<
 *             self.shape = self.detector.max_shape
 *         self.shape = tuple([int(i) for i in self.shape])
 */
  }
  __pyx_L4:;

  /* "pyFAI/ext/_distortion.pyx":1673
 *         elif "max_shape" in dir(self.detector):
 *             self.shape = self.detector.max_shape
 *         self.shape = tuple([int(i) for i in self.shape])             # <<<<<<<<<<<<<<
 *         self._sem = threading.Semaphore()
 *         self.lut_size = None
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1673, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1673, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1673, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1673, __pyx_L7_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1673, __pyx_L7_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1673, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1673, __pyx_L7_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1673, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1673, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_i, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_8genexpr1__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1673, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1673, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i); __pyx_8genexpr1__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 1673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1674
 *             self.shape = self.detector.max_shape
 *         self.shape = tuple([int(i) for i in self.shape])
 *         self._sem = threading.Semaphore()             # <<<<<<<<<<<<<<
 *         self.lut_size = None
 *         self.pos = None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Semaphore); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_sem, __pyx_t_5) < 0) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1675
 *         self.shape = tuple([int(i) for i in self.shape])
 *         self._sem = threading.Semaphore()
 *         self.lut_size = None             # <<<<<<<<<<<<<<
 *         self.pos = None
 *         self.LUT = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_lut_size, Py_None) < 0) __PYX_ERR(0, 1675, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1676
 *         self._sem = threading.Semaphore()
 *         self.lut_size = None
 *         self.pos = None             # <<<<<<<<<<<<<<
 *         self.LUT = None
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pos, Py_None) < 0) __PYX_ERR(0, 1676, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1677
 *         self.lut_size = None
 *         self.pos = None
 *         self.LUT = None             # <<<<<<<<<<<<<<
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_LUT, Py_None) < 0) __PYX_ERR(0, 1677, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1678
 *         self.pos = None
 *         self.LUT = None
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_delta0, Py_None) < 0) __PYX_ERR(0, 1678, __pyx_L1_error)
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_delta1, Py_None) < 0) __PYX_ERR(0, 1678, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1661
 * 
 *     """
 *     def __init__(self, detector="detector", shape=None):             # <<<<<<<<<<<<<<
 *         """
 *         :param detector: detector instance or detector name
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1680
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return os.linesep.join(["Distortion correction for detector:",
 *                                 self.detector.__repr__()])
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_2__repr__[] = "Distortion.__repr__(self)";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_3__repr__ = {"__repr__", (PyCFunction)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_3__repr__, METH_O, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_2__repr__};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_2__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyFAI/ext/_distortion.pyx":1681
 * 
 *     def __repr__(self):
 *         return os.linesep.join(["Distortion correction for detector:",             # <<<<<<<<<<<<<<
 *                                 self.detector.__repr__()])
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_linesep); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_join); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1682
 *     def __repr__(self):
 *         return os.linesep.join(["Distortion correction for detector:",
 *                                 self.detector.__repr__()])             # <<<<<<<<<<<<<<
 * 
 *     def calc_pos(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_repr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1681
 * 
 *     def __repr__(self):
 *         return os.linesep.join(["Distortion correction for detector:",             # <<<<<<<<<<<<<<
 *                                 self.detector.__repr__()])
 * 
 */
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_u_Distortion_correction_for_detect);
  __Pyx_GIVEREF(__pyx_kp_u_Distortion_correction_for_detect);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Distortion_correction_for_detect);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1680
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return os.linesep.join(["Distortion correction for detector:",
 *                                 self.detector.__repr__()])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1684
 *                                 self.detector.__repr__()])
 * 
 *     def calc_pos(self):             # <<<<<<<<<<<<<<
 *         if self.pos is None:
 *             with self._sem:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_5calc_pos(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_4calc_pos[] = "Distortion.calc_pos(self)";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_5calc_pos = {"calc_pos", (PyCFunction)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_5calc_pos, METH_O, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_4calc_pos};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_5calc_pos(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_pos (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_4calc_pos(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_4calc_pos(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_pos_corners = NULL;
  PyObject *__pyx_v_d1 = NULL;
  PyObject *__pyx_v_d2 = NULL;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_pos", 0);

  /* "pyFAI/ext/_distortion.pyx":1685
 * 
 *     def calc_pos(self):
 *         if self.pos is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.pos is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1686
 *     def calc_pos(self):
 *         if self.pos is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.pos is None:
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 */
    /*with:*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sem); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1686, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1686, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1686, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1686, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "pyFAI/ext/_distortion.pyx":1687
 *         if self.pos is None:
 *             with self._sem:
 *                 if self.pos is None:             # <<<<<<<<<<<<<<
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1687, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = (__pyx_t_1 == Py_None);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_2 = (__pyx_t_3 != 0);
            if (__pyx_t_2) {

              /* "pyFAI/ext/_distortion.pyx":1688
 *             with self._sem:
 *                 if self.pos is None:
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5
 */
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
              __Pyx_INCREF(__pyx_int_2);
              __Pyx_GIVEREF(__pyx_int_2);
              PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_int_2);
              __pyx_t_1 = 0;
              __pyx_t_6 = 0;
              __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1688, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_v_pos_corners = __pyx_t_11;
              __pyx_t_11 = 0;

              /* "pyFAI/ext/_distortion.pyx":1689
 *                 if self.pos is None:
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5             # <<<<<<<<<<<<<<
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5
 *                     p = self.detector.calc_cartesian_positions(d1, d2)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_expand2d); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyFloat_AddObjC(__pyx_t_12, __pyx_float_1_0, 1.0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
                if (likely(__pyx_t_12)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                  __Pyx_INCREF(__pyx_t_12);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_1, function);
                }
              }
              __pyx_t_6 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_12, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_6, __pyx_t_1, Py_False};
                __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1689, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_6, __pyx_t_1, Py_False};
                __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1689, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              {
                __pyx_t_12 = PyTuple_New(3+__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1689, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_12);
                if (__pyx_t_5) {
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_5); __pyx_t_5 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_13, __pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_13, __pyx_t_1);
                __Pyx_INCREF(Py_False);
                __Pyx_GIVEREF(Py_False);
                PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_13, Py_False);
                __pyx_t_6 = 0;
                __pyx_t_1 = 0;
                __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_12, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1689, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              }
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyFloat_SubtractObjC(__pyx_t_11, __pyx_float_0_5, 0.5, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1689, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_v_d1 = __pyx_t_7;
              __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1690
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5             # <<<<<<<<<<<<<<
 *                     p = self.detector.calc_cartesian_positions(d1, d2)
 *                     if p[2] is not None:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_expand2d); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyFloat_AddObjC(__pyx_t_5, __pyx_float_1_0, 1.0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_12 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_11)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_12, __pyx_t_6, Py_True};
                __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1690, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_12, __pyx_t_6, Py_True};
                __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1690, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              {
                __pyx_t_5 = PyTuple_New(3+__pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1690, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_5);
                if (__pyx_t_1) {
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_12);
                PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_13, __pyx_t_12);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_13, __pyx_t_6);
                __Pyx_INCREF(Py_True);
                __Pyx_GIVEREF(Py_True);
                PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_13, Py_True);
                __pyx_t_12 = 0;
                __pyx_t_6 = 0;
                __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_5, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1690, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              }
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyFloat_SubtractObjC(__pyx_t_7, __pyx_float_0_5, 0.5, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1690, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_v_d2 = __pyx_t_11;
              __pyx_t_11 = 0;

              /* "pyFAI/ext/_distortion.pyx":1691
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5
 *                     p = self.detector.calc_cartesian_positions(d1, d2)             # <<<<<<<<<<<<<<
 *                     if p[2] is not None:
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 */
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1691, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_calc_cartesian_positions); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1691, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_d1, __pyx_v_d2};
                __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1691, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_11);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_d1, __pyx_v_d2};
                __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1691, __pyx_L8_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_11);
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1691, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (__pyx_t_7) {
                  __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
                }
                __Pyx_INCREF(__pyx_v_d1);
                __Pyx_GIVEREF(__pyx_v_d1);
                PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_13, __pyx_v_d1);
                __Pyx_INCREF(__pyx_v_d2);
                __Pyx_GIVEREF(__pyx_v_d2);
                PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_13, __pyx_v_d2);
                __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1691, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_v_p = __pyx_t_11;
              __pyx_t_11 = 0;

              /* "pyFAI/ext/_distortion.pyx":1692
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5
 *                     p = self.detector.calc_cartesian_positions(d1, d2)
 *                     if p[2] is not None:             # <<<<<<<<<<<<<<
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 *                                        self.detector)
 */
              __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_p, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1692, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_2 = (__pyx_t_11 != Py_None);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_3 = (__pyx_t_2 != 0);
              if (__pyx_t_3) {

                /* "pyFAI/ext/_distortion.pyx":1693
 *                     p = self.detector.calc_cartesian_positions(d1, d2)
 *                     if p[2] is not None:
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",             # <<<<<<<<<<<<<<
 *                                        self.detector)
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]
 */
                __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_logger); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1693, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warning); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1693, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                /* "pyFAI/ext/_distortion.pyx":1694
 *                     if p[2] is not None:
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 *                                        self.detector)             # <<<<<<<<<<<<<<
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]
 *                     pos_corners[:, :, 0] /= self.detector.pixel1
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1694, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_7 = NULL;
                __pyx_t_13 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                    __pyx_t_13 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_u_makes_little_sense_to_correct_fo, __pyx_t_5};
                  __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1693, __pyx_L8_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_u_makes_little_sense_to_correct_fo, __pyx_t_5};
                  __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1693, __pyx_L8_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                {
                  __pyx_t_12 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1693, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_12);
                  if (__pyx_t_7) {
                    __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_7); __pyx_t_7 = NULL;
                  }
                  __Pyx_INCREF(__pyx_kp_u_makes_little_sense_to_correct_fo);
                  __Pyx_GIVEREF(__pyx_kp_u_makes_little_sense_to_correct_fo);
                  PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_13, __pyx_kp_u_makes_little_sense_to_correct_fo);
                  __Pyx_GIVEREF(__pyx_t_5);
                  PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_13, __pyx_t_5);
                  __pyx_t_5 = 0;
                  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_12, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1693, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                }
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

                /* "pyFAI/ext/_distortion.pyx":1692
 *                     d2 = expand2d(numpy.arange(self.shape[1] + 1.0), self.shape[0] + 1, True) - 0.5
 *                     p = self.detector.calc_cartesian_positions(d1, d2)
 *                     if p[2] is not None:             # <<<<<<<<<<<<<<
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 *                                        self.detector)
 */
              }

              /* "pyFAI/ext/_distortion.pyx":1695
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 *                                        self.detector)
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]             # <<<<<<<<<<<<<<
 *                     pos_corners[:, :, 0] /= self.detector.pixel1
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 */
              __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_v_p, 0, 2, NULL, NULL, &__pyx_slice__25, 0, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1695, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              if ((likely(PyTuple_CheckExact(__pyx_t_11))) || (PyList_CheckExact(__pyx_t_11))) {
                PyObject* sequence = __pyx_t_11;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 1695, __pyx_L8_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_12);
                #else
                __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1695, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1695, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_12);
                #endif
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_5 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1695, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __pyx_t_14 = Py_TYPE(__pyx_t_5)->tp_iternext;
                index = 0; __pyx_t_6 = __pyx_t_14(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L16_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_6);
                index = 1; __pyx_t_12 = __pyx_t_14(__pyx_t_5); if (unlikely(!__pyx_t_12)) goto __pyx_L16_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_12);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1695, __pyx_L8_error)
                __pyx_t_14 = NULL;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                goto __pyx_L17_unpacking_done;
                __pyx_L16_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_14 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 1695, __pyx_L8_error)
                __pyx_L17_unpacking_done:;
              }
              if (unlikely(PyObject_SetItem(__pyx_v_pos_corners, __pyx_tuple__26, __pyx_t_6) < 0)) __PYX_ERR(0, 1695, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(PyObject_SetItem(__pyx_v_pos_corners, __pyx_tuple__27, __pyx_t_12) < 0)) __PYX_ERR(0, 1695, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "pyFAI/ext/_distortion.pyx":1696
 *                                        self.detector)
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]
 *                     pos_corners[:, :, 0] /= self.detector.pixel1             # <<<<<<<<<<<<<<
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 */
              __Pyx_INCREF(__pyx_tuple__26);
              __pyx_t_15 = __pyx_tuple__26;
              __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1696, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1696, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_pixel1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1696, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = __Pyx_PyNumber_InPlaceDivide(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1696, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(PyObject_SetItem(__pyx_v_pos_corners, __pyx_t_15, __pyx_t_12) < 0)) __PYX_ERR(0, 1696, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

              /* "pyFAI/ext/_distortion.pyx":1697
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]
 *                     pos_corners[:, :, 0] /= self.detector.pixel1
 *                     pos_corners[:, :, 1] /= self.detector.pixel2             # <<<<<<<<<<<<<<
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 */
              __Pyx_INCREF(__pyx_tuple__27);
              __pyx_t_15 = __pyx_tuple__27;
              __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1697, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_detector); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1697, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_pixel2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1697, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyNumber_InPlaceDivide(__pyx_t_12, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1697, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(PyObject_SetItem(__pyx_v_pos_corners, __pyx_t_15, __pyx_t_6) < 0)) __PYX_ERR(0, 1697, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

              /* "pyFAI/ext/_distortion.pyx":1698
 *                     pos_corners[:, :, 0] /= self.detector.pixel1
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 */
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
              __Pyx_INCREF(__pyx_int_4);
              __Pyx_GIVEREF(__pyx_int_4);
              PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_int_4);
              __Pyx_INCREF(__pyx_int_2);
              __Pyx_GIVEREF(__pyx_int_2);
              PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_int_2);
              __pyx_t_12 = 0;
              __pyx_t_5 = 0;
              __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
              __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1698, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_pos = __pyx_t_7;
              __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1699
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 */
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__29); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1699, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(PyObject_SetItem(__pyx_v_pos, __pyx_tuple__30, __pyx_t_7) < 0)) __PYX_ERR(0, 1699, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1700
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 */
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1700, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(PyObject_SetItem(__pyx_v_pos, __pyx_tuple__33, __pyx_t_7) < 0)) __PYX_ERR(0, 1700, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1701
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 *                     self.pos = pos
 */
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__34); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1701, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(PyObject_SetItem(__pyx_v_pos, __pyx_tuple__35, __pyx_t_7) < 0)) __PYX_ERR(0, 1701, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1702
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]             # <<<<<<<<<<<<<<
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 */
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__36); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1702, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (unlikely(PyObject_SetItem(__pyx_v_pos, __pyx_tuple__37, __pyx_t_7) < 0)) __PYX_ERR(0, 1702, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "pyFAI/ext/_distortion.pyx":1703
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 *                     self.pos = pos             # <<<<<<<<<<<<<<
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pos, __pyx_v_pos) < 0) __PYX_ERR(0, 1703, __pyx_L8_error)

              /* "pyFAI/ext/_distortion.pyx":1704
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())             # <<<<<<<<<<<<<<
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())
 *         return self.pos
 */
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ceil); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__38); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_12 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_12)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_12);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                }
              }
              __pyx_t_6 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_floor); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__39); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_12);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_12, function);
                }
              }
              __pyx_t_11 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = PyNumber_Subtract(__pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_max); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_12)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_12);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                }
              }
              __pyx_t_7 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyNumber_Int(__pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_delta0, __pyx_t_11) < 0) __PYX_ERR(0, 1704, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "pyFAI/ext/_distortion.pyx":1705
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())             # <<<<<<<<<<<<<<
 *         return self.pos
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_ceil); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__40); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_12);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_floor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_v_pos_corners, __pyx_tuple__41); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_12);
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = PyNumber_Subtract(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_11 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_delta1, __pyx_t_6) < 0) __PYX_ERR(0, 1705, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "pyFAI/ext/_distortion.pyx":1687
 *         if self.pos is None:
 *             with self._sem:
 *                 if self.pos is None:             # <<<<<<<<<<<<<<
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 *                     d1 = expand2d(numpy.arange(self.shape[0] + 1.0), self.shape[1] + 1, False) - 0.5
 */
            }

            /* "pyFAI/ext/_distortion.pyx":1686
 *     def calc_pos(self):
 *         if self.pos is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.pos is None:
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L13_try_end;
          __pyx_L8_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_11, &__pyx_t_5) < 0) __PYX_ERR(0, 1686, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_11, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1686, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1686, __pyx_L10_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (__pyx_t_3 < 0) __PYX_ERR(0, 1686, __pyx_L10_except_error)
            __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_5);
              __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_11, __pyx_t_5);
              __pyx_t_6 = 0; __pyx_t_11 = 0; __pyx_t_5 = 0; 
              __PYX_ERR(0, 1686, __pyx_L10_except_error)
            }
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L9_exception_handled;
          }
          __pyx_L10_except_error:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L9_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          __pyx_L13_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_4) {
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__42, NULL);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1686, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
      goto __pyx_L21;
      __pyx_L4_error:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L1_error;
      __pyx_L21:;
    }

    /* "pyFAI/ext/_distortion.pyx":1685
 * 
 *     def calc_pos(self):
 *         if self.pos is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.pos is None:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1706
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())
 *         return self.pos             # <<<<<<<<<<<<<<
 * 
 *     def calc_LUT_size(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1684
 *                                 self.detector.__repr__()])
 * 
 *     def calc_pos(self):             # <<<<<<<<<<<<<<
 *         if self.pos is None:
 *             with self._sem:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pos_corners);
  __Pyx_XDECREF(__pyx_v_d1);
  __Pyx_XDECREF(__pyx_v_d2);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1708
 *         return self.pos
 * 
 *     def calc_LUT_size(self):             # <<<<<<<<<<<<<<
 *         """
 *         Considering the "half-CCD" spline from ID11 which describes a (1025,2048) detector,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_7calc_LUT_size(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_6calc_LUT_size[] = "Distortion.calc_LUT_size(self)\n\n        Considering the \"half-CCD\" spline from ID11 which describes a (1025,2048) detector,\n        the physical location of pixels should go from:\n        [-17.48634 : 1027.0543, -22.768829 : 2028.3689]\n        We chose to discard pixels falling outside the [0:1025,0:2048] range with a lose of intensity\n\n        We keep self.pos: pos_corners will not be compatible with systems showing non adjacent pixels (like some xpads)\n\n        ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_7calc_LUT_size = {"calc_LUT_size", (PyCFunction)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_7calc_LUT_size, METH_O, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_6calc_LUT_size};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_7calc_LUT_size(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_LUT_size (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_6calc_LUT_size(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_6calc_LUT_size(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_l;
  int __pyx_v_shape0;
  int __pyx_v_shape1;
  __Pyx_memviewslice __pyx_v_pos0min = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_pos1min = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_pos0max = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_pos1max = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lut_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_v_np_lut_size = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  int __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  PyObject *__pyx_t_35 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_LUT_size", 0);

  /* "pyFAI/ext/_distortion.pyx":1721
 *         cdef int[:, ::1] pos0min, pos1min, pos0max, pos1max
 *         cdef int32_t[:, ::1] lut_size
 *         if self.pos is None:             # <<<<<<<<<<<<<<
 *             pos = self.calc_pos()
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1722
 *         cdef int32_t[:, ::1] lut_size
 *         if self.pos is None:
 *             pos = self.calc_pos()             # <<<<<<<<<<<<<<
 *         else:
 *             pos = self.pos
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_calc_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_pos = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyFAI/ext/_distortion.pyx":1721
 *         cdef int[:, ::1] pos0min, pos1min, pos0max, pos1max
 *         cdef int32_t[:, ::1] lut_size
 *         if self.pos is None:             # <<<<<<<<<<<<<<
 *             pos = self.calc_pos()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyFAI/ext/_distortion.pyx":1724
 *             pos = self.calc_pos()
 *         else:
 *             pos = self.pos             # <<<<<<<<<<<<<<
 *         if self.lut_size is None:
 *             with self._sem:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_pos = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "pyFAI/ext/_distortion.pyx":1725
 *         else:
 *             pos = self.pos
 *         if self.lut_size is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.lut_size is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "pyFAI/ext/_distortion.pyx":1726
 *             pos = self.pos
 *         if self.lut_size is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.lut_size is None:
 *                     shape0, shape1 = self.shape
 */
    /*with:*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sem); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1726, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1726, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "pyFAI/ext/_distortion.pyx":1727
 *         if self.lut_size is None:
 *             with self._sem:
 *                 if self.lut_size is None:             # <<<<<<<<<<<<<<
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1727, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = (__pyx_t_1 == Py_None);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_3 = (__pyx_t_2 != 0);
            if (__pyx_t_3) {

              /* "pyFAI/ext/_distortion.pyx":1728
 *             with self._sem:
 *                 if self.lut_size is None:
 *                     shape0, shape1 = self.shape             # <<<<<<<<<<<<<<
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])
 */
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1728, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
                PyObject* sequence = __pyx_t_1;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 1728, __pyx_L9_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                #else
                __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1728, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1728, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1728, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext;
                index = 0; __pyx_t_4 = __pyx_t_11(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L16_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_4);
                index = 1; __pyx_t_5 = __pyx_t_11(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L16_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_5);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_7), 2) < 0) __PYX_ERR(0, 1728, __pyx_L9_error)
                __pyx_t_11 = NULL;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                goto __pyx_L17_unpacking_done;
                __pyx_L16_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_11 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 1728, __pyx_L9_error)
                __pyx_L17_unpacking_done:;
              }
              __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_v_shape0 = __pyx_t_12;
              __pyx_v_shape1 = __pyx_t_13;

              /* "pyFAI/ext/_distortion.pyx":1729
 *                 if self.lut_size is None:
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])             # <<<<<<<<<<<<<<
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 */
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_floor); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos, __pyx_tuple__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_min); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1729, __pyx_L9_error)
              __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_7, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_16);
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_astype); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_4, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_16);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_clip); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_14)) {
                PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_0, __pyx_t_16};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1729, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
                PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_int_0, __pyx_t_16};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1729, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1729, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_5) {
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
                }
                __Pyx_INCREF(__pyx_int_0);
                __Pyx_GIVEREF(__pyx_int_0);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_13, __pyx_int_0);
                __Pyx_GIVEREF(__pyx_t_16);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_13, __pyx_t_16);
                __pyx_t_16 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1729, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1729, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_v_pos0min = __pyx_t_17;
              __pyx_t_17.memview = NULL;
              __pyx_t_17.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1730
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])             # <<<<<<<<<<<<<<
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 */
              __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_numpy); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_floor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_16 = __Pyx_PyObject_GetItem(__pyx_v_pos, __pyx_tuple__44); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_min); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1730, __pyx_L9_error)
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_16 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_4 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15);
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_astype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_14 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_clip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_14, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_15};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1730, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_15};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1730, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1730, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_14) {
                  __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14); __pyx_t_14 = NULL;
                }
                __Pyx_INCREF(__pyx_int_0);
                __Pyx_GIVEREF(__pyx_int_0);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_13, __pyx_int_0);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_13, __pyx_t_15);
                __pyx_t_15 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1730, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1730, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_v_pos1min = __pyx_t_17;
              __pyx_t_17.memview = NULL;
              __pyx_t_17.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1731
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])             # <<<<<<<<<<<<<<
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_numpy); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_ceil); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_v_pos, __pyx_tuple__43); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_max); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1731, __pyx_L9_error)
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_empty_tuple, __pyx_t_15); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_15)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_15);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_4 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_15, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_7);
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_astype); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_7);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_clip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1731, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1731, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1731, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_14) {
                  __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14); __pyx_t_14 = NULL;
                }
                __Pyx_INCREF(__pyx_int_0);
                __Pyx_GIVEREF(__pyx_int_0);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_13, __pyx_int_0);
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_13, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1731, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1731, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_v_pos0max = __pyx_t_17;
              __pyx_t_17.memview = NULL;
              __pyx_t_17.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1732
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])             # <<<<<<<<<<<<<<
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)
 *                     with nogil:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ceil); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pos, __pyx_tuple__44); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_max); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1732, __pyx_L9_error)
              __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_7, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_16);
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_astype); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_4, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_16);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_clip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_14, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = NULL;
              __pyx_t_13 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                  __pyx_t_13 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_16};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_int_0, __pyx_t_16};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L9_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1732, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_14) {
                  __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14); __pyx_t_14 = NULL;
                }
                __Pyx_INCREF(__pyx_int_0);
                __Pyx_GIVEREF(__pyx_int_0);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_13, __pyx_int_0);
                __Pyx_GIVEREF(__pyx_t_16);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_13, __pyx_t_16);
                __pyx_t_16 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1732, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_v_pos1max = __pyx_t_17;
              __pyx_t_17.memview = NULL;
              __pyx_t_17.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1733
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *                     with nogil:
 *                         for i in range(shape0):
 */
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_numpy); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_int32); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_14) < 0) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1733, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_v_lut_size = __pyx_t_18;
              __pyx_t_18.memview = NULL;
              __pyx_t_18.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1734
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         for i in range(shape0):
 *                             for j in range(shape1):
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {

                    /* "pyFAI/ext/_distortion.pyx":1735
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)
 *                     with nogil:
 *                         for i in range(shape0):             # <<<<<<<<<<<<<<
 *                             for j in range(shape1):
 *                                 for k in range(pos0min[i, j], pos0max[i, j]):
 */
                    __pyx_t_13 = __pyx_v_shape0;
                    __pyx_t_12 = __pyx_t_13;
                    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_12; __pyx_t_19+=1) {
                      __pyx_v_i = __pyx_t_19;

                      /* "pyFAI/ext/_distortion.pyx":1736
 *                     with nogil:
 *                         for i in range(shape0):
 *                             for j in range(shape1):             # <<<<<<<<<<<<<<
 *                                 for k in range(pos0min[i, j], pos0max[i, j]):
 *                                     for l in range(pos1min[i, j], pos1max[i, j]):
 */
                      __pyx_t_20 = __pyx_v_shape1;
                      __pyx_t_21 = __pyx_t_20;
                      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                        __pyx_v_j = __pyx_t_22;

                        /* "pyFAI/ext/_distortion.pyx":1737
 *                         for i in range(shape0):
 *                             for j in range(shape1):
 *                                 for k in range(pos0min[i, j], pos0max[i, j]):             # <<<<<<<<<<<<<<
 *                                     for l in range(pos1min[i, j], pos1max[i, j]):
 *                                         lut_size[k, l] += 1
 */
                        __pyx_t_23 = __pyx_v_i;
                        __pyx_t_24 = __pyx_v_j;
                        __pyx_t_25 = (*((int *) ( /* dim=1 */ ((char *) (((int *) ( /* dim=0 */ (__pyx_v_pos0max.data + __pyx_t_23 * __pyx_v_pos0max.strides[0]) )) + __pyx_t_24)) )));
                        __pyx_t_24 = __pyx_v_i;
                        __pyx_t_23 = __pyx_v_j;
                        __pyx_t_26 = __pyx_t_25;
                        for (__pyx_t_27 = (*((int *) ( /* dim=1 */ ((char *) (((int *) ( /* dim=0 */ (__pyx_v_pos0min.data + __pyx_t_24 * __pyx_v_pos0min.strides[0]) )) + __pyx_t_23)) ))); __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
                          __pyx_v_k = __pyx_t_27;

                          /* "pyFAI/ext/_distortion.pyx":1738
 *                             for j in range(shape1):
 *                                 for k in range(pos0min[i, j], pos0max[i, j]):
 *                                     for l in range(pos1min[i, j], pos1max[i, j]):             # <<<<<<<<<<<<<<
 *                                         lut_size[k, l] += 1
 *                     np_lut_size = numpy.asarray(lut_size)
 */
                          __pyx_t_28 = __pyx_v_i;
                          __pyx_t_29 = __pyx_v_j;
                          __pyx_t_30 = (*((int *) ( /* dim=1 */ ((char *) (((int *) ( /* dim=0 */ (__pyx_v_pos1max.data + __pyx_t_28 * __pyx_v_pos1max.strides[0]) )) + __pyx_t_29)) )));
                          __pyx_t_29 = __pyx_v_i;
                          __pyx_t_28 = __pyx_v_j;
                          __pyx_t_31 = __pyx_t_30;
                          for (__pyx_t_32 = (*((int *) ( /* dim=1 */ ((char *) (((int *) ( /* dim=0 */ (__pyx_v_pos1min.data + __pyx_t_29 * __pyx_v_pos1min.strides[0]) )) + __pyx_t_28)) ))); __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
                            __pyx_v_l = __pyx_t_32;

                            /* "pyFAI/ext/_distortion.pyx":1739
 *                                 for k in range(pos0min[i, j], pos0max[i, j]):
 *                                     for l in range(pos1min[i, j], pos1max[i, j]):
 *                                         lut_size[k, l] += 1             # <<<<<<<<<<<<<<
 *                     np_lut_size = numpy.asarray(lut_size)
 *                     self.lut_size = np_lut_size.max()
 */
                            __pyx_t_33 = __pyx_v_k;
                            __pyx_t_34 = __pyx_v_l;
                            *((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_lut_size.data + __pyx_t_33 * __pyx_v_lut_size.strides[0]) )) + __pyx_t_34)) )) += 1;
                          }
                        }
                      }
                    }
                  }

                  /* "pyFAI/ext/_distortion.pyx":1734
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 *                     lut_size = numpy.zeros(self.shape, dtype=numpy.int32)
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         for i in range(shape0):
 *                             for j in range(shape1):
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L20;
                    }
                    __pyx_L20:;
                  }
              }

              /* "pyFAI/ext/_distortion.pyx":1740
 *                                     for l in range(pos1min[i, j], pos1max[i, j]):
 *                                         lut_size[k, l] += 1
 *                     np_lut_size = numpy.asarray(lut_size)             # <<<<<<<<<<<<<<
 *                     self.lut_size = np_lut_size.max()
 *                     return np_lut_size
 */
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1740, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1740, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_lut_size, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_int32_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1740, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              __pyx_t_14 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1740, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_v_np_lut_size = __pyx_t_14;
              __pyx_t_14 = 0;

              /* "pyFAI/ext/_distortion.pyx":1741
 *                                         lut_size[k, l] += 1
 *                     np_lut_size = numpy.asarray(lut_size)
 *                     self.lut_size = np_lut_size.max()             # <<<<<<<<<<<<<<
 *                     return np_lut_size
 * 
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_np_lut_size, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1741, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              __pyx_t_14 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1741, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_lut_size, __pyx_t_14) < 0) __PYX_ERR(0, 1741, __pyx_L9_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "pyFAI/ext/_distortion.pyx":1742
 *                     np_lut_size = numpy.asarray(lut_size)
 *                     self.lut_size = np_lut_size.max()
 *                     return np_lut_size             # <<<<<<<<<<<<<<
 * 
 *     def calc_LUT(self):
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_np_lut_size);
              __pyx_r = __pyx_v_np_lut_size;
              goto __pyx_L13_try_return;

              /* "pyFAI/ext/_distortion.pyx":1727
 *         if self.lut_size is None:
 *             with self._sem:
 *                 if self.lut_size is None:             # <<<<<<<<<<<<<<
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 */
            }

            /* "pyFAI/ext/_distortion.pyx":1726
 *             pos = self.pos
 *         if self.lut_size is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.lut_size is None:
 *                     shape0, shape1 = self.shape
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L14_try_end;
          __pyx_L9_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
          __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_LUT_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 1726, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = PyTuple_Pack(3, __pyx_t_14, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1726, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_35 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1726, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_35);
            __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_35);
            __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
            if (__pyx_t_3 < 0) __PYX_ERR(0, 1726, __pyx_L11_except_error)
            __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_4);
              __Pyx_XGIVEREF(__pyx_t_1);
              __Pyx_ErrRestoreWithState(__pyx_t_14, __pyx_t_4, __pyx_t_1);
              __pyx_t_14 = 0; __pyx_t_4 = 0; __pyx_t_1 = 0; 
              __PYX_ERR(0, 1726, __pyx_L11_except_error)
            }
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L10_exception_handled;
          }
          __pyx_L11_except_error:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L13_try_return:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L6_return;
          __pyx_L10_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          __pyx_L14_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_6) {
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__42, NULL);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1726, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          goto __pyx_L8;
        }
        __pyx_L6_return: {
          __pyx_t_10 = __pyx_r;
          __pyx_r = 0;
          if (__pyx_t_6) {
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__42, NULL);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1726, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __pyx_r = __pyx_t_10;
          __pyx_t_10 = 0;
          goto __pyx_L0;
        }
        __pyx_L8:;
      }
      goto __pyx_L32;
      __pyx_L5_error:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L1_error;
      __pyx_L32:;
    }

    /* "pyFAI/ext/_distortion.pyx":1725
 *         else:
 *             pos = self.pos
 *         if self.lut_size is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.lut_size is None:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1708
 *         return self.pos
 * 
 *     def calc_LUT_size(self):             # <<<<<<<<<<<<<<
 *         """
 *         Considering the "half-CCD" spline from ID11 which describes a (1025,2048) detector,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_LUT_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos0min, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos1min, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos0max, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos1max, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut_size, 1);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_np_lut_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1744
 *                     return np_lut_size
 * 
 *     def calc_LUT(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, j, ms, ml, ns, nl, shape0, shape1, size
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_9calc_LUT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_8calc_LUT[] = "Distortion.calc_LUT(self)";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_9calc_LUT = {"calc_LUT", (PyCFunction)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_9calc_LUT, METH_O, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_8calc_LUT};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_9calc_LUT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_LUT (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_8calc_LUT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_8calc_LUT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_ms;
  int __pyx_v_ml;
  int __pyx_v_ns;
  int __pyx_v_nl;
  int __pyx_v_shape0;
  int __pyx_v_shape1;
  int __pyx_v_size;
  int __pyx_v_offset0;
  int __pyx_v_offset1;
  int __pyx_v_box_size0;
  int __pyx_v_box_size1;
  int32_t __pyx_v_k;
  int32_t __pyx_v_idx;
  float __pyx_v_A0;
  float __pyx_v_A1;
  float __pyx_v_B0;
  float __pyx_v_B1;
  float __pyx_v_C0;
  float __pyx_v_C1;
  float __pyx_v_D0;
  float __pyx_v_D1;
  float __pyx_v_pAB;
  float __pyx_v_pBC;
  float __pyx_v_pCD;
  float __pyx_v_pDA;
  float __pyx_v_cAB;
  float __pyx_v_cBC;
  float __pyx_v_cCD;
  float __pyx_v_cDA;
  float __pyx_v_area;
  float __pyx_v_value;
  __Pyx_memviewslice __pyx_v_pos = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lut = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_outMax = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_buffer = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_17 = NULL;
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  __Pyx_memviewslice __pyx_t_23 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_24;
  int __pyx_t_25;
  int __pyx_t_26;
  int __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  float __pyx_t_29;
  float __pyx_t_30;
  float __pyx_t_31;
  float __pyx_t_32;
  float __pyx_t_33;
  int __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  PyObject *__pyx_t_40 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_LUT", 0);

  /* "pyFAI/ext/_distortion.pyx":1748
 *             int i, j, ms, ml, ns, nl, shape0, shape1, size
 *             int offset0, offset1, box_size0, box_size1
 *             int32_t k, idx = 0             # <<<<<<<<<<<<<<
 *             float A0, A1, B0, B1, C0, C1, D0, D1, pAB, pBC, pCD, pDA, cAB, cBC, cCD, cDA, area, value
 *             float32_t[:, :, :, ::1] pos
 */
  __pyx_v_idx = 0;

  /* "pyFAI/ext/_distortion.pyx":1752
 *             float32_t[:, :, :, ::1] pos
 *             lut_t[:, :, ::1] lut
 *             int32_t[:, ::1] outMax = numpy.zeros(self.shape, dtype=numpy.int32)             # <<<<<<<<<<<<<<
 *             float32_t[:, ::1] buffer
 *         shape0, shape1 = self.shape
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_outMax = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1754
 *             int32_t[:, ::1] outMax = numpy.zeros(self.shape, dtype=numpy.int32)
 *             float32_t[:, ::1] buffer
 *         shape0, shape1 = self.shape             # <<<<<<<<<<<<<<
 * 
 *         if self.lut_size is None:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
    PyObject* sequence = __pyx_t_5;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1754, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_2), 2) < 0) __PYX_ERR(0, 1754, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1754, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_shape0 = __pyx_t_8;
  __pyx_v_shape1 = __pyx_t_9;

  /* "pyFAI/ext/_distortion.pyx":1756
 *         shape0, shape1 = self.shape
 * 
 *         if self.lut_size is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT_size()
 *         if self.LUT is None:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = (__pyx_t_5 == Py_None);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = (__pyx_t_10 != 0);
  if (__pyx_t_11) {

    /* "pyFAI/ext/_distortion.pyx":1757
 * 
 *         if self.lut_size is None:
 *             self.calc_LUT_size()             # <<<<<<<<<<<<<<
 *         if self.LUT is None:
 *             with self._sem:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_calc_LUT_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyFAI/ext/_distortion.pyx":1756
 *         shape0, shape1 = self.shape
 * 
 *         if self.lut_size is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT_size()
 *         if self.LUT is None:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1758
 *         if self.lut_size is None:
 *             self.calc_LUT_size()
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.LUT is None:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = (__pyx_t_5 == Py_None);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = (__pyx_t_11 != 0);
  if (__pyx_t_10) {

    /* "pyFAI/ext/_distortion.pyx":1759
 *             self.calc_LUT_size()
 *         if self.LUT is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.LUT is None:
 *                     pos = self.pos
 */
    /*with:*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sem); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1759, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1759, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          /*try:*/ {

            /* "pyFAI/ext/_distortion.pyx":1760
 *         if self.LUT is None:
 *             with self._sem:
 *                 if self.LUT is None:             # <<<<<<<<<<<<<<
 *                     pos = self.pos
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1760, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_10 = (__pyx_t_5 == Py_None);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_11 = (__pyx_t_10 != 0);
            if (__pyx_t_11) {

              /* "pyFAI/ext/_distortion.pyx":1761
 *             with self._sem:
 *                 if self.LUT is None:
 *                     pos = self.pos             # <<<<<<<<<<<<<<
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)
 *                     size = self.shape[0] * self.shape[1] * self.lut_size * sizeof(lut_t)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1761, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 1761, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_v_pos = __pyx_t_16;
              __pyx_t_16.memview = NULL;
              __pyx_t_16.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1762
 *                 if self.LUT is None:
 *                     pos = self.pos
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)             # <<<<<<<<<<<<<<
 *                     size = self.shape[0] * self.shape[1] * self.lut_size * sizeof(lut_t)
 *                     memset(&lut[0, 0, 0], 0, size)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_17 = PyTuple_New(3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_t_1);
              __pyx_t_2 = 0;
              __pyx_t_4 = 0;
              __pyx_t_1 = 0;
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_shape, __pyx_t_17) < 0) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_lut_d); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_17) < 0) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(__pyx_t_17, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1762, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_v_lut = __pyx_t_18;
              __pyx_t_18.memview = NULL;
              __pyx_t_18.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1763
 *                     pos = self.pos
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)
 *                     size = self.shape[0] * self.shape[1] * self.lut_size * sizeof(lut_t)             # <<<<<<<<<<<<<<
 *                     memset(&lut[0, 0, 0], 0, size)
 *                     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], size / 1.0e6))
 */
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = PyNumber_Multiply(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = PyNumber_Multiply(__pyx_t_17, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyInt_FromSize_t((sizeof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_17 = PyNumber_Multiply(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_17); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1763, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_v_size = __pyx_t_9;

              /* "pyFAI/ext/_distortion.pyx":1764
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)
 *                     size = self.shape[0] * self.shape[1] * self.lut_size * sizeof(lut_t)
 *                     memset(&lut[0, 0, 0], 0, size)             # <<<<<<<<<<<<<<
 *                     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], size / 1.0e6))
 *                     buffer = numpy.empty((self.delta0, self.delta1), dtype=numpy.float32)
 */
              __pyx_t_19 = 0;
              __pyx_t_20 = 0;
              __pyx_t_21 = 0;
              (void)(memset((&(*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_19 * __pyx_v_lut.strides[0]) ) + __pyx_t_20 * __pyx_v_lut.strides[1]) )) + __pyx_t_21)) )))), 0, __pyx_v_size));

              /* "pyFAI/ext/_distortion.pyx":1765
 *                     size = self.shape[0] * self.shape[1] * self.lut_size * sizeof(lut_t)
 *                     memset(&lut[0, 0, 0], 0, size)
 *                     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], size / 1.0e6))             # <<<<<<<<<<<<<<
 *                     buffer = numpy.empty((self.delta0, self.delta1), dtype=numpy.float32)
 *                     #buffer_size = self.delta0 * self.delta1 * sizeof(float)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_lut.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_lut.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_lut.shape[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_2 = PyFloat_FromDouble((((double)__pyx_v_size) / 1.0e6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_22 = PyTuple_New(4); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_22, 2, __pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_22, 3, __pyx_t_2);
              __pyx_t_3 = 0;
              __pyx_t_1 = 0;
              __pyx_t_4 = 0;
              __pyx_t_2 = 0;
              __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_LUT_shape_i_i_i_3f_MByte, __pyx_t_22); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              __pyx_t_22 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_22)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_22);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_17 = (__pyx_t_22) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_22, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
              __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1765, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

              /* "pyFAI/ext/_distortion.pyx":1766
 *                     memset(&lut[0, 0, 0], 0, size)
 *                     logger.info("LUT shape: (%i,%i,%i) %.3f MByte" % (lut.shape[0], lut.shape[1], lut.shape[2], size / 1.0e6))
 *                     buffer = numpy.empty((self.delta0, self.delta1), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *                     #buffer_size = self.delta0 * self.delta1 * sizeof(float)
 *                     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], self.lut_size))
 */
              __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_numpy); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_delta0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_delta1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_17);
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2);
              __pyx_t_17 = 0;
              __pyx_t_2 = 0;
              __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_22);
              PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_22);
              __pyx_t_22 = 0;
              __pyx_t_22 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_numpy); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_float32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              __pyx_t_23 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_23.memview)) __PYX_ERR(0, 1766, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_v_buffer = __pyx_t_23;
              __pyx_t_23.memview = NULL;
              __pyx_t_23.data = NULL;

              /* "pyFAI/ext/_distortion.pyx":1768
 *                     buffer = numpy.empty((self.delta0, self.delta1), dtype=numpy.float32)
 *                     #buffer_size = self.delta0 * self.delta1 * sizeof(float)
 *                     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], self.lut_size))             # <<<<<<<<<<<<<<
 *                     with nogil:
 *                         # i,j, idx are indexes of the raw image uncorrected
 */
              __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_logger); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_n_s_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              __pyx_t_22 = PyInt_FromSsize_t((__pyx_v_buffer.shape[1])); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_buffer.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_22);
              PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_22);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_17);
              __pyx_t_22 = 0;
              __pyx_t_5 = 0;
              __pyx_t_17 = 0;
              __pyx_t_17 = PyUnicode_Format(__pyx_kp_u_Max_pixel_size_ix_i_Max_source_p, __pyx_t_1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                }
              }
              __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_17);
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1768, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "pyFAI/ext/_distortion.pyx":1769
 *                     #buffer_size = self.delta0 * self.delta1 * sizeof(float)
 *                     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], self.lut_size))
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         # i,j, idx are indexes of the raw image uncorrected
 *                         for i in range(shape0):
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {

                    /* "pyFAI/ext/_distortion.pyx":1771
 *                     with nogil:
 *                         # i,j, idx are indexes of the raw image uncorrected
 *                         for i in range(shape0):             # <<<<<<<<<<<<<<
 *                             for j in range(shape1):
 *                                 # reinit of buffer
 */
                    __pyx_t_9 = __pyx_v_shape0;
                    __pyx_t_8 = __pyx_t_9;
                    for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_8; __pyx_t_24+=1) {
                      __pyx_v_i = __pyx_t_24;

                      /* "pyFAI/ext/_distortion.pyx":1772
 *                         # i,j, idx are indexes of the raw image uncorrected
 *                         for i in range(shape0):
 *                             for j in range(shape1):             # <<<<<<<<<<<<<<
 *                                 # reinit of buffer
 *                                 buffer[:, :] = 0
 */
                      __pyx_t_25 = __pyx_v_shape1;
                      __pyx_t_26 = __pyx_t_25;
                      for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
                        __pyx_v_j = __pyx_t_27;

                        /* "pyFAI/ext/_distortion.pyx":1774
 *                             for j in range(shape1):
 *                                 # reinit of buffer
 *                                 buffer[:, :] = 0             # <<<<<<<<<<<<<<
 *                                 A0 = pos[i, j, 0, 0]
 *                                 A1 = pos[i, j, 0, 1]
 */
                        {
                            __pyx_t_5pyFAI_3ext_12shared_types_float32_t __pyx_temp_scalar = 0.0;
                            {
                                Py_ssize_t __pyx_temp_extent = __pyx_v_buffer.shape[0] * __pyx_v_buffer.shape[1];
                                Py_ssize_t __pyx_temp_idx;
                                __pyx_t_5pyFAI_3ext_12shared_types_float32_t *__pyx_temp_pointer = (__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_buffer.data;
                                for (__pyx_temp_idx = 0; __pyx_temp_idx < __pyx_temp_extent; __pyx_temp_idx++) {
                                  *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_temp_pointer) = __pyx_temp_scalar;
                                  __pyx_temp_pointer += 1;
                                }
                            }
                        }

                        /* "pyFAI/ext/_distortion.pyx":1775
 *                                 # reinit of buffer
 *                                 buffer[:, :] = 0
 *                                 A0 = pos[i, j, 0, 0]             # <<<<<<<<<<<<<<
 *                                 A1 = pos[i, j, 0, 1]
 *                                 B0 = pos[i, j, 1, 0]
 */
                        __pyx_t_21 = __pyx_v_i;
                        __pyx_t_20 = __pyx_v_j;
                        __pyx_t_19 = 0;
                        __pyx_t_28 = 0;
                        __pyx_v_A0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_21 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_19 * __pyx_v_pos.strides[2]) )) + __pyx_t_28)) )));

                        /* "pyFAI/ext/_distortion.pyx":1776
 *                                 buffer[:, :] = 0
 *                                 A0 = pos[i, j, 0, 0]
 *                                 A1 = pos[i, j, 0, 1]             # <<<<<<<<<<<<<<
 *                                 B0 = pos[i, j, 1, 0]
 *                                 B1 = pos[i, j, 1, 1]
 */
                        __pyx_t_28 = __pyx_v_i;
                        __pyx_t_19 = __pyx_v_j;
                        __pyx_t_20 = 0;
                        __pyx_t_21 = 1;
                        __pyx_v_A1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_28 * __pyx_v_pos.strides[0]) ) + __pyx_t_19 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_21)) )));

                        /* "pyFAI/ext/_distortion.pyx":1777
 *                                 A0 = pos[i, j, 0, 0]
 *                                 A1 = pos[i, j, 0, 1]
 *                                 B0 = pos[i, j, 1, 0]             # <<<<<<<<<<<<<<
 *                                 B1 = pos[i, j, 1, 1]
 *                                 C0 = pos[i, j, 2, 0]
 */
                        __pyx_t_21 = __pyx_v_i;
                        __pyx_t_20 = __pyx_v_j;
                        __pyx_t_19 = 1;
                        __pyx_t_28 = 0;
                        __pyx_v_B0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_21 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_19 * __pyx_v_pos.strides[2]) )) + __pyx_t_28)) )));

                        /* "pyFAI/ext/_distortion.pyx":1778
 *                                 A1 = pos[i, j, 0, 1]
 *                                 B0 = pos[i, j, 1, 0]
 *                                 B1 = pos[i, j, 1, 1]             # <<<<<<<<<<<<<<
 *                                 C0 = pos[i, j, 2, 0]
 *                                 C1 = pos[i, j, 2, 1]
 */
                        __pyx_t_28 = __pyx_v_i;
                        __pyx_t_19 = __pyx_v_j;
                        __pyx_t_20 = 1;
                        __pyx_t_21 = 1;
                        __pyx_v_B1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_28 * __pyx_v_pos.strides[0]) ) + __pyx_t_19 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_21)) )));

                        /* "pyFAI/ext/_distortion.pyx":1779
 *                                 B0 = pos[i, j, 1, 0]
 *                                 B1 = pos[i, j, 1, 1]
 *                                 C0 = pos[i, j, 2, 0]             # <<<<<<<<<<<<<<
 *                                 C1 = pos[i, j, 2, 1]
 *                                 D0 = pos[i, j, 3, 0]
 */
                        __pyx_t_21 = __pyx_v_i;
                        __pyx_t_20 = __pyx_v_j;
                        __pyx_t_19 = 2;
                        __pyx_t_28 = 0;
                        __pyx_v_C0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_21 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_19 * __pyx_v_pos.strides[2]) )) + __pyx_t_28)) )));

                        /* "pyFAI/ext/_distortion.pyx":1780
 *                                 B1 = pos[i, j, 1, 1]
 *                                 C0 = pos[i, j, 2, 0]
 *                                 C1 = pos[i, j, 2, 1]             # <<<<<<<<<<<<<<
 *                                 D0 = pos[i, j, 3, 0]
 *                                 D1 = pos[i, j, 3, 1]
 */
                        __pyx_t_28 = __pyx_v_i;
                        __pyx_t_19 = __pyx_v_j;
                        __pyx_t_20 = 2;
                        __pyx_t_21 = 1;
                        __pyx_v_C1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_28 * __pyx_v_pos.strides[0]) ) + __pyx_t_19 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_21)) )));

                        /* "pyFAI/ext/_distortion.pyx":1781
 *                                 C0 = pos[i, j, 2, 0]
 *                                 C1 = pos[i, j, 2, 1]
 *                                 D0 = pos[i, j, 3, 0]             # <<<<<<<<<<<<<<
 *                                 D1 = pos[i, j, 3, 1]
 *                                 offset0 = (<int> floor(min(A0, B0, C0, D0)))
 */
                        __pyx_t_21 = __pyx_v_i;
                        __pyx_t_20 = __pyx_v_j;
                        __pyx_t_19 = 3;
                        __pyx_t_28 = 0;
                        __pyx_v_D0 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_21 * __pyx_v_pos.strides[0]) ) + __pyx_t_20 * __pyx_v_pos.strides[1]) ) + __pyx_t_19 * __pyx_v_pos.strides[2]) )) + __pyx_t_28)) )));

                        /* "pyFAI/ext/_distortion.pyx":1782
 *                                 C1 = pos[i, j, 2, 1]
 *                                 D0 = pos[i, j, 3, 0]
 *                                 D1 = pos[i, j, 3, 1]             # <<<<<<<<<<<<<<
 *                                 offset0 = (<int> floor(min(A0, B0, C0, D0)))
 *                                 offset1 = (<int> floor(min(A1, B1, C1, D1)))
 */
                        __pyx_t_28 = __pyx_v_i;
                        __pyx_t_19 = __pyx_v_j;
                        __pyx_t_20 = 3;
                        __pyx_t_21 = 1;
                        __pyx_v_D1 = (*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=3 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_pos.data + __pyx_t_28 * __pyx_v_pos.strides[0]) ) + __pyx_t_19 * __pyx_v_pos.strides[1]) ) + __pyx_t_20 * __pyx_v_pos.strides[2]) )) + __pyx_t_21)) )));

                        /* "pyFAI/ext/_distortion.pyx":1783
 *                                 D0 = pos[i, j, 3, 0]
 *                                 D1 = pos[i, j, 3, 1]
 *                                 offset0 = (<int> floor(min(A0, B0, C0, D0)))             # <<<<<<<<<<<<<<
 *                                 offset1 = (<int> floor(min(A1, B1, C1, D1)))
 *                                 box_size0 = (<int> ceil(max(A0, B0, C0, D0))) - offset0
 */
                        __pyx_t_29 = __pyx_v_B0;
                        __pyx_t_30 = __pyx_v_C0;
                        __pyx_t_31 = __pyx_v_D0;
                        __pyx_t_32 = __pyx_v_A0;
                        if (((__pyx_t_29 < __pyx_t_32) != 0)) {
                          __pyx_t_33 = __pyx_t_29;
                        } else {
                          __pyx_t_33 = __pyx_t_32;
                        }
                        __pyx_t_32 = __pyx_t_33;
                        if (((__pyx_t_30 < __pyx_t_32) != 0)) {
                          __pyx_t_33 = __pyx_t_30;
                        } else {
                          __pyx_t_33 = __pyx_t_32;
                        }
                        __pyx_t_32 = __pyx_t_33;
                        if (((__pyx_t_31 < __pyx_t_32) != 0)) {
                          __pyx_t_33 = __pyx_t_31;
                        } else {
                          __pyx_t_33 = __pyx_t_32;
                        }
                        __pyx_v_offset0 = ((int)floor(__pyx_t_33));

                        /* "pyFAI/ext/_distortion.pyx":1784
 *                                 D1 = pos[i, j, 3, 1]
 *                                 offset0 = (<int> floor(min(A0, B0, C0, D0)))
 *                                 offset1 = (<int> floor(min(A1, B1, C1, D1)))             # <<<<<<<<<<<<<<
 *                                 box_size0 = (<int> ceil(max(A0, B0, C0, D0))) - offset0
 *                                 box_size1 = (<int> ceil(max(A1, B1, C1, D1))) - offset1
 */
                        __pyx_t_33 = __pyx_v_B1;
                        __pyx_t_29 = __pyx_v_C1;
                        __pyx_t_30 = __pyx_v_D1;
                        __pyx_t_31 = __pyx_v_A1;
                        if (((__pyx_t_33 < __pyx_t_31) != 0)) {
                          __pyx_t_32 = __pyx_t_33;
                        } else {
                          __pyx_t_32 = __pyx_t_31;
                        }
                        __pyx_t_31 = __pyx_t_32;
                        if (((__pyx_t_29 < __pyx_t_31) != 0)) {
                          __pyx_t_32 = __pyx_t_29;
                        } else {
                          __pyx_t_32 = __pyx_t_31;
                        }
                        __pyx_t_31 = __pyx_t_32;
                        if (((__pyx_t_30 < __pyx_t_31) != 0)) {
                          __pyx_t_32 = __pyx_t_30;
                        } else {
                          __pyx_t_32 = __pyx_t_31;
                        }
                        __pyx_v_offset1 = ((int)floor(__pyx_t_32));

                        /* "pyFAI/ext/_distortion.pyx":1785
 *                                 offset0 = (<int> floor(min(A0, B0, C0, D0)))
 *                                 offset1 = (<int> floor(min(A1, B1, C1, D1)))
 *                                 box_size0 = (<int> ceil(max(A0, B0, C0, D0))) - offset0             # <<<<<<<<<<<<<<
 *                                 box_size1 = (<int> ceil(max(A1, B1, C1, D1))) - offset1
 *                                 A0 -= <float> offset0
 */
                        __pyx_t_32 = __pyx_v_B0;
                        __pyx_t_33 = __pyx_v_C0;
                        __pyx_t_29 = __pyx_v_D0;
                        __pyx_t_30 = __pyx_v_A0;
                        if (((__pyx_t_32 > __pyx_t_30) != 0)) {
                          __pyx_t_31 = __pyx_t_32;
                        } else {
                          __pyx_t_31 = __pyx_t_30;
                        }
                        __pyx_t_30 = __pyx_t_31;
                        if (((__pyx_t_33 > __pyx_t_30) != 0)) {
                          __pyx_t_31 = __pyx_t_33;
                        } else {
                          __pyx_t_31 = __pyx_t_30;
                        }
                        __pyx_t_30 = __pyx_t_31;
                        if (((__pyx_t_29 > __pyx_t_30) != 0)) {
                          __pyx_t_31 = __pyx_t_29;
                        } else {
                          __pyx_t_31 = __pyx_t_30;
                        }
                        __pyx_v_box_size0 = (((int)ceil(__pyx_t_31)) - __pyx_v_offset0);

                        /* "pyFAI/ext/_distortion.pyx":1786
 *                                 offset1 = (<int> floor(min(A1, B1, C1, D1)))
 *                                 box_size0 = (<int> ceil(max(A0, B0, C0, D0))) - offset0
 *                                 box_size1 = (<int> ceil(max(A1, B1, C1, D1))) - offset1             # <<<<<<<<<<<<<<
 *                                 A0 -= <float> offset0
 *                                 A1 -= <float> offset1
 */
                        __pyx_t_31 = __pyx_v_B1;
                        __pyx_t_32 = __pyx_v_C1;
                        __pyx_t_33 = __pyx_v_D1;
                        __pyx_t_29 = __pyx_v_A1;
                        if (((__pyx_t_31 > __pyx_t_29) != 0)) {
                          __pyx_t_30 = __pyx_t_31;
                        } else {
                          __pyx_t_30 = __pyx_t_29;
                        }
                        __pyx_t_29 = __pyx_t_30;
                        if (((__pyx_t_32 > __pyx_t_29) != 0)) {
                          __pyx_t_30 = __pyx_t_32;
                        } else {
                          __pyx_t_30 = __pyx_t_29;
                        }
                        __pyx_t_29 = __pyx_t_30;
                        if (((__pyx_t_33 > __pyx_t_29) != 0)) {
                          __pyx_t_30 = __pyx_t_33;
                        } else {
                          __pyx_t_30 = __pyx_t_29;
                        }
                        __pyx_v_box_size1 = (((int)ceil(__pyx_t_30)) - __pyx_v_offset1);

                        /* "pyFAI/ext/_distortion.pyx":1787
 *                                 box_size0 = (<int> ceil(max(A0, B0, C0, D0))) - offset0
 *                                 box_size1 = (<int> ceil(max(A1, B1, C1, D1))) - offset1
 *                                 A0 -= <float> offset0             # <<<<<<<<<<<<<<
 *                                 A1 -= <float> offset1
 *                                 B0 -= <float> offset0
 */
                        __pyx_v_A0 = (__pyx_v_A0 - ((float)__pyx_v_offset0));

                        /* "pyFAI/ext/_distortion.pyx":1788
 *                                 box_size1 = (<int> ceil(max(A1, B1, C1, D1))) - offset1
 *                                 A0 -= <float> offset0
 *                                 A1 -= <float> offset1             # <<<<<<<<<<<<<<
 *                                 B0 -= <float> offset0
 *                                 B1 -= <float> offset1
 */
                        __pyx_v_A1 = (__pyx_v_A1 - ((float)__pyx_v_offset1));

                        /* "pyFAI/ext/_distortion.pyx":1789
 *                                 A0 -= <float> offset0
 *                                 A1 -= <float> offset1
 *                                 B0 -= <float> offset0             # <<<<<<<<<<<<<<
 *                                 B1 -= <float> offset1
 *                                 C0 -= <float> offset0
 */
                        __pyx_v_B0 = (__pyx_v_B0 - ((float)__pyx_v_offset0));

                        /* "pyFAI/ext/_distortion.pyx":1790
 *                                 A1 -= <float> offset1
 *                                 B0 -= <float> offset0
 *                                 B1 -= <float> offset1             # <<<<<<<<<<<<<<
 *                                 C0 -= <float> offset0
 *                                 C1 -= <float> offset1
 */
                        __pyx_v_B1 = (__pyx_v_B1 - ((float)__pyx_v_offset1));

                        /* "pyFAI/ext/_distortion.pyx":1791
 *                                 B0 -= <float> offset0
 *                                 B1 -= <float> offset1
 *                                 C0 -= <float> offset0             # <<<<<<<<<<<<<<
 *                                 C1 -= <float> offset1
 *                                 D0 -= <float> offset0
 */
                        __pyx_v_C0 = (__pyx_v_C0 - ((float)__pyx_v_offset0));

                        /* "pyFAI/ext/_distortion.pyx":1792
 *                                 B1 -= <float> offset1
 *                                 C0 -= <float> offset0
 *                                 C1 -= <float> offset1             # <<<<<<<<<<<<<<
 *                                 D0 -= <float> offset0
 *                                 D1 -= <float> offset1
 */
                        __pyx_v_C1 = (__pyx_v_C1 - ((float)__pyx_v_offset1));

                        /* "pyFAI/ext/_distortion.pyx":1793
 *                                 C0 -= <float> offset0
 *                                 C1 -= <float> offset1
 *                                 D0 -= <float> offset0             # <<<<<<<<<<<<<<
 *                                 D1 -= <float> offset1
 *                                 if B0 != A0:
 */
                        __pyx_v_D0 = (__pyx_v_D0 - ((float)__pyx_v_offset0));

                        /* "pyFAI/ext/_distortion.pyx":1794
 *                                 C1 -= <float> offset1
 *                                 D0 -= <float> offset0
 *                                 D1 -= <float> offset1             # <<<<<<<<<<<<<<
 *                                 if B0 != A0:
 *                                     pAB = (B1 - A1) / (B0 - A0)
 */
                        __pyx_v_D1 = (__pyx_v_D1 - ((float)__pyx_v_offset1));

                        /* "pyFAI/ext/_distortion.pyx":1795
 *                                 D0 -= <float> offset0
 *                                 D1 -= <float> offset1
 *                                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                                     pAB = (B1 - A1) / (B0 - A0)
 *                                     cAB = A1 - pAB * A0
 */
                        __pyx_t_11 = ((__pyx_v_B0 != __pyx_v_A0) != 0);
                        if (__pyx_t_11) {

                          /* "pyFAI/ext/_distortion.pyx":1796
 *                                 D1 -= <float> offset1
 *                                 if B0 != A0:
 *                                     pAB = (B1 - A1) / (B0 - A0)             # <<<<<<<<<<<<<<
 *                                     cAB = A1 - pAB * A0
 *                                 else:
 */
                          __pyx_v_pAB = ((__pyx_v_B1 - __pyx_v_A1) / (__pyx_v_B0 - __pyx_v_A0));

                          /* "pyFAI/ext/_distortion.pyx":1797
 *                                 if B0 != A0:
 *                                     pAB = (B1 - A1) / (B0 - A0)
 *                                     cAB = A1 - pAB * A0             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     pAB = cAB = 0.0
 */
                          __pyx_v_cAB = (__pyx_v_A1 - (__pyx_v_pAB * __pyx_v_A0));

                          /* "pyFAI/ext/_distortion.pyx":1795
 *                                 D0 -= <float> offset0
 *                                 D1 -= <float> offset1
 *                                 if B0 != A0:             # <<<<<<<<<<<<<<
 *                                     pAB = (B1 - A1) / (B0 - A0)
 *                                     cAB = A1 - pAB * A0
 */
                          goto __pyx_L25;
                        }

                        /* "pyFAI/ext/_distortion.pyx":1799
 *                                     cAB = A1 - pAB * A0
 *                                 else:
 *                                     pAB = cAB = 0.0             # <<<<<<<<<<<<<<
 *                                 if C0 != B0:
 *                                     pBC = (C1 - B1) / (C0 - B0)
 */
                        /*else*/ {
                          __pyx_v_pAB = 0.0;
                          __pyx_v_cAB = 0.0;
                        }
                        __pyx_L25:;

                        /* "pyFAI/ext/_distortion.pyx":1800
 *                                 else:
 *                                     pAB = cAB = 0.0
 *                                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                                     pBC = (C1 - B1) / (C0 - B0)
 *                                     cBC = B1 - pBC * B0
 */
                        __pyx_t_11 = ((__pyx_v_C0 != __pyx_v_B0) != 0);
                        if (__pyx_t_11) {

                          /* "pyFAI/ext/_distortion.pyx":1801
 *                                     pAB = cAB = 0.0
 *                                 if C0 != B0:
 *                                     pBC = (C1 - B1) / (C0 - B0)             # <<<<<<<<<<<<<<
 *                                     cBC = B1 - pBC * B0
 *                                 else:
 */
                          __pyx_v_pBC = ((__pyx_v_C1 - __pyx_v_B1) / (__pyx_v_C0 - __pyx_v_B0));

                          /* "pyFAI/ext/_distortion.pyx":1802
 *                                 if C0 != B0:
 *                                     pBC = (C1 - B1) / (C0 - B0)
 *                                     cBC = B1 - pBC * B0             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     pBC = cBC = 0.0
 */
                          __pyx_v_cBC = (__pyx_v_B1 - (__pyx_v_pBC * __pyx_v_B0));

                          /* "pyFAI/ext/_distortion.pyx":1800
 *                                 else:
 *                                     pAB = cAB = 0.0
 *                                 if C0 != B0:             # <<<<<<<<<<<<<<
 *                                     pBC = (C1 - B1) / (C0 - B0)
 *                                     cBC = B1 - pBC * B0
 */
                          goto __pyx_L26;
                        }

                        /* "pyFAI/ext/_distortion.pyx":1804
 *                                     cBC = B1 - pBC * B0
 *                                 else:
 *                                     pBC = cBC = 0.0             # <<<<<<<<<<<<<<
 *                                 if D0 != C0:
 *                                     pCD = (D1 - C1) / (D0 - C0)
 */
                        /*else*/ {
                          __pyx_v_pBC = 0.0;
                          __pyx_v_cBC = 0.0;
                        }
                        __pyx_L26:;

                        /* "pyFAI/ext/_distortion.pyx":1805
 *                                 else:
 *                                     pBC = cBC = 0.0
 *                                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                                     pCD = (D1 - C1) / (D0 - C0)
 *                                     cCD = C1 - pCD * C0
 */
                        __pyx_t_11 = ((__pyx_v_D0 != __pyx_v_C0) != 0);
                        if (__pyx_t_11) {

                          /* "pyFAI/ext/_distortion.pyx":1806
 *                                     pBC = cBC = 0.0
 *                                 if D0 != C0:
 *                                     pCD = (D1 - C1) / (D0 - C0)             # <<<<<<<<<<<<<<
 *                                     cCD = C1 - pCD * C0
 *                                 else:
 */
                          __pyx_v_pCD = ((__pyx_v_D1 - __pyx_v_C1) / (__pyx_v_D0 - __pyx_v_C0));

                          /* "pyFAI/ext/_distortion.pyx":1807
 *                                 if D0 != C0:
 *                                     pCD = (D1 - C1) / (D0 - C0)
 *                                     cCD = C1 - pCD * C0             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     pCD = cCD = 0.0
 */
                          __pyx_v_cCD = (__pyx_v_C1 - (__pyx_v_pCD * __pyx_v_C0));

                          /* "pyFAI/ext/_distortion.pyx":1805
 *                                 else:
 *                                     pBC = cBC = 0.0
 *                                 if D0 != C0:             # <<<<<<<<<<<<<<
 *                                     pCD = (D1 - C1) / (D0 - C0)
 *                                     cCD = C1 - pCD * C0
 */
                          goto __pyx_L27;
                        }

                        /* "pyFAI/ext/_distortion.pyx":1809
 *                                     cCD = C1 - pCD * C0
 *                                 else:
 *                                     pCD = cCD = 0.0             # <<<<<<<<<<<<<<
 *                                 if A0 != D0:
 *                                     pDA = (A1 - D1) / (A0 - D0)
 */
                        /*else*/ {
                          __pyx_v_pCD = 0.0;
                          __pyx_v_cCD = 0.0;
                        }
                        __pyx_L27:;

                        /* "pyFAI/ext/_distortion.pyx":1810
 *                                 else:
 *                                     pCD = cCD = 0.0
 *                                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                                     pDA = (A1 - D1) / (A0 - D0)
 *                                     cDA = D1 - pDA * D0
 */
                        __pyx_t_11 = ((__pyx_v_A0 != __pyx_v_D0) != 0);
                        if (__pyx_t_11) {

                          /* "pyFAI/ext/_distortion.pyx":1811
 *                                     pCD = cCD = 0.0
 *                                 if A0 != D0:
 *                                     pDA = (A1 - D1) / (A0 - D0)             # <<<<<<<<<<<<<<
 *                                     cDA = D1 - pDA * D0
 *                                 else:
 */
                          __pyx_v_pDA = ((__pyx_v_A1 - __pyx_v_D1) / (__pyx_v_A0 - __pyx_v_D0));

                          /* "pyFAI/ext/_distortion.pyx":1812
 *                                 if A0 != D0:
 *                                     pDA = (A1 - D1) / (A0 - D0)
 *                                     cDA = D1 - pDA * D0             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     pDA = cDA = 0.0
 */
                          __pyx_v_cDA = (__pyx_v_D1 - (__pyx_v_pDA * __pyx_v_D0));

                          /* "pyFAI/ext/_distortion.pyx":1810
 *                                 else:
 *                                     pCD = cCD = 0.0
 *                                 if A0 != D0:             # <<<<<<<<<<<<<<
 *                                     pDA = (A1 - D1) / (A0 - D0)
 *                                     cDA = D1 - pDA * D0
 */
                          goto __pyx_L28;
                        }

                        /* "pyFAI/ext/_distortion.pyx":1814
 *                                     cDA = D1 - pDA * D0
 *                                 else:
 *                                     pDA = cDA = 0.0             # <<<<<<<<<<<<<<
 *                                 # ABCD is ANTI-trigonometric order: order input position accordingly
 *                                 integrate(buffer, B0, A0, pAB, cAB)
 */
                        /*else*/ {
                          __pyx_v_pDA = 0.0;
                          __pyx_v_cDA = 0.0;
                        }
                        __pyx_L28:;

                        /* "pyFAI/ext/_distortion.pyx":1816
 *                                     pDA = cDA = 0.0
 *                                 # ABCD is ANTI-trigonometric order: order input position accordingly
 *                                 integrate(buffer, B0, A0, pAB, cAB)             # <<<<<<<<<<<<<<
 *                                 integrate(buffer, A0, D0, pDA, cDA)
 *                                 integrate(buffer, D0, C0, pCD, cCD)
 */
                        __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_B0, __pyx_v_A0, __pyx_v_pAB, __pyx_v_cAB);

                        /* "pyFAI/ext/_distortion.pyx":1817
 *                                 # ABCD is ANTI-trigonometric order: order input position accordingly
 *                                 integrate(buffer, B0, A0, pAB, cAB)
 *                                 integrate(buffer, A0, D0, pDA, cDA)             # <<<<<<<<<<<<<<
 *                                 integrate(buffer, D0, C0, pCD, cCD)
 *                                 integrate(buffer, C0, B0, pBC, cBC)
 */
                        __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_A0, __pyx_v_D0, __pyx_v_pDA, __pyx_v_cDA);

                        /* "pyFAI/ext/_distortion.pyx":1818
 *                                 integrate(buffer, B0, A0, pAB, cAB)
 *                                 integrate(buffer, A0, D0, pDA, cDA)
 *                                 integrate(buffer, D0, C0, pCD, cCD)             # <<<<<<<<<<<<<<
 *                                 integrate(buffer, C0, B0, pBC, cBC)
 *                                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 */
                        __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_D0, __pyx_v_C0, __pyx_v_pCD, __pyx_v_cCD);

                        /* "pyFAI/ext/_distortion.pyx":1819
 *                                 integrate(buffer, A0, D0, pDA, cDA)
 *                                 integrate(buffer, D0, C0, pCD, cCD)
 *                                 integrate(buffer, C0, B0, pBC, cBC)             # <<<<<<<<<<<<<<
 *                                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                                 for ms in range(box_size0):
 */
                        __pyx_f_5pyFAI_3ext_11_distortion_integrate(__pyx_v_buffer, __pyx_v_C0, __pyx_v_B0, __pyx_v_pBC, __pyx_v_cBC);

                        /* "pyFAI/ext/_distortion.pyx":1820
 *                                 integrate(buffer, D0, C0, pCD, cCD)
 *                                 integrate(buffer, C0, B0, pBC, cBC)
 *                                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))             # <<<<<<<<<<<<<<
 *                                 for ms in range(box_size0):
 *                                     ml = ms + offset0
 */
                        __pyx_v_area = (0.5 * (((__pyx_v_C0 - __pyx_v_A0) * (__pyx_v_D1 - __pyx_v_B1)) - ((__pyx_v_C1 - __pyx_v_A1) * (__pyx_v_D0 - __pyx_v_B0))));

                        /* "pyFAI/ext/_distortion.pyx":1821
 *                                 integrate(buffer, C0, B0, pBC, cBC)
 *                                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                                 for ms in range(box_size0):             # <<<<<<<<<<<<<<
 *                                     ml = ms + offset0
 *                                     if ml < 0 or ml >= shape0:
 */
                        __pyx_t_34 = __pyx_v_box_size0;
                        __pyx_t_35 = __pyx_t_34;
                        for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
                          __pyx_v_ms = __pyx_t_36;

                          /* "pyFAI/ext/_distortion.pyx":1822
 *                                 area = 0.5 * ((C0 - A0) * (D1 - B1) - (C1 - A1) * (D0 - B0))
 *                                 for ms in range(box_size0):
 *                                     ml = ms + offset0             # <<<<<<<<<<<<<<
 *                                     if ml < 0 or ml >= shape0:
 *                                         continue
 */
                          __pyx_v_ml = (__pyx_v_ms + __pyx_v_offset0);

                          /* "pyFAI/ext/_distortion.pyx":1823
 *                                 for ms in range(box_size0):
 *                                     ml = ms + offset0
 *                                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                                         continue
 *                                     for ns in range(box_size1):
 */
                          __pyx_t_10 = ((__pyx_v_ml < 0) != 0);
                          if (!__pyx_t_10) {
                          } else {
                            __pyx_t_11 = __pyx_t_10;
                            goto __pyx_L32_bool_binop_done;
                          }
                          __pyx_t_10 = ((__pyx_v_ml >= __pyx_v_shape0) != 0);
                          __pyx_t_11 = __pyx_t_10;
                          __pyx_L32_bool_binop_done:;
                          if (__pyx_t_11) {

                            /* "pyFAI/ext/_distortion.pyx":1824
 *                                     ml = ms + offset0
 *                                     if ml < 0 or ml >= shape0:
 *                                         continue             # <<<<<<<<<<<<<<
 *                                     for ns in range(box_size1):
 *                                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 */
                            goto __pyx_L29_continue;

                            /* "pyFAI/ext/_distortion.pyx":1823
 *                                 for ms in range(box_size0):
 *                                     ml = ms + offset0
 *                                     if ml < 0 or ml >= shape0:             # <<<<<<<<<<<<<<
 *                                         continue
 *                                     for ns in range(box_size1):
 */
                          }

                          /* "pyFAI/ext/_distortion.pyx":1825
 *                                     if ml < 0 or ml >= shape0:
 *                                         continue
 *                                     for ns in range(box_size1):             # <<<<<<<<<<<<<<
 *                                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                                         nl = ns + offset1
 */
                          __pyx_t_37 = __pyx_v_box_size1;
                          __pyx_t_38 = __pyx_t_37;
                          for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
                            __pyx_v_ns = __pyx_t_39;

                            /* "pyFAI/ext/_distortion.pyx":1827
 *                                     for ns in range(box_size1):
 *                                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                                         nl = ns + offset1             # <<<<<<<<<<<<<<
 *                                         if nl < 0 or nl >= shape1:
 *                                             continue
 */
                            __pyx_v_nl = (__pyx_v_ns + __pyx_v_offset1);

                            /* "pyFAI/ext/_distortion.pyx":1828
 *                                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                                         nl = ns + offset1
 *                                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         value = buffer[ms, ns] / area
 */
                            __pyx_t_10 = ((__pyx_v_nl < 0) != 0);
                            if (!__pyx_t_10) {
                            } else {
                              __pyx_t_11 = __pyx_t_10;
                              goto __pyx_L37_bool_binop_done;
                            }
                            __pyx_t_10 = ((__pyx_v_nl >= __pyx_v_shape1) != 0);
                            __pyx_t_11 = __pyx_t_10;
                            __pyx_L37_bool_binop_done:;
                            if (__pyx_t_11) {

                              /* "pyFAI/ext/_distortion.pyx":1829
 *                                         nl = ns + offset1
 *                                         if nl < 0 or nl >= shape1:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         value = buffer[ms, ns] / area
 *                                         if value <= 0:
 */
                              goto __pyx_L34_continue;

                              /* "pyFAI/ext/_distortion.pyx":1828
 *                                         # ms,ns are indexes of the corrected image in short form, ml & nl are the same
 *                                         nl = ns + offset1
 *                                         if nl < 0 or nl >= shape1:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         value = buffer[ms, ns] / area
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1830
 *                                         if nl < 0 or nl >= shape1:
 *                                             continue
 *                                         value = buffer[ms, ns] / area             # <<<<<<<<<<<<<<
 *                                         if value <= 0:
 *                                             continue
 */
                            __pyx_t_21 = __pyx_v_ms;
                            __pyx_t_20 = __pyx_v_ns;
                            __pyx_v_value = ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ (__pyx_v_buffer.data + __pyx_t_21 * __pyx_v_buffer.strides[0]) )) + __pyx_t_20)) ))) / ((__pyx_t_5pyFAI_3ext_12shared_types_float32_t)__pyx_v_area));

                            /* "pyFAI/ext/_distortion.pyx":1831
 *                                             continue
 *                                         value = buffer[ms, ns] / area
 *                                         if value <= 0:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         k = outMax[ml, nl]
 */
                            __pyx_t_11 = ((__pyx_v_value <= 0.0) != 0);
                            if (__pyx_t_11) {

                              /* "pyFAI/ext/_distortion.pyx":1832
 *                                         value = buffer[ms, ns] / area
 *                                         if value <= 0:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         k = outMax[ml, nl]
 *                                         lut[ml, nl, k].idx = idx
 */
                              goto __pyx_L34_continue;

                              /* "pyFAI/ext/_distortion.pyx":1831
 *                                             continue
 *                                         value = buffer[ms, ns] / area
 *                                         if value <= 0:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         k = outMax[ml, nl]
 */
                            }

                            /* "pyFAI/ext/_distortion.pyx":1833
 *                                         if value <= 0:
 *                                             continue
 *                                         k = outMax[ml, nl]             # <<<<<<<<<<<<<<
 *                                         lut[ml, nl, k].idx = idx
 *                                         lut[ml, nl, k].coef = value
 */
                            __pyx_t_20 = __pyx_v_ml;
                            __pyx_t_21 = __pyx_v_nl;
                            __pyx_v_k = (*((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_20 * __pyx_v_outMax.strides[0]) )) + __pyx_t_21)) )));

                            /* "pyFAI/ext/_distortion.pyx":1834
 *                                             continue
 *                                         k = outMax[ml, nl]
 *                                         lut[ml, nl, k].idx = idx             # <<<<<<<<<<<<<<
 *                                         lut[ml, nl, k].coef = value
 *                                         outMax[ml, nl] = k + 1
 */
                            __pyx_t_21 = __pyx_v_ml;
                            __pyx_t_20 = __pyx_v_nl;
                            __pyx_t_19 = __pyx_v_k;
                            (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_21 * __pyx_v_lut.strides[0]) ) + __pyx_t_20 * __pyx_v_lut.strides[1]) )) + __pyx_t_19)) ))).idx = __pyx_v_idx;

                            /* "pyFAI/ext/_distortion.pyx":1835
 *                                         k = outMax[ml, nl]
 *                                         lut[ml, nl, k].idx = idx
 *                                         lut[ml, nl, k].coef = value             # <<<<<<<<<<<<<<
 *                                         outMax[ml, nl] = k + 1
 *                                 idx += 1
 */
                            __pyx_t_19 = __pyx_v_ml;
                            __pyx_t_20 = __pyx_v_nl;
                            __pyx_t_21 = __pyx_v_k;
                            (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=2 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lut.data + __pyx_t_19 * __pyx_v_lut.strides[0]) ) + __pyx_t_20 * __pyx_v_lut.strides[1]) )) + __pyx_t_21)) ))).coef = __pyx_v_value;

                            /* "pyFAI/ext/_distortion.pyx":1836
 *                                         lut[ml, nl, k].idx = idx
 *                                         lut[ml, nl, k].coef = value
 *                                         outMax[ml, nl] = k + 1             # <<<<<<<<<<<<<<
 *                                 idx += 1
 *                     self.LUT = numpy.asarray(lut).reshape(self.shape[0] * self.shape[1], self.lut_size)
 */
                            __pyx_t_21 = __pyx_v_ml;
                            __pyx_t_20 = __pyx_v_nl;
                            *((int32_t *) ( /* dim=1 */ ((char *) (((int32_t *) ( /* dim=0 */ (__pyx_v_outMax.data + __pyx_t_21 * __pyx_v_outMax.strides[0]) )) + __pyx_t_20)) )) = (__pyx_v_k + 1);
                            __pyx_L34_continue:;
                          }
                          __pyx_L29_continue:;
                        }

                        /* "pyFAI/ext/_distortion.pyx":1837
 *                                         lut[ml, nl, k].coef = value
 *                                         outMax[ml, nl] = k + 1
 *                                 idx += 1             # <<<<<<<<<<<<<<
 *                     self.LUT = numpy.asarray(lut).reshape(self.shape[0] * self.shape[1], self.lut_size)
 *         return self.LUT
 */
                        __pyx_v_idx = (__pyx_v_idx + 1);
                      }
                    }
                  }

                  /* "pyFAI/ext/_distortion.pyx":1769
 *                     #buffer_size = self.delta0 * self.delta1 * sizeof(float)
 *                     logger.info("Max pixel size: %ix%i; Max source pixel in target: %i" % (buffer.shape[1], buffer.shape[0], self.lut_size))
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         # i,j, idx are indexes of the raw image uncorrected
 *                         for i in range(shape0):
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L20;
                    }
                    __pyx_L20:;
                  }
              }

              /* "pyFAI/ext/_distortion.pyx":1838
 *                                         outMax[ml, nl] = k + 1
 *                                 idx += 1
 *                     self.LUT = numpy.asarray(lut).reshape(self.shape[0] * self.shape[1], self.lut_size)             # <<<<<<<<<<<<<<
 *         return self.LUT
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_numpy); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_asarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = __pyx_memoryview_fromslice(__pyx_v_lut, 3, (PyObject *(*)(char *)) __pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, 0);; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_1, function);
                }
              }
              __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_17);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reshape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Multiply(__pyx_t_17, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lut_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_17 = NULL;
              __pyx_t_9 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_1);
                if (likely(__pyx_t_17)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                  __Pyx_INCREF(__pyx_t_17);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_1, function);
                  __pyx_t_9 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_t_2, __pyx_t_5};
                __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1838, __pyx_L11_error)
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_t_2, __pyx_t_5};
                __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1838, __pyx_L11_error)
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              } else
              #endif
              {
                __pyx_t_22 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1838, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_22);
                if (__pyx_t_17) {
                  __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_17); __pyx_t_17 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_22, 0+__pyx_t_9, __pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_5);
                PyTuple_SET_ITEM(__pyx_t_22, 1+__pyx_t_9, __pyx_t_5);
                __pyx_t_2 = 0;
                __pyx_t_5 = 0;
                __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_22, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1838, __pyx_L11_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              }
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_LUT, __pyx_t_4) < 0) __PYX_ERR(0, 1838, __pyx_L11_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "pyFAI/ext/_distortion.pyx":1760
 *         if self.LUT is None:
 *             with self._sem:
 *                 if self.LUT is None:             # <<<<<<<<<<<<<<
 *                     pos = self.pos
 *                     lut = numpy.recarray(shape=(self.shape[0], self.shape[1], self.lut_size), dtype=lut_d)
 */
            }

            /* "pyFAI/ext/_distortion.pyx":1759
 *             self.calc_LUT_size()
 *         if self.LUT is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.LUT is None:
 *                     pos = self.pos
 */
          }
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L16_try_end;
          __pyx_L11_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_23, 1);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
          /*except:*/ {
            __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_22) < 0) __PYX_ERR(0, 1759, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_22);
            __pyx_t_5 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_1, __pyx_t_22); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1759, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_40 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_5, NULL);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1759, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_40);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_40);
            __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
            if (__pyx_t_11 < 0) __PYX_ERR(0, 1759, __pyx_L13_except_error)
            __pyx_t_10 = ((!(__pyx_t_11 != 0)) != 0);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_22);
              __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_22 = 0; 
              __PYX_ERR(0, 1759, __pyx_L13_except_error)
            }
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
            goto __pyx_L12_exception_handled;
          }
          __pyx_L13_except_error:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          goto __pyx_L1_error;
          __pyx_L12_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_L16_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_12) {
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_tuple__42, NULL);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1759, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L43;
      __pyx_L7_error:;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L1_error;
      __pyx_L43:;
    }

    /* "pyFAI/ext/_distortion.pyx":1758
 *         if self.lut_size is None:
 *             self.calc_LUT_size()
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             with self._sem:
 *                 if self.LUT is None:
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1839
 *                                 idx += 1
 *                     self.LUT = numpy.asarray(lut).reshape(self.shape[0] * self.shape[1], self.lut_size)
 *         return self.LUT             # <<<<<<<<<<<<<<
 * 
 * ################################################################################
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_r = __pyx_t_22;
  __pyx_t_22 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1744
 *                     return np_lut_size
 * 
 *     def calc_LUT(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, j, ms, ml, ns, nl, shape0, shape1, size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_XDECREF(__pyx_t_17);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __Pyx_XDECREF(__pyx_t_22);
  __PYX_XDEC_MEMVIEW(&__pyx_t_23, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.calc_LUT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_pos, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lut, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_outMax, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_buffer, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1856
 * #         return ab
 * 
 *     def correct(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Correct an image based on the look-up table calculated ...
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_11correct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_10correct[] = "Distortion.correct(self, image)\n\n        Correct an image based on the look-up table calculated ...\n\n        :param image: 2D-array with the image\n        :return: corrected 2D image\n        ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_11correct = {"correct", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_11correct, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_10correct};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_11correct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_image = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("correct (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_image,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("correct", 1, 2, 2, 1); __PYX_ERR(0, 1856, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "correct") < 0)) __PYX_ERR(0, 1856, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_image = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("correct", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1856, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.correct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_10correct(__pyx_self, __pyx_v_self, __pyx_v_image);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_10correct(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_image) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_idx;
  int __pyx_v_size;
  float __pyx_v_coef;
  __Pyx_memviewslice __pyx_v_LUT = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lout = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lin = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_img_shape = NULL;
  PyObject *__pyx_v_new_image = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  __pyx_t_5pyFAI_3ext_11_distortion_index_t __pyx_t_20;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("correct", 0);
  __Pyx_INCREF(__pyx_v_image);

  /* "pyFAI/ext/_distortion.pyx":1868
 *             lut_t[:, ::1] LUT
 *             float32_t[::1] lout, lin
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT()
 *         LUT = self.LUT
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1869
 *             float32_t[::1] lout, lin
 *         if self.LUT is None:
 *             self.calc_LUT()             # <<<<<<<<<<<<<<
 *         LUT = self.LUT
 *         img_shape = image.shape
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_calc_LUT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyFAI/ext/_distortion.pyx":1868
 *             lut_t[:, ::1] LUT
 *             float32_t[::1] lout, lin
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT()
 *         LUT = self.LUT
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1870
 *         if self.LUT is None:
 *             self.calc_LUT()
 *         LUT = self.LUT             # <<<<<<<<<<<<<<
 *         img_shape = image.shape
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 1870, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_LUT = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1871
 *             self.calc_LUT()
 *         LUT = self.LUT
 *         img_shape = image.shape             # <<<<<<<<<<<<<<
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_img_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1872
 *         LUT = self.LUT
 *         img_shape = image.shape
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):             # <<<<<<<<<<<<<<
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)
 *             new_image[:img_shape[0], :img_shape[1]] = image
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_img_shape, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_img_shape, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1873
 *         img_shape = image.shape
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *             new_image[:img_shape[0], :img_shape[1]] = image
 *             image = new_image
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_new_image = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "pyFAI/ext/_distortion.pyx":1874
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)
 *             new_image[:img_shape[0], :img_shape[1]] = image             # <<<<<<<<<<<<<<
 *             image = new_image
 *             logger.warning("Patching image as image is %ix%i and spline is %ix%i" % (img_shape[1], img_shape[0], self.shape[1], self.shape[0]))
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_img_shape, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = PySlice_New(Py_None, __pyx_t_8, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_img_shape, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = PySlice_New(Py_None, __pyx_t_8, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_new_image, __pyx_t_8, __pyx_v_image) < 0)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pyFAI/ext/_distortion.pyx":1875
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)
 *             new_image[:img_shape[0], :img_shape[1]] = image
 *             image = new_image             # <<<<<<<<<<<<<<
 *             logger.warning("Patching image as image is %ix%i and spline is %ix%i" % (img_shape[1], img_shape[0], self.shape[1], self.shape[0]))
 * 
 */
    __Pyx_INCREF(__pyx_v_new_image);
    __Pyx_DECREF_SET(__pyx_v_image, __pyx_v_new_image);

    /* "pyFAI/ext/_distortion.pyx":1876
 *             new_image[:img_shape[0], :img_shape[1]] = image
 *             image = new_image
 *             logger.warning("Patching image as image is %ix%i and spline is %ix%i" % (img_shape[1], img_shape[0], self.shape[1], self.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_img_shape, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_img_shape, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_7, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_10);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = PyUnicode_Format(__pyx_kp_u_Patching_image_as_image_is_ix_i, __pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_8 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pyFAI/ext/_distortion.pyx":1872
 *         LUT = self.LUT
 *         img_shape = image.shape
 *         if (img_shape[0] < self.shape[0]) or (img_shape[1] < self.shape[1]):             # <<<<<<<<<<<<<<
 *             new_image = numpy.zeros(self.shape, dtype=numpy.float32)
 *             new_image[:img_shape[0], :img_shape[1]] = image
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1878
 *             logger.warning("Patching image as image is %ix%i and spline is %ix%i" % (img_shape[1], img_shape[0], self.shape[1], self.shape[0]))
 * 
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *         lout = out.ravel()
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float32); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_out = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":1879
 * 
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 *         lout = out.ravel()             # <<<<<<<<<<<<<<
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *         size = lin.size
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lout = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1880
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 *         lout = out.ravel()
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *         size = lin.size
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_ascontiguousarray); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_9 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":1881
 *         lout = out.ravel()
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *         size = lin.size             # <<<<<<<<<<<<<<
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *             for j in range(LUT.shape[1]):
 */
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_lin, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_size = __pyx_t_12;

  /* "pyFAI/ext/_distortion.pyx":1882
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *         size = lin.size
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *             for j in range(LUT.shape[1]):
 *                 idx = LUT[i, j].idx
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_13 = (__pyx_v_LUT.shape[0]);
        if ((1 == 0)) abort();
        {
            float __pyx_parallel_temp0 = ((float)__PYX_NAN());
            int __pyx_parallel_temp1 = ((int)0xbad0bad0);
            int __pyx_parallel_temp2 = ((int)0xbad0bad0);
            int __pyx_parallel_temp3 = ((int)0xbad0bad0);
            const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
            PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
            int __pyx_parallel_why;
            __pyx_parallel_why = 0;
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_15 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_3) firstprivate(__pyx_t_10, __pyx_t_5, __pyx_t_7, __pyx_t_8, __pyx_t_9) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    Py_BEGIN_ALLOW_THREADS
                    #endif /* _OPENMP */
                    #ifdef _OPENMP
                    #pragma omp for lastprivate(__pyx_v_coef) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_idx) lastprivate(__pyx_v_j) schedule(static)
                    #endif /* _OPENMP */
                    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                        if (__pyx_parallel_why < 2)
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_14);
                            /* Initialize private variables to invalid values */
                            __pyx_v_coef = ((float)__PYX_NAN());
                            __pyx_v_idx = ((int)0xbad0bad0);
                            __pyx_v_j = ((int)0xbad0bad0);

                            /* "pyFAI/ext/_distortion.pyx":1883
 *         size = lin.size
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *             for j in range(LUT.shape[1]):             # <<<<<<<<<<<<<<
 *                 idx = LUT[i, j].idx
 *                 coef = LUT[i, j].coef
 */
                            __pyx_t_16 = (__pyx_v_LUT.shape[1]);
                            __pyx_t_17 = __pyx_t_16;
                            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_17; __pyx_t_12+=1) {
                              __pyx_v_j = __pyx_t_12;

                              /* "pyFAI/ext/_distortion.pyx":1884
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):
 *             for j in range(LUT.shape[1]):
 *                 idx = LUT[i, j].idx             # <<<<<<<<<<<<<<
 *                 coef = LUT[i, j].coef
 *                 if coef <= 0:
 */
                              __pyx_t_18 = __pyx_v_i;
                              __pyx_t_19 = __pyx_v_j;
                              __pyx_t_20 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_18 * __pyx_v_LUT.strides[0]) )) + __pyx_t_19)) ))).idx;
                              __pyx_v_idx = __pyx_t_20;

                              /* "pyFAI/ext/_distortion.pyx":1885
 *             for j in range(LUT.shape[1]):
 *                 idx = LUT[i, j].idx
 *                 coef = LUT[i, j].coef             # <<<<<<<<<<<<<<
 *                 if coef <= 0:
 *                     continue
 */
                              __pyx_t_19 = __pyx_v_i;
                              __pyx_t_18 = __pyx_v_j;
                              __pyx_t_21 = (*((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=1 */ ((char *) (((struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) ( /* dim=0 */ (__pyx_v_LUT.data + __pyx_t_19 * __pyx_v_LUT.strides[0]) )) + __pyx_t_18)) ))).coef;
                              __pyx_v_coef = __pyx_t_21;

                              /* "pyFAI/ext/_distortion.pyx":1886
 *                 idx = LUT[i, j].idx
 *                 coef = LUT[i, j].coef
 *                 if coef <= 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 if idx >= size:
 */
                              __pyx_t_3 = ((__pyx_v_coef <= 0.0) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1887
 *                 coef = LUT[i, j].coef
 *                 if coef <= 0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 if idx >= size:
 *                     with gil:
 */
                                goto __pyx_L14_continue;

                                /* "pyFAI/ext/_distortion.pyx":1886
 *                 idx = LUT[i, j].idx
 *                 coef = LUT[i, j].coef
 *                 if coef <= 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 if idx >= size:
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1888
 *                 if coef <= 0:
 *                     continue
 *                 if idx >= size:             # <<<<<<<<<<<<<<
 *                     with gil:
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              __pyx_t_3 = ((__pyx_v_idx >= __pyx_v_size) != 0);
                              if (__pyx_t_3) {

                                /* "pyFAI/ext/_distortion.pyx":1889
 *                     continue
 *                 if idx >= size:
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                         continue
 */
                                {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    /*try:*/ {

                                      /* "pyFAI/ext/_distortion.pyx":1890
 *                 if idx >= size:
 *                     with gil:
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))             # <<<<<<<<<<<<<<
 *                         continue
 *                 lout[i] += lin[idx] * coef
 */
                                      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warning); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_10);
                                      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                                      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_7);
                                      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_8);
                                      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_5);
                                      __Pyx_GIVEREF(__pyx_t_7);
                                      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
                                      __Pyx_GIVEREF(__pyx_t_8);
                                      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
                                      __pyx_t_7 = 0;
                                      __pyx_t_8 = 0;
                                      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_Accessing_i_i, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_8);
                                      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      __pyx_t_5 = NULL;
                                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
                                        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
                                        if (likely(__pyx_t_5)) {
                                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                                          __Pyx_INCREF(__pyx_t_5);
                                          __Pyx_INCREF(function);
                                          __Pyx_DECREF_SET(__pyx_t_10, function);
                                        }
                                      }
                                      __pyx_t_9 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_8);
                                      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                                      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                                      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1890, __pyx_L21_error)
                                      __Pyx_GOTREF(__pyx_t_9);
                                      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                                      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

                                      /* "pyFAI/ext/_distortion.pyx":1891
 *                     with gil:
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                         continue             # <<<<<<<<<<<<<<
 *                 lout[i] += lin[idx] * coef
 *         return out[:img_shape[0], :img_shape[1]]
 */
                                      goto __pyx_L18_continue;
                                    }

                                    /* "pyFAI/ext/_distortion.pyx":1889
 *                     continue
 *                 if idx >= size:
 *                     with gil:             # <<<<<<<<<<<<<<
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                         continue
 */
                                    /*finally:*/ {
                                      __pyx_L18_continue: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L14_continue;
                                      }
                                      __pyx_L21_error: {
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        goto __pyx_L12_error;
                                      }
                                    }
                                }

                                /* "pyFAI/ext/_distortion.pyx":1888
 *                 if coef <= 0:
 *                     continue
 *                 if idx >= size:             # <<<<<<<<<<<<<<
 *                     with gil:
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 */
                              }

                              /* "pyFAI/ext/_distortion.pyx":1892
 *                         logger.warning("Accessing %i >= %i !!!" % (idx, size))
 *                         continue
 *                 lout[i] += lin[idx] * coef             # <<<<<<<<<<<<<<
 *         return out[:img_shape[0], :img_shape[1]]
 * 
 */
                              __pyx_t_18 = __pyx_v_idx;
                              __pyx_t_19 = __pyx_v_i;
                              *((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lout.data) + __pyx_t_19)) )) += ((*((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) ( /* dim=0 */ ((char *) (((__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) __pyx_v_lin.data) + __pyx_t_18)) ))) * __pyx_v_coef);
                              __pyx_L14_continue:;
                            }
                            goto __pyx_L24;
                            __pyx_L12_error:;
                            {
                                #ifdef WITH_THREAD
                                PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                #endif
                                #ifdef _OPENMP
                                #pragma omp flush(__pyx_parallel_exc_type)
                                #endif /* _OPENMP */
                                if (!__pyx_parallel_exc_type) {
                                  __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                  __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                  __Pyx_GOTREF(__pyx_parallel_exc_type);
                                }
                                #ifdef WITH_THREAD
                                __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                #endif
                            }
                            __pyx_parallel_why = 4;
                            goto __pyx_L23;
                            __pyx_L23:;
                            #ifdef _OPENMP
                            #pragma omp critical(__pyx_parallel_lastprivates6)
                            #endif /* _OPENMP */
                            {
                                __pyx_parallel_temp0 = __pyx_v_coef;
                                __pyx_parallel_temp1 = __pyx_v_i;
                                __pyx_parallel_temp2 = __pyx_v_idx;
                                __pyx_parallel_temp3 = __pyx_v_j;
                            }
                            __pyx_L24:;
                            #ifdef _OPENMP
                            #pragma omp flush(__pyx_parallel_why)
                            #endif /* _OPENMP */
                        }
                    }
                    #ifdef _OPENMP
                    Py_END_ALLOW_THREADS
                    #else
{
#ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    #endif /* _OPENMP */
                    /* Clean up any temporaries */
                    __Pyx_XDECREF(__pyx_t_10);
                    __pyx_t_10 = NULL;
                    __Pyx_XDECREF(__pyx_t_5);
                    __pyx_t_5 = NULL;
                    __Pyx_XDECREF(__pyx_t_7);
                    __pyx_t_7 = NULL;
                    __Pyx_XDECREF(__pyx_t_8);
                    __pyx_t_8 = NULL;
                    __Pyx_XDECREF(__pyx_t_9);
                    __pyx_t_9 = NULL;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                    #ifndef _OPENMP
}
#endif /* _OPENMP */
                }
            }
            if (__pyx_parallel_exc_type) {
              /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
              __pyx_parallel_why = 4;
            }
            if (__pyx_parallel_why) {
              __pyx_v_coef = __pyx_parallel_temp0;
              __pyx_v_i = __pyx_parallel_temp1;
              __pyx_v_idx = __pyx_parallel_temp2;
              __pyx_v_j = __pyx_parallel_temp3;
              switch (__pyx_parallel_why) {
                    case 4:
                {
                    #ifdef WITH_THREAD
                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                    #endif
                    __Pyx_GIVEREF(__pyx_parallel_exc_type);
                    __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                    __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                    #ifdef WITH_THREAD
                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                    #endif
                }
                goto __pyx_L8_error;
              }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "pyFAI/ext/_distortion.pyx":1882
 *         lin = numpy.ascontiguousarray(image.ravel(), dtype=numpy.float32)
 *         size = lin.size
 *         for i in prange(LUT.shape[0], nogil=True, schedule="static"):             # <<<<<<<<<<<<<<
 *             for j in range(LUT.shape[1]):
 *                 idx = LUT[i, j].idx
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L8_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L9:;
      }
  }

  /* "pyFAI/ext/_distortion.pyx":1893
 *                         continue
 *                 lout[i] += lin[idx] * coef
 *         return out[:img_shape[0], :img_shape[1]]             # <<<<<<<<<<<<<<
 * 
 *     def uncorrect(self, image):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_img_shape, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PySlice_New(Py_None, __pyx_t_9, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_img_shape, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PySlice_New(Py_None, __pyx_t_9, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8);
  __pyx_t_10 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_out, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1856
 * #         return ab
 * 
 *     def correct(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Correct an image based on the look-up table calculated ...
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.correct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_LUT, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lout, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lin, 1);
  __Pyx_XDECREF(__pyx_v_img_shape);
  __Pyx_XDECREF(__pyx_v_new_image);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyFAI/ext/_distortion.pyx":1895
 *         return out[:img_shape[0], :img_shape[1]]
 * 
 *     def uncorrect(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Take an image which has been corrected and transform it into it's raw (with loss of information)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_13uncorrect(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_12uncorrect[] = "Distortion.uncorrect(self, image)\n\n        Take an image which has been corrected and transform it into it's raw (with loss of information)\n\n        :param image: 2D-array with the image\n        :return: uncorrected 2D image and a mask (pixels in raw image\n        ";
static PyMethodDef __pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_13uncorrect = {"uncorrect", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_13uncorrect, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_10Distortion_12uncorrect};
static PyObject *__pyx_pw_5pyFAI_3ext_11_distortion_10Distortion_13uncorrect(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_image = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uncorrect (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_image,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("uncorrect", 1, 2, 2, 1); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "uncorrect") < 0)) __PYX_ERR(0, 1895, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_image = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uncorrect", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.uncorrect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_12uncorrect(__pyx_self, __pyx_v_self, __pyx_v_image);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyFAI_3ext_11_distortion_10Distortion_12uncorrect(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_image) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_lmask = NULL;
  PyObject *__pyx_v_lout = NULL;
  PyObject *__pyx_v_lin = NULL;
  PyObject *__pyx_v_tot = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uncorrect", 0);

  /* "pyFAI/ext/_distortion.pyx":1902
 *         :return: uncorrected 2D image and a mask (pixels in raw image
 *         """
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT()
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "pyFAI/ext/_distortion.pyx":1903
 *         """
 *         if self.LUT is None:
 *             self.calc_LUT()             # <<<<<<<<<<<<<<
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 *         mask = numpy.zeros(self.shape, dtype=numpy.int8)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_calc_LUT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyFAI/ext/_distortion.pyx":1902
 *         :return: uncorrected 2D image and a mask (pixels in raw image
 *         """
 *         if self.LUT is None:             # <<<<<<<<<<<<<<
 *             self.calc_LUT()
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 */
  }

  /* "pyFAI/ext/_distortion.pyx":1904
 *         if self.LUT is None:
 *             self.calc_LUT()
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)             # <<<<<<<<<<<<<<
 *         mask = numpy.zeros(self.shape, dtype=numpy.int8)
 *         lmask = mask.ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_out = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyFAI/ext/_distortion.pyx":1905
 *             self.calc_LUT()
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 *         mask = numpy.zeros(self.shape, dtype=numpy.int8)             # <<<<<<<<<<<<<<
 *         lmask = mask.ravel()
 *         lout = out.ravel()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_mask = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pyFAI/ext/_distortion.pyx":1906
 *         out = numpy.zeros(self.shape, dtype=numpy.float32)
 *         mask = numpy.zeros(self.shape, dtype=numpy.int8)
 *         lmask = mask.ravel()             # <<<<<<<<<<<<<<
 *         lout = out.ravel()
 *         lin = image.ravel()
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_ravel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lmask = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pyFAI/ext/_distortion.pyx":1907
 *         mask = numpy.zeros(self.shape, dtype=numpy.int8)
 *         lmask = mask.ravel()
 *         lout = out.ravel()             # <<<<<<<<<<<<<<
 *         lin = image.ravel()
 *         tot = self.LUT.coef.sum(axis=-1)
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ravel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lout = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pyFAI/ext/_distortion.pyx":1908
 *         lmask = mask.ravel()
 *         lout = out.ravel()
 *         lin = image.ravel()             # <<<<<<<<<<<<<<
 *         tot = self.LUT.coef.sum(axis=-1)
 *         for idx in range(self.LUT.shape[0]):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ravel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_lin = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pyFAI/ext/_distortion.pyx":1909
 *         lout = out.ravel()
 *         lin = image.ravel()
 *         tot = self.LUT.coef.sum(axis=-1)             # <<<<<<<<<<<<<<
 *         for idx in range(self.LUT.shape[0]):
 *             t = tot[idx]
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_coef); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1909, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_tot = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1910
 *         lin = image.ravel()
 *         tot = self.LUT.coef.sum(axis=-1)
 *         for idx in range(self.LUT.shape[0]):             # <<<<<<<<<<<<<<
 *             t = tot[idx]
 *             if t <= 0:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_shape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_5 = __pyx_t_7; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1910, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1910, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1910, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1910, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1910, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1910, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "pyFAI/ext/_distortion.pyx":1911
 *         tot = self.LUT.coef.sum(axis=-1)
 *         for idx in range(self.LUT.shape[0]):
 *             t = tot[idx]             # <<<<<<<<<<<<<<
 *             if t <= 0:
 *                 lmask[idx] = 1
 */
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_tot, __pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "pyFAI/ext/_distortion.pyx":1912
 *         for idx in range(self.LUT.shape[0]):
 *             t = tot[idx]
 *             if t <= 0:             # <<<<<<<<<<<<<<
 *                 lmask[idx] = 1
 *                 continue
 */
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_t, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1912, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1912, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_3) {

      /* "pyFAI/ext/_distortion.pyx":1913
 *             t = tot[idx]
 *             if t <= 0:
 *                 lmask[idx] = 1             # <<<<<<<<<<<<<<
 *                 continue
 *             val = lin[idx] / t
 */
      if (unlikely(PyObject_SetItem(__pyx_v_lmask, __pyx_v_idx, __pyx_int_1) < 0)) __PYX_ERR(0, 1913, __pyx_L1_error)

      /* "pyFAI/ext/_distortion.pyx":1914
 *             if t <= 0:
 *                 lmask[idx] = 1
 *                 continue             # <<<<<<<<<<<<<<
 *             val = lin[idx] / t
 *             lout[self.LUT[idx].idx] += val * self.LUT[idx].coef
 */
      goto __pyx_L4_continue;

      /* "pyFAI/ext/_distortion.pyx":1912
 *         for idx in range(self.LUT.shape[0]):
 *             t = tot[idx]
 *             if t <= 0:             # <<<<<<<<<<<<<<
 *                 lmask[idx] = 1
 *                 continue
 */
    }

    /* "pyFAI/ext/_distortion.pyx":1915
 *                 lmask[idx] = 1
 *                 continue
 *             val = lin[idx] / t             # <<<<<<<<<<<<<<
 *             lout[self.LUT[idx].idx] += val * self.LUT[idx].coef
 *         return out, mask
 */
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_lin, __pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_v_t); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "pyFAI/ext/_distortion.pyx":1916
 *                 continue
 *             val = lin[idx] / t
 *             lout[self.LUT[idx].idx] += val * self.LUT[idx].coef             # <<<<<<<<<<<<<<
 *         return out, mask
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_idx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_lout, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_LUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_idx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_coef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_val, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_lout, __pyx_t_6, __pyx_t_1) < 0)) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pyFAI/ext/_distortion.pyx":1910
 *         lin = image.ravel()
 *         tot = self.LUT.coef.sum(axis=-1)
 *         for idx in range(self.LUT.shape[0]):             # <<<<<<<<<<<<<<
 *             t = tot[idx]
 *             if t <= 0:
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/_distortion.pyx":1917
 *             val = lin[idx] / t
 *             lout[self.LUT[idx].idx] += val * self.LUT[idx].coef
 *         return out, mask             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_out);
  __Pyx_GIVEREF(__pyx_v_out);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_out);
  __Pyx_INCREF(__pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_mask);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyFAI/ext/_distortion.pyx":1895
 *         return out[:img_shape[0], :img_shape[1]]
 * 
 *     def uncorrect(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Take an image which has been corrected and transform it into it's raw (with loss of information)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyFAI.ext._distortion.Distortion.uncorrect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_lmask);
  __Pyx_XDECREF(__pyx_v_lout);
  __Pyx_XDECREF(__pyx_v_lin);
  __Pyx_XDECREF(__pyx_v_tot);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

static struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *__pyx_v_obj) {
  struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t __pyx_v_result;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __pyx_t_5pyFAI_3ext_11_distortion_index_t __pyx_t_10;
  __pyx_t_5pyFAI_3ext_11_distortion_data_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t", 0);

  /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  __pyx_t_1 = ((!(PyMapping_Check(__pyx_v_obj) != 0)) != 0);
  if (__pyx_t_1) {

    /* "FromPyStructUtility":14
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_2 = PyErr_Format(__pyx_builtin_TypeError, ((char const *)"Expected %.16s, got %.200s"), ((char *)"a mapping"), Py_TYPE(__pyx_v_obj)->tp_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  }

  /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['idx']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":17
 * 
 *     try:
 *         value = obj['idx']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'idx'")
 */
      __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_idx); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_value = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['idx']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":18
 *     try:
 *         value = obj['idx']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'idx'")
 *     result.idx = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(1, 18, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "FromPyStructUtility":19
 *         value = obj['idx']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'idx'")             # <<<<<<<<<<<<<<
 *     result.idx = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 19, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 19, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['idx']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "FromPyStructUtility":20
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'idx'")
 *     result.idx = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['coef']
 */
  __pyx_t_10 = __Pyx_PyInt_As_int32_t(__pyx_v_value); if (unlikely((__pyx_t_10 == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_result.idx = __pyx_t_10;

  /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'idx'")
 *     result.idx = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['coef']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":22
 *     result.idx = value
 *     try:
 *         value = obj['coef']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'coef'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_coef); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 22, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'idx'")
 *     result.idx = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['coef']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":23
 *     try:
 *         value = obj['coef']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'coef'")
 *     result.coef = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(1, 23, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "FromPyStructUtility":24
 *         value = obj['coef']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'coef'")             # <<<<<<<<<<<<<<
 *     result.coef = value
 *     return result
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 24, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 24, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;
    __pyx_L14_except_error:;

    /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'idx'")
 *     result.idx = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['coef']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "FromPyStructUtility":25
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'coef'")
 *     result.coef = value             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_11 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_11 == (float)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L1_error)
  __pyx_v_result.coef = __pyx_t_11;

  /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'coef'")
 *     result.coef = value
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(1, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(1, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(1, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      __pyx_t_1 = (__pyx_v_self->len / __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(1, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(1, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 445, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 446, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 456, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__58, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 622, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 628, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__20);
            __Pyx_GIVEREF(__pyx_slice__20);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__20);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__20); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(1, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__20);
        __Pyx_GIVEREF(__pyx_slice__20);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__20);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__61, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(1, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    __pyx_v_shape = (__pyx_v_view->len / __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = (__pyx_v_ndim / 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__63, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1179
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1181
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1182
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1184
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1197
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1198
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1199
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1201
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1202
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1203
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1205
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1219
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1220
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1222
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1224
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1224, __pyx_L1_error)

    /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1227
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1228
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1229
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1230
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1231
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1233
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1237
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1239
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1242
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1244
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1246
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1254
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1253
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1253, __pyx_L1_error)

  /* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1258
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 1258, __pyx_L1_error)

  /* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1263
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 1263, __pyx_L1_error)

    /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1265
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(1, 1265, __pyx_L1_error)
  }

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1276
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1277
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1279
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1280
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1281
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1285
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1289
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1291
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1294
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1295
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1297
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1297, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1300
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1300, __pyx_L1_error)

      /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1305
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1307
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1307, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1308
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1314
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1321
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1322
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1323
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1329
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1329, __pyx_L1_error)

    /* "View.MemoryView":1330
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1330, __pyx_L1_error)

    /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1333
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1334
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1337
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1344
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1346
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1347
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1348
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1349
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1351
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1352
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1353
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1354
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1367
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1374
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1381
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1384
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1386
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1388
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1389
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1391
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1400
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1401
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1403
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1411
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1412
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1415
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1416
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1417
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1419
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1420
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1422
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyFAI.ext._distortion.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyFAI.ext._distortion.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyFAI.ext._distortion.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyFAI.ext._distortion._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"calc_area", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_1calc_area, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_calc_area},
  {"clip", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5pyFAI_3ext_11_distortion_3clip, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyFAI_3ext_11_distortion_2clip},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__distortion(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__distortion},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_distortion",
    __pyx_k_Distortion_correction_are_corre, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_14_01_2021, __pyx_k_14_01_2021, sizeof(__pyx_k_14_01_2021), 0, 1, 0, 0},
  {&__pyx_kp_u_1_5, __pyx_k_1_5, sizeof(__pyx_k_1_5), 0, 1, 0, 0},
  {&__pyx_kp_u_2011_2021_ESRF, __pyx_k_2011_2021_ESRF, sizeof(__pyx_k_2011_2021_ESRF), 0, 1, 0, 0},
  {&__pyx_kp_u_A0, __pyx_k_A0, sizeof(__pyx_k_A0), 0, 1, 0, 0},
  {&__pyx_n_s_A0_2, __pyx_k_A0_2, sizeof(__pyx_k_A0_2), 0, 0, 1, 1},
  {&__pyx_kp_u_A1, __pyx_k_A1, sizeof(__pyx_k_A1), 0, 1, 0, 0},
  {&__pyx_n_s_A1_2, __pyx_k_A1_2, sizeof(__pyx_k_A1_2), 0, 0, 1, 1},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_u_Accessing_i_i, __pyx_k_Accessing_i_i, sizeof(__pyx_k_Accessing_i_i), 0, 1, 0, 0},
  {&__pyx_kp_u_B0, __pyx_k_B0, sizeof(__pyx_k_B0), 0, 1, 0, 0},
  {&__pyx_n_s_B0_2, __pyx_k_B0_2, sizeof(__pyx_k_B0_2), 0, 0, 1, 1},
  {&__pyx_kp_u_B1, __pyx_k_B1, sizeof(__pyx_k_B1), 0, 1, 0, 0},
  {&__pyx_n_s_B1_2, __pyx_k_B1_2, sizeof(__pyx_k_B1_2), 0, 0, 1, 1},
  {&__pyx_n_s_BIG, __pyx_k_BIG, sizeof(__pyx_k_BIG), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_u_C0, __pyx_k_C0, sizeof(__pyx_k_C0), 0, 1, 0, 0},
  {&__pyx_n_s_C0_2, __pyx_k_C0_2, sizeof(__pyx_k_C0_2), 0, 0, 1, 1},
  {&__pyx_kp_u_C1, __pyx_k_C1, sizeof(__pyx_k_C1), 0, 1, 0, 0},
  {&__pyx_n_s_C1_2, __pyx_k_C1_2, sizeof(__pyx_k_C1_2), 0, 0, 1, 1},
  {&__pyx_kp_u_CSR_matrix_3f_MByte, __pyx_k_CSR_matrix_3f_MByte, sizeof(__pyx_k_CSR_matrix_3f_MByte), 0, 1, 0, 0},
  {&__pyx_kp_u_CSR_matrix_3f_MByte_Max_source_p, __pyx_k_CSR_matrix_3f_MByte_Max_source_p, sizeof(__pyx_k_CSR_matrix_3f_MByte_Max_source_p), 0, 1, 0, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_kp_u_D0, __pyx_k_D0, sizeof(__pyx_k_D0), 0, 1, 0, 0},
  {&__pyx_n_s_D0_2, __pyx_k_D0_2, sizeof(__pyx_k_D0_2), 0, 0, 1, 1},
  {&__pyx_kp_u_D1, __pyx_k_D1, sizeof(__pyx_k_D1), 0, 1, 0, 0},
  {&__pyx_n_s_D1_2, __pyx_k_D1_2, sizeof(__pyx_k_D1_2), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion, __pyx_k_Distortion, sizeof(__pyx_k_Distortion), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion___init, __pyx_k_Distortion___init, sizeof(__pyx_k_Distortion___init), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion___repr, __pyx_k_Distortion___repr, sizeof(__pyx_k_Distortion___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion_calc_LUT, __pyx_k_Distortion_calc_LUT, sizeof(__pyx_k_Distortion_calc_LUT), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion_calc_LUT_size, __pyx_k_Distortion_calc_LUT_size, sizeof(__pyx_k_Distortion_calc_LUT_size), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion_calc_pos, __pyx_k_Distortion_calc_pos, sizeof(__pyx_k_Distortion_calc_pos), 0, 0, 1, 1},
  {&__pyx_n_s_Distortion_correct, __pyx_k_Distortion_correct, sizeof(__pyx_k_Distortion_correct), 0, 0, 1, 1},
  {&__pyx_kp_u_Distortion_correction_for_detect, __pyx_k_Distortion_correction_for_detect, sizeof(__pyx_k_Distortion_correction_for_detect), 0, 1, 0, 0},
  {&__pyx_n_s_Distortion_uncorrect, __pyx_k_Distortion_uncorrect, sizeof(__pyx_k_Distortion_uncorrect), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_I1, __pyx_k_I1, sizeof(__pyx_k_I1), 0, 0, 1, 1},
  {&__pyx_n_s_I2, __pyx_k_I2, sizeof(__pyx_k_I2), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_Integrate1dtpl, __pyx_k_Integrate1dtpl, sizeof(__pyx_k_Integrate1dtpl), 0, 0, 1, 1},
  {&__pyx_n_s_Integrate2dtpl, __pyx_k_Integrate2dtpl, sizeof(__pyx_k_Integrate2dtpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_kp_u_Jerome_Kieffer, __pyx_k_Jerome_Kieffer, sizeof(__pyx_k_Jerome_Kieffer), 0, 1, 0, 0},
  {&__pyx_kp_u_Jerome_kieffer_esrf_fr, __pyx_k_Jerome_kieffer_esrf_fr, sizeof(__pyx_k_Jerome_kieffer_esrf_fr), 0, 1, 0, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_LUT, __pyx_k_LUT, sizeof(__pyx_k_LUT), 0, 0, 1, 1},
  {&__pyx_n_s_LUT_ITEMSIZE, __pyx_k_LUT_ITEMSIZE, sizeof(__pyx_k_LUT_ITEMSIZE), 0, 0, 1, 1},
  {&__pyx_kp_u_LUT_matrix_3f_MByte_Max_source_p, __pyx_k_LUT_matrix_3f_MByte_Max_source_p, sizeof(__pyx_k_LUT_matrix_3f_MByte_Max_source_p), 0, 1, 0, 0},
  {&__pyx_kp_u_LUT_shape_i_i_i_3f_MByte, __pyx_k_LUT_shape_i_i_i_3f_MByte, sizeof(__pyx_k_LUT_shape_i_i_i_3f_MByte), 0, 1, 0, 0},
  {&__pyx_n_u_MIT, __pyx_k_MIT, sizeof(__pyx_k_MIT), 0, 1, 0, 1},
  {&__pyx_kp_u_Max_pixel_size_ix_i_Max_source_p, __pyx_k_Max_pixel_size_ix_i_Max_source_p, sizeof(__pyx_k_Max_pixel_size_ix_i_Max_source_p), 0, 1, 0, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_kp_u_Mismatch_between_shape_of_detect, __pyx_k_Mismatch_between_shape_of_detect, sizeof(__pyx_k_Mismatch_between_shape_of_detect), 0, 1, 0, 0},
  {&__pyx_n_s_NaN, __pyx_k_NaN, sizeof(__pyx_k_NaN), 0, 0, 1, 1},
  {&__pyx_n_s_Nan, __pyx_k_Nan, sizeof(__pyx_k_Nan), 0, 0, 1, 1},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at, __pyx_k_No_value_specified_for_struct_at, sizeof(__pyx_k_No_value_specified_for_struct_at), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_2, __pyx_k_No_value_specified_for_struct_at_2, sizeof(__pyx_k_No_value_specified_for_struct_at_2), 0, 0, 1, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_kp_u_Patching_image_as_image_is_ix_i, __pyx_k_Patching_image_as_image_is_ix_i, sizeof(__pyx_k_Patching_image_as_image_is_ix_i), 0, 1, 0, 0},
  {&__pyx_kp_u_Patching_image_of_shape_ix_i_on, __pyx_k_Patching_image_of_shape_ix_i_on, sizeof(__pyx_k_Patching_image_of_shape_ix_i_on), 0, 1, 0, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_kp_u_Pixel_size_cannot_be_null_Zero_d, __pyx_k_Pixel_size_cannot_be_null_Zero_d, sizeof(__pyx_k_Pixel_size_cannot_be_null_Zero_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Please_increase_bins_per_pixel_s, __pyx_k_Please_increase_bins_per_pixel_s, sizeof(__pyx_k_Please_increase_bins_per_pixel_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Provided_temporary_space_for_sto, __pyx_k_Provided_temporary_space_for_sto, sizeof(__pyx_k_Provided_temporary_space_for_sto), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_Semaphore, __pyx_k_Semaphore, sizeof(__pyx_k_Semaphore), 0, 0, 1, 1},
  {&__pyx_kp_u_Temporary_storage_3fMB, __pyx_k_Temporary_storage_3fMB, sizeof(__pyx_k_Temporary_storage_3fMB), 0, 1, 0, 0},
  {&__pyx_kp_u_The_look_up_table_has_dimension, __pyx_k_The_look_up_table_has_dimension, sizeof(__pyx_k_The_look_up_table_has_dimension), 0, 1, 0, 0},
  {&__pyx_kp_u_The_suggested_value_is_i_or_grea, __pyx_k_The_suggested_value_is_i_or_grea, sizeof(__pyx_k_The_suggested_value_is_i_or_grea), 0, 1, 0, 0},
  {&__pyx_kp_s_This_class_applies_a_distortion, __pyx_k_This_class_applies_a_distortion, sizeof(__pyx_k_This_class_applies_a_distortion), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_kp_u_Unimplemented_sparse_matrix_form, __pyx_k_Unimplemented_sparse_matrix_form, sizeof(__pyx_k_Unimplemented_sparse_matrix_form), 0, 1, 0, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_u__64, __pyx_k__64, sizeof(__pyx_k__64), 0, 1, 0, 0},
  {&__pyx_n_s_acc_d, __pyx_k_acc_d, sizeof(__pyx_k_acc_d), 0, 0, 1, 1},
  {&__pyx_n_s_all_max0, __pyx_k_all_max0, sizeof(__pyx_k_all_max0), 0, 0, 1, 1},
  {&__pyx_n_s_all_max1, __pyx_k_all_max1, sizeof(__pyx_k_all_max1), 0, 0, 1, 1},
  {&__pyx_n_s_all_min0, __pyx_k_all_min0, sizeof(__pyx_k_all_min0), 0, 0, 1, 1},
  {&__pyx_n_s_all_min1, __pyx_k_all_min1, sizeof(__pyx_k_all_min1), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_kp_u_and_shape_of_mask, __pyx_k_and_shape_of_mask, sizeof(__pyx_k_and_shape_of_mask), 0, 1, 0, 0},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_area, __pyx_k_area, sizeof(__pyx_k_area), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_ascontiguousarray, __pyx_k_ascontiguousarray, sizeof(__pyx_k_ascontiguousarray), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_author, __pyx_k_author, sizeof(__pyx_k_author), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_bin_number, __pyx_k_bin_number, sizeof(__pyx_k_bin_number), 0, 0, 1, 1},
  {&__pyx_n_s_bin_size, __pyx_k_bin_size, sizeof(__pyx_k_bin_size), 0, 0, 1, 1},
  {&__pyx_n_s_bins, __pyx_k_bins, sizeof(__pyx_k_bins), 0, 0, 1, 1},
  {&__pyx_n_s_bins_per_pixel, __pyx_k_bins_per_pixel, sizeof(__pyx_k_bins_per_pixel), 0, 0, 1, 1},
  {&__pyx_n_s_block_size, __pyx_k_block_size, sizeof(__pyx_k_block_size), 0, 0, 1, 1},
  {&__pyx_n_s_box_size0, __pyx_k_box_size0, sizeof(__pyx_k_box_size0), 0, 0, 1, 1},
  {&__pyx_n_s_box_size1, __pyx_k_box_size1, sizeof(__pyx_k_box_size1), 0, 0, 1, 1},
  {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_builder, __pyx_k_builder, sizeof(__pyx_k_builder), 0, 0, 1, 1},
  {&__pyx_n_s_builder_config, __pyx_k_builder_config, sizeof(__pyx_k_builder_config), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_cAB, __pyx_k_cAB, sizeof(__pyx_k_cAB), 0, 0, 1, 1},
  {&__pyx_n_s_cBC, __pyx_k_cBC, sizeof(__pyx_k_cBC), 0, 0, 1, 1},
  {&__pyx_n_s_cCD, __pyx_k_cCD, sizeof(__pyx_k_cCD), 0, 0, 1, 1},
  {&__pyx_n_s_cDA, __pyx_k_cDA, sizeof(__pyx_k_cDA), 0, 0, 1, 1},
  {&__pyx_n_s_calc_CSR, __pyx_k_calc_CSR, sizeof(__pyx_k_calc_CSR), 0, 0, 1, 1},
  {&__pyx_n_s_calc_LUT, __pyx_k_calc_LUT, sizeof(__pyx_k_calc_LUT), 0, 0, 1, 1},
  {&__pyx_n_s_calc_LUT_size, __pyx_k_calc_LUT_size, sizeof(__pyx_k_calc_LUT_size), 0, 0, 1, 1},
  {&__pyx_n_s_calc_cartesian_positions, __pyx_k_calc_cartesian_positions, sizeof(__pyx_k_calc_cartesian_positions), 0, 0, 1, 1},
  {&__pyx_n_s_calc_pos, __pyx_k_calc_pos, sizeof(__pyx_k_calc_pos), 0, 0, 1, 1},
  {&__pyx_n_s_calc_size, __pyx_k_calc_size, sizeof(__pyx_k_calc_size), 0, 0, 1, 1},
  {&__pyx_n_s_calc_sparse, __pyx_k_calc_sparse, sizeof(__pyx_k_calc_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_calc_sparse_v2, __pyx_k_calc_sparse_v2, sizeof(__pyx_k_calc_sparse_v2), 0, 0, 1, 1},
  {&__pyx_n_s_cdelta_dummy, __pyx_k_cdelta_dummy, sizeof(__pyx_k_cdelta_dummy), 0, 0, 1, 1},
  {&__pyx_n_s_cdummy, __pyx_k_cdummy, sizeof(__pyx_k_cdummy), 0, 0, 1, 1},
  {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clip, __pyx_k_clip, sizeof(__pyx_k_clip), 0, 0, 1, 1},
  {&__pyx_n_s_cmask, __pyx_k_cmask, sizeof(__pyx_k_cmask), 0, 0, 1, 1},
  {&__pyx_n_b_coef, __pyx_k_coef, sizeof(__pyx_k_coef), 0, 0, 0, 1},
  {&__pyx_n_s_coef, __pyx_k_coef, sizeof(__pyx_k_coef), 0, 0, 1, 1},
  {&__pyx_n_u_coef, __pyx_k_coef, sizeof(__pyx_k_coef), 0, 1, 0, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
  {&__pyx_n_s_contact, __pyx_k_contact, sizeof(__pyx_k_contact), 0, 0, 1, 1},
  {&__pyx_n_s_containers, __pyx_k_containers, sizeof(__pyx_k_containers), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_copyright, __pyx_k_copyright, sizeof(__pyx_k_copyright), 0, 0, 1, 1},
  {&__pyx_n_s_core, __pyx_k_core, sizeof(__pyx_k_core), 0, 0, 1, 1},
  {&__pyx_n_s_correct, __pyx_k_correct, sizeof(__pyx_k_correct), 0, 0, 1, 1},
  {&__pyx_n_s_correct_CSR, __pyx_k_correct_CSR, sizeof(__pyx_k_correct_CSR), 0, 0, 1, 1},
  {&__pyx_n_s_correct_CSR_double, __pyx_k_correct_CSR_double, sizeof(__pyx_k_correct_CSR_double), 0, 0, 1, 1},
  {&__pyx_n_s_correct_CSR_kahan, __pyx_k_correct_CSR_kahan, sizeof(__pyx_k_correct_CSR_kahan), 0, 0, 1, 1},
  {&__pyx_n_s_correct_CSR_preproc_double, __pyx_k_correct_CSR_preproc_double, sizeof(__pyx_k_correct_CSR_preproc_double), 0, 0, 1, 1},
  {&__pyx_n_s_correct_LUT, __pyx_k_correct_LUT, sizeof(__pyx_k_correct_LUT), 0, 0, 1, 1},
  {&__pyx_n_s_correct_LUT_double, __pyx_k_correct_LUT_double, sizeof(__pyx_k_correct_LUT_double), 0, 0, 1, 1},
  {&__pyx_n_s_correct_LUT_kahan, __pyx_k_correct_LUT_kahan, sizeof(__pyx_k_correct_LUT_kahan), 0, 0, 1, 1},
  {&__pyx_n_s_correct_LUT_preproc_double, __pyx_k_correct_LUT_preproc_double, sizeof(__pyx_k_correct_LUT_preproc_double), 0, 0, 1, 1},
  {&__pyx_n_b_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 0, 1},
  {&__pyx_n_u_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 1, 0, 1},
  {&__pyx_n_s_counter, __pyx_k_counter, sizeof(__pyx_k_counter), 0, 0, 1, 1},
  {&__pyx_n_u_csr, __pyx_k_csr, sizeof(__pyx_k_csr), 0, 1, 0, 1},
  {&__pyx_n_s_cumsum, __pyx_k_cumsum, sizeof(__pyx_k_cumsum), 0, 0, 1, 1},
  {&__pyx_n_s_d1, __pyx_k_d1, sizeof(__pyx_k_d1), 0, 0, 1, 1},
  {&__pyx_n_s_d2, __pyx_k_d2, sizeof(__pyx_k_d2), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_d, __pyx_k_data_d, sizeof(__pyx_k_data_d), 0, 0, 1, 1},
  {&__pyx_n_s_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_delta0, __pyx_k_delta0, sizeof(__pyx_k_delta0), 0, 0, 1, 1},
  {&__pyx_n_s_delta1, __pyx_k_delta1, sizeof(__pyx_k_delta1), 0, 0, 1, 1},
  {&__pyx_n_s_delta_dummy, __pyx_k_delta_dummy, sizeof(__pyx_k_delta_dummy), 0, 0, 1, 1},
  {&__pyx_n_s_detector, __pyx_k_detector, sizeof(__pyx_k_detector), 0, 0, 1, 1},
  {&__pyx_n_u_detector, __pyx_k_detector, sizeof(__pyx_k_detector), 0, 1, 0, 1},
  {&__pyx_n_s_detector_factory, __pyx_k_detector_factory, sizeof(__pyx_k_detector_factory), 0, 0, 1, 1},
  {&__pyx_n_s_detectors, __pyx_k_detectors, sizeof(__pyx_k_detectors), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dim0, __pyx_k_dim0, sizeof(__pyx_k_dim0), 0, 0, 1, 1},
  {&__pyx_n_s_dim1, __pyx_k_dim1, sizeof(__pyx_k_dim1), 0, 0, 1, 1},
  {&__pyx_n_s_do_dummy, __pyx_k_do_dummy, sizeof(__pyx_k_do_dummy), 0, 0, 1, 1},
  {&__pyx_n_s_do_mask, __pyx_k_do_mask, sizeof(__pyx_k_do_mask), 0, 0, 1, 1},
  {&__pyx_n_s_do_shape, __pyx_k_do_shape, sizeof(__pyx_k_do_shape), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
  {&__pyx_n_u_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 1, 0, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_dummy, __pyx_k_dummy, sizeof(__pyx_k_dummy), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eps, __pyx_k_eps, sizeof(__pyx_k_eps), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_err_cnt, __pyx_k_err_cnt, sizeof(__pyx_k_err_cnt), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_expand2d, __pyx_k_expand2d, sizeof(__pyx_k_expand2d), 0, 0, 1, 1},
  {&__pyx_n_s_fabio, __pyx_k_fabio, sizeof(__pyx_k_fabio), 0, 0, 1, 1},
  {&__pyx_n_s_finfo, __pyx_k_finfo, sizeof(__pyx_k_finfo), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
  {&__pyx_n_s_foffset0, __pyx_k_foffset0, sizeof(__pyx_k_foffset0), 0, 0, 1, 1},
  {&__pyx_n_s_foffset1, __pyx_k_foffset1, sizeof(__pyx_k_foffset1), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_heap_size, __pyx_k_heap_size, sizeof(__pyx_k_heap_size), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_b_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 0, 1},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_u_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 1, 0, 1},
  {&__pyx_n_s_idx_bin, __pyx_k_idx_bin, sizeof(__pyx_k_idx_bin), 0, 0, 1, 1},
  {&__pyx_n_s_idx_pixel, __pyx_k_idx_pixel, sizeof(__pyx_k_idx_pixel), 0, 0, 1, 1},
  {&__pyx_n_u_if, __pyx_k_if, sizeof(__pyx_k_if), 0, 1, 0, 1},
  {&__pyx_n_s_image, __pyx_k_image, sizeof(__pyx_k_image), 0, 0, 1, 1},
  {&__pyx_kp_u_image_is_2D, __pyx_k_image_is_2D, sizeof(__pyx_k_image_is_2D), 0, 1, 0, 0},
  {&__pyx_kp_u_image_is_3D, __pyx_k_image_is_3D, sizeof(__pyx_k_image_is_3D), 0, 1, 0, 0},
  {&__pyx_n_s_img_shape, __pyx_k_img_shape, sizeof(__pyx_k_img_shape), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index_d, __pyx_k_index_d, sizeof(__pyx_k_index_d), 0, 0, 1, 1},
  {&__pyx_n_s_indices, __pyx_k_indices, sizeof(__pyx_k_indices), 0, 0, 1, 1},
  {&__pyx_n_s_indptr, __pyx_k_indptr, sizeof(__pyx_k_indptr), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_int8, __pyx_k_int8, sizeof(__pyx_k_int8), 0, 0, 1, 1},
  {&__pyx_n_s_intercept, __pyx_k_intercept, sizeof(__pyx_k_intercept), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_kp_u_jerome_kieffer_esrf_fr, __pyx_k_jerome_kieffer_esrf_fr, sizeof(__pyx_k_jerome_kieffer_esrf_fr), 0, 1, 0, 0},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_u_kahan, __pyx_k_kahan, sizeof(__pyx_k_kahan), 0, 1, 0, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_large_data, __pyx_k_large_data, sizeof(__pyx_k_large_data), 0, 0, 1, 1},
  {&__pyx_n_s_large_size, __pyx_k_large_size, sizeof(__pyx_k_large_size), 0, 0, 1, 1},
  {&__pyx_n_s_lerr, __pyx_k_lerr, sizeof(__pyx_k_lerr), 0, 0, 1, 1},
  {&__pyx_n_s_license, __pyx_k_license, sizeof(__pyx_k_license), 0, 0, 1, 1},
  {&__pyx_n_s_lin, __pyx_k_lin, sizeof(__pyx_k_lin), 0, 0, 1, 1},
  {&__pyx_n_s_linesep, __pyx_k_linesep, sizeof(__pyx_k_linesep), 0, 0, 1, 1},
  {&__pyx_n_s_lmask, __pyx_k_lmask, sizeof(__pyx_k_lmask), 0, 0, 1, 1},
  {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_s_lout, __pyx_k_lout, sizeof(__pyx_k_lout), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_lprop, __pyx_k_lprop, sizeof(__pyx_k_lprop), 0, 0, 1, 1},
  {&__pyx_n_s_lut, __pyx_k_lut, sizeof(__pyx_k_lut), 0, 0, 1, 1},
  {&__pyx_n_u_lut, __pyx_k_lut, sizeof(__pyx_k_lut), 0, 1, 0, 1},
  {&__pyx_n_s_lut_d, __pyx_k_lut_d, sizeof(__pyx_k_lut_d), 0, 0, 1, 1},
  {&__pyx_n_s_lut_size, __pyx_k_lut_size, sizeof(__pyx_k_lut_size), 0, 0, 1, 1},
  {&__pyx_n_s_lut_total_size, __pyx_k_lut_total_size, sizeof(__pyx_k_lut_total_size), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_major, __pyx_k_major, sizeof(__pyx_k_major), 0, 0, 1, 1},
  {&__pyx_kp_u_makes_little_sense_to_correct_fo, __pyx_k_makes_little_sense_to_correct_fo, sizeof(__pyx_k_makes_little_sense_to_correct_fo), 0, 1, 0, 0},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask_d, __pyx_k_mask_d, sizeof(__pyx_k_mask_d), 0, 0, 1, 1},
  {&__pyx_kp_u_mask_shape_dim0, __pyx_k_mask_shape_dim0, sizeof(__pyx_k_mask_shape_dim0), 0, 1, 0, 0},
  {&__pyx_kp_u_mask_shape_dim1, __pyx_k_mask_shape_dim1, sizeof(__pyx_k_mask_shape_dim1), 0, 1, 0, 0},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max0, __pyx_k_max0, sizeof(__pyx_k_max0), 0, 0, 1, 1},
  {&__pyx_n_s_max1, __pyx_k_max1, sizeof(__pyx_k_max1), 0, 0, 1, 1},
  {&__pyx_n_s_max_pixel_size, __pyx_k_max_pixel_size, sizeof(__pyx_k_max_pixel_size), 0, 0, 1, 1},
  {&__pyx_n_s_max_shape, __pyx_k_max_shape, sizeof(__pyx_k_max_shape), 0, 0, 1, 1},
  {&__pyx_n_u_max_shape, __pyx_k_max_shape, sizeof(__pyx_k_max_shape), 0, 1, 0, 1},
  {&__pyx_n_s_max_val, __pyx_k_max_val, sizeof(__pyx_k_max_val), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_min0, __pyx_k_min0, sizeof(__pyx_k_min0), 0, 0, 1, 1},
  {&__pyx_n_s_min1, __pyx_k_min1, sizeof(__pyx_k_min1), 0, 0, 1, 1},
  {&__pyx_n_s_min_val, __pyx_k_min_val, sizeof(__pyx_k_min_val), 0, 0, 1, 1},
  {&__pyx_n_s_ml, __pyx_k_ml, sizeof(__pyx_k_ml), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_ms, __pyx_k_ms, sizeof(__pyx_k_ms), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
  {&__pyx_n_s_nb_corners, __pyx_k_nb_corners, sizeof(__pyx_k_nb_corners), 0, 0, 1, 1},
  {&__pyx_n_s_nbins, __pyx_k_nbins, sizeof(__pyx_k_nbins), 0, 0, 1, 1},
  {&__pyx_n_s_nbytes, __pyx_k_nbytes, sizeof(__pyx_k_nbytes), 0, 0, 1, 1},
  {&__pyx_n_s_nchan, __pyx_k_nchan, sizeof(__pyx_k_nchan), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_new_image, __pyx_k_new_image, sizeof(__pyx_k_new_image), 0, 0, 1, 1},
  {&__pyx_n_s_nl, __pyx_k_nl, sizeof(__pyx_k_nl), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_b_norm, __pyx_k_norm, sizeof(__pyx_k_norm), 0, 0, 0, 1},
  {&__pyx_n_u_norm, __pyx_k_norm, sizeof(__pyx_k_norm), 0, 1, 0, 1},
  {&__pyx_n_s_np_lut_size, __pyx_k_np_lut_size, sizeof(__pyx_k_np_lut_size), 0, 0, 1, 1},
  {&__pyx_n_s_ns, __pyx_k_ns, sizeof(__pyx_k_ns), 0, 0, 1, 1},
  {&__pyx_kp_u_number_of_elements_s_average_per, __pyx_k_number_of_elements_s_average_per, sizeof(__pyx_k_number_of_elements_s_average_per), 0, 1, 0, 0},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_numpy_1_12_py2_bug, __pyx_k_numpy_1_12_py2_bug, sizeof(__pyx_k_numpy_1_12_py2_bug), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_offset0, __pyx_k_offset0, sizeof(__pyx_k_offset0), 0, 0, 1, 1},
  {&__pyx_n_s_offset1, __pyx_k_offset1, sizeof(__pyx_k_offset1), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_outMax, __pyx_k_outMax, sizeof(__pyx_k_outMax), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_p0, __pyx_k_p0, sizeof(__pyx_k_p0), 0, 0, 1, 1},
  {&__pyx_n_s_p1, __pyx_k_p1, sizeof(__pyx_k_p1), 0, 0, 1, 1},
  {&__pyx_n_s_pAB, __pyx_k_pAB, sizeof(__pyx_k_pAB), 0, 0, 1, 1},
  {&__pyx_n_s_pBC, __pyx_k_pBC, sizeof(__pyx_k_pBC), 0, 0, 1, 1},
  {&__pyx_n_s_pCD, __pyx_k_pCD, sizeof(__pyx_k_pCD), 0, 0, 1, 1},
  {&__pyx_n_s_pDA, __pyx_k_pDA, sizeof(__pyx_k_pDA), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pixel1, __pyx_k_pixel1, sizeof(__pyx_k_pixel1), 0, 0, 1, 1},
  {&__pyx_n_s_pixel2, __pyx_k_pixel2, sizeof(__pyx_k_pixel2), 0, 0, 1, 1},
  {&__pyx_n_s_pixel_corners, __pyx_k_pixel_corners, sizeof(__pyx_k_pixel_corners), 0, 0, 1, 1},
  {&__pyx_n_s_pixel_count, __pyx_k_pixel_count, sizeof(__pyx_k_pixel_count), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_pos0max, __pyx_k_pos0max, sizeof(__pyx_k_pos0max), 0, 0, 1, 1},
  {&__pyx_n_s_pos0min, __pyx_k_pos0min, sizeof(__pyx_k_pos0min), 0, 0, 1, 1},
  {&__pyx_n_s_pos1max, __pyx_k_pos1max, sizeof(__pyx_k_pos1max), 0, 0, 1, 1},
  {&__pyx_n_s_pos1min, __pyx_k_pos1min, sizeof(__pyx_k_pos1min), 0, 0, 1, 1},
  {&__pyx_n_s_pos_corners, __pyx_k_pos_corners, sizeof(__pyx_k_pos_corners), 0, 0, 1, 1},
  {&__pyx_n_s_position_d, __pyx_k_position_d, sizeof(__pyx_k_position_d), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_preprocessed_data, __pyx_k_preprocessed_data, sizeof(__pyx_k_preprocessed_data), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_prod, __pyx_k_prod, sizeof(__pyx_k_prod), 0, 0, 1, 1},
  {&__pyx_n_s_prop, __pyx_k_prop, sizeof(__pyx_k_prop), 0, 0, 1, 1},
  {&__pyx_n_s_prop_d, __pyx_k_prop_d, sizeof(__pyx_k_prop_d), 0, 0, 1, 1},
  {&__pyx_n_s_pyFAI_ext__distortion, __pyx_k_pyFAI_ext__distortion, sizeof(__pyx_k_pyFAI_ext__distortion), 0, 0, 1, 1},
  {&__pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_k_pyFAI_ext__distortion_pyx, sizeof(__pyx_k_pyFAI_ext__distortion_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ravel, __pyx_k_ravel, sizeof(__pyx_k_ravel), 0, 0, 1, 1},
  {&__pyx_n_s_recarray, __pyx_k_recarray, sizeof(__pyx_k_recarray), 0, 0, 1, 1},
  {&__pyx_n_s_records, __pyx_k_records, sizeof(__pyx_k_records), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_resize_image_2D, __pyx_k_resize_image_2D, sizeof(__pyx_k_resize_image_2D), 0, 0, 1, 1},
  {&__pyx_n_s_resize_image_3D, __pyx_k_resize_image_3D, sizeof(__pyx_k_resize_image_3D), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_sem, __pyx_k_sem, sizeof(__pyx_k_sem), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_shape0, __pyx_k_shape0, sizeof(__pyx_k_shape0), 0, 0, 1, 1},
  {&__pyx_n_s_shape1, __pyx_k_shape1, sizeof(__pyx_k_shape1), 0, 0, 1, 1},
  {&__pyx_n_s_shape_img0, __pyx_k_shape_img0, sizeof(__pyx_k_shape_img0), 0, 0, 1, 1},
  {&__pyx_n_s_shape_img1, __pyx_k_shape_img1, sizeof(__pyx_k_shape_img1), 0, 0, 1, 1},
  {&__pyx_n_s_shape_in, __pyx_k_shape_in, sizeof(__pyx_k_shape_in), 0, 0, 1, 1},
  {&__pyx_n_s_shape_in0, __pyx_k_shape_in0, sizeof(__pyx_k_shape_in0), 0, 0, 1, 1},
  {&__pyx_kp_u_shape_in0_mask_shape_0, __pyx_k_shape_in0_mask_shape_0, sizeof(__pyx_k_shape_in0_mask_shape_0), 0, 1, 0, 0},
  {&__pyx_n_s_shape_in1, __pyx_k_shape_in1, sizeof(__pyx_k_shape_in1), 0, 0, 1, 1},
  {&__pyx_kp_u_shape_in1_mask_shape_1, __pyx_k_shape_in1_mask_shape_1, sizeof(__pyx_k_shape_in1_mask_shape_1), 0, 1, 0, 0},
  {&__pyx_n_s_shape_out, __pyx_k_shape_out, sizeof(__pyx_k_shape_out), 0, 0, 1, 1},
  {&__pyx_n_s_shape_out0, __pyx_k_shape_out0, sizeof(__pyx_k_shape_out0), 0, 0, 1, 1},
  {&__pyx_kp_u_shape_out0_shape_out1_LUT_shape, __pyx_k_shape_out0_shape_out1_LUT_shape, sizeof(__pyx_k_shape_out0_shape_out1_LUT_shape), 0, 1, 0, 0},
  {&__pyx_kp_u_shape_out0_shape_out1_indptr_siz, __pyx_k_shape_out0_shape_out1_indptr_siz, sizeof(__pyx_k_shape_out0_shape_out1_indptr_siz), 0, 1, 0, 0},
  {&__pyx_n_s_shape_out1, __pyx_k_shape_out1, sizeof(__pyx_k_shape_out1), 0, 0, 1, 1},
  {&__pyx_n_b_signal, __pyx_k_signal, sizeof(__pyx_k_signal), 0, 0, 0, 1},
  {&__pyx_n_u_signal, __pyx_k_signal, sizeof(__pyx_k_signal), 0, 1, 0, 1},
  {&__pyx_n_s_signatures, __pyx_k_signatures, sizeof(__pyx_k_signatures), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_size_in, __pyx_k_size_in, sizeof(__pyx_k_size_in), 0, 0, 1, 1},
  {&__pyx_n_s_slope, __pyx_k_slope, sizeof(__pyx_k_slope), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_u_stable, __pyx_k_stable, sizeof(__pyx_k_stable), 0, 1, 0, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_sum_norm, __pyx_k_sum_norm, sizeof(__pyx_k_sum_norm), 0, 0, 1, 1},
  {&__pyx_n_s_sum_sig, __pyx_k_sum_sig, sizeof(__pyx_k_sum_sig), 0, 0, 1, 1},
  {&__pyx_n_s_sum_var, __pyx_k_sum_var, sizeof(__pyx_k_sum_var), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_ary, __pyx_k_tmp_ary, sizeof(__pyx_k_tmp_ary), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_index, __pyx_k_tmp_index, sizeof(__pyx_k_tmp_index), 0, 0, 1, 1},
  {&__pyx_n_s_to_csr, __pyx_k_to_csr, sizeof(__pyx_k_to_csr), 0, 0, 1, 1},
  {&__pyx_n_s_tot, __pyx_k_tot, sizeof(__pyx_k_tot), 0, 0, 1, 1},
  {&__pyx_n_s_total, __pyx_k_total, sizeof(__pyx_k_total), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_uncorrect, __pyx_k_uncorrect, sizeof(__pyx_k_uncorrect), 0, 0, 1, 1},
  {&__pyx_n_s_uncorrect_CSR, __pyx_k_uncorrect_CSR, sizeof(__pyx_k_uncorrect_CSR), 0, 0, 1, 1},
  {&__pyx_n_s_uncorrect_LUT, __pyx_k_uncorrect_LUT, sizeof(__pyx_k_uncorrect_LUT), 0, 0, 1, 1},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_utils, __pyx_k_utils, sizeof(__pyx_k_utils), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_b_variance, __pyx_k_variance, sizeof(__pyx_k_variance), 0, 0, 0, 1},
  {&__pyx_n_s_variance, __pyx_k_variance, sizeof(__pyx_k_variance), 0, 0, 1, 1},
  {&__pyx_n_u_variance, __pyx_k_variance, sizeof(__pyx_k_variance), 0, 1, 0, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
  {&__pyx_n_s_view, __pyx_k_view, sizeof(__pyx_k_view), 0, 0, 1, 1},
  {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 1653, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 969, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 148, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 151, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 613, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(1, 832, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "pyFAI/ext/_distortion.pyx":264
 * 
 *     if (pixel1 == 0.0) or (pixel2 == 0.0):
 *         raise RuntimeError("Pixel size cannot be null -> Zero division error")             # <<<<<<<<<<<<<<
 * 
 *     dim0 = pixel_corners.shape[0]
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_Pixel_size_cannot_be_null_Zero_d); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pyFAI/ext/_distortion.pyx":302
 *     res = numpy.asarray(pos), int(delta0), int(delta1), \
 *         (int(ceil(all_max0 - all_min0)), int(ceil(all_max1 - all_min1))) if do_shape else shape_out, \
 *         (float(all_min0), float(all_min1)) if do_shape else (0.0, 0.0)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "pyFAI/ext/_distortion.pyx":392
 *     #buffer_nbytes = buffer.nbytes
 *     if (size == 0):  # fix 271
 *         raise RuntimeError("The look-up table has dimension 0 which is a non-sense." +             # <<<<<<<<<<<<<<
 *                            " Did you mask out all pixel or is your image out of the geometry range?")
 *     lut = view.array(shape=(shape0, shape1, size), itemsize=sizeof(lut_t), format="if")
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_The_look_up_table_has_dimension); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "pyFAI/ext/_distortion.pyx":632
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,
 *                 shape,
 *                 max_pixel_size=(8, 8),             # <<<<<<<<<<<<<<
 *                 int8_t[:, ::1] mask=None,
 *                 format="csr",
 */
  __pyx_tuple__15 = PyTuple_Pack(2, __pyx_int_8, __pyx_int_8); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "pyFAI/ext/_distortion.pyx":969
 *         res = builder.to_csr()
 *     elif format == "lut":
 *         raise NotImplementedError("")             # <<<<<<<<<<<<<<
 *     else:
 *         raise RuntimeError("Unimplemented sparse matrix format: %s", format)
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u__18); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "pyFAI/ext/_distortion.pyx":997
 *             new_image[:shape_img0, :shape_img1] = image
 *         else:
 *             new_image[:shape_img0, :] = image[:, :shape_in1]             # <<<<<<<<<<<<<<
 *     else:
 *         if shape_img1 < shape_in1:
 */
  __pyx_slice__20 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__20)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__20);
  __Pyx_GIVEREF(__pyx_slice__20);

  /* "pyFAI/ext/_distortion.pyx":1002
 *             new_image[:, :shape_img1] = image[:shape_in0, :]
 *         else:
 *             new_image[:, :] = image[:shape_in0, :shape_in1]             # <<<<<<<<<<<<<<
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 */
  __pyx_tuple__21 = PyTuple_Pack(2, __pyx_slice__20, __pyx_slice__20); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "pyFAI/ext/_distortion.pyx":1037
 *             new_image[:, :shape_img1, :] = image[:shape_in0, :, :]
 *         else:
 *             new_image[:, :, :] = image[:shape_in0, :shape_in1, :]             # <<<<<<<<<<<<<<
 *     logger.warning("Patching image of shape %ix%i on expected size of %ix%i",
 *                    shape_img1, shape_img0, shape_in1, shape_in0)
 */
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_slice__20, __pyx_slice__20, __pyx_slice__20); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "pyFAI/ext/_distortion.pyx":1695
 *                         logger.warning("makes little sense to correct for distortion non-flat detectors: %s",
 *                                        self.detector)
 *                     pos_corners[:, :, 0], pos_corners[:, :, 1] = p[:2]             # <<<<<<<<<<<<<<
 *                     pos_corners[:, :, 0] /= self.detector.pixel1
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 */
  __pyx_slice__25 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__25)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__25);
  __Pyx_GIVEREF(__pyx_slice__25);
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_slice__20, __pyx_slice__20, __pyx_int_0); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_tuple__27 = PyTuple_Pack(3, __pyx_slice__20, __pyx_slice__20, __pyx_int_1); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "pyFAI/ext/_distortion.pyx":1699
 *                     pos_corners[:, :, 1] /= self.detector.pixel2
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 */
  __pyx_slice__28 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__28)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__28);
  __Pyx_GIVEREF(__pyx_slice__28);
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_slice__28, __pyx_slice__28); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_tuple__30 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_int_0, __pyx_slice__20); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "pyFAI/ext/_distortion.pyx":1700
 *                     pos = numpy.empty((self.shape[0], self.shape[1], 4, 2), dtype=numpy.float32)
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 */
  __pyx_slice__31 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__31)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__31);
  __Pyx_GIVEREF(__pyx_slice__31);
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_slice__28, __pyx_slice__31); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_tuple__33 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_int_1, __pyx_slice__20); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "pyFAI/ext/_distortion.pyx":1701
 *                     pos[:, :, 0, :] = pos_corners[:-1, :-1]
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]             # <<<<<<<<<<<<<<
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 *                     self.pos = pos
 */
  __pyx_tuple__34 = PyTuple_Pack(2, __pyx_slice__31, __pyx_slice__31); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_tuple__35 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_int_2, __pyx_slice__20); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "pyFAI/ext/_distortion.pyx":1702
 *                     pos[:, :, 1, :] = pos_corners[:-1, 1:]
 *                     pos[:, :, 2, :] = pos_corners[1:, 1:]
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]             # <<<<<<<<<<<<<<
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 */
  __pyx_tuple__36 = PyTuple_Pack(2, __pyx_slice__31, __pyx_slice__28); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 1702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_tuple__37 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_int_3, __pyx_slice__20); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 1702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "pyFAI/ext/_distortion.pyx":1704
 *                     pos[:, :, 3, :] = pos_corners[1:, :-1]
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())             # <<<<<<<<<<<<<<
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())
 *         return self.pos
 */
  __pyx_tuple__38 = PyTuple_Pack(3, __pyx_slice__31, __pyx_slice__20, __pyx_int_0); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_tuple__39 = PyTuple_Pack(3, __pyx_slice__28, __pyx_slice__20, __pyx_int_0); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "pyFAI/ext/_distortion.pyx":1705
 *                     self.pos = pos
 *                     self.delta0 = int((numpy.ceil(pos_corners[1:, :, 0]) - numpy.floor(pos_corners[:-1, :, 0])).max())
 *                     self.delta1 = int((numpy.ceil(pos_corners[:, 1:, 1]) - numpy.floor(pos_corners[:, :-1, 1])).max())             # <<<<<<<<<<<<<<
 *         return self.pos
 * 
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_slice__20, __pyx_slice__31, __pyx_int_1); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 1705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_tuple__41 = PyTuple_Pack(3, __pyx_slice__20, __pyx_slice__28, __pyx_int_1); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 1705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "pyFAI/ext/_distortion.pyx":1686
 *     def calc_pos(self):
 *         if self.pos is None:
 *             with self._sem:             # <<<<<<<<<<<<<<
 *                 if self.pos is None:
 *                     pos_corners = numpy.empty((self.shape[0] + 1, self.shape[1] + 1, 2), dtype=numpy.float64)
 */
  __pyx_tuple__42 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 1686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "pyFAI/ext/_distortion.pyx":1729
 *                 if self.lut_size is None:
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])             # <<<<<<<<<<<<<<
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 */
  __pyx_tuple__43 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_slice__20, __pyx_int_0); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 1729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "pyFAI/ext/_distortion.pyx":1730
 *                     shape0, shape1 = self.shape
 *                     pos0min = numpy.floor(pos[:, :, :, 0].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[0])
 *                     pos1min = numpy.floor(pos[:, :, :, 1].min(axis=-1)).astype(numpy.int32).clip(0, self.shape[1])             # <<<<<<<<<<<<<<
 *                     pos0max = (numpy.ceil(pos[:, :, :, 0].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[0])
 *                     pos1max = (numpy.ceil(pos[:, :, :, 1].max(axis=-1)).astype(numpy.int32) + 1).clip(0, self.shape[1])
 */
  __pyx_tuple__44 = PyTuple_Pack(4, __pyx_slice__20, __pyx_slice__20, __pyx_slice__20, __pyx_int_1); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "FromPyStructUtility":19
 *         value = obj['idx']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'idx'")             # <<<<<<<<<<<<<<
 *     result.idx = value
 *     try:
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "FromPyStructUtility":24
 *         value = obj['coef']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'coef'")             # <<<<<<<<<<<<<<
 *     result.coef = value
 *     return result
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_2); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(1, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(1, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__58 = PyTuple_New(1); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__58, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(1, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "pyFAI/ext/regrid_common.pxi":48
 * 
 * _numpy_1_12_py2_bug = ((sys.version_info.major == 2) and
 *                        ([1, 12] >= [int(i) for i in numpy.version.version.split(".", 2)[:2]]))             # <<<<<<<<<<<<<<
 * 
 * # Imports at the C level
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_kp_u__64, __pyx_int_2); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(4, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "pyFAI/ext/_distortion.pyx":63
 * from .sparse_builder cimport SparseBuilder
 * 
 * cdef bint NEED_DECREF = sys.version_info < (2, 7) and numpy.version.version < "1.5"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__90 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_7); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */
  __pyx_tuple__91 = PyTuple_Pack(26, __pyx_n_s_pixel_corners, __pyx_n_s_pixel1, __pyx_n_s_pixel2, __pyx_n_s_shape_out, __pyx_n_s_pos, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_dim0, __pyx_n_s_dim1, __pyx_n_s_nb_corners, __pyx_n_s_do_shape, __pyx_n_s_BIG, __pyx_n_s_min0, __pyx_n_s_min1, __pyx_n_s_max0, __pyx_n_s_max1, __pyx_n_s_delta0, __pyx_n_s_delta1, __pyx_n_s_all_min0, __pyx_n_s_all_max0, __pyx_n_s_all_max1, __pyx_n_s_all_min1, __pyx_n_s_p0, __pyx_n_s_p1, __pyx_n_s_res); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(4, 0, 26, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_pos, 245, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 245, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */
  __pyx_tuple__93 = PyTuple_Pack(30, __pyx_n_s_pos, __pyx_n_s_shape, __pyx_n_s_mask, __pyx_n_s_offset, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_l, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1, __pyx_n_s_shape_in0, __pyx_n_s_shape_in1, __pyx_n_s_min0, __pyx_n_s_min1, __pyx_n_s_max0, __pyx_n_s_max1, __pyx_n_s_lut_size, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_do_mask, __pyx_n_s_cmask, __pyx_n_s_err); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(4, 0, 30, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_size, 306, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 306, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":362
 * 
 * 
 * def calc_LUT(float32_t[:, :, :, ::1] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, :] mask=None):
 *     """
 */
  __pyx_tuple__95 = PyTuple_Pack(49, __pyx_n_s_pos, __pyx_n_s_shape, __pyx_n_s_bin_size, __pyx_n_s_max_pixel_size, __pyx_n_s_mask, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_ms, __pyx_n_s_ml, __pyx_n_s_ns, __pyx_n_s_nl, __pyx_n_s_shape0, __pyx_n_s_shape1, __pyx_n_s_delta0, __pyx_n_s_delta1, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_box_size0, __pyx_n_s_box_size1, __pyx_n_s_size, __pyx_n_s_k, __pyx_n_s_idx, __pyx_n_s_err_cnt, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_pAB, __pyx_n_s_pBC, __pyx_n_s_pCD, __pyx_n_s_pDA, __pyx_n_s_cAB, __pyx_n_s_cBC, __pyx_n_s_cCD, __pyx_n_s_cDA, __pyx_n_s_area, __pyx_n_s_value, __pyx_n_s_foffset0, __pyx_n_s_foffset1, __pyx_n_s_lut, __pyx_n_s_do_mask, __pyx_n_s_buffer, __pyx_n_s_outMax, __pyx_n_s_lut_total_size, __pyx_n_s_tmp_ary); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(5, 0, 49, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_LUT, 362, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 362, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":500
 * 
 * 
 * def calc_CSR(float32_t[:, :, :, :] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, ::1] mask=None):
 *     """Calculate the Look-up table as CSR format
 */
  __pyx_tuple__97 = PyTuple_Pack(51, __pyx_n_s_pos, __pyx_n_s_shape, __pyx_n_s_bin_size, __pyx_n_s_max_pixel_size, __pyx_n_s_mask, __pyx_n_s_shape0, __pyx_n_s_shape1, __pyx_n_s_delta0, __pyx_n_s_delta1, __pyx_n_s_bins, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_ms, __pyx_n_s_ml, __pyx_n_s_ns, __pyx_n_s_nl, __pyx_n_s_idx, __pyx_n_s_tmp_index, __pyx_n_s_err_cnt, __pyx_n_s_lut_size, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_box_size0, __pyx_n_s_box_size1, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_pAB, __pyx_n_s_pBC, __pyx_n_s_pCD, __pyx_n_s_pDA, __pyx_n_s_cAB, __pyx_n_s_cBC, __pyx_n_s_cCD, __pyx_n_s_cDA, __pyx_n_s_area, __pyx_n_s_value, __pyx_n_s_foffset0, __pyx_n_s_foffset1, __pyx_n_s_indptr, __pyx_n_s_indices, __pyx_n_s_data, __pyx_n_s_outMax, __pyx_n_s_buffer, __pyx_n_s_do_mask); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(5, 0, 51, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_CSR, 500, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 500, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":630
 * 
 * 
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                 shape,
 *                 max_pixel_size=(8, 8),
 */
  __pyx_tuple__99 = PyTuple_Pack(61, __pyx_n_s_pos, __pyx_n_s_shape, __pyx_n_s_max_pixel_size, __pyx_n_s_mask, __pyx_n_s_format, __pyx_n_s_bins_per_pixel, __pyx_n_s_shape_in0, __pyx_n_s_shape_in1, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1, __pyx_n_s_size_in, __pyx_n_s_delta0, __pyx_n_s_delta1, __pyx_n_s_bins, __pyx_n_s_large_size, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_ms, __pyx_n_s_ml, __pyx_n_s_ns, __pyx_n_s_nl, __pyx_n_s_lut_size, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_box_size0, __pyx_n_s_box_size1, __pyx_n_s_counter, __pyx_n_s_bin_number, __pyx_n_s_idx, __pyx_n_s_err_cnt, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_pAB, __pyx_n_s_pBC, __pyx_n_s_pCD, __pyx_n_s_pDA, __pyx_n_s_cAB, __pyx_n_s_cBC, __pyx_n_s_cCD, __pyx_n_s_cDA, __pyx_n_s_area, __pyx_n_s_value, __pyx_n_s_foffset0, __pyx_n_s_foffset1, __pyx_n_s_indptr, __pyx_n_s_indices, __pyx_n_s_idx_bin, __pyx_n_s_idx_pixel, __pyx_n_s_pixel_count, __pyx_n_s_data, __pyx_n_s_large_data, __pyx_n_s_buffer, __pyx_n_s_do_mask, __pyx_n_s_lut, __pyx_n_s_res); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(6, 0, 61, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_sparse, 630, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 630, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":823
 * 
 * 
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                    shape,
 *                    max_pixel_size=(8, 8),
 */
  __pyx_tuple__101 = PyTuple_Pack(54, __pyx_n_s_pos, __pyx_n_s_shape, __pyx_n_s_max_pixel_size, __pyx_n_s_mask, __pyx_n_s_format, __pyx_n_s_bins_per_pixel, __pyx_n_s_builder_config, __pyx_n_s_shape_in0, __pyx_n_s_shape_in1, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1, __pyx_n_s_size_in, __pyx_n_s_delta0, __pyx_n_s_delta1, __pyx_n_s_bins, __pyx_n_s_large_size, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_ms, __pyx_n_s_ml, __pyx_n_s_ns, __pyx_n_s_nl, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_box_size0, __pyx_n_s_box_size1, __pyx_n_s_counter, __pyx_n_s_bin_number, __pyx_n_s_idx, __pyx_n_s_err_cnt, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_pAB, __pyx_n_s_pBC, __pyx_n_s_pCD, __pyx_n_s_pDA, __pyx_n_s_cAB, __pyx_n_s_cBC, __pyx_n_s_cCD, __pyx_n_s_cDA, __pyx_n_s_area, __pyx_n_s_value, __pyx_n_s_foffset0, __pyx_n_s_foffset1, __pyx_n_s_buffer, __pyx_n_s_do_mask, __pyx_n_s_builder, __pyx_n_s_res); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(7, 0, 54, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_sparse_v2, 823, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 823, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":975
 * 
 * 
 * def resize_image_2D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */
  __pyx_tuple__103 = PyTuple_Pack(7, __pyx_n_s_image, __pyx_n_s_shape, __pyx_n_s_shape_in0, __pyx_n_s_shape_in1, __pyx_n_s_shape_img0, __pyx_n_s_shape_img1, __pyx_n_s_new_image); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_resize_image_2D, 975, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 975, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1008
 * 
 * 
 * def resize_image_3D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */
  __pyx_tuple__105 = PyTuple_Pack(8, __pyx_n_s_image, __pyx_n_s_shape, __pyx_n_s_shape_in0, __pyx_n_s_shape_in1, __pyx_n_s_shape_img0, __pyx_n_s_shape_img1, __pyx_n_s_nchan, __pyx_n_s_new_image); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_resize_image_3D, 1008, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 1008, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1043
 * 
 * 
 * def correct(image, shape_in, shape_out, LUT not None, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *             method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__107 = PyTuple_Pack(10, __pyx_n_s_image, __pyx_n_s_shape_in, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_method, __pyx_n_s_preprocessed_data, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(7, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct, 1043, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 1043, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1087
 * 
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                 dummy=None, delta_dummy=None, method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__109 = PyTuple_Pack(9, __pyx_n_s_image, __pyx_n_s_shape_in, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_method, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(7, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__109, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_LUT, 1087, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 1087, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1112
 * 
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                       dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__111 = PyTuple_Pack(21, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_coef, __pyx_n_s_sum, __pyx_n_s_error, __pyx_n_s_t, __pyx_n_s_y, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_do_dummy, __pyx_n_s_out); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(5, 0, 21, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_LUT_kahan, 1112, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 1112, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1166
 * 
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                        dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__113 = PyTuple_Pack(18, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_sum, __pyx_n_s_coef, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_do_dummy, __pyx_n_s_out); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(5, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__113, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_LUT_double, 1166, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 1166, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1220
 * 
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                 variance=None, method="double"):
 *     """
 */
  __pyx_tuple__115 = PyTuple_Pack(8, __pyx_n_s_image, __pyx_n_s_shape_in, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_variance, __pyx_n_s_method); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(8, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__115, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_CSR, 1220, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 1220, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1246
 * 
 * 
 * def correct_CSR_kahan(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__117 = PyTuple_Pack(25, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_bins, __pyx_n_s_coef, __pyx_n_s_error, __pyx_n_s_sum, __pyx_n_s_y, __pyx_n_s_t, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_data, __pyx_n_s_indices, __pyx_n_s_indptr, __pyx_n_s_do_dummy, __pyx_n_s_out); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(5, 0, 25, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_CSR_kahan, 1246, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 1246, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1307
 * 
 * 
 * def correct_CSR_double(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__119 = PyTuple_Pack(22, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_bins, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_coef, __pyx_n_s_sum, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_data, __pyx_n_s_indices, __pyx_n_s_indptr, __pyx_n_s_do_dummy, __pyx_n_s_out); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(5, 0, 22, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_CSR_double, 1307, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(0, 1307, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1365
 * 
 * 
 * def correct_LUT_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 */
  __pyx_tuple__121 = PyTuple_Pack(28, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_empty, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_nchan, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_sum_sig, __pyx_n_s_sum_var, __pyx_n_s_sum_norm, __pyx_n_s_coef, __pyx_n_s_lout, __pyx_n_s_lerr, __pyx_n_s_lin, __pyx_n_s_lprop, __pyx_n_s_do_dummy, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1, __pyx_n_s_prop, __pyx_n_s_out, __pyx_n_s_err); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(6, 0, 28, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__121, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_LUT_preproc_double, 1365, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 1365, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1466
 * 
 * 
 * def correct_CSR_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 */
  __pyx_tuple__123 = PyTuple_Pack(32, __pyx_n_s_image, __pyx_n_s_shape_out, __pyx_n_s_LUT, __pyx_n_s_dummy, __pyx_n_s_delta_dummy, __pyx_n_s_empty, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_bins, __pyx_n_s_nchan, __pyx_n_s_value, __pyx_n_s_cdummy, __pyx_n_s_cdelta_dummy, __pyx_n_s_sum_sig, __pyx_n_s_sum_var, __pyx_n_s_sum_norm, __pyx_n_s_coef, __pyx_n_s_lout, __pyx_n_s_lerr, __pyx_n_s_data, __pyx_n_s_lin, __pyx_n_s_lprop, __pyx_n_s_indices, __pyx_n_s_indptr, __pyx_n_s_do_dummy, __pyx_n_s_shape_out0, __pyx_n_s_shape_out1, __pyx_n_s_prop, __pyx_n_s_out, __pyx_n_s_err); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(6, 0, 32, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct_CSR_preproc_double, 1466, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 1466, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1573
 * 
 * 
 * def uncorrect_LUT(image, shape, lut_t[:, :]LUT):             # <<<<<<<<<<<<<<
 *     """
 *     Take an image which has been corrected and transform it into it's raw (with loss of information)
 */
  __pyx_tuple__125 = PyTuple_Pack(13, __pyx_n_s_image, __pyx_n_s_shape, __pyx_n_s_LUT, __pyx_n_s_idx, __pyx_n_s_j, __pyx_n_s_total, __pyx_n_s_coef, __pyx_n_s_lmask, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_val); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(3, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_uncorrect_LUT, 1573, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(0, 1573, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1610
 * 
 * 
 * def uncorrect_CSR(image, shape, LUT):             # <<<<<<<<<<<<<<
 *     """Take an image which has been corrected and transform it into it's raw (with loss of information)
 * 
 */
  __pyx_tuple__127 = PyTuple_Pack(17, __pyx_n_s_image, __pyx_n_s_shape, __pyx_n_s_LUT, __pyx_n_s_idx, __pyx_n_s_j, __pyx_n_s_nbins, __pyx_n_s_total, __pyx_n_s_coef, __pyx_n_s_lmask, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_data, __pyx_n_s_indices, __pyx_n_s_indptr, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_val); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 1610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(3, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_uncorrect_CSR, 1610, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 1610, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1653
 * 
 * 
 * class Distortion(object):             # <<<<<<<<<<<<<<
 *     """
 * 
 */
  __pyx_tuple__129 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "pyFAI/ext/_distortion.pyx":1661
 * 
 *     """
 *     def __init__(self, detector="detector", shape=None):             # <<<<<<<<<<<<<<
 *         """
 *         :param detector: detector instance or detector name
 */
  __pyx_tuple__130 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_detector, __pyx_n_s_shape, __pyx_n_s_i); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_init, 1661, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __pyx_tuple__132 = PyTuple_Pack(2, ((PyObject*)__pyx_n_u_detector), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "pyFAI/ext/_distortion.pyx":1680
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return os.linesep.join(["Distortion correction for detector:",
 *                                 self.detector.__repr__()])
 */
  __pyx_tuple__133 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_repr, 1680, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(0, 1680, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1684
 *                                 self.detector.__repr__()])
 * 
 *     def calc_pos(self):             # <<<<<<<<<<<<<<
 *         if self.pos is None:
 *             with self._sem:
 */
  __pyx_tuple__135 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_pos_corners, __pyx_n_s_d1, __pyx_n_s_d2, __pyx_n_s_p, __pyx_n_s_pos); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 1684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_pos, 1684, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 1684, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1708
 *         return self.pos
 * 
 *     def calc_LUT_size(self):             # <<<<<<<<<<<<<<
 *         """
 *         Considering the "half-CCD" spline from ID11 which describes a (1025,2048) detector,
 */
  __pyx_tuple__137 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_l, __pyx_n_s_shape0, __pyx_n_s_shape1, __pyx_n_s_pos0min, __pyx_n_s_pos1min, __pyx_n_s_pos0max, __pyx_n_s_pos1max, __pyx_n_s_lut_size, __pyx_n_s_pos, __pyx_n_s_np_lut_size); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(1, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_LUT_size, 1708, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 1708, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1744
 *                     return np_lut_size
 * 
 *     def calc_LUT(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, j, ms, ml, ns, nl, shape0, shape1, size
 */
  __pyx_tuple__139 = PyTuple_Pack(38, __pyx_n_s_self, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_ms, __pyx_n_s_ml, __pyx_n_s_ns, __pyx_n_s_nl, __pyx_n_s_shape0, __pyx_n_s_shape1, __pyx_n_s_size, __pyx_n_s_offset0, __pyx_n_s_offset1, __pyx_n_s_box_size0, __pyx_n_s_box_size1, __pyx_n_s_k, __pyx_n_s_idx, __pyx_n_s_A0_2, __pyx_n_s_A1_2, __pyx_n_s_B0_2, __pyx_n_s_B1_2, __pyx_n_s_C0_2, __pyx_n_s_C1_2, __pyx_n_s_D0_2, __pyx_n_s_D1_2, __pyx_n_s_pAB, __pyx_n_s_pBC, __pyx_n_s_pCD, __pyx_n_s_pDA, __pyx_n_s_cAB, __pyx_n_s_cBC, __pyx_n_s_cCD, __pyx_n_s_cDA, __pyx_n_s_area, __pyx_n_s_value, __pyx_n_s_pos, __pyx_n_s_lut, __pyx_n_s_outMax, __pyx_n_s_buffer); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 1744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(1, 0, 38, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_calc_LUT, 1744, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 1744, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1856
 * #         return ab
 * 
 *     def correct(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Correct an image based on the look-up table calculated ...
 */
  __pyx_tuple__141 = PyTuple_Pack(13, __pyx_n_s_self, __pyx_n_s_image, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_idx, __pyx_n_s_size, __pyx_n_s_coef, __pyx_n_s_LUT, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_img_shape, __pyx_n_s_new_image, __pyx_n_s_out); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(2, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_correct, 1856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 1856, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":1895
 *         return out[:img_shape[0], :img_shape[1]]
 * 
 *     def uncorrect(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Take an image which has been corrected and transform it into it's raw (with loss of information)
 */
  __pyx_tuple__143 = PyTuple_Pack(11, __pyx_n_s_self, __pyx_n_s_image, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_lmask, __pyx_n_s_lout, __pyx_n_s_lin, __pyx_n_s_tot, __pyx_n_s_idx, __pyx_n_s_t, __pyx_n_s_val); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyFAI_ext__distortion_pyx, __pyx_n_s_uncorrect, 1895, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 1895, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__145 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__146 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__148 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__149 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__150 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #ifdef WITH_THREAD
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1e6 = PyFloat_FromDouble(1e6); if (unlikely(!__pyx_float_1e6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0e6 = PyFloat_FromDouble(1.0e6); if (unlikely(!__pyx_float_1_0e6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i = Py_None; Py_INCREF(Py_None);
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyFAI.ext.sparse_builder"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder = __Pyx_ImportType(__pyx_t_1, "pyFAI.ext.sparse_builder", "SparseBuilder", sizeof(struct __pyx_obj_5pyFAI_3ext_14sparse_builder_SparseBuilder), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder) __PYX_ERR(3, 44, __pyx_L1_error)
  __pyx_vtabptr_5pyFAI_3ext_14sparse_builder_SparseBuilder = (struct __pyx_vtabstruct_5pyFAI_3ext_14sparse_builder_SparseBuilder*)__Pyx_GetVtable(__pyx_ptype_5pyFAI_3ext_14sparse_builder_SparseBuilder->tp_dict); if (unlikely(!__pyx_vtabptr_5pyFAI_3ext_14sparse_builder_SparseBuilder)) __PYX_ERR(3, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_distortion(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_distortion(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__distortion(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__distortion(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__distortion(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  double __pyx_t_10;
  int __pyx_t_11;
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  static PyThread_type_lock __pyx_t_14[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_distortion' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__distortion(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_distortion", __pyx_methods, __pyx_k_Distortion_correction_are_corre, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyFAI__ext___distortion) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyFAI.ext._distortion")) {
      if (unlikely(PyDict_SetItemString(modules, "pyFAI.ext._distortion", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pyFAI/ext/_distortion.pyx":36
 * """
 * 
 * __author__ = "Jerome Kieffer"             # <<<<<<<<<<<<<<
 * __license__ = "MIT"
 * __date__ = "14/01/2021"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_author, __pyx_kp_u_Jerome_Kieffer) < 0) __PYX_ERR(0, 36, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":37
 * 
 * __author__ = "Jerome Kieffer"
 * __license__ = "MIT"             # <<<<<<<<<<<<<<
 * __date__ = "14/01/2021"
 * __copyright__ = "2011-2021, ESRF"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_license, __pyx_n_u_MIT) < 0) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":38
 * __author__ = "Jerome Kieffer"
 * __license__ = "MIT"
 * __date__ = "14/01/2021"             # <<<<<<<<<<<<<<
 * __copyright__ = "2011-2021, ESRF"
 * __contact__ = "jerome.kieffer@esrf.fr"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_date, __pyx_kp_u_14_01_2021) < 0) __PYX_ERR(0, 38, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":39
 * __license__ = "MIT"
 * __date__ = "14/01/2021"
 * __copyright__ = "2011-2021, ESRF"             # <<<<<<<<<<<<<<
 * __contact__ = "jerome.kieffer@esrf.fr"
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copyright, __pyx_kp_u_2011_2021_ESRF) < 0) __PYX_ERR(0, 39, __pyx_L1_error)

  /* "pyFAI/ext/_distortion.pyx":40
 * __date__ = "14/01/2021"
 * __copyright__ = "2011-2021, ESRF"
 * __contact__ = "jerome.kieffer@esrf.fr"             # <<<<<<<<<<<<<<
 * 
 * include "regrid_common.pxi"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_contact, __pyx_kp_u_jerome_kieffer_esrf_fr) < 0) __PYX_ERR(0, 40, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":33
 * """
 * 
 * __author__ = "Jerome Kieffer"             # <<<<<<<<<<<<<<
 * __contact__ = "Jerome.kieffer@esrf.fr"
 * __date__ = "14/01/2021"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_author, __pyx_kp_u_Jerome_Kieffer) < 0) __PYX_ERR(4, 33, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":34
 * 
 * __author__ = "Jerome Kieffer"
 * __contact__ = "Jerome.kieffer@esrf.fr"             # <<<<<<<<<<<<<<
 * __date__ = "14/01/2021"
 * __status__ = "stable"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_contact, __pyx_kp_u_Jerome_kieffer_esrf_fr) < 0) __PYX_ERR(4, 34, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":35
 * __author__ = "Jerome Kieffer"
 * __contact__ = "Jerome.kieffer@esrf.fr"
 * __date__ = "14/01/2021"             # <<<<<<<<<<<<<<
 * __status__ = "stable"
 * __license__ = "MIT"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_date, __pyx_kp_u_14_01_2021) < 0) __PYX_ERR(4, 35, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":36
 * __contact__ = "Jerome.kieffer@esrf.fr"
 * __date__ = "14/01/2021"
 * __status__ = "stable"             # <<<<<<<<<<<<<<
 * __license__ = "MIT"
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_status, __pyx_n_u_stable) < 0) __PYX_ERR(4, 36, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":37
 * __date__ = "14/01/2021"
 * __status__ = "stable"
 * __license__ = "MIT"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_license, __pyx_n_u_MIT) < 0) __PYX_ERR(4, 37, __pyx_L1_error)

  /* "pyFAI/ext/regrid_common.pxi":42
 * # Imports at the Python level
 * import cython
 * import numpy             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":43
 * import cython
 * import numpy
 * import sys             # <<<<<<<<<<<<<<
 * 
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(4, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":47
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 * 
 * _numpy_1_12_py2_bug = ((sys.version_info.major == 2) and             # <<<<<<<<<<<<<<
 *                        ([1, 12] >= [int(i) for i in numpy.version.version.split(".", 2)[:2]]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_version_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_major); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(4, 47, __pyx_L1_error)
  if (__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L2_bool_binop_done;
  }

  /* "pyFAI/ext/regrid_common.pxi":48
 * 
 * _numpy_1_12_py2_bug = ((sys.version_info.major == 2) and
 *                        ([1, 12] >= [int(i) for i in numpy.version.version.split(".", 2)[:2]]))             # <<<<<<<<<<<<<<
 * 
 * # Imports at the C level
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_1);
  __Pyx_INCREF(__pyx_int_12);
  __Pyx_GIVEREF(__pyx_int_12);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_int_12);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_version); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_split); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_t_5, 0, 2, NULL, NULL, &__pyx_slice__25, 0, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 48, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(4, 48, __pyx_L6_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(4, 48, __pyx_L6_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_5);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(4, 48, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XGOTREF(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i);
      __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 48, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(4, 48, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i);
    __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i, Py_None);
    goto __pyx_L9_exit_scope;
    __pyx_L6_error:;
    __Pyx_GOTREF(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i);
    __Pyx_DECREF_SET(__pyx_7genexpr__pyx_v_5pyFAI_3ext_11_distortion_i, Py_None);
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_1 = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_L2_bool_binop_done:;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy_1_12_py2_bug, __pyx_t_1) < 0) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":61
 * # How position are stored
 * ctypedef float64_t position_t
 * position_d = numpy.float64             # <<<<<<<<<<<<<<
 * 
 * # How weights or data are stored
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_position_d, __pyx_t_5) < 0) __PYX_ERR(4, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/regrid_common.pxi":65
 * # How weights or data are stored
 * ctypedef float32_t data_t
 * data_d = numpy.float32             # <<<<<<<<<<<<<<
 * 
 * # how data are accumulated
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_data_d, __pyx_t_1) < 0) __PYX_ERR(4, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":69
 * # how data are accumulated
 * ctypedef float64_t acc_t
 * acc_d = numpy.float64             # <<<<<<<<<<<<<<
 * 
 * # type of the mask:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_acc_d, __pyx_t_5) < 0) __PYX_ERR(4, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/regrid_common.pxi":73
 * # type of the mask:
 * ctypedef int8_t mask_t
 * mask_d = numpy.int8             # <<<<<<<<<<<<<<
 * 
 * # type of the indexes:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int8); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mask_d, __pyx_t_1) < 0) __PYX_ERR(4, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":77
 * # type of the indexes:
 * ctypedef int32_t index_t
 * index_d = numpy.int32             # <<<<<<<<<<<<<<
 * 
 * cdef struct lut_t:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int32); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_index_d, __pyx_t_5) < 0) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyFAI/ext/regrid_common.pxi":83
 *     data_t coef
 * 
 * LUT_ITEMSIZE = int(sizeof(lut_t))             # <<<<<<<<<<<<<<
 * 
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 */
  __pyx_t_5 = __Pyx_PyInt_FromSize_t((sizeof(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t))); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LUT_ITEMSIZE, __pyx_t_1) < 0) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":86
 * 
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 * if _numpy_1_12_py2_bug:             # <<<<<<<<<<<<<<
 *     lut_d = numpy.dtype([(b"idx", index_d), (b"coef", data_d)])
 * else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy_1_12_py2_bug); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(4, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "pyFAI/ext/regrid_common.pxi":87
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 * if _numpy_1_12_py2_bug:
 *     lut_d = numpy.dtype([(b"idx", index_d), (b"coef", data_d)])             # <<<<<<<<<<<<<<
 * else:
 *     lut_d = numpy.dtype([("idx", index_d), ("coef", numpy.float32)])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_index_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_b_idx);
    __Pyx_GIVEREF(__pyx_n_b_idx);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_b_idx);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_data_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_b_coef);
    __Pyx_GIVEREF(__pyx_n_b_coef);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_b_coef);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_lut_d, __pyx_t_3) < 0) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyFAI/ext/regrid_common.pxi":86
 * 
 * # Work around for issue similar to : https://github.com/pandas-dev/pandas/issues/16358
 * if _numpy_1_12_py2_bug:             # <<<<<<<<<<<<<<
 *     lut_d = numpy.dtype([(b"idx", index_d), (b"coef", data_d)])
 * else:
 */
    goto __pyx_L10;
  }

  /* "pyFAI/ext/regrid_common.pxi":89
 *     lut_d = numpy.dtype([(b"idx", index_d), (b"coef", data_d)])
 * else:
 *     lut_d = numpy.dtype([("idx", index_d), ("coef", numpy.float32)])             # <<<<<<<<<<<<<<
 * 
 * # Type used for propagating variance
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_index_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_u_idx);
    __Pyx_GIVEREF(__pyx_n_u_idx);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_idx);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_u_coef);
    __Pyx_GIVEREF(__pyx_n_u_coef);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_coef);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_lut_d, __pyx_t_3) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L10:;

  /* "pyFAI/ext/regrid_common.pxi":92
 * 
 * # Type used for propagating variance
 * if _numpy_1_12_py2_bug:             # <<<<<<<<<<<<<<
 *     prop_d = numpy.dtype([(b'signal', acc_d),
 *                           (b'variance', acc_d),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy_1_12_py2_bug); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(4, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pyFAI/ext/regrid_common.pxi":93
 * # Type used for propagating variance
 * if _numpy_1_12_py2_bug:
 *     prop_d = numpy.dtype([(b'signal', acc_d),             # <<<<<<<<<<<<<<
 *                           (b'variance', acc_d),
 *                           (b'norm', acc_d),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_b_signal);
    __Pyx_GIVEREF(__pyx_n_b_signal);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_b_signal);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyFAI/ext/regrid_common.pxi":94
 * if _numpy_1_12_py2_bug:
 *     prop_d = numpy.dtype([(b'signal', acc_d),
 *                           (b'variance', acc_d),             # <<<<<<<<<<<<<<
 *                           (b'norm', acc_d),
 *                           (b'count', acc_d)])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_b_variance);
    __Pyx_GIVEREF(__pyx_n_b_variance);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_b_variance);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyFAI/ext/regrid_common.pxi":95
 *     prop_d = numpy.dtype([(b'signal', acc_d),
 *                           (b'variance', acc_d),
 *                           (b'norm', acc_d),             # <<<<<<<<<<<<<<
 *                           (b'count', acc_d)])
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_n_b_norm);
    __Pyx_GIVEREF(__pyx_n_b_norm);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_b_norm);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyFAI/ext/regrid_common.pxi":96
 *                           (b'variance', acc_d),
 *                           (b'norm', acc_d),
 *                           (b'count', acc_d)])             # <<<<<<<<<<<<<<
 * else:
 *     prop_d = numpy.dtype([('signal', acc_d),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_b_count);
    __Pyx_GIVEREF(__pyx_n_b_count);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_b_count);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyFAI/ext/regrid_common.pxi":93
 * # Type used for propagating variance
 * if _numpy_1_12_py2_bug:
 *     prop_d = numpy.dtype([(b'signal', acc_d),             # <<<<<<<<<<<<<<
 *                           (b'variance', acc_d),
 *                           (b'norm', acc_d),
 */
    __pyx_t_3 = PyList_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_3, 3, __pyx_t_9);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_prop_d, __pyx_t_9) < 0) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "pyFAI/ext/regrid_common.pxi":92
 * 
 * # Type used for propagating variance
 * if _numpy_1_12_py2_bug:             # <<<<<<<<<<<<<<
 *     prop_d = numpy.dtype([(b'signal', acc_d),
 *                           (b'variance', acc_d),
 */
    goto __pyx_L11;
  }

  /* "pyFAI/ext/regrid_common.pxi":98
 *                           (b'count', acc_d)])
 * else:
 *     prop_d = numpy.dtype([('signal', acc_d),             # <<<<<<<<<<<<<<
 *                           ('variance', acc_d),
 *                           ('norm', acc_d),
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_u_signal);
    __Pyx_GIVEREF(__pyx_n_u_signal);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_signal);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyFAI/ext/regrid_common.pxi":99
 * else:
 *     prop_d = numpy.dtype([('signal', acc_d),
 *                           ('variance', acc_d),             # <<<<<<<<<<<<<<
 *                           ('norm', acc_d),
 *                          ('count', acc_d)])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_n_u_variance);
    __Pyx_GIVEREF(__pyx_n_u_variance);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_variance);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyFAI/ext/regrid_common.pxi":100
 *     prop_d = numpy.dtype([('signal', acc_d),
 *                           ('variance', acc_d),
 *                           ('norm', acc_d),             # <<<<<<<<<<<<<<
 *                          ('count', acc_d)])
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_u_norm);
    __Pyx_GIVEREF(__pyx_n_u_norm);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_norm);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyFAI/ext/regrid_common.pxi":101
 *                           ('variance', acc_d),
 *                           ('norm', acc_d),
 *                          ('count', acc_d)])             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused any_int_t:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_acc_d); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_u_count);
    __Pyx_GIVEREF(__pyx_n_u_count);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_count);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyFAI/ext/regrid_common.pxi":98
 *                           (b'count', acc_d)])
 * else:
 *     prop_d = numpy.dtype([('signal', acc_d),             # <<<<<<<<<<<<<<
 *                           ('variance', acc_d),
 *                           ('norm', acc_d),
 */
    __pyx_t_9 = PyList_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_6);
    PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_prop_d, __pyx_t_1) < 0) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L11:;

  /* "pyFAI/ext/regrid_common.pxi":121
 *         data_t count
 * 
 *     float pi = <float> M_PI             # <<<<<<<<<<<<<<
 *     float piover2 = <float> (pi * 0.5)
 *     float onef = <float> 1.0
 */
  __pyx_v_5pyFAI_3ext_11_distortion_pi = ((float)M_PI);

  /* "pyFAI/ext/regrid_common.pxi":122
 * 
 *     float pi = <float> M_PI
 *     float piover2 = <float> (pi * 0.5)             # <<<<<<<<<<<<<<
 *     float onef = <float> 1.0
 *     float zerof = <float> 1.0
 */
  __pyx_v_5pyFAI_3ext_11_distortion_piover2 = ((float)(__pyx_v_5pyFAI_3ext_11_distortion_pi * 0.5));

  /* "pyFAI/ext/regrid_common.pxi":123
 *     float pi = <float> M_PI
 *     float piover2 = <float> (pi * 0.5)
 *     float onef = <float> 1.0             # <<<<<<<<<<<<<<
 *     float zerof = <float> 1.0
 *     double EPS32 = (1.0 + numpy.finfo(numpy.float32).eps)
 */
  __pyx_v_5pyFAI_3ext_11_distortion_onef = ((float)1.0);

  /* "pyFAI/ext/regrid_common.pxi":124
 *     float piover2 = <float> (pi * 0.5)
 *     float onef = <float> 1.0
 *     float zerof = <float> 1.0             # <<<<<<<<<<<<<<
 *     double EPS32 = (1.0 + numpy.finfo(numpy.float32).eps)
 * 
 */
  __pyx_v_5pyFAI_3ext_11_distortion_zerof = ((float)1.0);

  /* "pyFAI/ext/regrid_common.pxi":125
 *     float onef = <float> 1.0
 *     float zerof = <float> 1.0
 *     double EPS32 = (1.0 + numpy.finfo(numpy.float32).eps)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_finfo); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_eps); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyFloat_AddCObj(__pyx_float_1_0, __pyx_t_3, 1.0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(4, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_5pyFAI_3ext_11_distortion_EPS32 = __pyx_t_10;

  /* "pyFAI/ext/regrid_common.pxi":128
 * 
 * 
 * from collections import namedtuple             # <<<<<<<<<<<<<<
 * from ..containers import Integrate1dtpl, Integrate2dtpl
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_namedtuple);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_1) < 0) __PYX_ERR(4, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/regrid_common.pxi":129
 * 
 * from collections import namedtuple
 * from ..containers import Integrate1dtpl, Integrate2dtpl             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Integrate1dtpl);
  __Pyx_GIVEREF(__pyx_n_s_Integrate1dtpl);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Integrate1dtpl);
  __Pyx_INCREF(__pyx_n_s_Integrate2dtpl);
  __Pyx_GIVEREF(__pyx_n_s_Integrate2dtpl);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_Integrate2dtpl);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_containers, __pyx_t_3, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Integrate1dtpl); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Integrate1dtpl, __pyx_t_3) < 0) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Integrate2dtpl); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Integrate2dtpl, __pyx_t_3) < 0) __PYX_ERR(4, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/regrid_common.pxi":156
 * 
 * 
 * cdef inline bint preproc_value_inplace(preproc_t* result,             # <<<<<<<<<<<<<<
 *                                        floating data,
 *                                        floating variance=0.0,
 */
  __pyx_k__66 = 0.0;
  __pyx_k__67 = 0.0;
  __pyx_k__68 = 1.0;
  __pyx_k__69 = 1.0;
  __pyx_k__70 = 1.0;
  __pyx_k__71 = 1.0;
  __pyx_k__72 = 0;
  __pyx_k__73 = 0.0;
  __pyx_k__74 = 0.0;
  __pyx_k__75 = 0;
  __pyx_k__76 = 1.0;
  __pyx_k__77 = 0.0;
  __pyx_k__78 = 0.0;
  __pyx_k__79 = 0.0;
  __pyx_k__80 = 1.0;
  __pyx_k__81 = 1.0;
  __pyx_k__82 = 1.0;
  __pyx_k__83 = 1.0;
  __pyx_k__84 = 0;
  __pyx_k__85 = 0.0;
  __pyx_k__86 = 0.0;
  __pyx_k__87 = 0;
  __pyx_k__88 = 1.0;
  __pyx_k__89 = 0.0;

  /* "pyFAI/ext/_distortion.pyx":44
 * include "regrid_common.pxi"
 * import cython
 * import numpy             # <<<<<<<<<<<<<<
 * from cython cimport view, floating
 * from cython.parallel import prange
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":50
 * from libc.string cimport memset, memcpy
 * from libc.math cimport floor, ceil, fabs, copysign, sqrt
 * import logging             # <<<<<<<<<<<<<<
 * import threading
 * import types
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_logging, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_1) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":51
 * from libc.math cimport floor, ceil, fabs, copysign, sqrt
 * import logging
 * import threading             # <<<<<<<<<<<<<<
 * import types
 * import os
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_threading, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_threading, __pyx_t_1) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":52
 * import logging
 * import threading
 * import types             # <<<<<<<<<<<<<<
 * import os
 * import sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_types, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_types, __pyx_t_1) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":53
 * import threading
 * import types
 * import os             # <<<<<<<<<<<<<<
 * import sys
 * import time
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":54
 * import types
 * import os
 * import sys             # <<<<<<<<<<<<<<
 * import time
 * logger = logging.getLogger(__name__)
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":55
 * import os
 * import sys
 * import time             # <<<<<<<<<<<<<<
 * logger = logging.getLogger(__name__)
 * from ..detectors import detector_factory
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":56
 * import sys
 * import time
 * logger = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * from ..detectors import detector_factory
 * from ..utils import expand2d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_9) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":57
 * import time
 * logger = logging.getLogger(__name__)
 * from ..detectors import detector_factory             # <<<<<<<<<<<<<<
 * from ..utils import expand2d
 * import fabio
 */
  __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_detector_factory);
  __Pyx_GIVEREF(__pyx_n_s_detector_factory);
  PyList_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_detector_factory);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_detectors, __pyx_t_9, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_detector_factory); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_detector_factory, __pyx_t_9) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":58
 * logger = logging.getLogger(__name__)
 * from ..detectors import detector_factory
 * from ..utils import expand2d             # <<<<<<<<<<<<<<
 * import fabio
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_expand2d);
  __Pyx_GIVEREF(__pyx_n_s_expand2d);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_expand2d);
  __pyx_t_9 = __Pyx_Import(__pyx_n_s_utils, __pyx_t_1, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_9, __pyx_n_s_expand2d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_expand2d, __pyx_t_1) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":59
 * from ..detectors import detector_factory
 * from ..utils import expand2d
 * import fabio             # <<<<<<<<<<<<<<
 * 
 * from .sparse_builder cimport SparseBuilder
 */
  __pyx_t_9 = __Pyx_Import(__pyx_n_s_fabio, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fabio, __pyx_t_9) < 0) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":63
 * from .sparse_builder cimport SparseBuilder
 * 
 * cdef bint NEED_DECREF = sys.version_info < (2, 7) and numpy.version.version < "1.5"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_sys); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_version_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyObject_RichCompare(__pyx_t_1, __pyx_tuple__90, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__pyx_t_11) {
  } else {
    __pyx_t_4 = __pyx_t_11;
    goto __pyx_L12_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_version); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_9, __pyx_kp_u_1_5, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_11;
  __pyx_L12_bool_binop_done:;
  __pyx_v_5pyFAI_3ext_11_distortion_NEED_DECREF = __pyx_t_4;

  /* "pyFAI/ext/_distortion.pyx":245
 * ################################################################################
 * 
 * def calc_pos(floating[:, :, :, ::1] pixel_corners not None,             # <<<<<<<<<<<<<<
 *              float pixel1, float pixel2, shape_out=None):
 *     """Calculate the pixel boundary position on the regular grid
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None);
  __Pyx_INCREF(Py_None);
  __pyx_k_ = Py_None;
  __Pyx_GIVEREF(Py_None);
  __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_5pyFAI_3ext_11_distortion_43calc_pos, 0, __pyx_n_s_calc_pos, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_3, sizeof(__pyx_defaults2), 1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_INCREF(Py_None);
  __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_t_3)->__pyx_arg_shape_out = Py_None;
  __Pyx_GIVEREF(Py_None);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_3, __pyx_pf_5pyFAI_3ext_11_distortion_58__defaults__);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_float, __pyx_t_3) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_5pyFAI_3ext_11_distortion_45calc_pos, 0, __pyx_n_s_calc_pos, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_3, sizeof(__pyx_defaults3), 1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_INCREF(Py_None);
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_3)->__pyx_arg_shape_out = Py_None;
  __Pyx_GIVEREF(Py_None);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_3, __pyx_pf_5pyFAI_3ext_11_distortion_60__defaults__);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_double, __pyx_t_3) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_FusedFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_5calc_pos, 0, __pyx_n_s_calc_pos, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_1);
  ((__pyx_FusedFunctionObject *) __pyx_t_3)->__signatures__ = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_pos, __pyx_t_3) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);
  __Pyx_INCREF(Py_None);
  __pyx_k__8 = Py_None;
  __Pyx_GIVEREF(Py_None);
  __Pyx_INCREF(Py_None);
  __pyx_k__9 = Py_None;
  __Pyx_GIVEREF(Py_None);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_5pyFAI_3ext_11_distortion_49calc_size, 0, __pyx_n_s_calc_size, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults6), 1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_INCREF(Py_None);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_9)->__pyx_arg_offset = Py_None;
  __Pyx_GIVEREF(Py_None);

  /* "pyFAI/ext/_distortion.pyx":308
 * def calc_size(floating[:, :, :, ::1] pos not None,
 *               shape,
 *               int8_t[:, ::1] mask=None,             # <<<<<<<<<<<<<<
 *               offset=None):
 *     """Calculate the number of items per output pixel
 */
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_9)->__pyx_arg_mask = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_t_1);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_5pyFAI_3ext_11_distortion_66__defaults__);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_float, __pyx_t_9) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */
  __pyx_t_9 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_5pyFAI_3ext_11_distortion_51calc_size, 0, __pyx_n_s_calc_size, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults7), 1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_INCREF(Py_None);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_9)->__pyx_arg_offset = Py_None;
  __Pyx_GIVEREF(Py_None);

  /* "pyFAI/ext/_distortion.pyx":308
 * def calc_size(floating[:, :, :, ::1] pos not None,
 *               shape,
 *               int8_t[:, ::1] mask=None,             # <<<<<<<<<<<<<<
 *               offset=None):
 *     """Calculate the number of items per output pixel
 */
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_9)->__pyx_arg_mask = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_t_1);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_5pyFAI_3ext_11_distortion_68__defaults__);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_double, __pyx_t_9) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":306
 * 
 * 
 * def calc_size(floating[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *               shape,
 *               int8_t[:, ::1] mask=None,
 */
  __pyx_t_9 = __pyx_FusedFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_7calc_size, 0, __pyx_n_s_calc_size, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_t_1);
  ((__pyx_FusedFunctionObject *) __pyx_t_9)->__signatures__ = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_size, __pyx_t_9) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":363
 * 
 * def calc_LUT(float32_t[:, :, :, ::1] pos not None, shape, bin_size, max_pixel_size,
 *              int8_t[:, :] mask=None):             # <<<<<<<<<<<<<<
 *     """
 *     :param pos: 4D position array
 */
  __pyx_t_13 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_13.memview)) __PYX_ERR(0, 363, __pyx_L1_error)
  __pyx_k__12 = __pyx_t_13;
  __pyx_t_13.memview = NULL;
  __pyx_t_13.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":362
 * 
 * 
 * def calc_LUT(float32_t[:, :, :, ::1] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, :] mask=None):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_9calc_LUT, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_LUT, __pyx_t_1) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":501
 * 
 * def calc_CSR(float32_t[:, :, :, :] pos not None, shape, bin_size, max_pixel_size,
 *              int8_t[:, ::1] mask=None):             # <<<<<<<<<<<<<<
 *     """Calculate the Look-up table as CSR format
 * 
 */
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_k__14 = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":500
 * 
 * 
 * def calc_CSR(float32_t[:, :, :, :] pos not None, shape, bin_size, max_pixel_size,             # <<<<<<<<<<<<<<
 *              int8_t[:, ::1] mask=None):
 *     """Calculate the Look-up table as CSR format
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_11calc_CSR, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_CSR, __pyx_t_1) < 0) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":633
 *                 shape,
 *                 max_pixel_size=(8, 8),
 *                 int8_t[:, ::1] mask=None,             # <<<<<<<<<<<<<<
 *                 format="csr",
 *                 int bins_per_pixel=8):
 */
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 633, __pyx_L1_error)
  __pyx_k__16 = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":630
 * 
 * 
 * def calc_sparse(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                 shape,
 *                 max_pixel_size=(8, 8),
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_13calc_sparse, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_sparse, __pyx_t_1) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":826
 *                    shape,
 *                    max_pixel_size=(8, 8),
 *                    int8_t[:, ::1] mask=None,             # <<<<<<<<<<<<<<
 *                    format="csr",
 *                    int bins_per_pixel=8,
 */
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 826, __pyx_L1_error)
  __pyx_k__17 = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "pyFAI/ext/_distortion.pyx":823
 * 
 * 
 * def calc_sparse_v2(float32_t[:, :, :, ::1] pos not None,             # <<<<<<<<<<<<<<
 *                    shape,
 *                    max_pixel_size=(8, 8),
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_15calc_sparse_v2, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_sparse_v2, __pyx_t_1) < 0) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":975
 * 
 * 
 * def resize_image_2D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_17resize_image_2D, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_resize_image_2D, __pyx_t_1) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1008
 * 
 * 
 * def resize_image_3D(image not None,             # <<<<<<<<<<<<<<
 *                     shape=None):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_19resize_image_3D, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_resize_image_3D, __pyx_t_1) < 0) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1043
 * 
 * 
 * def correct(image, shape_in, shape_out, LUT not None, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *             method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_21correct, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct, __pyx_t_1) < 0) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1087
 * 
 * 
 * def correct_LUT(image, shape_in, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                 dummy=None, delta_dummy=None, method="double"):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_23correct_LUT, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_LUT, __pyx_t_1) < 0) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1112
 * 
 * 
 * def correct_LUT_kahan(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                       dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_25correct_LUT_kahan, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_LUT_kahan, __pyx_t_1) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1166
 * 
 * 
 * def correct_LUT_double(image, shape_out, lut_t[:, ::1] LUT not None,             # <<<<<<<<<<<<<<
 *                        dummy=None, delta_dummy=None):
 *     """Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_27correct_LUT_double, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_LUT_double, __pyx_t_1) < 0) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1220
 * 
 * 
 * def correct_CSR(image, shape_in, shape_out, LUT, dummy=None, delta_dummy=None,             # <<<<<<<<<<<<<<
 *                 variance=None, method="double"):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_29correct_CSR, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_CSR, __pyx_t_1) < 0) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1246
 * 
 * 
 * def correct_CSR_kahan(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_31correct_CSR_kahan, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_CSR_kahan, __pyx_t_1) < 0) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1307
 * 
 * 
 * def correct_CSR_double(image, shape_out, LUT, dummy=None, delta_dummy=None):             # <<<<<<<<<<<<<<
 *     """
 *     Correct an image based on the look-up table calculated ...
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_33correct_CSR_double, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_CSR_double, __pyx_t_1) < 0) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1368
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 *                                empty=numpy.NaN):             # <<<<<<<<<<<<<<
 *     """Correct an image based on the look-up table calculated ...
 *     implementation using double precision accumulator
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_NaN); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_k__23 = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":1365
 * 
 * 
 * def correct_LUT_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                lut_t[:, ::1] LUT not None,
 *                                dummy=None, delta_dummy=None,
 */
  __pyx_t_9 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_35correct_LUT_preproc_double, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_LUT_preproc_double, __pyx_t_9) < 0) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pyFAI/ext/_distortion.pyx":1469
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 *                                empty=numpy.NaN):             # <<<<<<<<<<<<<<
 *     """Correct an image based on the look-up table calculated ...
 *     implementation using double precision accumulator
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_NaN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_k__24 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1466
 * 
 * 
 * def correct_CSR_preproc_double(image, shape_out,             # <<<<<<<<<<<<<<
 *                                LUT not None,
 *                                dummy=None, delta_dummy=None,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_37correct_CSR_preproc_double, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_CSR_preproc_double, __pyx_t_1) < 0) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1573
 * 
 * 
 * def uncorrect_LUT(image, shape, lut_t[:, :]LUT):             # <<<<<<<<<<<<<<
 *     """
 *     Take an image which has been corrected and transform it into it's raw (with loss of information)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_39uncorrect_LUT, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uncorrect_LUT, __pyx_t_1) < 0) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1610
 * 
 * 
 * def uncorrect_CSR(image, shape, LUT):             # <<<<<<<<<<<<<<
 *     """Take an image which has been corrected and transform it into it's raw (with loss of information)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyFAI_3ext_11_distortion_41uncorrect_CSR, NULL, __pyx_n_s_pyFAI_ext__distortion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uncorrect_CSR, __pyx_t_1) < 0) __PYX_ERR(0, 1610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1653
 * 
 * 
 * class Distortion(object):             # <<<<<<<<<<<<<<
 *     """
 * 
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__129); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple__129, __pyx_n_s_Distortion, __pyx_n_s_Distortion, (PyObject *) NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_kp_s_This_class_applies_a_distortion); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pyFAI/ext/_distortion.pyx":1661
 * 
 *     """
 *     def __init__(self, detector="detector", shape=None):             # <<<<<<<<<<<<<<
 *         """
 *         :param detector: detector instance or detector name
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_1__init__, 0, __pyx_n_s_Distortion___init, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__131)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__132);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_init, __pyx_t_3) < 0) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1680
 *         self.delta0 = self.delta1 = None  # max size of an pixel on a regular grid ...
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return os.linesep.join(["Distortion correction for detector:",
 *                                 self.detector.__repr__()])
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_3__repr__, 0, __pyx_n_s_Distortion___repr, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_repr, __pyx_t_3) < 0) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1684
 *                                 self.detector.__repr__()])
 * 
 *     def calc_pos(self):             # <<<<<<<<<<<<<<
 *         if self.pos is None:
 *             with self._sem:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_5calc_pos, 0, __pyx_n_s_Distortion_calc_pos, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_calc_pos, __pyx_t_3) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1708
 *         return self.pos
 * 
 *     def calc_LUT_size(self):             # <<<<<<<<<<<<<<
 *         """
 *         Considering the "half-CCD" spline from ID11 which describes a (1025,2048) detector,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_7calc_LUT_size, 0, __pyx_n_s_Distortion_calc_LUT_size, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_calc_LUT_size, __pyx_t_3) < 0) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1744
 *                     return np_lut_size
 * 
 *     def calc_LUT(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, j, ms, ml, ns, nl, shape0, shape1, size
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_9calc_LUT, 0, __pyx_n_s_Distortion_calc_LUT, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_calc_LUT, __pyx_t_3) < 0) __PYX_ERR(0, 1744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1856
 * #         return ab
 * 
 *     def correct(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Correct an image based on the look-up table calculated ...
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_11correct, 0, __pyx_n_s_Distortion_correct, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_correct, __pyx_t_3) < 0) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1895
 *         return out[:img_shape[0], :img_shape[1]]
 * 
 *     def uncorrect(self, image):             # <<<<<<<<<<<<<<
 *         """
 *         Take an image which has been corrected and transform it into it's raw (with loss of information)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5pyFAI_3ext_11_distortion_10Distortion_13uncorrect, 0, __pyx_n_s_Distortion_uncorrect, NULL, __pyx_n_s_pyFAI_ext__distortion, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_uncorrect, __pyx_t_3) < 0) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyFAI/ext/_distortion.pyx":1653
 * 
 * 
 * class Distortion(object):             # <<<<<<<<<<<<<<
 *     """
 * 
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Distortion, __pyx_tuple__129, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Distortion, __pyx_t_3) < 0) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyFAI/ext/_distortion.pyx":1
 * # -*- coding: utf-8 -*-             # <<<<<<<<<<<<<<
 * #cython: embedsignature=True, language_level=3
 * #cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__145, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__146, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__148, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__149, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_14[0] = PyThread_allocate_lock();
  __pyx_t_14[1] = PyThread_allocate_lock();
  __pyx_t_14[2] = PyThread_allocate_lock();
  __pyx_t_14[3] = PyThread_allocate_lock();
  __pyx_t_14[4] = PyThread_allocate_lock();
  __pyx_t_14[5] = PyThread_allocate_lock();
  __pyx_t_14[6] = PyThread_allocate_lock();
  __pyx_t_14[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_14, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pyFAI.ext._distortion", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyFAI.ext._distortion");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", c->ob_type->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyIntFromDouble */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value) {
    if (value >= (double)LONG_MIN && value <= (double)LONG_MAX) {
        return PyInt_FromLong((long)value);
    }
    return PyLong_FromDouble(value);
}
#endif

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyObjectFormatAndDecref */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* BytesEquals */
  static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
  static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyIntCompare */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyObjectSetAttrStr */
  #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyFloatBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
    }
        
        PyFPE_START_PROTECT("add", return NULL)
        result = a + b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* SliceObject */
      static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetTopmostException */
      #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
      #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrExceptionMatches */
      #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* ArgTypeTest */
      static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetAttr */
      static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* decode_c_string */
      static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
      static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* ExtTypeTest */
      static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* SwapException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
      static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
      #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* ImportFrom */
      static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
      static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
      #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
      #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
      static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
      static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
      static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
      #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
      static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyFloatBinop */
      #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: b = 0.0; break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
    }
        
        PyFPE_START_PROTECT("add", return NULL)
        result = a + b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* FetchCommonType */
        static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
        #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
        static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         self->ob_type->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CalculateMetaclass */
        static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* CythonFunction */
        static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3ClassCreate */
        static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


        /* MemviewSliceIsContig */
        static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
        static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
        static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* IsLittleEndian */
        static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
        static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
          static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
          static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 4,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 4,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int8_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_int8_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value) {
    const int8_t neg_one = (int8_t) ((int8_t) 0 - (int8_t) 1), const_zero = (int8_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int8_t),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_nn_int8_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_int8_t(*(int8_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn_int8_t(const char *itemp, PyObject *obj) {
    int8_t value = __Pyx_PyInt_As_int8_t(obj);
    if ((value == ((int8_t)-1)) && PyErr_Occurred())
        return 0;
    *(int8_t *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 4,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int8_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_int8_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 4,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[2];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[2];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(const char *itemp, PyObject *obj) {
    __pyx_t_5pyFAI_3ext_12shared_types_float32_t value = __pyx_PyFloat_AsFloat(obj);
    if ((value == (float)-1) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5pyFAI_3ext_12shared_types_float32_t *) itemp = value;
    return 1;
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
    const int32_t neg_one = (int32_t) ((int32_t) 0 - (int32_t) 1), const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_nn_int32_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_int32_t(*(int32_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn_int32_t(const char *itemp, PyObject *obj) {
    int32_t value = __Pyx_PyInt_As_int32_t(obj);
    if ((value == ((int32_t)-1)) && PyErr_Occurred())
        return 0;
    *(int32_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(__pyx_t_5pyFAI_3ext_12shared_types_float64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5pyFAI_3ext_12shared_types_float64_t value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5pyFAI_3ext_12shared_types_float64_t *) itemp = value;
    return 1;
}

static PyObject* __pyx_convert__to_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t s) {
            PyObject* res;
            PyObject* member;
            res = __Pyx_PyDict_NewPresized(2); if (unlikely(!res)) return NULL;
            member = __Pyx_PyInt_From_int32_t(s.idx); if (unlikely(!member)) goto bad;
            if (unlikely(PyDict_SetItem(res, __pyx_n_s_idx, member) < 0)) goto bad;
            Py_DECREF(member);
            member = PyFloat_FromDouble(s.coef); if (unlikely(!member)) goto bad;
            if (unlikely(PyDict_SetItem(res, __pyx_n_s_coef, member) < 0)) goto bad;
            Py_DECREF(member);
            return res;
            bad:
            Py_XDECREF(member);
            Py_DECREF(res);
            return NULL;
          }
          /* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(const char *itemp) {
    return (PyObject *) __pyx_convert__to_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(*(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(const char *itemp, PyObject *obj) {
    struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t value = __pyx_convert__from_py_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(obj);
    if (PyErr_Occurred())
        return 0;
    *(struct __pyx_t_5pyFAI_3ext_11_distortion_lut_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
          static CYTHON_INLINE PyObject *__pyx_memview_get_int(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_int(*(int *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_int(const char *itemp, PyObject *obj) {
    int value = __Pyx_PyInt_As_int(obj);
    if ((value == (int)-1) && PyErr_Occurred())
        return 0;
    *(int *) itemp = value;
    return 1;
}

/* MemviewSliceCopyTemplate */
          static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *x) {
    const int8_t neg_one = (int8_t) ((int8_t) 0 - (int8_t) 1), const_zero = (int8_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 2 * PyLong_SHIFT) {
                            return (int8_t) (((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 3 * PyLong_SHIFT) {
                            return (int8_t) (((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 4 * PyLong_SHIFT) {
                            return (int8_t) (((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) ((((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) ((((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) ((((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int8_t) -1;
        }
    } else {
        int8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int8_t) -1;
        val = __Pyx_PyInt_As_int8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int8_t");
    return (int8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int8_t");
    return (int8_t) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
    const int32_t neg_one = (int32_t) ((int32_t) 0 - (int32_t) 1), const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 2 * PyLong_SHIFT) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 3 * PyLong_SHIFT) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 4 * PyLong_SHIFT) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* BytesContains */
          static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character) {
    const Py_ssize_t length = PyBytes_GET_SIZE(bytes);
    char* char_start = PyBytes_AS_STRING(bytes);
    return memchr(char_start, (unsigned char)character, (size_t)length) != NULL;
}

/* ImportNumPyArray */
          static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_int32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_int32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn_int32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_int32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[2];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 3,
                                                 &__Pyx_TypeInfo_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_int32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5pyFAI_3ext_12shared_types_float32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_int(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_int, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_nn_int8_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_int8_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_int8_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_int8_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_int(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_int, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
          static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[2];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_nn_struct____pyx_t_5pyFAI_3ext_11_distortion_lut_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CheckBinaryVersion */
          static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
          static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
